;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-version.el,v 3.190 2002/12/15 22:18:54 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;; DO NOT TOUCH THIS LINE: ZZZXXX>>>1039990721<<<XXXZZZ

;;;
;;; Version number constants
;;;

(defconst irchat-client-version-major "4"
  "*Major version number. Major functional changes change this number.")

(defconst irchat-client-version-minor "00"
  "*Minor version number. Server Protocol changes and bug fixes change this number.")

(defconst irchat-client-version-beta-p t
  "*Is this version a beta version?")

(defconst irchat-client-version-alpha-p t
  "*Is this version an alpha version?")

(defconst irchat-client-version-rcs-snap t
  "*Is this just a snapshot from rcs?")

(defconst irchat-client-version-release "$Date: 2002/12/15 22:18:54 $"
  "*Version release date")

(defconst irchat-client-name "Irchat"
  "*Name of this program.")

(defconst irchat-client-distribution "none"
  "*What is the buld name of this distribution?")

(defconst irchat-emacs-version (let ((e-v-s (if (boundp 'emacs-version) 
						emacs-version
					      nil))
				     (e-v-l (if (fboundp 'emacs-version) 
						(emacs-version)
					      nil)))
				 (cond ((and (boundp 'emacs-major-version)
					     (integerp emacs-major-version))
					emacs-major-version)
				       ((and (stringp e-v-l)
					     (string-match 
					      "\\([1-9][0-9]*\\)\.[0-9][0-9]*"
					      e-v-l))
					(string-to-int
					 (substring e-v-l
						    (match-beginning 1)
						    (match-end 1))))
				       ((and (stringp e-v-s)
					     (string-match 
					      "\\([1-9][0-9]*\\)\.[0-9][0-9]*"
					      e-v-s))
					(string-to-int
					 (substring e-v-s
						    (match-beginning 1)
						    (match-end 1))))
				       (t 0)))
  "*Major number of emacs running this irchat.")


(defconst irchat-emacs-version-name
  (let ((e-v (emacs-version))
	(case-fold-search t)
	irchat-emacs-subtype
	irchat-emacs-version-number)
    (setq irchat-emacs-subtype
	(cond ((string-match "xemacs" e-v) "XEmacs")
	      ((string-match "lucid" e-v) "Lucid Emacs")
	      ((string-match "epoch" e-v) "Epoch")
	      ((or (string-match "gnu.emacs" e-v)
		   (string-match "gnuemacs" e-v)) "GNU Emacs")
	      ((string-match "mule" e-v) "Mule")
	      ((string-match "emacs" e-v) "Emacs")
	      (t "?????")))
    (setq irchat-emacs-version-number
	  (if (and (boundp 'emacs-major-version)
		   (boundp 'emacs-minor-version))
	      (format "%d.%d" emacs-major-version emacs-minor-version)
	    (if (string-match "[1-9][0-9]*\.[0-9][0-9]*" e-v)
		(substring e-v (match-beginning 0) (match-end 0))
	      "##.##")))
    (concat irchat-emacs-subtype " " irchat-emacs-version-number))
  "*Version of Emacs running this Irchat")

(defconst irchat-version
  (format "%s %s.%s%s %s" 
	  irchat-client-name 
	  irchat-client-version-major
	  irchat-client-version-minor
	  (cond (irchat-client-version-rcs-snap "current")
		(irchat-client-version-alpha-p  "alpha")
		(irchat-client-version-beta-p   "beta")
		(t ""))
	  (if (string-match ".*: \\([^ ]*\\) .*"
			    irchat-client-version-release)
	      (concat "("
		      (substring irchat-client-version-release 
				 (match-beginning 1)
				 (match-end 1))
		      ")")
	    ""))
  "*The version of irchat you are using.")

(defun irchat-version (&optional int)
  "Irchat version identifier string."
  (interactive '(t))
  (if int
      (message irchat-version))
  irchat-version)

(provide 'irchat-version)

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-globals.el,v 3.9 1998/05/26 14:07:29 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-and-compile
  (require 'irchat-version))

;;;
;;; Do not change these
;;;

(defconst irchat-author-nickname "tml"
  "The nickname used by the author of IRCHAT.")

(defvar irchat-server-process nil)

(defvar irchat-status-message-string nil)

(defun irchat-buffer-disable-undo (buffer)
  "Disabling UNDO"
  (if (fboundp 'buffer-disable-undo)
      (buffer-disable-undo buffer)
    (buffer-flush-undo buffer)))

(defun irchat-get-buffer-window (buffer)
  "Getting the window for a buffer"
  (if (> irchat-emacs-version 18)
      (get-buffer-window buffer t)
    (get-buffer-window buffer)))

(defun irchat-get-buffer-window-list (buf)
  "Getting the window list for a buffer."
  (if (string-match "XEmacs" emacs-version)
      (windows-of-buffer buf)
    (get-buffer-window-list buf nil t)))

(defconst irchat-client-error-msg "Unrecognized command: '%s'"
  "*Error message given to anyone asking wrong CLIENT data.")

(defvar irchat-channel-filter "" 
  "*Enables use of \\[universal-argument] with NAMES and TOPIC.")

(defvar irchat-invited-channel)
(setq irchat-invited-channel nil)

(defvar irchat-old-nickname nil
  "A place to keep old nickname in case it returs thas
we cannot change to new nick")

(defvar irchat-current-channel nil
  "The channel you currently have joined.")

(defvar irchat-current-topic nil
  "The topic of current channel.")

(defvar irchat-current-channels nil
  "The channels you have currently joined.")

(defvar irchat-channel-indicator "No channel"
  "The current channel, \"pretty-printed.\"")

(defvar irchat-private-indicator nil
  "A string displayed in the mode line indicating that user is
currently engaged in a one-to-one conversation.")

(defvar irchat-polling 0
  "T when we are automatically polling the server.")

(defvar irchat-ownfreeze nil
  "If non-nil the Dialogue window will not be scrolled automatically to bring
new entries into view when the user sends messages.")

(defvar irchat-privmsg-partner nil
  "The person who got your last private message.")

(defvar irchat-current-chat-partner nil
  "The person you are in a private conversation with.")

(defvar irchat-current-chat-partners nil
  "An list containing nics user is chatting with.")

(defvar irchat-chat-partner-alist nil
  "An alist containing nics user is chatting with.")

(defvar irchat-command-buffer-mode 'channel
  "symbol chat or channel depending on which is current mode at 
command buffer.")

(defvar irchat-nick-alist nil
  "An alist containing the nicknames of users known to currently be on IRC.
Each element in the list is a list containing a nickname.")

(defvar irchat-channel-alist nil 
  "An alist containing the channels on IRC.  Each element in the list is 
a list containing a channel name.")

(defvar irchat-greet-author nil
  "T until we notice that the author of IRCHAT is present, and send him
a greeting telling what version of IRCHAT this is.")

(defvar irchat-save-vars-is-dirty nil
  "Non nil if irchat_vars.el is changed but not saved.")

(defvar irchat-userathost ""
  "The user@host for the current line.")

(defvar irchat-userathost-type nil
  "The authentication of uerathost.  'ok 'not-verified 'fake or 'invalid")

(defvar irchat-awaymsg ""
  "Current away message.")


(defvar irchat-debug-buffer nil)
(defvar irchat-server-buffer nil)
(defvar irchat-server-name nil)

(defvar irchat-away-indicator "-")
(defvar irchat-freeze-indicator "-")
(defvar irchat-ownfreeze-indicator "-")
(defvar irchat-crypt-indicator "-")

(defconst irchat-msg 'msg)
(defconst irchat-privmsg 'privmsg)
(defconst irchat-killmsg 'kill)
(defconst irchat-ignoredmsg 'ignore)
(defvar irchat-nick-accepted nil)

(defconst irchat-obarray-size 1327
  "*The size of obarray used by irchat on channelname and username space.
For efficiency this should be prime. See documentation of intern and 
make-vector for more information. Here is a list of some small primes...    

13, 29, 37, 47, 59, 71, 89, 107, 131, 163, 197, 239, 293, 353, 431, 521, 
631, 761, 919, 1103, 1327, 1597, 1931, 2333, 2801, 3371, 4049, 4861, 5839, 
7013, 8419, 10103, 12143, 14591, 17519, 21023, 25229, 30293, 36353, 
43627, 52361, 62851, 75431, 90523, 108631, 130363, 156437, 187751, 
225307, 270371, 324449, 389357, 467237, 560689, 672827, 807403, 968897,
1162687, 1395263, 1674319, 2009191, 2411033, 2893249.")

(defvar irchat-obarray (make-vector irchat-obarray-size nil))

(defvar irchat-ctcp-ping-time (list 0 0 0))


(defvar irchat-query-client-lastcommand nil
  "*Place to keep last entered command")

(defvar irchat-query-client-nick nil
  "*Place to keep last queried nick")

(defvar irchat-query-client-alist
  '(("VERSION") ("CLIENTINFO") ("HELP") ("DCC") ("USERINFO") ("PING")
    ("X-FACE")))

(defconst irchat-query-client-insert-to-generic
  "")

(defconst irchat-query-client-version
  (concat "VERSION" irchat-query-client-insert-to-generic))

(defconst irchat-query-client-userinfo
  (concat "USERINFO" irchat-query-client-insert-to-generic))

(defconst irchat-query-client-help
  (concat "HELP" irchat-query-client-insert-to-generic))

(defconst irchat-query-client-clientinfo
  (concat "CLIENTINFO" irchat-query-client-insert-to-generic))

(defconst irchat-query-client-ping
  (concat "PING" irchat-query-client-insert-to-generic))

(defconst irchat-query-client-x-face
  (concat "X-FACE" irchat-query-client-insert-to-generic))

(eval-and-compile (provide 'irchat-globals))
;;;
;;;  eof
;;; 
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-vars.el,v 3.39 1998/11/04 10:54:33 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; user modifiable variables
;;;

(defvar irchat-saved-forms 
  (purecopy '(irchat-ignore-nickname irchat-ignore-message-regexp))
  "*Variables whose values are saved via command irchat-Command-save-vars")

(defvar irchat-command-window-height 4
  "*How large should Command window be on startup.")

(defvar irchat-use-full-window t
  "*If non-nil, IRCHAT will use whole emacs window. Annoying for GNUS-
users, therefore added by nam.")

(defvar irchat-want-traditional nil
  "*Do we want /commands.")

(defvar irchat-command-window-on-top nil
  "*If non-nil, the Command window will be put at the top of the screen.")

(defvar irchat-one-buffer-mode nil
  "*When non-nil, irchat will put up only a dialogue-buffer (on the
screen). Useful for those (perverts) who use 24 line terminals.")

(defvar irchat-variables-file "~/.irchat_vars.el"
  "*Where to look for variables. Helps to remove clutter from your .emacs.
This feature is most likely to dissappear in near future. The preferred 
way is to put irchat variables on .emacs or file loaded from there.")

(defvar irchat-backup-irchat-vars nil
  "*Do we want to backup irchat-vars file when updating it.  Nil inhibits.")

;;;
;;;  Where to connect
;;;
(defvar irchat-server (or (getenv "IRCSERVER") "ircd.eunet.fi")
  "*Name of the host running the IRC server. 
Value initialized from the IRCSERVER environment variable if one is set")

(defvar irchat-service 
  (let ((ircport-env (getenv "IRCPORT")))
    (if ircport-env
	(if (> (string-to-int ircport-env) 0)
	    (string-to-int ircport-env)
	  ircport-env)
      6667))
  "*IRC service name or (port) number.")

(defvar irchat-nickname (or (getenv "IRCNICK") (user-real-login-name))
  "*The nickname you want to use in IRC.
Default is the environment variable IRCNICK, or your login name.")

(defvar irchat-autoremove-virtual-penis t
  "*If t, virtual penis `_' is removed from nickname on reconnect")

(defvar irchat-real-nickname nil
  "Internal IRCHAT nickname.")

(defvar irchat-startup-channel nil
  "*The channel to join automatically at startup.
If nil, do not join any channel.")

(defvar irchat-reconnect-automagic t
  "*Automatic reconnection, default is disabled")

(defvar irchat-ask-for-nickname t
  "*Ask for nickname if irchat was entered with \\[universal-argument].")

(defvar irchat-grow-tail "_"
  "*Add irchat-grow-tail to nick when reconnecting. Otherwise you might get
killed again if automagic reconnect is too fast.")

(defvar irchat-message-length-limit 256
  "*Messages longer than this are split to pieces")

(defvar irchat-message-split-separator ""
  "String to used as separator in splitted messages.")

(defvar irchat-message-split-^C-compat t
  "Also recognize ^C as split separator?")

(defvar irchat-default-kick-message nil
  "String to use as message in channel KICK commands.")

(defvar irchat-ask-kick-message t
  "If non-nil, the KICK message is asked from user with every kick.")

;;;
;;;  Hooks, each message type has its own hook. These are used to setup
;;;  variables, buffers &c
;;;
(defvar irchat-Command-mode-hook nil
  "*A hook for IRCHAT Command mode.")

(defvar irchat-Dialogue-mode-hook nil
  "*A hook for IRCHAT Dialogue mode.")

(defvar irchat-Exit-hook nil
  "*A hook executed when signing off IRC.")

;;;
;;;  MULE coding systems
;;;
(defvar irchat-send-coding-system 'iso-8859-1
  "String coding system for sending.")

(defvar irchat-receive-coding-system 'iso-8859-1
   "String coding system for reading.")

;;;
;;;  DCC
;;;
(defvar irchat-dcc-program
;;; set dcc program as "dcc", "dcc.perl" or nil, depending which binary found.
;;; current version takes the one which is earlier in path (if many).
  (let* ((result nil) 
	 (list exec-path)
	 (path (car list)))
    (while (and path (not result))
      (setq path (file-name-as-directory path))
      (if (file-executable-p (concat path "dcc")) (setq result "dcc")
	(if (file-executable-p (concat path "dcc.perl")) 
	    (setq result "dcc.perl")))
      (setq list (cdr list)
	    path (car list)))
    result)
  "*Name of the external dcc-program.")

(defvar irchat-dcc-directory "~/tmp" 
  "*Directory where irchat-dcc puts its files.")

(defvar irchat-dcc-port 1200 
  "*Default port for DCC operations.")

;;;
;;;  Format strings for various messages
;;;
(defvar irchat-myformat-string ">"
  "*Format for own messages.")

(defvar irchat-format-string "-> *%s*"
  "*Format string for private messages being sent.")

(defvar irchat-format-string0 "-%s-"
  "*Format string for arriving NOTICE messages.")

(defvar irchat-format-string1 "*%s*"
  "*Format string for arriving private messages.")

(defvar irchat-format-string2 "<%s>"
  "*Format string for arriving messages to current channel.")

(defvar irchat-format-string3 "<%s:%s>"
  "*Format string for arriving messages to current channel from outside the channel.")

(defvar irchat-format-string4 "(%s)"
  "*Format string for arriving messages to other channel from outside the channel.")

(defvar irchat-format-string5 ">%s:%s<"
  "*Format string for arriving messages to other channel from outside the channel.")

(defvar irchat-change-prefix "*** Change: "
  "*String to add before any change msg, used for customisation of
IRCHAT to suit old users of the irc-loser-client.")

(defvar irchat-notice-prefix "*** Notice: "
  "*String to add before any notice message.")

(defvar irchat-broadcast-prefix "*** Broadcast: "
  "*String to add before any Broadcast message")

(defvar irchat-wallops-prefix "*** Notice: "
  "*String to add before any WALLOPS message")

(defvar irchat-error-prefix "*** Error: "
  "*String to add before any ERROR message")

(defvar irchat-info-prefix "*** Info: "
  "*String to add before any informational message")

;;;
;;;  Buffers
;;;
(defvar irchat-Command-buffer "*IRC Commands*"
  "*Name of Command input buffer")
(defvar irchat-Dialogue-buffer "*IRC Dialogue*"
  "*Name of Dialogue output buffer")
(defvar irchat-Private-buffer "*IRC Private*"
  "*Name of Private message buffer*")

(defvar irchat-KILLS-buffer "*IRC KILLS*")
(defvar irchat-IGNORED-buffer "*IRC IGNORED*")
(defvar irchat-WALLOPS-buffer "*IRC WALLOPS*")

;;;
;;;  Misc
;;;
(defvar irchat-send-delayed-allow-duplicates nil
  "*Allow duplicate strings to be queued in send-delayed?")

(defvar irchat-send-delayed-interval 3
  "Seconds between delayed sends.")

(defvar irchat-use-delayed-privmsg nil
  "*Should privmsg messages be delayed too?")

(defvar irchat-blink-parens nil
  "*Should we blink matching parenthesis in irchat command buffer?")

(defvar irchat-show-wallops t
  "*Show wallops messages if usermode +ws")

(defvar irchat-ignore-extra-notices t
  "*Don't show NOTICEs with \"as being away\" unless they come from
the local server.")

(defvar irchat-ignore-changes-from-ignored nil
  "*Don't show changes from ignored users.")

(defvar irchat-timestamp-irc-I-buffer-p nil
  "*Timestamps lines inserted into *IRC IGNORED* buffer.")

(defvar irchat-shorten-kills t
  "*Shorten KILL messages to about one line.")

(defvar irchat-ignore-changes nil
  "*Ignore changes? Good in topic-wars/link troubles.")

(defvar irchat-ignore-fakes nil
  "*If non nil, ignore fake notices if receiving them.")

(defvar irchat-signoff-msg ""
  "*Default signoff message")

(defvar irchat-confirm-bell-on-channel-message t
  "*If non-nil, confirmation is asked before bell is sent to the channel.")

(defvar irchat-beep-on-bells 'always
  "*If non-nil, and the IRC Dialogue buffer is not selected in a window,
an IRC message arriving containing a bell character, will cause you
to be notified.
If value is 'always, an arriving bell will always cause a beep (or flash).")

(defvar irchat-system-fqdname (system-name)
  "*The fully qualified domain name of the system.
Default is what (system-name) returns.")

(defvar irchat-use-smiley nil
  "*Display smileys in dialogue buffer.")

(defvar irchat-use-ancient-version-format nil
  "*Use ancient ultra-long and ultra-ugly version query reply.")

;;;
;;; Global-kill system
;;;
(defvar irchat-global-kill-valid-senders '()
  "*List of strings or string-int pairs containing valid IGNORE senders.")

(defvar irchat-global-kill-use-uah-p nil
  "*Use user-at-host in ignore command if possible")

(defvar irchat-global-kill-use-silent-ignore nil
  "*If non-nil, no expire messages are shown.")

(defvar irchat-global-kill-default-timeout-limit 99
  "*Default timeout limit for valid senders.")

(defvar irchat-global-kill-report-ignored nil
  "*How to report ignored kill request. (nil, t, 'minibuffer or 'dialogue)")

;;;
;;;  Send/Receive files
;;;
(defvar irchat-file-accept nil
  "*Do we accept files.")

(defvar irchat-file-confirm-save nil
  "*Do we want confirmation on saving files.")

;;;
;;;  Userinfos
;;;
(defvar irchat-client-userinfo "No user information given."
  "*Userinfo message given to anyone asking.")

(defvar irchat-client-x-face ""
  "*X-Face message given to anyone asking.  Proper form is \"X-Face: ....\".")

(defvar irchat-keepalive-interval nil
  "*Interval on seconds the existence of server connection is checked")

(defvar irchat-pollnames-interval nil
  "*Interval the names are polled from server")

(defvar irchat-timestamp-interval (* 60 10)
  "*Interval in seconds between timestamps in dialogue-buffer, nil for none.")

(defvar irchat-timestamp-format "*** Time: %s"
  "*Format-string for timestamp.")

;;;
;;; External version identification.
;;;
(defvar irchat-external-version-string nil
  "Version id given to VERSION queries. nil is default. 'none inhibits reply.")

;;;
;;;  Conversions: Not used.
;;;
(defvar irchat-send-convert-list nil 
  "*Convert characters before sending to server." )

(defvar irchat-receive-convert-list nil
  "*Convert characters after receiving from server." )

;;;
;;;  Ignores
;;;
(defvar irchat-ignore-nickname nil
  "*A list of nicknames, as symbols, to ignore.  Messages from these people
won't be displayed.")

(defvar irchat-ignore-message-regexp nil
  "*A list of regexps to ignore.  Messages matching to them 
won't be displayed.")

;(defvar irchat-ignore-realname nil
;  "*A list of real names of people to ignore. Messages from them
;won't be displayed.")

;(defvar irchat-ignore-logon nil
;  "*A list of logon names (user@host.dom.ain). Messages from them
;won't be displayed.")

;;;
;;; New Ignore
;;;
(defvar irchat-no-ignore-buffers nil
  "*A list of buffers that don't adhere to ignore-list.")

(defvar irchat-ignore-list nil
  "*A list of lists that contain ignore-info.
It is list of lists that contain two elements,
first element may be two item list of
(match-for-messagetype match-for-string-in-message)
match-for-messagetype may be nil,
match-for-string-in-message may be nil,
or
name of function that returns this list, the function takes
one argument, the string we are asked to ignore.
The other argument is
(list-of-users) 
list-of-users may be nil (in case it does not match),
or 
name of function returning that returns similar list, the
function takes one argument, the string we are asked to ignore.

(setq irchat-ignore-list
      (list
       (list (list \"msg\" nil) (list \"!user\" \"@host\"))
       (list (list \"privmsg\" nil) 'irchat-own-ignore-who)
       (list 'irchat-own-ignore-check (list \"com\"))
       (list (list nil \"fuck\") (list \".*\"))
       )
      )
The previous ignores normal msg`s (not used so much anymore)
from username matching 'user' or from host 'host'.
Then any privmsg from list that irchat-own-ignore-who return.
Then if the user is from .com (or has com somewhere in name
and the irchat-own-ignore-check returns message-type and/or string.
The last ignores all messages that contain the word `fuck`.
")

;;;
;;; Buffer display variables
;;;
(defvar irchat-buffer-list nil
  "*A list of buffers used in displaying messages.")

(defvar irchat-D-buffer (list irchat-Dialogue-buffer)
  "*A list of buffer where normal Dialogue is sent.")
(defvar irchat-P-buffer (list irchat-Dialogue-buffer irchat-Private-buffer)
  "*A list of buffers where private messages to me are sent.")
(defvar irchat-I-buffer (list irchat-IGNORED-buffer)
  "*A list of buffers where private messages to me are sent.")
(defvar irchat-W-buffer (list irchat-WALLOPS-buffer)
  "*A list of buffers where WALLOPS messages to me are sent.")
(defvar irchat-K-buffer (list irchat-KILLS-buffer)
  "*A list of buffers where KILL messages to me are sent.")
(defvar irchat-000-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where 000 messages to me are sent.")
(defvar irchat-200-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where 200 messages to me are sent.")
(defvar irchat-300-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where 300 messages to me are sent.")
(defvar irchat-400-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where 400 messages to me are sent.")
(defvar irchat-500-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where 500 messages to me are sent.")

(defvar irchat-debugmsg nil
  "*String to add in front of debug messages.")

(defvar irchat-compress-changes t
  "*Set to t if we want instant compressed messages in the old format.")
(defvar irchat-compress-treshold 1
  "*Number of lines to search back for a previous matching join/part/quit/mode.")

(defvar irchat-buffer-maxsize 4242424
  "*Maximum size (in bytes) of any irchat buffer.")
(defvar irchat-buffer-defsize 4042424
  "*Size to shrink buffer if it grows too big.")
(defvar irchat-checkbuffer-interval (* 60 10)
  "*Interval between buffer-size checks.")

(defvar irchat-private-window-height 4
  "*How tall is the window for private messages when shown.")

(defvar irchat-format-time-function (function irchat-compose-servertimestring)
  "*Function to convert result of (current-time-string) to human readable 
form. If you like to have the short format, set this to nil or to a 
funtion (defun my-format (s) s).")

(defvar irchat-keep-buffers t
  "*When starting IRC process, instructs whether to erase contents of old
buffers if NIL or to keep them if T.")

(defvar irchat-ownfreeze nil
  "*Set this to t if you don't want your own messages to scroll the
Dialogue window.")

(defvar irchat-scroll-step 1
  "*Set this to the number of lines you want to have empty when irchat scrolls. If nil, scrolls half a page.");

(defvar irchat-swap-private nil
  "*When showing private is it between under Dialogue or not.")

(defvar irchat-buffer-preferences
  (list 
   (list ".*" irchat-D-buffer))

  "*List of lists that contain regexp to match and buffer-list to
insert. Setting this as 
(list 
 (list \"#report\" (list \"*report-buffer*\")) 
 (list \".*\" irchat-D-buffer))
would cause messages from and to channel #report to be displayed on
buffer named *report-buffer* and all other messages are displayed on
Dialogue-buffer.")

(eval-and-compile (provide 'irchat-vars))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-inlines.el,v 3.6 1997/03/13 23:40:07 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;; these must be loaded in the file that requires this since this
;;; won't work otherwise.

(eval-and-compile  (require 'irchat-globals)  (require 'irchat-vars))


(defsubst matching-substring (string arg)
  (substring string (match-beginning arg) (match-end arg)))


(defsubst string-ci-equal (s1 s2)
  (string-equal (upcase s1) (upcase s2)))


(defsubst string-list-ci-memberp (thing list)
  "returns t if thing is member of list, not funcallable"
  (let ((item (car list))
	(uthing (upcase thing)))
    (while (and item (not (string-equal uthing (upcase item))))
      (setq item (car list)
	    list (cdr list)))
    item))


(defsubst string-list-memberp (thing list)
  "returns t if thing is member of list, not funcallable"
  (let ((item (car list)))
    (while (and item (not (string-equal thing item)))
      (setq item (car list)
	    list (cdr list)))
    item))


(defsubst string-list-ci-delete (thing list)
  (let ((uthing (upcase thing))
	(item (car list))
	(result nil))
    (while item
      (if (listp item)
	  (if (not (string-equal (upcase (car item)) uthing))
	      (setq result (nconc result (list item))))
	(if (not (string-equal (upcase item) uthing))
	    (setq result (nconc result (list item)))))
      (setq list (cdr list)
	    item (car list)))
    result))


(defsubst string-list-delete (thing list)
  (let ((item (car list))
	(result nil))
    (while item
      (if (listp item)
	  (if (not (string-equal (car item) thing))
	      (setq result (nconc result (list item))))
	(if (not (string-equal item thing))
	    (setq result (nconc result (list item)))))
      (setq list (cdr list)
	    item (car list)))
    result))


(defsubst list-to-assoclist (list)
  (let ((result nil) (item (car list)))
    (while item
      (setq result (cons (list item) result)
	    list (cdr list)
	    item (car list)))
    result))


(defsubst assoc-ci-string (key lst)
  (assoc-if
   (function
    (lambda (x)
      (string-ci-equal x key)))
   lst))

(defsubst assoc-ci-regexp (key lst)
  "Assoc with REGEXP-KEY from LIST."
  (assoc-if
   (function
    (lambda (x)
      (not (null (string-match (concat "^" (upcase key) "$") (upcase x))))))
   lst))

(defsubst assoc-ci-regexp-rev (key lst)
  "Assoc with KEY from LIST, in which keys are regexps."
  (assoc-if
   (function
    (lambda (x)
      (not (null (string-match (concat "^" (upcase x) "$") (upcase key))))))
   lst))

;;;;;

(defsubst irchat-scan-channels (chnl)
  (setq irchat-channel-alist 
	(if (assoc chnl irchat-channel-alist)
	    irchat-channel-alist 
	  (cons (list chnl) irchat-channel-alist))))


(defsubst irchat-user-on-this-channel (user chnl)
  "return T if USER is on channel CHNL"
  (let ((u (intern user irchat-obarray)))
    (string-list-ci-memberp chnl (get u 'chnl))))


(defsubst irchat-update-user (chnl user)
  "Add CHNL to list of channels USER belongs to"
  (if (not (string= user ""))
      (let* ((u (if (or (string= (substring user 0 1) "@")
			(string= (substring user 0 1) "+"))
                    (intern (substring user 1 (length user)) irchat-obarray)
                  (intern user irchat-obarray)))
             (chnls (get u 'chnl)))
        (progn
          (if (get u 'irchat-waited-for)
              (irchat-greet-user user chnl))
          (if (and irchat-greet-author 
                   (string= user irchat-author-nickname))
              (irchat-greet-author))

	  (if (not (assoc (prin1-to-string u) irchat-nick-alist))
	      (setq irchat-nick-alist (cons 
				       (list (prin1-to-string u)) 
				       irchat-nick-alist)))

          (if (not (string-list-ci-memberp chnl chnls))
              (put u 'chnl (nconc chnls (list chnl))))))))


(defsubst irchat-add-to-channel (user chnl)
  "Add users info to his chnl"
  (irchat-update-user chnl user))


(defsubst irchat-remove-from-thischannel (user chnl)
  "Remove users info from his chnl"
  (let ((u (intern user irchat-obarray)))
    (put u 'chnl (string-list-ci-delete chnl (get u 'chnl)))))


(defsubst irchat-update-thischannel (chnl origusers)
  "Update our copy of users on channel chnl."
  (let ((users origusers))
    (while (string-match "^\\([^ ]*\\) \\(.*\\)" users)
      (irchat-update-user chnl (matching-substring users 1))
      (setq users (matching-substring users 2)))
    (irchat-update-user chnl users)))


(defsubst irchat-next-line (&optional n)
  (if (= (point) (point-max))
      (newline))
  (forward-line n))


(defsubst irchat-Dialogue-insert (msg)
  (save-excursion 
    (irchat-w-insert irchat-D-buffer (format "%s\n" msg))))

(eval-and-compile (provide 'irchat-inlines))

;;;
;;;  eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-filter.el,v 3.12 2002/06/09 16:43:51 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  These are defsubst just for speed, as it is expensive to call funtions at
;;;  emacs lisp (also evals are expensive)
;;;
(defun irchat-run-message-hook-types (hook prefix rest-of-line)
  "Run either old fashion irchat hook variable or hook list."
  (let ((hook (if (and (symbolp hook)
		       (boundp hook)
		       (not (fboundp hook)))
		  (eval hook)
		hook)))
    (if (or (and (symbolp hook)
		 (fboundp hook))
	    (and (listp hook)
		 (eq (car hook) 'lambda))
	    (compiled-function-p hook))
	(eval (list hook prefix rest-of-line))
      (if (listp hook)
	  (let ((hooks hook)
		(r nil))
	    (while hooks
	      (if (irchat-run-message-hook-types (car hooks)
						 prefix 
						 rest-of-line)
		  (progn
		    (setq hooks '())
		    (setq r t))
		(setq hooks (cdr hooks))))
	    r)
	nil))))

(defsubst irchat-handle-message-2 (parsed-sender parsed-msg prefix message rest-of-line)
  "Helper function (actually a macro) for irchat-handle-message."
  (let ((hook (intern (concat "irchat-" message "-hook")))
	(after-hook (intern (concat "irchat-after-" message "-hook")))
	buffer-read-only fun)
    (if (irchat-run-message-hook-types hook prefix rest-of-line)
	;; If we have a hook, and it returns T, do nothing more
	nil
      ;; else call the handler

      (progn
	(if (string= message "msg")
	    (irchat-handle-msg-msg parsed-sender parsed-msg prefix rest-of-line)
	    
	  (if (fboundp (setq fun (intern
				  (concat "irchat-handle-" message "-msg"))))
	      (progn
		(apply fun (list parsed-sender parsed-msg prefix rest-of-line)))
	    (let* ((message-number (string-to-int message))
		   (default-number (/ message-number 100)))
	      (if (and (> message-number 0)
		       (fboundp (setq fun (intern
					   (concat "irchat-handle-"
						   (format "%d00" 
							   default-number)
						   "-msgs")))))
		  (progn
		    (apply fun (list message-number parsed-sender parsed-msg prefix rest-of-line)))
		(message "IRCHAT: Unknown IRC message \":%s %s %s\"" prefix
			 (upcase message) rest-of-line)
		(irchat-w-insert irchat-D-buffer 
				 (format "MESSAGE: %s, %s, %s\n" 
					 prefix message rest-of-line))
		))))
	(irchat-run-message-hook-types after-hook prefix rest-of-line)))))


(defsubst irchat-handle-message ()
  "Called when we have at least one line of output from the IRC server."
  (let ((obuf (current-buffer))
	beg end prefix message rest-of-line)
    (while (or
	    (looking-at 
	     "\\(:[^! \n]*\\)!\\([^ \n]*\\) \\([^ \n]+\\) :?\\(.*\\)\r\n")
	    (looking-at 
	     "\\(:[^ \n]*\\)?\\(\\) *\\([^ \n]+\\) :?\\(.*\\)\r\n")
	    (looking-at 
	     "\\(:[^! \n]*\\)!\\([^ \n]*\\) \\([^ \n]+\\) :?\\(.*\\)\n")
	    (looking-at 
	     "\\(:[^ \n]*\\)?\\(\\) *\\([^ \n]+\\) :?\\(.*\\)\n"))

      (setq beg (match-beginning 0)
	    end (match-end 0)
	    prefix (if (match-beginning 1)
		       (buffer-substring (1+ (match-beginning 1)) 
					 (match-end 1)))
	    irchat-userathost (buffer-substring (match-beginning 2)
						(match-end 2))
	    rest-of-line (buffer-substring (match-beginning 4) (match-end 4))
	    message (downcase 
		     (buffer-substring (match-beginning 3) (match-end 3))))

      (let ((cookie (if (and (stringp irchat-userathost)
			     (> (length irchat-userathost) 2))
			(substring irchat-userathost 0 1)
		      nil)))
	(cond ((null cookie)
	       (setq irchat-userathost-type 'invalid))
	      ((string= cookie "^")
	       (setq irchat-userathost (substring irchat-userathost
						  1
						  (length irchat-userathost)))
	       (setq irchat-userathost-type 'fake))
	      ((string= cookie "~")
	       (setq irchat-userathost (substring irchat-userathost
						  1
						  (length irchat-userathost)))
	       (setq irchat-userathost-type 'not-verified))
	      (t
	       (setq irchat-userathost-type 'ok))))

      (set-buffer irchat-Dialogue-buffer)
      (let ((parsed-sender nil)
	    (parsed-msg nil))
	(setq irchat-current-function (list parsed-sender parsed-msg prefix rest-of-line message))
	(irchat-handle-message-2 parsed-sender parsed-msg prefix message rest-of-line))
      (setq irchat-current-function (list "" "" "" "" ""))
      (set-buffer obuf)
      (delete-region beg end))))


(defun irchat-filter (process output)
  "Filter function for IRC server process."
  (let ((obuf (current-buffer))
	(data (match-data))
	(output (irchat-decode-coding-string output))
	bol)
    ;;
    ;; C-c C-d creates debug buffer for incoming messages...
    ;;
    (irchat-insert-to-debug output "<- ")

    (set-buffer (process-buffer process))
    (goto-char (point-max))
    ;;
    ;; convert input before it is processed: convert any occurrences of 
    ;; of heads of convert-list to corresponding tails. heads and tails 
    ;; may be functions in case they are evaluated, pattern-function is 
    ;; called with received message as input and target-function with 
    ;; output of pattern function: this enables user to create bots for 
    ;; example... BTW: long convert list slow thing down noticeably
    ;;
    (let ((conv-list irchat-receive-convert-list))
      (while (and conv-list (not irchat-polling))
	(let* ((i (car conv-list)) (f (car i)) (s (car (cdr i)))
	       (s1 (if (stringp f) f (funcall f output)))
	       (s2 (if (stringp s) s (funcall s s1))))
	  (setq output (irchat-replace-in-string output s1 s2)
		conv-list (cdr conv-list)))))

    (insert output)
    (goto-char (point-min))
    (while (re-search-forward "\n\n" (point-max) t)
      (delete-char -1)) ; This hack (from mta) is for 2.4 servers
    (goto-char (point-min))

    (if (string-match "\n" output)
	(irchat-handle-message))
    (set-buffer obuf)
    (store-match-data data)))


(defun irchat-sentinel (proc status)
  "Sentinel function for Irchat process."
  (if irchat-reconnect-automagic
      (progn 
	(condition-case err
	    (progn
	      (set-process-filter irchat-server-process nil))
	  (wrong-type-argument 
	   nil))
	(condition-case err
	    (progn
	      (set-process-sentinel irchat-server-process nil))
	  (wrong-type-argument 
	   nil))
	(setq irchat-server-process nil)
	(irchat))
    (message "IRCHAT: Connection closed.")))

(eval-and-compile (provide 'irchat-filter))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-dcc.el,v 3.5 1998/10/06 11:47:06 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info
;;;
;;; This file:
;;; Copyright 1994 Kim Nyberg <kny@tekla.fi>
;;; All Rights Reserved

(eval-when-compile (require 'irchat-inlines))

(defvar irchat-dcc-receive-list nil)
(defvar irchat-dcc-receive-direct t)

(defun irchat-ctl-a-dcc-msg (from chnl rest)
  (irchat-w-insert irchat-D-buffer (format "%sDCC from %s: %s\n"
					   irchat-info-prefix from rest))
  ;; old versions of ircii 2.2.1, do not send file length on DCC request.
  ;; so irchat can not do DCC with them. Sorry. ircii 2.2.9 does it on 
  ;; correctly. Why do people change protocol inside one minor version?
  (if (string-match 
       "\\(SEND\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\)" rest)
      (let ((command "receive")
	    (filename (file-name-nondirectory (matching-substring rest 2)))
	    (host (matching-substring rest 3))
	    (port (matching-substring rest 4))
	    (size (matching-substring rest 5)))
	(setq irchat-dcc-receive-list
	      (append irchat-dcc-receive-list
		      (list (list command host port size filename))))
	(if irchat-dcc-receive-direct
	    (progn 
	      (irchat-w-insert irchat-D-buffer 
			       (format "%sDCC autoreceive\n" irchat-info-prefix))
	      (irchat-Command-dcc-receive))))))


(defun irchat-Command-dcc-send (filename towhom)
  "Send file to user."
  (interactive
   (list (expand-file-name (read-file-name
			    "File to send: "
			    default-directory nil))
	 (irchat-completing-default-read 
	  "To whom: "
	  (append irchat-nick-alist irchat-channel-alist)
	  '(lambda (s) t) 
	  nil irchat-privmsg-partner)))

  (setq irchat-privmsg-partner towhom)
  (set-process-filter
   (start-process irchat-dcc-program nil 
		  irchat-dcc-program "send" 
		  (number-to-string irchat-dcc-port)
		  filename)
   'irchat-dcc-send-filter)
  (setq irchat-dcc-port (1+ irchat-dcc-port)))


(defun irchat-dcc-send-filter (process output)
  (cond ((string-match "DCC file [^ ]* sent" output)
	 (irchat-own-message (substring output 0 (1- (length output))))
	 (message ""))
	((string-match 
	  "DCC send [^ ]*/\\([^ ]*\\) \\([^ ]+\\) \\([^ ]+\\) \\(.+\\)" output)
	 (let ((filename (matching-substring output 1))
	       (port (matching-substring output 2))
	       (machine (matching-substring output 3))
	       (size (matching-substring output 4)))
	   (irchat-send-privmsg "PRIVMSG %s :DCC SEND %s %s %s %s"
			irchat-privmsg-partner filename machine port size)
	   (irchat-own-message
	    (format "%sSending file %s (%s bytes) to %s"
		    irchat-info-prefix filename size irchat-privmsg-partner))))))


(defun irchat-Command-dcc-receive (&optional number)
  "Receive next file from list."
  (interactive "P")
  (let ((dcc-object (car irchat-dcc-receive-list)))
    (setq irchat-dcc-receive-list (cdr irchat-dcc-receive-list))
    (if dcc-object
	(let ((command (car dcc-object))
	      (host (nth 1 dcc-object))
	      (port (nth 2 dcc-object))
	      (size (nth 3 dcc-object))
	      (default (nth 4 dcc-object)))
	  (irchat-own-message (format "%sGetting file %s (%s bytes)"
				      irchat-info-prefix default size))
	  (let ((filename (expand-file-name default irchat-dcc-directory)))
	    (set-process-filter
	     (start-process irchat-dcc-program nil irchat-dcc-program 
			    command host port size filename)
	     'irchat-dcc-receive-filter))))))


(defun irchat-dcc-receive-filter (process output)
  (if (string-match "DCC file [^ ]* received" output)
      (irchat-own-message (substring output 0 (1- (length output))))
    (message "%s" (substring output 0 (1- (length output))))))


(defun irchat-Command-dcc-list ()
  "List files in receive queue."
  (interactive)
  (let* ((dcc-list irchat-dcc-receive-list)
	 dcc-object)
    (while dcc-list
      (setq dcc-object (car dcc-list)
	    dcc-list (cdr dcc-list))
      (let ((filename (cdr (cdr (cdr (cdr dcc-object)))))
	    (size (car (cdr (cdr (cdr dcc-object))))))
	(irchat-own-message (format "%sDCC file %s (%s bytes)"
				    irchat-info-prefix filename size))))))


(defun irchat-dcc-compare-hostnames (h1 h2)
  "Compare two internet domain hostnames. Return true iff they resolve to the 
same IP-address."
  (if (string-ci-equal h1 h2)
      t
    (if irchat-dcc-program
	(let ((pob (get-buffer-create "*IRC DCC resolve*"))
	      (output) (domatch nil))
	  (save-excursion
	    (call-process irchat-dcc-program nil pob nil "resolve" h1 h2)
	    (set-buffer pob)
	    (goto-char (point-min))
	    (setq output (buffer-substring (point-min) (point-max)))
	    (if (string-match "\\([^ ]+\\)\n\\([^ ]+\\)\n" output)
		(if (string= (matching-substring output 1) 
			     (matching-substring output 2))
		    (setq domatch t))))
	  (kill-buffer pob)
	  domatch)
      (string-ci-equal h1 h2))))
      
(eval-and-compile (provide 'irchat-dcc))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-caesar.el,v 3.3 1998/08/08 14:56:08 tri Exp $
;;;
;;; caesar-region written by phr@prep.ai.mit.edu  Nov 86
;;; Modified by tower@prep Nov 86
;;; Modified by umerin@flab.flab.Fujitsu.JUNET for ROT47.
;;; Modified by tri@iki.fi for irchat (japanese is now optional)

(eval-and-compile (provide 'irchat-caesar))

(defun irchat-caesar-string (str &optional n copy japanese)
  "Caesar rotation of STRING by N, default 13, for decrypting netnews.
ROT47 will be performed for Japanese text in any case."
  (cond ((not (numberp n)) (setq n 13))
	((< n 0) (setq n (- 26 (% (- n) 26))))
	(t (setq n (% n 26))))		;canonicalize N
  (if (not (zerop n))			; no action needed for a rot of 0
      (progn 
	(if (and (not japanese)
		 (or (not (boundp 'caesar-translate-table-1))
		     (/= (aref caesar-translate-table-1 ?a) (+ ?a n))))
	    (let ((i 0) (lower "abcdefghijklmnopqrstuvwxyz") upper)
	      (message "Building caesar-translate-table-1...")
	      (setq caesar-translate-table-1 (make-vector 256 0))
	      (while (< i 256)
		(aset caesar-translate-table-1 i i)
		(setq i (1+ i)))
	      (setq lower (concat lower lower) upper (upcase lower) i 0)
	      (while (< i 26)
		(aset caesar-translate-table-1 (+ ?a i) (aref lower (+ i n)))
		(aset caesar-translate-table-1 (+ ?A i) (aref upper (+ i n)))
		(setq i (1+ i)))
	      (message "Building caesar-translate-table 1... done")))
	(if (and japanese
		 (or (not (boundp 'caesar-translate-table-2))
		     (/= (aref caesar-translate-table-2 ?a) (+ ?a n))))
	    (let ((i 0) (lower "abcdefghijklmnopqrstuvwxyz") upper)
	      (message "Building caesar-translate-table-2...")
	      (setq caesar-translate-table-2 (make-vector 256 0))
	      (while (< i 256)
		(aset caesar-translate-table-2 i i)
		(setq i (1+ i)))
	      (setq lower (concat lower lower) upper (upcase lower) i 0)
	      (while (< i 26)
		(aset caesar-translate-table-2 (+ ?a i) (aref lower (+ i n)))
		(aset caesar-translate-table-2 (+ ?A i) (aref upper (+ i n)))
		(setq i (1+ i)))
	      ;; ROT47 for Japanese text.
	      ;; Thanks to ichikawa@flab.fujitsu.junet.
	      (setq i 161)
	      (let ((t1 (logior ?O 128))
		    (t2 (logior ?! 128))
		    (t3 (logior ?~ 128)))
		(while (< i 256)
		  (aset caesar-translate-table-2 i
			(let ((v (aref caesar-translate-table-2 i)))
			  (if (<= v t1) (if (< v t2) v (+ v 47))
			    (if (<= v t3) (- v 47) v))))
		  (setq i (1+ i))))
	      (message "Building caesar-translate-table 2... done")))
	(let ((caesar-translate-table (if japanese 
					  caesar-translate-table-2
					caesar-translate-table-1))
	      (str (if copy (substring str 0 (length str)) str))
	      (i 0) 
	      len)
	  (setq len (length str))
	  (while (< i len)
	    (aset str i (aref caesar-translate-table (aref str i)))
	    (setq i (1+ i)))
	  str))))


(defun irchat-caesar-region (str &optional n &optional japanese)
  "Caesar rotation of region by N, default 13, for decrypting netnews.
ROT47 will be performed for Japanese text in any case."
  (interactive (if current-prefix-arg	; Was there a prefix arg?
		   (list (prefix-numeric-value current-prefix-arg))
		 (list nil nil)))
  (let* ((from (region-beginning))
	 (to (region-end))
	 (str (buffer-substring from to)))
    (setq str (irchat-caesar-string str n nil japanese))
    (goto-char from)
    (delete-region from to)
    (insert str)))
  
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-000.el,v 3.3 2002/06/09 15:16:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  000 replies -- what is the author of ircd thinking
;;;
(defun irchat-handle-000-msgs (number parsed-sender parsed-msg prefix rest)
  (setq irchat-nick-accepted 'ok)
  (if (string-match "[^ ]* \\([^ :]*\\) *\\([^ :]*\\) *:\\(.*\\)" rest)
      (let ((target1 (matching-substring rest 1))
	    (target2 (matching-substring rest 2))
	    (msg (matching-substring rest 3)))
	(cond ((string-equal target1 "")
	       (irchat-w-insert irchat-000-buffer 
				(format "%s%s\n" irchat-info-prefix msg)))
	      ((string-equal target2 "")
	       (irchat-w-insert irchat-000-buffer 
				(format "%s%s (%s)\n"
					irchat-info-prefix msg target1)))
	      (t
	       (irchat-w-insert irchat-000-buffer 
				(format "%s%s %s (%s)\n" irchat-info-prefix
					target1 msg target2)))))
    (message "IRCHAT: Strange %s reply" number)))


(defun irchat-handle-004-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-000-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 004 reply")))

;;;
;;;  eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-200.el,v 3.4 2002/06/09 15:16:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  200 replies
;;;
(defun irchat-handle-200-msgs (number parsed-sender parsed-msg prefix rest)
  (if (string-match "[^ ]* \\([^ :]*\\) *\\([^ :]*\\) *:\\(.*\\)" rest)
      (let ((target1 (matching-substring rest 1))
	    (target2 (matching-substring rest 2))
	    (msg (matching-substring rest 3)))
	(cond ((string-equal target1 "")
	       (irchat-w-insert irchat-200-buffer 
				(format "%s%s\n" irchat-info-prefix msg)))
	      ((string-equal target2 "")
	       (irchat-w-insert irchat-200-buffer 
				(format "%s%s %s\n"
					irchat-info-prefix target1 msg)))
	      (t
	       (irchat-w-insert irchat-200-buffer 
				(format "%s%s %s (%s)\n" 
					irchat-info-prefix target1 msg target2)))))
    (message "IRCHAT: Strange %s reply" number)))


(defun irchat-handle-200-msg (parsed-sender parsed-msg prefix rest)
  "200 TRACELINK Link <version & debug level> <destination> <next server>"
  (if (string-match "Link \\([^ ]*\\)[ :]*\\([^ ]*\\)[ :]*\\(.*\\)" rest)
      (let ((version (matching-substring rest 1))
	    (dest (matching-substring rest 2))
	    (next (matching-substring rest 3)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%sLink %s (%s) ==> %s (next %s)\n"
				 irchat-info-prefix prefix version dest next)))
    (message "IRCHAT: Strange 200 message")))


(defun irchat-handle-201-msg (parsed-sender parsed-msg prefix rest)
  "201 TRACECONNECTING Try. <class> <server>"
  (if (string-match "[^ ]* [^ ]* \\([0-9]*\\)[ :]*\\(.*\\)" rest)
      (let ((class (matching-substring rest 1))
	    (server (matching-substring rest 2)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%s%s Trying to connect to %s (class %s)\n" 
				 irchat-info-prefix prefix server class)))
    (message "IRCHAT: Strange 201 message")))


(defun irchat-handle-202-msg (parsed-sender parsed-msg prefix rest)
  "202 RPL_TRACEHANDSHAKE H.S. <class> <server>"
  (if (string-match "[^ ]* [^ ]* \\([0-9]*\\)[ :]*\\(.*\\)" rest)
      (let ((class (matching-substring rest 1))
	    (server (matching-substring rest 2)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s Handshaking with %s (class: %s)\n" 
				 irchat-info-prefix prefix server class)))
    (message "IRCHAT: Strange 202 message")))


(defun irchat-handle-203-msg (parsed-sender parsed-msg prefix rest)
  "203 RPL_TRACEUNKNOWN ???? <class> [<client IP address in dot form>]"
  (if (string-match "\\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\)[ :]+\\(.*\\)" rest)
      (let ((status (matching-substring rest 2))
	    (class (matching-substring rest 3))
	    (who (matching-substring rest 4)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s Class[%s] ==> %s\n"
				 irchat-info-prefix status class who)))
    (message "IRCHAT: Strange 203 message")))


(defun irchat-handle-204-msg (parsed-sender parsed-msg prefix rest)
  "204 RPL_TRACEOPERATOR Oper <class> <nick>"
  (if (string-match "\\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\)[ :]+\\(.*\\)" rest)
      (let ((status (matching-substring rest 2))
	    (class (matching-substring rest 3))
	    (who (matching-substring rest 4)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s Class[%s] ==> %s\n" 
				 irchat-info-prefix status class who)))
  (message "IRCHAT: Strange 204 message")))


(defun irchat-handle-205-msg (parsed-sender parsed-msg prefix rest)
  "205 RPL_TRACEUSER User %d %s"
  (if (string-match "[^ ]* \\([^ ]*\\) \\([0-9]*\\)[ :]*\\(.*\\)" rest)
      (let ((kind (matching-substring rest 1))
	    (hops (matching-substring rest 2))
	    (where (matching-substring rest 3)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s Class[%s] ==> %s\n"
				 irchat-info-prefix kind hops where)))
    (message "IRCHAT: Strange 205 message")))


(defun irchat-handle-206-msg (parsed-sender parsed-msg prefix rest)
  "206 RPL_TRACESERVER Serv %d %dS %dC %s %s!%s@%s"
  (if (string-match "Serv \\([^ ]*\\) \\(.*\\)" rest)
      (let ((class (matching-substring rest 1))
	    (pars (matching-substring rest 2)))
	(if (string-match "^[ :]*\\(.*\\)" pars)
	    (irchat-w-insert irchat-200-buffer
			     (format "%sServ %s (%s) ==> %s\n" irchat-info-prefix
				     prefix class (matching-substring pars 1)))
	  (if (string-match 
	       "[ :]*\\([0-9]*\\)*C \\([0-9]*\\)*S[ :]*\\(.*\\)"
	       pars)
	      (irchat-w-insert irchat-200-buffer
			       (format
				"%sServ %s (%s) ==> %s (%sC, %sS)\n"
				irchat-info-prefix prefix class
				(matching-substring pars 3)
				(matching-substring pars 1)
				(matching-substring pars 2)))
	    (irchat-w-insert irchat-200-buffer 
			     (format "%sServ %s (%s) ==> %s\n" 
				     irchat-info-prefix prefix class pars)))))
    (message "IRCHAT: Strange 206 message")))


(defun irchat-handle-207-msg (parsed-sender parsed-msg prefix rest)
  "207 RPL_TRACESERVICE Service %d %s"
  (if (string-match "[^ ]* Service \\([0-9]*\\) \\(.*\\)" rest)
      (let ((class (matching-substring rest 1))
	    (service (matching-substring rest 2)))
	(irchat-w-insert irchat-200-buffer (format
		 "%sService %s (class %s)\n" irchat-info-prefix service class)))
    (message "IRCHAT: Strange 207 message")))


(defun irchat-handle-208-msg (parsed-sender parsed-msg prefix rest)
  "208 RPL_TRACENEWTYPE <newtype> 0 %s"
  (irchat-w-insert irchat-200-buffer 
		   (format "%s%s: RPL_TRACENEWTYPE: Why this?\n"
			   irchat-info-prefix prefix))
  nil)


(defun irchat-handle-209-msg (parsed-sender parsed-msg prefix rest)
  "RPL_TRACECLASS Class %d %d"
  (if (string-match "[^ ]* Class \\([0-9]*\\) \\([0-9]*\\)" rest)
      (let ((class (matching-substring rest 1))
	    (entries (matching-substring rest 2)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%sClass %s Entries linked: %s\n" 
				 irchat-info-prefix class entries)))
    (message "IRCHAT: Strange 209 message")))


(defun irchat-handle-211-msg (parsed-sender parsed-msg prefix rest)
  "NOTICE %s :%-15.15s%5u%7u%10u%7u%10u %s"
  (if (string-match "\\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\)[ :]+\\(.*\\)" rest)
      (let ((link (matching-substring rest 2))
	    (sendq (matching-substring rest 3))
	    (sendm (matching-substring rest 4))
	    (sendb (matching-substring rest 5))
	    (rcvem (matching-substring rest 6))
	    (rcveb (matching-substring rest 7))
	    (open (matching-substring rest 8)))
	(irchat-w-insert irchat-200-buffer 
	 (format "%s%-35s %s: %5s%7s%10s%7s%10s %s\n" irchat-info-prefix link
		 prefix sendq sendm sendb rcvem rcveb open)))
    (message "IRCHAT: Strance 211 message")))


(defun irchat-handle-212-msg (parsed-sender parsed-msg prefix rest)
  "212 RPL_STATSCOMMANDS %s %u %u"
  (if (string-match "[^ ]* \\([^ ]*\\) \\([0-9]*\\)" rest)
      (let ((cmd (matching-substring rest 1))
	    (times (matching-substring rest 2)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%s%s has been used %s times after startup\n"
				 irchat-info-prefix cmd times)))
    (message "IRCHAT: Strange 212 message")))


(defun irchat-handle-213-msg (parsed-sender parsed-msg prefix rest)
  "213 RPL_STATSCLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.\\) \\([^ ]*\\) \\(.\\) \\([^ ]*\\) \\([0-9]*\\) \\([0-9]*\\)" rest)
      (let ((cn (matching-substring rest 1))
	    (canon (matching-substring rest 2))
	    (pass (matching-substring rest 3))
	    (name (matching-substring rest 4))
	    (port (matching-substring rest 5))
	    (hmmm (matching-substring rest 6)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%s%s:%s:%s:%s:%s:%s\n" 
				 irchat-info-prefix cn canon pass name port hmmm)))
    (message "IRCHAT: Strange 213 message")))


(defun irchat-handle-214-msg (parsed-sender parsed-msg prefix rest)
  "214 RPL_STATSNLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.\\) \\([^ ]*\\) \\(.\\) \\([^ ]*\\) \\([0-9]*\\) \\([0-9]*\\)" rest)
      (let ((cn (matching-substring rest 1))
	    (canon (matching-substring rest 2))
	    (pass (matching-substring rest 3))
	    (name (matching-substring rest 4))
	    (port (matching-substring rest 5))
	    (hmmm (matching-substring rest 6)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%s%s:%s:%s:%s:%s:%s\n" 
				 irchat-info-prefix cn canon pass name port hmmm)))
    (message "IRCHAT: Strange 214 message")))


(defun irchat-handle-215-msg (parsed-sender parsed-msg prefix rest)
  "215 RPL_STATSILINE k2 I * * * 0 0"
  (if (string-match "[^ ]* I \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\)" rest)
      (let ((domain (matching-substring rest 1))
            (passwd (matching-substring rest 2))
            (redomain (matching-substring rest 3)))
        (irchat-w-insert irchat-200-buffer 
			 (format "%sI:%s:%s:%s\n"
				 irchat-info-prefix domain passwd redomain)))
    (message "IRCHAT: Strange 215 message")))


(defun irchat-handle-216-msg (parsed-sender parsed-msg prefix rest)
  "216 RPL_STATSKLINE k2 K *.hut.fi * tsh 0 -1"
  (if (or
       (string-match "[^ ]* K \\([^ ]*\\) \\(.\\) \\([^ ]*\\) 0 -1" rest)
       (string-match "[^ ]* K \\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) 0 [:]*-1" rest))
      (let ((host (matching-substring rest 1))
	    (pass (matching-substring rest 2))
	    (user (matching-substring rest 3)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%sK:%s:%s:%s\n"
				 irchat-info-prefix host pass user)))
    (message "IRCHAT: Strange 216 message")))


(defun irchat-handle-217-msg (parsed-sender parsed-msg prefix rest)
  "217 RPL_STATSQLINE k2 Q <NULL> * eris.berkeley.edu 0 -1"
  (if (string-match "[^ ]* Q \\([^ ]*\\) \\(.\\) \\([^ ]*\\) \\(.*\\)" rest)
      (let ((reason (matching-substring rest 1))
	    (star (matching-substring rest 2))
	    (host (matching-substring rest 3))
	    (stuff (matching-substring rest 4)))
	(irchat-w-insert irchat-200-buffer 
			 (format "%sQ:%s:%s:%s:%s\n"
				 irchat-info-prefix reason star host stuff)))
    (message "IRCHAT: Strange 217 message")))


(defun irchat-handle-218-msg (parsed-sender parsed-msg prefix rest)
  "218 RPL_STATSYLINE k2 Y 1 90 300 10"
  (if (string-match "[^ ]* Y \\([0-9]*\\) \\([0-9]*\\) \\([0-9]*\\) \\([0-9]*\\)" 
		    rest)
      (let ((class (matching-substring rest 1))
	    (pingfreq (matching-substring rest 2))
	    (confreq (matching-substring rest 3))
	    (maxlinks (matching-substring rest 4)))
	(irchat-w-insert irchat-200-buffer
			 (format "%sClass %s: PingFreq %s, ConFreq %s, MaxLinks %s\n"
				 irchat-info-prefix class
				 pingfreq confreq maxlinks)))
    (message "IRCHAT: Strange 218 message")))

    
(defun irchat-handle-219-msg (parsed-sender parsed-msg prefix rest)
  "219 RPL_ENDOFSTATS %c :End of /STATS report" 
  nil)


(defun irchat-handle-221-msg (parsed-sender parsed-msg prefix rest) 
  "221 RPL_UMODEIS %s"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((str (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%sMode for you is %s\n"
				 irchat-info-prefix str)))
    (message (format "IRCHAT: Strange 324 reply '%s'" rest))))


;;;
;;; 230 series not implemented as 7/94
;;;
(defun irchat-handle-231-msg (parsed-sender parsed-msg prefix rest)
  "231 RPL_SERVICEINFO"
  nil)


(defun irchat-handle-232-msg (parsed-sender parsed-msg prefix rest)
  "232 RPL_ENDOFSERVICES"
  nil)


(defun irchat-handle-233-msg (parsed-sender parsed-msg prefix rest)
  "233 RPL_SERVICE"
  nil)


(defun irchat-handle-234-msg (parsed-sender parsed-msg prefix rest)
  "234 RPL_SERVLIST"
  nil)


(defun irchat-handle-235-msg (parsed-sender parsed-msg prefix rest)
  "235 RPL_SERVLISTEND"
  nil)


(defun irchat-handle-241-msg (parsed-sender parsed-msg prefix rest)
  "241 RPL_STATSLLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 241 reply")))


(defun irchat-handle-242-msg (parsed-sender parsed-msg prefix rest)
  "242 RPL_STATSUPTIME :Server Up %d days, %d:%02d:%02d"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 242 reply")))


(defun irchat-handle-243-msg (parsed-sender parsed-msg prefix rest)
  "243 RPL_STATSOLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 243 reply")))


(defun irchat-handle-244-msg (parsed-sender parsed-msg prefix rest)
  "244 RPL_STATSHLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 244 reply")))


(defun irchat-handle-245-msg (parsed-sender parsed-msg prefix rest)
  "245 RPL_STATSSLINE %c %s * %s %d %d"
  (if (string-match "[^ ]* \\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-200-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 245 reply")))

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-300.el,v 3.9 2002/06/09 15:35:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;; 300 replies
;;;

(defvar irchat-recursing-whois nil)
(defvar irchat-recursing-whowas nil)
(defvar irchat-recursing-whois-for-host-mask nil)

(defun irchat-handle-300-msgs (number parsed-sender parsed-msg prefix rest)
  "Generic handler for 3?? messages. 
This is called if no specific handler exists"
  (if (string-match "[^ ]* \\([^ :]*\\) *\\([^ :]*\\) *:\\(.*\\)" rest)
      (let ((target1 (matching-substring rest 1))
	    (target2 (matching-substring rest 2))
	    (msg (matching-substring rest 3)))
	(cond ((string-equal target1 "")
	       (irchat-w-insert irchat-300-buffer 
				(format "%s%s\n" irchat-info-prefix msg)))
	      ((string-equal target2 "")
	       (irchat-w-insert irchat-300-buffer 
				(format "%s%s (%s)\n"
					irchat-info-prefix msg target1)))
	      (t
	       (irchat-w-insert irchat-300-buffer 
				(format "%s%s %s (%s)\n" 
					irchat-info-prefix target1 msg target2))))
	)
    (message "IRCHAT: Strange %s reply" number)))


(defun irchat-handle-301-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 301 RPL_AWAY."
  (if (string-match "^[^ ]+ \\([^ ]+\\) +:\\(.*\\)" rest)
      (let ((who (matching-substring rest 1))
	    (iswhat (matching-substring rest 2)))
	(if (not irchat-recursing-whois)
	    (irchat-w-insert irchat-300-buffer 
	     (format 
	      "%s%s is marked as being AWAY, but left the message:\n%s\n"
	      irchat-info-prefix who iswhat))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 301 reply")))


(defun irchat-handle-302-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 302 USERHOST."
  (while (string-match 
	  "^[^ ]* :[ ]*\\([^*=]+\\)\\([*]*\\)=\\([+-]\\)\\([^ ]+\\)" rest)
    (let ((nick (matching-substring rest 1))
	  (oper (matching-substring rest 2))
	  (away (matching-substring rest 3))
	  (who (matching-substring rest 4)))
      (irchat-w-insert irchat-300-buffer 
		       (format "%sNick %s is %s [%s, %s]\n"
			       irchat-info-prefix
			       nick who 
			       (concat (if (string= oper "") "Not ")
				       "Operator")
			       (concat (if (string= away "+") "Not ")
				       "AWAY")))
      (setq rest (concat " :" (substring rest (match-end 4) nil))))))


(defun irchat-handle-303-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 303 ISON"
  (if (string-match "[^ ]+ :\\(.*\\)" rest)
      (if (string= (matching-substring rest 1) "")
	  (irchat-w-insert irchat-300-buffer 
			   (format "%sNo one you requested is on now.\n"
				   irchat-info-prefix))
	(irchat-w-insert irchat-300-buffer 
			 (format "%sFollowing people(s) are on: %s\n"
				 irchat-info-prefix
				 (matching-substring rest 1))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 303 reply")))


(defun irchat-handle-305-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 305 UNAWAY"
  (if (string-equal irchat-away-indicator "A")
      (progn
	(setq irchat-away-indicator "-")
	(irchat-maybe-poll)
	(if (string-match "[^:]:\\(.*\\)" rest)
	    (let ((msg (matching-substring rest 1)))
	      (irchat-w-insert irchat-300-buffer 
			       (format "%s%s (%s)\n" 
				       irchat-info-prefix msg 
				       (if irchat-format-time-function
					   (apply irchat-format-time-function
						  (list (current-time-string)))
					 (current-time-string)))))
	  (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 305 reply\n")))))


(defun irchat-handle-306-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 306 NOWAWAY"
  (setq irchat-away-indicator "A")
  (if (string-match "[^:]:\\(.*\\)" rest)
      (let ((msg (matching-substring rest 1)))
	(irchat-w-insert irchat-300-buffer 
			 (format "%s%s (%s)\n"
				 irchat-info-prefix msg
				 (if irchat-format-time-function
				     (apply irchat-format-time-function
					    (list (current-time-string)))
				   (current-time-string)))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 306 reply\n")))


(defun irchat-handle-311-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 311 WHOISUSER."
  (if (string-match "[^ ]+ \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" rest)
      (let ((nick (matching-substring rest 1))
	    (username (matching-substring rest 2))
	    (machine (matching-substring rest 3))
	    (channel (matching-substring rest 4))
	    (realname (matching-substring rest 5)))
	(if (not irchat-recursing-whois)
	 (irchat-w-insert irchat-300-buffer 
			  (format "%s%s is %s (%s) at %s\n"
				  irchat-info-prefix
				  nick username realname machine))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 311 reply")))


(defun irchat-handle-312-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 312 WHOISSERVER."
  (if (string-match "^[^ ]+ \\(\\([^ ]+\\) \\)?\\([^ ]+\\) :\\(.*\\)" rest)
      (let ((who (matching-substring rest 2))
            (server (matching-substring rest 3))
            (real (matching-substring rest 4)))
        (if (and (not (irchat-dcc-compare-hostnames server irchat-server))
		 (not irchat-recursing-whois-for-host-mask)
		 (not irchat-recursing-whois)
		 (not irchat-recursing-whowas))
            (progn
	      (if (string-match "\*\..*" server)
		  (setq irchat-recursing-whois-for-host-mask t))
              (setq irchat-recursing-whois t)
              (irchat-send "WHOIS %s %s" server who))
	  (if irchat-recursing-whois-for-host-mask
	      (progn
		(setq irchat-recursing-whois-for-host-mask nil)
		(irchat-send "WHOIS %s %s" server who))
	    (progn
	      (setq irchat-recursing-whois nil)
	      (setq irchat-recursing-whois-for-host-mask nil)
	      (irchat-w-insert irchat-300-buffer 
			       (format "%son via server %s (%s)\n"
				       irchat-info-prefix server real))))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 312 reply")))


(defun irchat-handle-313-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 313 WHOISOPERATOR."
  (if (string-match "^[^ ]+ \\([^ ]+\\) :\\(.*\\)" rest)
      (let ((who (matching-substring rest 1))
	    (iswhat (matching-substring rest 2)))
	(if (not irchat-recursing-whois)
	     (irchat-w-insert irchat-300-buffer 
			      (format "%sSTATUS: %s\n"
				      irchat-info-prefix iswhat))))
    (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 313 reply")))


(defun irchat-handle-316-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 316 WHOISCHANOP."
  (if (string-match "^\\([^ ]+\\) :\\(.*\\)" rest)
      (let ((who (matching-substring rest 1))
	    (iswhat (matching-substring rest 2)))
	(if (not irchat-recursing-whois)
	     (irchat-w-insert irchat-300-buffer 
			      (format "STATUS: %s\n" iswhat))))
    (if (string-match "^\\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" rest)
	(let ((who (matching-substring rest 2))
	      (iswhat (matching-substring rest 3)))
	  (if (not irchat-recursing-whois)
	       (irchat-w-insert irchat-300-buffer 
				(format "%sSTATUS: %s\n"
					irchat-info-prefix iswhat))))
      (irchat-w-insert irchat-300-buffer "IRCHAT: Strange 316 reply"))))


(defun irchat-handle-319-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 319 reply (what channels user is on)."
  (if (string-match "^\\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" rest)
      (let ((who (matching-substring rest 2))
	    (isonchannels (matching-substring rest 3)))
	(if (not irchat-recursing-whois)
	    (irchat-w-insert irchat-300-buffer
			     (format "%schannels: %s\n"
				     irchat-info-prefix isonchannels))))))


(defun irchat-handle-314-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 314 WHOWASUSER."
  (if (string-match "[^ ]+ \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" rest)
      (let ((nick (matching-substring rest 1))
	    (username (matching-substring rest 2))
	    (machine (matching-substring rest 3))
	    (channel (matching-substring rest 4))
	    (realname (matching-substring rest 5)))
	(message "")
	(setq irchat-recursing-whowas t)
	(irchat-w-insert irchat-300-buffer (format "%s%s [%s] was %s (%s) at %s\n"
			irchat-info-prefix
			nick
			(if (string= channel "*") "Priv" channel)
			username
			realname
			machine)))
    (message "IRCHAT: Strange 314 reply"))
  )

(defun irchat-handle-315-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 315 ENDOFWHO."
  nil)


(defun irchat-317-args-parse (str) 
  (let* ((al (irchat-split-protocol-args str)) 
         (nl (if (> (length al) 1) 
                 (split-string (nth (- (length al) 1) al)  
                               ",[ \t]*") 
               '())) 
         (r '())) 
    (if (eq (length nl) (- (length al) 1)) 
        (while nl 
          (setq r (cons (cons (car nl) (car al)) r)) 
          (setq nl (cdr nl)) 
          (setq al (cdr al)))) 
    r)) 
 
(defun irchat-317-args-lookup (name parsed-args)  
  (let ((r nil)) 
    (while parsed-args 
      (if (string-ci-equal name (car (car parsed-args))) 
          (setq r (cdr (car parsed-args)) 
                parsed-args '()) 
        (setq parsed-args (cdr parsed-args)))) 
    r)) 

 
(defun irchat-handle-317-msg (parsed-sender parsed-msg prefix rest) 
  "Handle the 317 WHOISIDLE." 
  (let ((s nil)
	(l nil))
    (if (string-match "^[^ ]+ [^ ]+ \\(.*\\)" rest)   
	(let ((p (irchat-317-args-parse (matching-substring rest 1))))
	  (setq s (irchat-317-args-lookup "seconds idle" p)
		l (irchat-317-args-lookup "signon time" p))))
    (if (and (null s) 
             (string-match "^[^ ]+ \\([0-9]*\\) :\\(.*\\)" rest)) 
        (setq s (matching-substring rest 1))) 
    (if (and (null s) 
             (string-match 
              "^\\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) +\\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" 
              rest)) 
        (let ((dofw (matching-substring rest 2)) 
              (month (matching-substring rest 3))    
              (date (matching-substring rest 4)) 
              (time (matching-substring rest 5)) 
              (year (matching-substring rest 6))) 
          (irchat-w-insert irchat-300-buffer
                           (format "%sLast input received %s.\n"  
                                   irchat-info-prefix  
                                   time)) 
          (setq s 0))) 
    (if (not (null l))
	(let ((c (irchat-time-val-string-to-current-time-format l)))
	  (if (not (null c))
	      (irchat-w-insert irchat-300-buffer
			       (format "%sSign-on time is %s\n"
				       irchat-info-prefix
				       (if irchat-format-time-function
					   (apply irchat-format-time-function
						  (list (current-time-string
							 c)))
					 (current-time-string c)))))))
    (if (not (null s)) 
        (if (stringp s)
            (irchat-w-insert irchat-300-buffer
                             (format "%sIDLE for %s\n"
                                     irchat-info-prefix
                                     (irchat-convert-seconds s))))
      (message "IRCHAT: Strange 317 reply")))) 


(defun irchat-handle-318-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 318 ENDOFWHOIS."
  nil)


(defun irchat-handle-321-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 321 LISTSTART. (first of names)"
  (irchat-w-insert irchat-300-buffer
   (format "%s%-10s%6s        %s\n"
	   irchat-info-prefix "Channel" "Users" "Topic")))


(defun irchat-handle-322-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 322 LIST.(from NAMES)."
  (if (string-match "^\\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) :\\(.*\\)" rest)
      (let ((chnl (matching-substring rest 2))
	    (users (matching-substring rest 3))
	    (topic (matching-substring rest 4)))
	(if (or (string= irchat-channel-filter (downcase chnl))
		(string= irchat-channel-filter "")
		(and (string= irchat-channel-filter "0")
		     (string= chnl "*")))
	    (progn
	      (put (intern chnl irchat-obarray) 'topic topic)
	      (irchat-w-insert (irchat-pick-buffer chnl)
			       (format "%s%-10s%6s user%s%s%s\n"
				       irchat-info-prefix
				       (if (string= chnl "*") "Priv"
					 chnl)
				       users
				       (if (> (string-to-int users) 1) "s" " ")
				       (if (string= "" topic) "" ": ")
				       topic)))))
    (message "IRCHAT: Strange 322 reply")))


(defun irchat-handle-323-msg (parsed-sender parsed-msg prefix rest) 
  "Handle the 323 RPL_LISTEND. (end of names)"
  nil)


(defun irchat-handle-324-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 324 CHANNELMODEIS."
  (if (string-match "[^ ]* \\([^ ]*\\) \\(.*\\)" rest)
      (let ((chnl (matching-substring rest 1))
	    (str (matching-substring rest 2)))
	(irchat-w-insert (irchat-pick-buffer chnl)
			 (format "%sMode for %s is %s\n"
				 irchat-info-prefix chnl str)))
    (message (format "IRCHAT: Strange 324 reply '%s'" rest))))


(defun irchat-handle-331-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 331 NOTOPIC"
  (if (string-match "[^ ]* \\([^ ]*\\) \\(.*\\)" rest)
      (let ((ichan (intern (matching-substring rest 1) irchat-obarray)))
	(put ichan 'topic nil)
 	(irchat-w-insert irchat-300-buffer
			 (format "%sIRCHAT: No topic is set\n"
				 irchat-info-prefix)))))


(defun irchat-handle-332-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 332 TOPIC."
  (if (string-match "[^ ]* \\([^ ]*\\) +:\\(.*\\)" rest)
      (let ((ichan (intern (matching-substring rest 1) irchat-obarray))
 	    (topic (matching-substring rest 2)))
 	(put ichan 'topic topic)
 	(irchat-w-insert irchat-300-buffer
			 (format "%sTopic: %s\n" irchat-info-prefix topic)))
    (message "IRCHAT: Strange 332 message")))


(defun irchat-handle-341-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 341 INVITING."
  (if (string-match "^\\([^ ]+\\) +\\([^ ]+\\) +\\([-#&0-9+][^ ]*\\)" rest)
      (let ((who (matching-substring rest 1))
	    (nick (matching-substring rest 2))
	    (chnl (matching-substring rest 3)))
	(irchat-w-insert (irchat-pick-buffer chnl)
			 (format "%sInviting user %s to channel %s\n"
				 irchat-info-prefix nick chnl)))
    (message "Strange 341 message")))


(defun irchat-handle-351-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 351 VERSION."
  (if (string-match "[^ ]+ \\([^ ]+\\) :*\\([^ ]+\\)[ :]*\\(.*\\)" rest)
      (let ((version (matching-substring rest 1))
	    (machine (matching-substring rest 2))
	    (comments (matching-substring rest 3)))
	(irchat-w-insert irchat-300-buffer
			 (format "%sMachine %s is running IRC version %s (%s)\n"
				 irchat-info-prefix machine version comments)))
    (message "IRCHAT: Strange 351 reply")))


(defun irchat-handle-whoreply-msg (parsed-sender parsed-msg prefix rest)
  (irchat-handle-352-msg parsed-sender parsed-msg prefix rest))


(defun irchat-handle-352-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 352 WHOREPLY."
  (if (string-match "\\([^ ]*\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) \\([^ ]+\\) :[0-9]* ?\\(.*\\)" rest)
      (let* ((chnl (matching-substring rest 1))
	     (user (matching-substring rest 2))
	     (host (matching-substring rest 3))
	     (nick (matching-substring rest 5))
	     (oper (matching-substring rest 6))
	     (name (matching-substring rest 7))
	     (chan-buffer (irchat-pick-buffer chnl)))

	(irchat-w-insert chan-buffer
			 (format "%s%3s %10s %9s %-29s%s\n"
				 irchat-info-prefix
				 oper ;; Kaizzu 06/03/90
				 (if (string= chnl "*") "Priv" ; *WORK* needed
				   (if (string= chnl "0")
				       "Priv"
				     chnl))
				 nick
				 (format "<%s@%s>" user
					 (irchat-clean-hostname host))
				 name)))
    (message "IRCHAT: Strange 352 message")))


(defun irchat-handle-namreply-msg (parsed-sender parsed-msg prefix rest)
  (irchat-handle-353-msg parsed-sender parsed-msg prefix rest))

(defun irchat-count-words-from-string (str)
  "count words from a string"
  (let* ((len (- (length str) 1)) ; string is from [0,len[
	 (i len) (seen nil) (words (if (< len 0) 0 1)))
    (while (not (< i 0))
      (progn
	(if (string= (char-to-string (aref str i)) " ")
	    (if seen
		(setq words (+ words 1)))
	  (setq seen t))
	(setq i (- i 1))))
    words))

(defun irchat-handle-346-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 346 invitation list."
  (if (string-match "[^ ]* \\([^ ]*\\) \\([^ ]*\\)" rest)
      (let ((chnl (matching-substring rest 1))
	    (regexp (matching-substring rest 2)))
	(irchat-w-insert (irchat-pick-buffer chnl)
			 (format "%s%s carries invitation to %s\n"
				 irchat-info-prefix regexp chnl)))
    (message "IRCHAT: Strange 346 message")))

(defun irchat-handle-347-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 347 ???????."
  nil)

(defun irchat-handle-348-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 348 ban exceptions."
  (if (string-match "[^ ]* \\([^ ]*\\) \\([^ ]*\\)" rest)
      (let ((chnl (matching-substring rest 1))
	    (regexp (matching-substring rest 2)))
	(irchat-w-insert (irchat-pick-buffer chnl)
			 (format "%s%s carries exception to ban on %s\n"
				 irchat-info-prefix regexp chnl)))
    (message "IRCHAT: Strange 348 message")))

(defun irchat-handle-349-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 347 ???????."
  nil)

(defvar irchat-353-nameschnl nil "")
(defvar irchat-353-nameslist "" "names list reply string")
(defvar irchat-353-namescount 0 "")

(defun irchat-handle-353-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 353 (NAMREPLY) message.   If we are just polling the server,
don't display anything."
  (if (string-match "[^ =*@]?[=*@] \\([^ ]*\\) :\\(.*\\)" rest)
      (let* ((chnl (matching-substring rest 1))
	     (users (matching-substring rest 2))
	     (numusers (irchat-count-words-from-string users)))
	(if (> irchat-polling 0)
	    nil
         (progn
           (setq irchat-353-nameslist (concat irchat-353-nameslist users)
                 irchat-353-nameschnl chnl
                 irchat-353-namescount (+ irchat-353-namescount numusers))))
	(irchat-scan-channels chnl)
	(irchat-update-thischannel chnl users))
    (message "IRCHAT: Strange 353 message")))

(defun irchat-handle-361-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 361 KILLDONE."
  (if (string-match "[^ ]+ \\([^ ]+\\) +:\\(.*\\)" rest)
      (let ((who (matching-substring rest 1))
	    (message (matching-substring rest 2)))
	(irchat-w-insert irchat-300-buffer 
			 (format "%sYou just KILLED %s. %s\n"
				 irchat-info-prefix who message)))
    (message "IRCHAT: Strange 361 reply")))

(defun irchat-handle-364-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match "^\\([^ ]+\\) +\\([^ ]*\\) +[^ ]* +:\\(.*\\)" rest)
      (progn
	(irchat-w-insert irchat-300-buffer 
			 (format "%s%-30s%s\n"
				 irchat-info-prefix
				 (matching-substring rest 2)
				 (matching-substring rest 3))))
    (message "IRCHAT: Strange 364 message")))


(defun irchat-handle-365-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 365 ENDOFLINKS."
  nil)


(defun irchat-handle-366-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 366 ENDOFNAMES."
  (let ((level (- irchat-polling 1)))
    (setq irchat-polling (if (< level 0) 0 level))
    (irchat-w-insert (irchat-pick-buffer irchat-353-nameschnl)
		     (format "%s%-10s%6d user%s: %s\n"
			     irchat-info-prefix
			     (if (string= irchat-353-nameschnl "*") 
				 "Priv" 
			       irchat-353-nameschnl)
			     irchat-353-namescount
			     (cond 
			      ((= irchat-353-namescount 0) "s")
			      ((= irchat-353-namescount 1) "")
			      (t "s"))
			     irchat-353-nameslist))
    (setq irchat-353-nameslist nil
	  irchat-353-nameschnl nil
	  irchat-353-namescount 0)))

(defun irchat-handle-367-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 367 BAN."
  (if (string-match "[^ ]* \\([^ ]*\\) \\([^ ]*\\)" rest)
      (let ((chnl (matching-substring rest 1))
	    (regexp (matching-substring rest 2)))
	(irchat-w-insert (irchat-pick-buffer chnl)
			 (format "%s%s has been banned on %s\n"
				 irchat-info-prefix regexp chnl)))
    (message "IRCHAT: Strange 367 message")))


(defun irchat-handle-368-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 368 ???????."
  nil)


(defun irchat-handle-369-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 369 WHOWAS."
  (setq irchat-recursing-whowas nil)
  nil)

(defun irchat-handle-371-msg (parsed-sender parsed-msg prefix rest) 
  "Handle the 371 INFO."
  (if (string-match "^\\([^ ]+\\) +:?\\(.*\\)" rest)
      (let ((msg (matching-substring rest 2)))
	(irchat-w-insert irchat-300-buffer
			 (format "%s%s\n" irchat-info-prefix msg)))
    (message "IRCHAT: Strange 371 message")))


(defun irchat-handle-372-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 372 MOTD."
  (string-match "^\\([^ ]+\\) +:?\\(.*\\)" rest)
  (let ((msg (matching-substring rest 2)))
    (irchat-w-insert irchat-300-buffer (format "%s%s\n" irchat-info-prefix msg))
    ))


(defun irchat-handle-381-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 381 YOUREOPER. ."
  (if (string-match "^\\([^ ]+\\) +:\\(.*\\)" rest)
      (let ((message (matching-substring rest 2)))
	(irchat-w-insert irchat-300-buffer
			 (format "%sOPER: %s\n" irchat-info-prefix message)))
    (message "IRCHAT: Strange 381 reply")))


(defun irchat-handle-382-msg (parsed-sender parsed-msg prefix rest) 
  "Handle the 382 REHASHING."
  (string-match "^\\([^ ]+\\) +:\\(.*\\)" rest)
  (let ((name (matching-substring rest 1))
	(msg (matching-substring rest 2)))
    (irchat-w-insert irchat-300-buffer
		     (format "%s%s: %s\n" irchat-info-prefix name msg))))


(defun irchat-handle-391-msg (parsed-sender parsed-msg prefix rest)
  "Handle the 391 TIME."
  (if (string-match "^\\([^ ]+\\) +\\(.*\\)" rest)
      (let ((time (matching-substring rest 2)))
	(irchat-w-insert irchat-300-buffer 
			 (format "%sServer time: %s\n" irchat-info-prefix
				 time)))
    (message "IRCHAT: Strange 391 message")))

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-400.el,v 3.11 2002/06/09 15:16:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  400 replies -- ERRORS
;;; 
(defun irchat-handle-400-msgs (number parsed-sender parsed-msg prefix rest)
  "Generic handler for 4?? messages. This is called if no specific handler exists"
  (if (string-match "[^ ]* \\([^ :]*\\) *\\([^ :]*\\) *:\\(.*\\)" rest)
      (let ((target1 (matching-substring rest 1))
	    (target2 (matching-substring rest 2))
	    (msg (matching-substring rest 3)))
	(cond ((string-equal target1 "")
	       (irchat-w-insert irchat-400-buffer
				(format "%s%s\n" 
					irchat-error-prefix
					msg)))
	      ((string-equal target2 "")
	       (irchat-w-insert irchat-400-buffer
				(format "%s%s (%s)\n" 
					irchat-error-prefix 
					msg 
					target1)))
	      (t
	       (irchat-w-insert irchat-400-buffer
				(format "%s%s %s (%s)\n"
					irchat-error-prefix 
					target1
					msg 
					target2))))
	)
    (message "IRCHAT: Strange %s reply" number)))


(defun irchat-handle-401-msg (parsed-sender parsed-msg prefix rest) ;;; ERR_NOSUCHNICK
  (if (string-match "[^ ]+ \\([^ ]+\\) +:\\(.*\\)" rest)
      (let ((name (matching-substring rest 1))
	    (error (matching-substring rest 2)))
	(irchat-change-nick-of name nil)
	(irchat-send "WHOWAS %s" name))
    (message "IRCHAT: Strange 401 reply")))


(defun irchat-handle-406-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match "[^ ]+ \\([^ ]+\\) +:\\(.*\\)" rest)
      (let ((nickorchan (matching-substring rest 1)))
	(irchat-change-nick-of nickorchan nil)
	(message "IRCHAT: No such user %s" nickorchan))))


(defun irchat-handle-412-msg (parsed-sender parsed-msg prefix rest)
  (message "IRCHAT: No text to send"))


(defun irchat-iterate-nick (nick)
  (let* ((fmt (format "%s_" nick))
	 (new (substring fmt 0 (min 9 (length fmt)))))
    (if (string= nick new)
	(irchat-iterate-nick (format "_%s" nick))
      new)))


(defun irchat-handle-432-msg (parsed-sender parsed-msg prefix rest) ;;; ERR_ERRONEUSNICKNAME
  "Handle the 432 reply (erroneus nickname)"
  (save-excursion
    (set-buffer irchat-Command-buffer)
    (beep)
    (let ((nick (cond ((string-match "^\\([^ ]+\\) +\\([^ ]+\\) +:\\(.*\\)" 
				     rest)
		       (matching-substring rest 2))
		      ((string-match "^ *\\([^ ]+\\) :.*" rest)
		       (matching-substring rest 1))
		      (t
		       "UNKNOWN (Could not figure out, contact developers)"))))
      (if (eq irchat-nick-accepted 'ok)
	  (setq irchat-real-nickname irchat-old-nickname))
      (message "IRCHAT: Erroneous nickname %s.  Choose a new one with %s."
	       nick
	       (substitute-command-keys "\\[irchat-Command-nickname]")))))
	

(defun irchat-handle-433-msg (parsed-sender parsed-msg prefix rest) ;;; ERR_NICKNAMEINUSE
  "Handle the 433 reply (nickname already in use)"
  (if (not (eq irchat-nick-accepted 'ok))
      (progn
	(setq irchat-real-nickname (irchat-iterate-nick irchat-real-nickname))
	(setq irchat-old-nickname irchat-real-nickname)
	(irchat-send "NICK %s" irchat-real-nickname)
	(setq irchat-nick-accepted 'sent))
    (save-excursion
      (set-buffer irchat-Command-buffer)
      (beep)
      (setq irchat-real-nickname irchat-old-nickname)
      (let ((nick (cond ((string-match "^\\([^ ]+\\) +\\([^ ]+\\) +:\\(.*\\)" 
				       rest)
			 (matching-substring rest 2))
			((string-match "^ *\\([^ ]+\\) :.*" rest)
			 (matching-substring rest 1))
			(t
			 "UNKNOWN (Could not figure out, contact developers)"))))
	(message 
	 "IRCHAT: Nickname %s already in use.  Choose a new one with %s."
	 nick
	 (substitute-command-keys "\\[irchat-Command-nickname]"))))))


(defun irchat-handle-437-msg (parsed-sender parsed-msg prefix rest) ;;; ERR_XXX
  "Handle the 437 reply (nick/channel unavailable)"
  (if (not (eq irchat-nick-accepted 'ok))
      (progn
	(setq irchat-real-nickname (irchat-iterate-nick irchat-real-nickname))
	(setq irchat-old-nickname irchat-real-nickname)
	(irchat-send "NICK %s" irchat-real-nickname)
	(setq irchat-nick-accepted 'sent))
    (save-excursion
      (set-buffer irchat-Command-buffer)
      (beep)
      (let ((onick irchat-old-nickname)
	    (wnick "UNKNOWN (Could not figure out, contact developers)"))
	(if (string-match "^\\([^ ]+\\) +\\([^ ]+\\) +:\\(.*\\)" rest)
	    (progn
	      (setq onick (matching-substring rest 1))
	      (setq wnick (matching-substring rest 2)))
	  (if (string-match "^ *\\([^ ]+\\) :.*" rest)
	      (setq onick (matching-substring rest 1))))
	(if (stringp onick)
	    (setq irchat-real-nickname onick))
	(message 
	 "IRCHAT: Nickname/channel %s unavailable." wnick)))))

(defun irchat-handle-482-msg (parsed-sender parsed-msg prefix rest)
  (message "IRCHAT: You are not a channel operator"))

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-500.el,v 3.3 2002/06/09 15:16:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  500 replies -- ERRORS
;;;
(defun irchat-handle-500-msgs (number parsed-sender parsed-msg prefix rest)
  (if (string-match "[^ ]* \\([^ :]*\\) *\\([^ :]*\\) *:\\(.*\\)" rest)
      (let ((target1 (matching-substring rest 1))
	    (target2 (matching-substring rest 2))
	    (msg (matching-substring rest 3)))
	(cond ((string-equal target1 "")
	       (irchat-w-insert 
		irchat-500-buffer
		(format "%s%s\n" irchat-error-prefix msg)))
	      ((string-equal target2 "")
	       (irchat-w-insert 
		irchat-500-buffer 
		(format "%s%s (%s)\n" irchat-error-prefix msg target1)))
	      (t
	       (irchat-w-insert 
		irchat-500-buffer 
		(format "%s%s %s (%s)\n" 
			irchat-error-prefix
			target1 
			msg
			target2))))
	)
    (message "IRCHAT: Strange %s reply" number)))
;;;
;;; eof
;;;


;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-commands.el,v 3.40 2002/06/04 15:12:51 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))
(eval-and-compile  
;  (require 'irchat-crypt)
  (require 'irchat-dcc)
  (require 'irchat-caesar))

(defun irchat-Command-describe-briefly ()
  (message (substitute-command-keys "Type \\[describe-mode] for help")))


(defun irchat-Command-redisplay (&optional center)
  "Un-freezes and re-selects the Dialogue buffer in another window.
   With argument, recenter with that argument."
  (interactive "P")
  (if (irchat-frozen (car irchat-D-buffer))
      (irchat-freeze-toggle (car irchat-D-buffer)))
  (if (irchat-ownfrozen (car irchat-D-buffer))
      (irchat-ownfreeze-toggle (car irchat-D-buffer)))
  (setq irchat-freeze-indicator "-")
  (setq irchat-ownfreeze nil)
  (setq irchat-ownfreeze-indicator "-")
  (if (irchat-get-buffer-window irchat-Dialogue-buffer)
      nil
    (if (one-window-p)
	(irchat-configure-windows)
      (display-buffer irchat-Dialogue-buffer)))
  (set-buffer irchat-Dialogue-buffer)
  (set-window-point (irchat-get-buffer-window irchat-Dialogue-buffer) 
		    (point-max)))


(defun irchat-Command-pollnames ()
  (setq irchat-polling (+ irchat-polling (length irchat-channel-alist)))
  (mapcar (function (lambda (channel)
		      (irchat-send "NAMES %s" (car channel))))
	  irchat-channel-alist))


(defun irchat-Command-find-timestamp ()
  (interactive)
  (save-excursion
    (let ((obuffer (current-buffer))
          (range ""))
      (switch-to-buffer irchat-Dialogue-buffer)
      (if (re-search-backward "^\\*\\*\\* Time: " (point-min) t)
          (let ((start (+ (point) 10)))
            (end-of-line)
            (setq range (format "%s   ---   "
                                (buffer-substring start (point))))))
      (if (re-search-forward "^\\*\\*\\* Time: " (point-max) t)
          (let ((start (point)))
            (end-of-line)
            (setq range (concat range (buffer-substring start (point))))))
      (message range)
      (switch-to-buffer obuffer))))


(defun irchat-Command-keepalive ()
  (if (not (irchat-server-opened))
      (irchat 'always))
  (irchat-ping-if-idle))


(defun irchat-compose-servertimestring (s)
  (let* ((w '(("Mon" "Monday") ("Tue" "Tuesday") ("Wed" "Wednesday") 
	      ("Thu" "Thursday") ("Fri" "Friday") ("Sat" "Saturday") 
	      ("Sun" "Sunday")))
	 (m '(("Jan" "January") ("Feb" "February") ("Mar" "March") 
	      ("Apr" "April") ("May" "May") ("Jun" "June")  ("Jul" "July") 
	      ("Aug" "August") ("Sep" "September") ("Oct" "October") 
	      ("Nov" "November") ("Dec" "December")))
	 (wdat) (mnam)
	 (wabb (substring s 0 3))
	 (mabb (substring s 4 7)))
    (setq wdat (car (cdr (assoc wabb w)))
	  mnam (car (cdr (assoc mabb m))))
    (format "%s %s %s %s %s" 
	    wdat mnam (substring s 8 10) (substring s 20 24)
	    (substring s 11 16))))


(defvar irchat-last-timestamp-time nil "Last time timestamp was inserted")
(defvar irchat-last-timestamp-no-cons-p nil "Last timestamp was no-cons")


(defun irchat-Command-timestamp-if-interval-expired (&optional no-cons)
  (interactive)
  (if (and (not (and no-cons
		     irchat-last-timestamp-no-cons-p))
	   (numberp irchat-timestamp-interval)
	   (> irchat-timestamp-interval 0)
	   (or (null irchat-last-timestamp-time)
	       (> (irchat-time-difference irchat-last-timestamp-time
					  (current-time))
		  irchat-timestamp-interval)))
      (progn
	(if irchat-save-vars-is-dirty
	    (irchat-Command-save-vars))
	(irchat-Command-timestamp)
	(setq irchat-last-timestamp-no-cons-p no-cons))))

(defun irchat-Command-timestamp ()
  (interactive)
  (let ((stamp (format "%s" (format irchat-timestamp-format
				    (if irchat-format-time-function
					(apply irchat-format-time-function
					       (list (current-time-string)))
				      (current-time-string))))))
    (let ((irchat-timestamp-interval 0))
      (irchat-Dialogue-insert stamp)))
  (setq irchat-last-timestamp-time (current-time)))

(defun irchat-Command-send-message (message &optional crypt-type
					              user-defined-key
						      own-message
						      do-not-split)
  "Send MESSAGE to current chat partner of current channel."
  (if (not irchat-crypt-mode-active) (setq crypt-type 'cleartext))
  (if (and (not do-not-split)
	   (not (eq irchat-command-buffer-mode 'chat))
	   (stringp irchat-message-split-separator)
	   (> (length message) irchat-message-length-limit)
	   (> irchat-message-length-limit 
	      (* 2 (length irchat-message-split-separator))))
      (let ((lst (irchat-split-string-with-separator 
		  message 
		  irchat-message-split-separator
		  irchat-message-length-limit))
	    (r nil))
	(while lst
	  (let ((msg (car lst)))
	    (setq lst (cdr lst))
	    (setq r (irchat-Command-send-message msg
						 crypt-type
						 user-defined-key
						 (if lst "" message)
						 t))))
	r)
    (if (> (length message) 0)
	(let* ((addr (if (eq irchat-command-buffer-mode 'chat)
			 irchat-current-chat-partner
		       irchat-current-channel))
	       (msg-encrypted-p nil)
	       (msg
		(cond ((eq irchat-command-buffer-mode 'chat)
		       (progn
			 ;; Chat mode is encrypted later if needed.
			 (setq msg-encrypted-p nil)
			 message))
		      ((equal crypt-type 'cleartext)
		       (progn
			 (setq msg-encrypted-p nil)
			 message))
		      ((and (equal crypt-type 'encrypted)
			    addr)
		       (progn
			 (setq msg-encrypted-p t)
			 (irchat-encrypt-message message addr t)))
		      ((equal crypt-type 'user-defined-key)
		       (progn
			 (setq msg-encrypted-p t)
			 (irchat-encrypt-message message user-defined-key t)))
		      (addr
		       (let ((cipher (irchat-encrypt-message message 
							     addr
							     nil)))
			 (if (not (string= cipher message))
			     (progn
			       (setq msg-encrypted-p t)
			       cipher)
			   (progn
			     (setq msg-encrypted-p nil)
			     message))))
		      (t (progn
			   (setq msg-encrypted-p nil)
			   message)))))
	  (if (eq irchat-command-buffer-mode 'chat)
	      (if irchat-current-chat-partner
		  (irchat-Command-message irchat-current-chat-partner 
					  msg
					  crypt-type
					  own-message
					  do-not-split)
		(message (substitute-command-keys 
			  "Type \\[irchat-Command-join] to start private conversation"))
		nil)
	    (if (not irchat-current-channel)
		(progn
		  (beep t)
		  (message 
		   (substitute-command-keys 
		    "Type \\[irchat-Command-join] to join a channel"))
		  nil)
	      (progn
		(irchat-send-privmsg "PRIVMSG %s :%s"
				     irchat-current-channel
				     msg)
		(cond ((null own-message)
		       (irchat-own-message
			(format (format (format "%s %%%%s" 
					    (irchat-format-string 
					     t
					     msg-encrypted-p))
					irchat-real-nickname) message)))
		      ((> (length own-message) 0)
		       (irchat-own-message
			(format (format (format "%s %%%%s" 
					    (irchat-format-string 
					     t
					     msg-encrypted-p))
					irchat-real-nickname) own-message)))
		      (t '())))))
	  t)
      (progn
	(message "IRCHAT: NO text to send")
	nil))))


(defun irchat-enter-message (crypt-type &optional key)
  "Enter the current line as an entry in the IRC dialogue on the
current channel."
  (interactive)
  (let (message start)
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq message (buffer-substring start (point)))
    (if (and irchat-confirm-bell-on-channel-message
	     (stringp irchat-current-channel)
	     (string-match "^#" irchat-current-channel)
	     (string-match "" message))
	(if (yes-or-no-p 
	     (format "Do you really want to send bell to the channel %s? "
		     irchat-current-channel))
	    (if (irchat-Command-send-message message crypt-type key)
		(irchat-next-line 1))
	  (message "Message not sent!"))
      (if (irchat-Command-send-message message crypt-type key)
	  (irchat-next-line 1)))))

(defun irchat-Command-enter-message ()
  (interactive)
  (irchat-enter-message nil))


(defun irchat-Command-enter-message-encrypted ()
  (interactive)
  (let ((irchat-crypt-mode-active t))
    (irchat-enter-message 'encrypted)))


(defun irchat-Command-enter-message-cleartext ()
  (interactive)
  (irchat-enter-message 'cleartext))


(defun irchat-Command-enter-message-opposite-crypt-mode ()
  (interactive)
  (let ((irchat-crypt-mode-active (not irchat-crypt-mode-active)))
    (irchat-enter-message nil)))

(defun irchat-Command-enter-message-with-key (&optional encryption-key)
  (interactive (let ((encryption-key nil)
		     (completion-ignore-case t))
		 (setq encryption-key
		       (completing-read "Encrypt message with key [RET for none]: "
					(cons (cons "" nil)
					      irchat-default-idea-key-list)))
		 (list encryption-key)))
  (if (> (length encryption-key) 0)
      (irchat-enter-message 'user-defined-key encryption-key)
    (irchat-enter-message 'cleartext)))

(defun irchat-Dialogue-enter-message ()
  "Ask for a line as an entry in the IRC dialogue on the current channel."
  (interactive)
  (let ((message "x"))
    (while (not (string= message ""))
      (setq message (read-string "> "))
      (if (not (string= message ""))
	  (irchat-Command-send-message message)))))


(defun irchat-Command-debug ()
  "Start debugging irchat."
  (interactive)
  (let ((buf (current-buffer))
	win)
    (if irchat-debug-buffer
	(progn
	  (if (eq buf irchat-debug-buffer)
	      (setq buf nil))
	  (delete-windows-on irchat-debug-buffer)
	  (setq irchat-debug-buffer nil)
	  (and buf (select-window (irchat-get-buffer-window buf)))
	  )
      (if irchat-use-full-window
	  (progn
	    (delete-other-windows)
	    (irchat-configure-windows)))
      (and (setq win (irchat-get-buffer-window irchat-Dialogue-buffer))
	   (select-window win))
      (and (< (window-height) (* window-min-height 2))
	   (enlarge-window (- (* window-min-height 2) (window-height))))
      (split-window)
      (setq irchat-debug-buffer (irchat-get-buffer-create "*IRC Debugging*"))
      (switch-to-buffer irchat-debug-buffer)
      (goto-char (point-max))
      (recenter)
      (and (setq win (irchat-get-buffer-window buf))
	   (select-window win)))))


(defun irchat-Command-inline ()
  "Send current line as a message to the IRC server."
  (interactive)
  (let (message start stop)
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq stop (point))
    (setq message (buffer-substring start stop))
    (newline)
    (irchat-send message)))


(defun irchat-Command-join (join-channel-var)
  "Join a channel or private conversation.
If user nicname is given, join the same set of channels as 
the specified user. 
If Command-buffer is in chat-mode, start private conversation 
with specified user."
  (interactive (let (join-channel-var (completion-ignore-case t))
		 (setq join-channel-var 
		       (if (eq irchat-command-buffer-mode 'chat)
			   (irchat-completing-default-read
			    "Start private conversation with: "
			    irchat-nick-alist
			    '(lambda (s) t) nil irchat-privmsg-partner)
			 (irchat-completing-default-read
			  "Join channel: "
			  (append irchat-channel-alist
				  irchat-nick-alist)
			  '(lambda (s) t) nil irchat-invited-channel)))
		 (list join-channel-var)))

  (if (eq irchat-command-buffer-mode 'chat)
      (progn
	(setq irchat-current-chat-partners (append 
					    (list join-channel-var)
					    (string-list-ci-delete
					     join-channel-var
					     irchat-current-chat-partners))
	      irchat-current-chat-partner join-channel-var
	      irchat-chat-partner-alist (list-to-assoclist 
					 irchat-current-chat-partners)
	      irchat-channel-indicator (if irchat-current-chat-partner
					   (format "Chatting with %s" 
						   irchat-current-chat-partner)
					 "No partner"))
		(irchat-set-crypt-indicator))
    (progn
      (let ((nicks irchat-nick-alist)
	    (found nil))
	(while (and (not found)	nicks)
	  (if (and (car (car nicks))
		   (string-ci-equal join-channel-var (car (car nicks))))
	      (setq join-channel-var
		    (or (car (get (intern (car (car nicks)) irchat-obarray) 
 				  'chnl))
			join-channel-var)
		    found t))
	  (setq nicks (cdr nicks))))
      (let ((found nil))
	(mapcar 
	 '(lambda (elem)
	    (if (string-ci-equal join-channel-var elem)
		(setq irchat-current-channel elem
		      found t)))
	 irchat-current-channels)
	(setq irchat-channel-indicator
	      (if irchat-current-channel
		  (format "Channel %s" irchat-current-channel)
		"No channel"))
	(irchat-set-crypt-indicator)
	(setq irchat-invited-channel nil)
	(or found (irchat-send "JOIN %s" join-channel-var))))))


(defun irchat-Command-part (part-channel-var)
  "Part a channel or private conversation."
  (interactive (let (part-channel-var (completion-ignore-case t))
		 (setq part-channel-var 
		       (if (eq irchat-command-buffer-mode 'chat)
			   (irchat-completing-default-read	
			    "End private conversation with: "
			    irchat-chat-partner-alist
			    '(lambda (s) t) nil 
			    irchat-current-chat-partner)
			 (irchat-completing-default-read	
			  "Part channel: "
			  (list-to-assoclist irchat-current-channels)
			  '(lambda (s) t) nil 
			  irchat-current-channel)))
		 (list part-channel-var)))
  (if (eq irchat-command-buffer-mode 'chat)
      (progn
	(setq irchat-current-chat-partners (string-list-ci-delete
					    part-channel-var
					    irchat-current-chat-partners)
	      irchat-current-chat-partner (car irchat-current-chat-partners)
	      irchat-chat-partner-alist (list-to-assoclist 
					 irchat-current-chat-partners)
	      irchat-channel-indicator (if irchat-current-chat-partner
					   (format "Chatting with %s" 
						   irchat-current-chat-partner)
					 "No partner"))
	(irchat-set-crypt-indicator))
    (progn
      (if (string-list-ci-memberp part-channel-var irchat-current-channels)
	  (setq irchat-current-channel part-channel-var)) ; just refocusing
      (irchat-send "PART %s" part-channel-var))))

(defun irchat-Command-ignore (kill-nickname-var &optional timeout silent)
  "Ignore messages from this user. Username can be given as case insensitive
regular expression of form \".*@.*\.sub.domain\". 
If already ignoring him/her, toggle.
If variable irchat-variables-file is defined and the file is writable, its
contents are updated future sessions."
  (interactive (let ((kill-nickname-var nil)
		     (timeout nil)
		     (completion-ignore-case t))
		 (setq kill-nickname-var 
		       (completing-read "Ignore nickname or regexp: " 
					(append irchat-nick-alist
						irchat-ignore-nickname)
					'(lambda (s) t) nil nil))
		 (if (and (not (string= "" kill-nickname-var))
			  (not (assoc-ci-string kill-nickname-var irchat-ignore-nickname)))
		     (setq timeout
			   (string-to-int
			    (read-from-minibuffer "Timeout [RET for none]: "))))
		 (list kill-nickname-var timeout)))
  
  ;; empty, just list them
  (if (string= "" kill-nickname-var)
      (let ((buf (current-buffer))
	    (buffer-read-only))
	(set-buffer irchat-Dialogue-buffer)
	(goto-char (point-max))
	(irchat-w-insert irchat-D-buffer
			 (format "%sCurrently ignoring:" irchat-info-prefix))
	(let ((mylist irchat-ignore-nickname)
	      (time (current-time)))
	  (while mylist
	    (let* ((expiretime (if (cdr (car mylist))
				   (/ (irchat-time-difference time (cdr (car mylist))) 60)
				 nil))
		   (expire (cond ((not expiretime) "")
				 ((>= expiretime 0)
				  (format " (%d min)" expiretime))
				 ((< expiretime 0)
				  (format " expired")))))
	      (irchat-w-insert irchat-D-buffer
			       (format " %s%s" (car (car mylist)) expire)))
	    (setq mylist (cdr mylist))))
	(irchat-w-insert irchat-D-buffer "\n")
	(set-buffer buf))
    ;; else not empty, check if exists
    (let ((elem (assoc-ci-string kill-nickname-var irchat-ignore-nickname)))
      (if elem
	  (progn
	    (setq irchat-ignore-nickname (remassoc (car elem)
						 irchat-ignore-nickname))
	    (irchat-w-insert irchat-D-buffer
			     (format "%sNo longer ignoring: %s.\n"
				     irchat-info-prefix (car elem))))
	;; did not find, add to ignored ones
	(let ((expire-time (if (> timeout 0)
			       (irchat-time-add (current-time)
						(* timeout 60)))))
	  (if (and silent (> timeout 0))
	      (setcar (cdr (cdr expire-time)) -1))
	  (setq irchat-ignore-nickname
		(cons (cons kill-nickname-var expire-time)
		      irchat-ignore-nickname))
	  (if (not silent)
	      (progn
		(irchat-w-insert irchat-D-buffer
				 (format "%sIgnoring %s"
					 irchat-info-prefix kill-nickname-var))
		(irchat-w-insert irchat-D-buffer
				 (if (> timeout 0)
				     (format " for %d minutes.\n" timeout)
				   (format ".\n"))))))))
    (setq irchat-save-vars-is-dirty t)))

(defun irchat-Command-ignore-by-regexp (kill-regexp-var 
					&optional timeout 
					          silent)
  "Ignore messages matcing regexp.  If already ignoring regexp, toggle."
  (interactive (let ((kill-regexp-var nil)
		     (timeout nil)
		     (completion-ignore-case t))
		 (setq kill-regexp-var 
		       (completing-read "Ignore messages matching: " 
					irchat-ignore-message-regexp
					'(lambda (s) t) nil nil))
		 (if (and (not (string= "" kill-regexp-var))
			  (not (assoc-ci-string kill-regexp-var
						irchat-ignore-message-regexp)))
		     (setq timeout
			   (string-to-int
			    (read-from-minibuffer
			     "Timeout [RET for none]: "))))
		 (list kill-regexp-var timeout)))
  
  ;; empty, just list them
  (if (string= "" kill-regexp-var)
      (let ((buf (current-buffer))
	    (buffer-read-only))
	(set-buffer irchat-Dialogue-buffer)
	(goto-char (point-max))
	(irchat-w-insert irchat-D-buffer
			 (format "%sCurrently ignoring messages matching:"
				 irchat-info-prefix))
	(let ((mylist irchat-ignore-message-regexp)
	      (time (current-time)))
	  (while mylist
	    (let* ((expiretime (if (cdr (car mylist))
				   (/ (irchat-time-difference time 
							      (cdr 
							       (car mylist)))
				      60)
				 nil))
		   (expire (cond ((not expiretime) "")
				 ((>= expiretime 0)
				  (format " (%d min)" expiretime))
				 ((< expiretime 0)
				  (format " expired")))))
	      (irchat-w-insert irchat-D-buffer
			       (format " \"%s\"%s" (car (car mylist)) expire)))
	    (setq mylist (cdr mylist))))
	(irchat-w-insert irchat-D-buffer "\n")
	(set-buffer buf))
    ;; else not empty, check if exists
    (let ((elem (assoc-ci-string kill-regexp-var irchat-ignore-message-regexp)))
      (if elem
	  (progn
	    (setq irchat-ignore-message-regexp 
		  (remassoc (car elem)
			    irchat-ignore-message-regexp))
	    (irchat-w-insert irchat-D-buffer
			     (format "%sNo longer ignoring: \"%s\".\n"
				     irchat-info-prefix (car elem))))
	;; did not find, add to ignored ones
	(let ((expire-time (if (> timeout 0)
			       (irchat-time-add (current-time)
						(* timeout 60)))))
	  (if (and silent (> timeout 0))
	      (setcar (cdr (cdr expire-time)) -1))
	  (setq irchat-ignore-message-regexp
		(cons (cons kill-regexp-var expire-time)
		      irchat-ignore-message-regexp))
	  (if (not silent)
	      (progn
		(irchat-w-insert irchat-D-buffer
				 (format "%sIgnoring messages matching \"%s\""
					 irchat-info-prefix kill-regexp-var))
		(irchat-w-insert irchat-D-buffer
				 (if (> timeout 0)
				     (format " for %d minutes.\n" timeout)
				   (format ".\n"))))))))
    (setq irchat-save-vars-is-dirty t)))

(defun irchat-Command-send-action (&optional private)
  "Send action ctcp - if on empty line, ask for the message"
  (interactive
   (if current-prefix-arg
       (list current-prefix-arg)
     nil))
  (let ((completion-ignore-case t) message start stop)
    (if private
	(setq irchat-privmsg-partner
	      (irchat-completing-default-read 
	       "To whom: "
	       (append irchat-nick-alist irchat-channel-alist)
	       '(lambda (s) t) 
	       nil irchat-privmsg-partner)))
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq stop (point))
    (if (eq start stop)
	(setq message (read-string "Action: "))
      (setq message (buffer-substring start stop))
      (irchat-next-line 1))
    (irchat-send-privmsg "PRIVMSG %s :ACTION %s"
			 (if private
			     irchat-privmsg-partner
			   irchat-current-channel)
			 message)
    (if private
	(irchat-own-private-message (format "%sAction to %s: %s %s"
					    irchat-info-prefix
					    irchat-privmsg-partner
					    irchat-real-nickname
					    message))
      (irchat-own-message (format "%sAction: %s %s"
				  irchat-info-prefix irchat-real-nickname
				  message)))))


(defun irchat-Command-kick (kick-nickname-var &optional why-var)
  "Kick this user out."
  (interactive (let (kick-nickname-var 
		     why-var
		     (completion-ignore-case t))
		 (setq kick-nickname-var 
		       (completing-read 
			"Kick out nickname: " 
			irchat-nick-alist
			'(lambda (s) t) nil nil))
		 (setq why-var (if irchat-ask-kick-message
				   (read-from-minibuffer 
				    "Kick message: "
				    (if (stringp irchat-default-kick-message)
					irchat-default-kick-message
				      nil))
				 (if (stringp irchat-default-kick-message)
				     irchat-default-kick-message
				   nil)))
		 (list kick-nickname-var (if (and why-var
						  (not (string= ""
								why-var)))
					     why-var
					   nil))))
  (irchat-send "KICK %s %s%s" 
	       irchat-current-channel 
	       kick-nickname-var
	       (if why-var (concat " :" why-var) "")))


(defun irchat-Command-list (&optional channel)
  "List the given channel and its topics.
If you enter only Control-U as argument, list the current channel.
With - as argument, list all channels."
  (interactive
   (if current-prefix-arg
       (if (eq current-prefix-arg '-)
	   (list current-prefix-arg)
	 nil)
     (list
      (let ((completion-ignore-case t))
	(completing-read 
	 "LIST channel: " 
	 irchat-channel-alist
	 '(lambda (s) t) nil nil)))))
  (if (not channel)
      (irchat-send "LIST %s" irchat-current-channel)
    (if (eq channel '-)
	(irchat-send "LIST")
      (irchat-send "LIST %s" channel))))


(defun irchat-Command-lusers ()
  "List the number of users and servers"
  (interactive)
  (irchat-send "LUSERS"))


(defun irchat-Command-modec (change)
  "Send a MODE command"
  (interactive "sMode for this channel: ")
  (irchat-send "MODE %s %s" irchat-current-channel change))


(defun irchat-Command-message-delayed (&rest args)
  "Send an always delayed private message to another user."
  (interactive (let (message-nick-var 
		     crypt-type-var 
		     (completion-ignore-case t))
		 (setq message-nick-var 
		       (irchat-completing-default-read 
			"Private message to: "
			(append irchat-nick-alist irchat-channel-alist)
			'(lambda (s) t) nil irchat-privmsg-partner))
		 (setq crypt-type-var nil)
		 (list message-nick-var 
		       (read-string 
			(format "Private message to %s: " message-nick-var))
		       crypt-type-var
		       nil)))
  (let ((irchat-use-delayed-privmsg t))
    (eval (cons (function irchat-Command-message) args))))


(defun irchat-Command-message (message-nick-var
			       message
			       &optional crypt-type-var
			                 own-message-var
					 do-not-split)
  "Send a private message to another user.  If you send a message that
is already encrypted use 'cleartext flag and put message as a cleartext
into own-message-var"
  (interactive (let (message-nick-var 
		     crypt-type-var 
		     (completion-ignore-case t))
		 (setq message-nick-var 
		       (irchat-completing-default-read 
			"Private message to: "
			(append irchat-nick-alist irchat-channel-alist)
			'(lambda (s) t) nil irchat-privmsg-partner))
		 (setq crypt-type-var nil)
		 (list message-nick-var 
		       (read-string 
			(format "Private message to %s: " message-nick-var))
		       crypt-type-var
		       nil)))
  (if (and (not do-not-split)
	   (stringp irchat-message-split-separator)
	   (> (length message) irchat-message-length-limit)
	   (> irchat-message-length-limit
	      (* 2 (length irchat-message-split-separator))))
      (let ((lst (irchat-split-string-with-separator 
		  message 
		  irchat-message-split-separator
		  irchat-message-length-limit))
	    (r nil))
	(while lst
	  (let ((msg (car lst)))
	    (setq lst (cdr lst))
	    (setq r (irchat-Command-message message-nick-var
					    msg
					    crypt-type-var
					    (if lst 
						""
					      (if own-message-var
						  own-message-var
						message))
					    t))))
	r)
    (let* ((msg-encrypted-p nil)
	   (msg (cond ((equal crypt-type-var 'cleartext)
		       message)
		      ((and (equal crypt-type-var 'encrypted)
			    message-nick-var)
		       (progn
			 (setq msg-encrypted-p t)
			 (irchat-encrypt-message message message-nick-var t)))
		      (message-nick-var
		       (let ((cipher (irchat-encrypt-message message 
							     message-nick-var
							     nil)))
			 (if (not (string= cipher message))
			     (progn
			       (setq msg-encrypted-p t)
			       cipher)
			   (progn
			     (setq msg-encrypted-p nil)
			     message))))
		      (t 
		       (progn
			 (setq msg-encrypted-p nil)
			 message)))))
      (setq irchat-privmsg-partner message-nick-var)
      (irchat-send-privmsg "PRIVMSG %s :%s" message-nick-var msg)
      (if (or (null own-message-var)
	      (> (length own-message-var) 0))
	  (irchat-own-private-message 
	   (format (format "%s %%s" 
			   (irchat-format-string 
			    nil
			    (or msg-encrypted-p
				(and (not (null own-message-var))
				     (string-match "^|\\*E\\*|" msg)))))
		   message-nick-var (if own-message-var
					own-message-var
				      message)))))))


;; Added at mta@tut.fi's request...
;; Does not support encryption (yet!?)

(defun irchat-Command-mta-private ()
  "Send a private message (current line) to another user."
  (interactive)
  (let ((completion-ignore-case t) message start stop)
    (setq irchat-privmsg-partner
	  (irchat-completing-default-read 
	   "To whom: "
	   (append irchat-nick-alist irchat-channel-alist)
	   '(lambda (s) t) 
	   nil irchat-privmsg-partner))
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq stop (point))
    (setq message (buffer-substring start stop))
    (irchat-next-line 1)
    (if (> (length message) 0)
	(irchat-Command-message irchat-privmsg-partner message)
      (message "IRCHAT: No text to send"))))


(defun irchat-Command-names (&optional channel)
  "List the nicknames of the current IRC users on given channel.
With an Control-U as argument, only the current channel is listed.
With - as argument, list all channels."
  (interactive
   (if current-prefix-arg
       (if (eq current-prefix-arg '-)
	   (list current-prefix-arg)
	 nil)
     (list
      (let ((completion-ignore-case t))
	(completing-read 
	 "Names on channel: " 
	 irchat-channel-alist
	 '(lambda (s) t) nil nil)))))
  (if (not channel)
      (irchat-send "NAMES %s" irchat-current-channel)
    (if (eq channel '-)
	(irchat-send "NAMES")
      (irchat-send "NAMES %s" channel))))


(defun irchat-Command-nickname (nick)
  "Set your nickname."
  (interactive "sEnter your nickname: ")
  (let ((nickname (irchat-read-nickname nick)))
    (if (not (= (length nickname) 0))
	(progn 
	  (setq irchat-old-nickname irchat-real-nickname)
	  (setq irchat-real-nickname nickname)
	  (irchat-send "NICK %s" nick))
      (message "IRCHAT: illegal nickname \"%s\"; not changed" nickname))))


(defun irchat-Command-who (&optional expr)
  "Lists tue users that match the given expression.
If you enter only Control-U as argument, list the current channel.
With - as argument, list all users."
  (interactive 
   (if current-prefix-arg
       (if (eq current-prefix-arg '-)
	   (list current-prefix-arg)
	 nil)
     (list
      (let ((completion-ignore-case t))
	(completing-read 
	 "WHO expression: " 
	 irchat-channel-alist
	 '(lambda (s) t) nil nil)))))
  (if (not expr)
      (irchat-send "WHO %s" irchat-current-channel)
    (if (eq expr '-)
	(irchat-send "WHO")
      (irchat-send "WHO %s" expr))))


(defun irchat-Command-wait (nick &optional greeting)
  "Wait for NICK to enter IRC.  When this person appears, you will
be informed. If the optional argument GREETING is non-nil, it should 
be a string to send NICK upon entering."
  (interactive 
   (progn (setq nick (read-string "Wait for: ")
		greeting (read-string 
			  (format "Message to send %s upon entering: " nick)))
	  (if (string= greeting "")
	      (setq greeting nil))
	  (list nick greeting)))
  (put (intern nick irchat-obarray) 'irchat-waited-for t)
  (if greeting 
      (put (intern nick irchat-obarray) 'irchat-greeting greeting)))


(defun irchat-Command-finger (finger-nick-var)
  "Get information about a specific user."
  (interactive (let (finger-nick-var (completion-ignore-case t))
		 (setq finger-nick-var 
		       (completing-read 
			"Finger whom: " irchat-nick-alist
			'(lambda (s) t) nil nil))
		 (list finger-nick-var)))
  (irchat-send "WHOIS %s" finger-nick-var))


(defun irchat-Command-topic (topic)
  "Change topic of channel."
  (interactive "sTopic: ")
  (irchat-send "TOPIC %s :%s" irchat-current-channel topic))


(defun irchat-Command-invite (&optional invite-channel-var invite-nick-var)
  "Invite user to channel."
  (interactive 
   (list
    (if current-prefix-arg
	(let ((completion-ignore-case t))
	  (completing-read 
	   "Invite channel: "
	   (mapcar '(lambda (x)
		      (list x))
		   irchat-current-channels)
	   '(lambda (s) t) nil nil)
	  nil))
    (let ((completion-ignore-case t)) 
      (completing-read "Invite whom: " 
		       irchat-nick-alist
		       '(lambda (s) t) nil nil))))
  (if (not invite-channel-var)
      (setq invite-channel-var irchat-current-channel))
  (irchat-send "INVITE %s %s" invite-nick-var invite-channel-var))


(defun irchat-Command-away (awaymsg)
  "Mark/unmark yourself as being away."
  (interactive "sAway message: ")
  (progn
    (irchat-send "AWAY :%s" awaymsg)
    (setq irchat-awaymsg awaymsg)))


(defun irchat-Command-scroll-down ()
  "Scroll Dialogue-buffer down from Command-buffer."
  (interactive)
  (let ((obuffer (current-buffer)) (owindow (selected-window)))
    (select-window (irchat-get-buffer-window irchat-Dialogue-buffer))
    (pop-to-buffer irchat-Dialogue-buffer)
    (if (pos-visible-in-window-p (point-min))
	(message "Beginning of buffer")
      (scroll-down))
    (select-window (irchat-get-buffer-window obuffer))
    (pop-to-buffer obuffer)))


(defun irchat-Command-scroll-up ()
  "Scroll Dialogue-buffer up from Command-buffer."
  (interactive)
  (let ((obuffer (current-buffer)))
    (select-window (irchat-get-buffer-window irchat-Dialogue-buffer))
    (pop-to-buffer irchat-Dialogue-buffer)
    (if (pos-visible-in-window-p (point-max))
	(progn
	  (goto-char (point-max))
	  (recenter 1))
      (scroll-up))
    (select-window (irchat-get-buffer-window obuffer))
    (pop-to-buffer obuffer)))


(defun irchat-Command-toggle-crypt ()
  (interactive)
  (if (irchat-crypt-support-p)
      (progn
	(if irchat-crypt-mode-active
	    (setq irchat-crypt-mode-active nil)
	  (setq irchat-crypt-mode-active t))
	(irchat-set-crypt-indicator)
	(switch-to-buffer (current-buffer)))
    (error "Crypto not supported in this version of Irchat.")))


(defun irchat-Command-freeze ()
  "Toggle the automatic scrolling of the Dialogue window."
  (interactive)
  (if (irchat-frozen (car irchat-D-buffer))
      (setq irchat-freeze-indicator "-")
    (setq irchat-freeze-indicator "F"))
  (switch-to-buffer (current-buffer))
  (irchat-freeze-toggle (car irchat-D-buffer))
  )


(defun irchat-Command-ownfreeze ()
  "Toggle the automatic scrolling of the Dialogue window when user sends messages."
  (interactive)
  (if (irchat-ownfrozen (car irchat-D-buffer))
      (setq irchat-ownfreeze-indicator "-")
    (setq irchat-ownfreeze-indicator "M"))
  (switch-to-buffer (current-buffer))
  (setq irchat-ownfreeze (not irchat-ownfreeze))
  (irchat-ownfreeze-toggle (car irchat-D-buffer))
  )


(defun irchat-Command-quit (&optional quit-msg)
  "Quit IRCHAT."
  (interactive "P")
  (if (or (not (irchat-server-opened))
	  quit-msg
	  (y-or-n-p "Quit IRCHAT? "))
      (progn
	(message "")
	(irchat-send-delayed-reset)
	(if (get-buffer-process irchat-server-buffer)
	    (if (and (irchat-server-opened)
		     quit-msg)
		(let ((quit-string (read-string "Signoff message: ")))
		  (irchat-send "QUIT :%s" quit-string))
	      (irchat-send "QUIT :%s" (or irchat-signoff-msg ""))))
	(irchat-clear-system)
	(setq irchat-timers
	      (mapcar (function (lambda (timer)
				  (if (nth 0 timer)
				      (irchat-cancel-timer (nth 0 timer)))
				  (list nil (nth 1 timer) (nth 2 timer))))
		      irchat-timers))
	(if irchat-use-full-window
	    (delete-other-windows))
	(irchat-close-server)
	(if irchat-save-vars-is-dirty
	    (irchat-Command-save-vars))
	(run-hooks 'irchat-Exit-hook)
	(setq irchat-polling 0
	      irchat-current-channel nil
	      irchat-obarray nil
	      irchat-current-channels nil))))


(defun irchat-Command-generic (message)
  "Enter a generic IRC message, which is sent to the server.
 A ? lists the useful generic messages."
  (interactive "sIRC Command: ")
  (if (string= message "?")
      (with-output-to-temp-buffer "*IRC Help*"
	(princ "The following generic IRC messages may be of interest to you:
TOPIC <new topic>		set the topic of your channel
INVITE <nickname>		invite another user to join your channel
LINKS				lists the currently reachable IRC servers
SUMMON <user@host>		invites an user not currently in IRC
USERS <host>			lists the users on a host
AWAY <reason>			marks you as not really actively using IRC
				(an empty reason clears it)
WALL <message>			send to everyone on IRC
NAMES <channel>			lists users per channel
")
	(message 
	 (substitute-command-keys 
	  "Type \\[irchat-Command-redisplay] to continue")))
    (irchat-send "%s" message)))


(defun irchat-Command-irc-compatible ()
  "If entered at column 0, allows you to enter a generic IRC message to
be sent to the server.  For a list of messages, see irchat-Command-generic."
  (interactive)
  (if (eq (current-column) 0)
      (call-interactively (function irchat-Command-generic))
    (self-insert-command 1)))


(defun irchat-Command-exec (command)
  "Execute command, stdout to dialogue."
  (interactive "sShell Command: ")
  (shell-command command t)
  (let ((opoint (point)))
    (while (< (point) (mark))
      (progn
	(irchat-Command-enter-message)
	(set-buffer irchat-Command-buffer)))
    (push-mark opoint t)))

;;;
;;; client-to-client queries
;;;
(defun irchat-Command-client-version (client-version-nick-var)
  "Ask about someones client version."
  (interactive (let (client-version-nick-var (completion-ignore-case t))
		 (setq client-version-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-version-nick-var)))
  (setq irchat-query-client-nick client-version-nick-var)
  (irchat-send-privmsg "PRIVMSG %s :%s" 
		       client-version-nick-var
		       irchat-query-client-version))


(defun irchat-Command-client-userinfo (client-userinfo-nick-var)
  "Ask about someones client userinfo."
  (interactive (let (client-userinfo-nick-var (completion-ignore-case t))
		 (setq client-userinfo-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-userinfo-nick-var)))
  (setq irchat-query-client-nick client-userinfo-nick-var)
  (irchat-send-privmsg "PRIVMSG %s :%s"
		       client-userinfo-nick-var
		       irchat-query-client-userinfo))


(defun irchat-Command-client-help (client-help-nick-var)
  "Ask about someones client help."
  (interactive (let (client-help-nick-var (completion-ignore-case t))
		 (setq client-help-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-help-nick-var)))
  (setq irchat-query-client-nick client-help-nick-var)
  (irchat-send-privmsg "PRIVMSG %s :%s" 
		       client-help-nick-var
		       irchat-query-client-help))


(defun irchat-Command-client-clientinfo (client-clientinfo-nick-var)
  "Ask about someones client clientinfo."
  (interactive (let (client-clientinfo-nick-var (completion-ignore-case t))
		 (setq client-clientinfo-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-clientinfo-nick-var)))
  (setq irchat-query-client-nick client-clientinfo-nick-var)
  (irchat-send-privmsg "PRIVMSG %s :%s" 
		       client-clientinfo-nick-var
		       irchat-query-client-clientinfo))


(defun irchat-Command-client-x-face (client-x-face-nick-var)
  "Ask about someones client x-face."
  (interactive (let (client-x-face-nick-var (completion-ignore-case t))
		 (setq client-x-face-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-x-face-nick-var)))
  (setq irchat-query-client-nick client-x-face-nick-var)
  (irchat-send-privmsg "PRIVMSG %s :%s"
		       client-x-face-nick-var
		       irchat-query-client-x-face))


(defun irchat-Command-client-generic (client-generic-nick-var
				      irchat-query-client-command)
  "Ask about someones client clientinfo."
  (interactive (let ((completion-ignore-case t))
		 (list (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick)
		       (irchat-completing-default-read 
			"What info: " irchat-query-client-alist 
			'(lambda (s) t) nil irchat-query-client-lastcommand)
		       )))
  (setq irchat-query-client-nick client-generic-nick-var
	irchat-query-client-lastcommand irchat-query-client-command)
  (if (string-ci-equal irchat-query-client-lastcommand "ping")
      (setq irchat-ctcp-ping-time (current-time)))
  (irchat-send-privmsg "PRIVMSG %s :%s%s"
		       client-generic-nick-var
		       irchat-query-client-lastcommand
		       irchat-query-client-insert-to-generic))


(defun irchat-Command-client-userinfo-from-minibuffer ()
  "Ask about someones client clientinfo."
  (interactive)
  (setq irchat-client-userinfo
	(read-from-minibuffer "New userinfo: "
			      irchat-client-userinfo)))


(defun irchat-Command-client-userinfo-from-commandbuffer ()
  "Ask about someones client clientinfo."
  (interactive)
  (let (start stop)
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq stop (point))
    (setq irchat-client-userinfo (buffer-substring start stop))
    (irchat-next-line 1)))


(defun irchat-Command-client-x-face-from-minibuffer ()
  "Ask about someones client clientinfo."
  (interactive)
  (setq irchat-client-x-face
	(read-from-minibuffer "New X-Face: "
			      irchat-client-x-face)))


(defun irchat-Command-client-x-face-from-commandbuffer ()
  "Ask about someones client clientinfo."
  (interactive)
  (let (start stop)
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq stop (point))
    (setq irchat-client-x-face (buffer-substring start stop))
    (irchat-next-line 1)))


(defun irchat-Command-client-ping (client-help-nick-var)
  "Ask about someones client help."
  (interactive (let (client-help-nick-var (completion-ignore-case t))
		 (setq client-help-nick-var 
		       (irchat-completing-default-read 
			"Whose client: " irchat-nick-alist
			'(lambda (s) t) nil irchat-query-client-nick))
		 (list client-help-nick-var)))
  (setq irchat-query-client-nick client-help-nick-var)
  (setq irchat-ctcp-ping-time (current-time))
  (irchat-send-privmsg "PRIVMSG %s :%s" 
		       client-help-nick-var 
		       irchat-query-client-ping))

;;;
;;; sending files (any files actually)
;;;
(defun irchat-Command-send-file (name to-who)
  "Send a file to given  user"
  (interactive "fFile name: \nsTo who: ")
  (save-excursion 
    (set-buffer (get-buffer-create (format "*IRC S_FILE_%s*" name)))
    (delete-region (point-min) (point-max))
    (insert-file name)
    (let (str)
      (setq str (buffer-string))
      (delete-region (point-min) (point-max))
      (insert (irchat-quote-encode str)))
    (goto-char (point-min))
    (irchat-send "NOTICE %s :FILE START %s :%s" to-who name
		 (buffer-substring (point)
				   (min (point-max)
					(+ 80 (point)))))
    (goto-char (min (point-max) (+ 80 (point))))
    (while (< (point) (point-max))
      (progn
	(if (=  1 (mod (point) 800))
	    (sleep-for 1))
	(irchat-send "NOTICE %s :FILE CONT %s :%s" to-who name
		     (buffer-substring (point)
				       (min (point-max) (+ 80 (point)))))
	(goto-char (min (point-max) (+ 80 (point))))))
    (irchat-send "NOTICE %s :FILE END %s : " to-who name)
    (kill-buffer (get-buffer-create (format "*IRC S_FILE_%s*" name)))))


;;;
;;; send text in kill-buffer
;;;
(defun irchat-Command-yank-send (&optional howmany)
  (interactive)
  (let ((beg (point)) end)
    (insert (car kill-ring-yank-pointer))
    (setq end (point))
    (goto-char beg)
    (while (< (point) end)
      (progn
	(irchat-Command-enter-message)
	(set-buffer irchat-Command-buffer)))))

;;;
;;; send rot-13 encrypted data
;;;
(defun irchat-Command-caesar-line (&optional n)
  "*Caesar encrypt current line. Rotate optional N characters."
  (interactive)
  (beginning-of-line nil)
  (push-mark (point))
  (end-of-line)
  (irchat-caesar-region n))

;;;
;;;
;;;
(defun get-word-left ()
  "Return word left from point."
  (save-excursion
    (let (point-now)
      (setq point-now (point))
      (backward-word 1)
      (buffer-substring (point) point-now))))


(defun irchat-Command-complete ()
  "Complete word before point from userlist."
  (interactive)
  (insert
   (save-excursion
     (let ((completion-ignore-case t) point-now word result)
       (setq point-now (point)
	     word (get-word-left)
	     result (try-completion word irchat-nick-alist))
       (backward-word 1)
       (delete-region (point) point-now)
       (if (or (eq result t) (eq result nil))
           word
         result)))))


(defun irchat-Command-load-vars ()
  "Load configuration from irchat-variables-file."
  (interactive)
  (let ((file (expand-file-name irchat-variables-file)))
    (if (file-exists-p file)
	(let ((nick irchat-real-nickname))
	  (load-file file)
	  (irchat-append-obsolete-vars)
	  (setq irchat-real-nickname nick)
	  (irchat-Command-reconfigure-windows)))))

(defun irchat-Command-save-vars ()
  (interactive)
  (let ((output-buffer (find-file-noselect
                        (expand-file-name irchat-variables-file)))
	output-marker)
    (save-excursion
      (set-buffer output-buffer)
      (goto-char (point-min))
      (if (re-search-forward "^;; Saved Settings *\n" nil 'move)
          (let ((p (match-beginning 0)))
            (goto-char p)
            (or (re-search-forward "^;; End of Saved Settings *\\(\n\\|\\'\\)"
				   nil t)
                (error (format 
			"can't find END of saved state in %s" 
			irchat-variables-file)))
	    (delete-region p (match-end 0)))
        (goto-char (point-max))
        (insert "\n"))
      (setq output-marker (point-marker))
      (let ((print-readably t) 
	    (print-escape-newlines t)
            (standard-output output-marker))
        (princ ";; Saved Settings\n")
        (mapcar (function 
		 (lambda (var)
		   (if (symbolp var)
		       (prin1 (list 'setq var
				    (let ((val (symbol-value var)))
				      (if (memq val '(t nil))
					  val
					(list 'quote val)))))
		     (setq var (eval var))
		     (cond ((eq (car-safe var) 'progn)
			    (while (setq var (cdr var))
			      (prin1 (car var))
			      (princ "\n")
			      (if (cdr var) (princ "  "))))
			   (var
			    (prin1 "xx")(prin1 var))))
		   (if var (princ "\n"))))
                irchat-saved-forms)
        (princ "\n")
        (princ ";; End of Saved Settings\n")))
    (set-marker output-marker nil)
    (save-excursion
      (set-buffer output-buffer)
      (let ((backup-inhibited (if irchat-backup-irchat-vars nil t)))
	(save-buffer)))
    (kill-buffer output-buffer))
  (setq irchat-save-vars-is-dirty nil))

(defun irchat-Command-reconfigure-windows ()
  (interactive)
  (let ((command-window (irchat-get-buffer-window irchat-Command-buffer))
	(dialogue-window (irchat-get-buffer-window irchat-Dialogue-buffer))
	(old-buffer (current-buffer)))
    (if (and command-window dialogue-window)
	(let ((c-height (window-height command-window))
	      (d-height (window-height dialogue-window)))
	  (delete-window command-window)
	  (pop-to-buffer irchat-Dialogue-buffer)
	  (enlarge-window (+ c-height d-height
			     (- (window-height dialogue-window)))))
      (pop-to-buffer irchat-Dialogue-buffer))
    (irchat-configure-windows)
    (if irchat-one-buffer-mode
	(pop-to-buffer irchat-Dialogue-buffer)
      (pop-to-buffer old-buffer))))

;;;
;;; command to get end of the Dialogue buffer
;;;
(defun irchat-Command-eod-buffer ()
  (interactive)
  (let ((saved-buffer (current-buffer))
	(dialogue-window (irchat-get-buffer-window irchat-Dialogue-buffer)))
    (set-buffer irchat-Dialogue-buffer)
    (goto-char (point-max))
    (set-window-point dialogue-window (point-max))
    (set-buffer saved-buffer)))

;;;
;;; commands to manage channel/conversation rings
;;;
(defun irchat-Command-private-conversation (arg)
  "Toggle between private conversation mode and channel mode. 
User can then join and part to a private conversation as he would 
join or part to a channel. 

If there are no private conversations or argument is given user is 
prompted the partner/channel (return as partner/channel means toggle 
mode, the current channel and current chat partner are not altered)"

  (interactive
   (list (if current-prefix-arg
       ;; prefixed, ask where to continue
       (if (eq irchat-command-buffer-mode 'chat)
	   (irchat-completing-default-read "Return to channel: "
					   (append irchat-channel-alist 
						   irchat-nick-alist) 
					   '(lambda (s) t) nil 
					   irchat-current-channel)
	 (completing-read "Start private conversation with: "
			  irchat-nick-alist '(lambda (s) t) nil))
     ;; no prefix, see if going to chat
     (if (eq irchat-command-buffer-mode 'channel)
	 ;; and if we have chat partner, select that
	 (if irchat-current-chat-partner
	     irchat-current-chat-partner
	   (completing-read "Start private conversation with: "
			    irchat-nick-alist '(lambda (s) t) nil)
	   )))))

  (setq irchat-command-buffer-mode (if (eq irchat-command-buffer-mode 'chat)
				       'channel
				     'chat))
  (if (and arg (not (string= arg "")))
      (irchat-Command-join arg))
  (setq irchat-channel-indicator (if (eq irchat-command-buffer-mode 'chat)
				     (if irchat-current-chat-partner
					 (format "Chatting with %s" 
						 irchat-current-chat-partner)
				       "No partner")
				   (if irchat-current-channel
				       (format "Channel %s" 
					       irchat-current-channel)
				     "No channel")))
  (irchat-set-crypt-indicator)
  ;; refresh mode line
  (set-buffer-modified-p (buffer-modified-p)))


(defun irchat-Command-push ()
  "Select next channel or chat partner."
  (interactive)
  (if (eq irchat-command-buffer-mode 'chat)
    (if irchat-current-chat-partners
	(let ((chat-partner (nth (1- (length irchat-current-chat-partners))
				 irchat-current-chat-partners)))
	  (if (string-list-ci-memberp chat-partner irchat-current-chat-partners)
	      (progn
		(setq irchat-current-chat-partner chat-partner
		      irchat-current-chat-partners
		      (cons chat-partner
			    (string-list-ci-delete 
			     chat-partner
			     irchat-current-chat-partners)))
		(setq irchat-channel-indicator
		      (if irchat-current-chat-partner
			  (format "Chatting with %s" 
				  irchat-current-chat-partner)
			"No partner"))
		(irchat-set-crypt-indicator)))
	  (set-buffer-modified-p (buffer-modified-p))))
    (if irchat-current-channels
	(let ((channel (nth (1- (length irchat-current-channels))
			    irchat-current-channels)))
	  (if (string-list-ci-memberp channel irchat-current-channels)
	      (progn
		(setq irchat-current-channel channel
		      irchat-current-channels
		      (cons channel
			    (string-list-ci-delete channel 
						   irchat-current-channels)))
		(setq irchat-channel-indicator
		      (if irchat-current-channel
			  (format "Channel %s" irchat-current-channel)
			"No channel"))
		(irchat-set-crypt-indicator)))
	  (set-buffer-modified-p (buffer-modified-p))))))


(defun irchat-Command-pop ()
  "Select previous channel or chat partner."
  (interactive)
  (if (eq irchat-command-buffer-mode 'chat)
      (let ((chat-partner irchat-current-chat-partner))
	(progn
	  (setq irchat-current-chat-partners (append
					      (string-list-ci-delete 
					       chat-partner 
					       irchat-current-chat-partners)
					      (list chat-partner))
		irchat-current-chat-partner (car irchat-current-chat-partners)
		irchat-channel-indicator 
		(if irchat-current-chat-partner
		    (format "Chatting with %s" 
			    irchat-current-chat-partner)
		  "No partner"))
	  (irchat-set-crypt-indicator)
	  (set-buffer-modified-p (buffer-modified-p))))

    (let ((channel irchat-current-channel))
      (progn
	(setq irchat-current-channels (append 
				       (string-list-ci-delete 
					channel 
					irchat-current-channels)
				       (list channel))
	      irchat-current-channel (car irchat-current-channels)
	      irchat-channel-indicator (format "Channel %s" 
					       irchat-current-channel))
	(irchat-set-crypt-indicator)
	(set-buffer-modified-p (buffer-modified-p))))))


(defun irchat-Command-select-channel-1 ()
  "Select current channel from channel list."
  (interactive)
  (irchat-switch-to-channel-no 0))


(defun irchat-Command-select-channel-2 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 1))


(defun irchat-Command-select-channel-3 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 2))


(defun irchat-Command-select-channel-4 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 3))


(defun irchat-Command-select-channel-5 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 4))


(defun irchat-Command-select-channel-6 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 5))


(defun irchat-Command-select-channel-7 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 6))


(defun irchat-Command-select-channel-8 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 7))


(defun irchat-Command-select-channel-9 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 8))


(defun irchat-Command-select-channel-10 ()
  "Select current channel from channel list"
  (interactive)
  (irchat-switch-to-channel-no 9))


(defun irchat-switch-to-channel-no (num)
  (if (and (integerp num)
	   (stringp (nth num irchat-current-channels)))
      (let ((chnl (nth num irchat-current-channels)))
	(message (format "Switching to \"%s\"." chnl))
	(setq irchat-current-channel chnl)
        (setq irchat-channel-indicator (format "Channel %s" 
                                               irchat-current-channel))
        (irchat-set-crypt-indicator)
        (set-buffer-modified-p (buffer-modified-p))
	t)
    (progn
      (message "Invalid channel!")
      nil)))


(defun irchat-Command-ping ()
  (interactive)
  (if (stringp irchat-server-name)
      (progn
	(irchat-send "PING %s" irchat-server-name)
	t)
    nil))
;;;
;;; Cannot use completing read, user may want to query many names
;;;
(defun irchat-Command-ison (nick)
  "IsON user."
  (interactive "sIsON: ")
  (irchat-send "ISON %s" nick))


(defun irchat-Command-userhost (nick)
  "Ask for userhost."
  (interactive "sUserhost nick(s): ")
  (irchat-send "USERHOST %s" nick))


;;;
;;; dig last kill from KILLS and show it
;;;

(defun irchat-Command-show-last-kill ()
  (interactive)
  (save-excursion
    (let ((buf (current-buffer))
	  (buffer-read-only)
	  str)
      (set-buffer irchat-KILLS-buffer)
      (goto-char (point-max))
      (forward-line -1)
      (setq str (buffer-substring (point) (point-max)))
      (goto-char (point-max))
      (set-buffer irchat-Dialogue-buffer)
      (goto-char (point-max))
      (irchat-w-insert irchat-D-buffer str)
      (goto-char (point-max))
      (set-window-point (irchat-get-buffer-window irchat-Dialogue-buffer)
			(point-max))
      (set-buffer buf))))


(defun irchat-Command-toggle-private ()
  (interactive)
  (let ((irchat-private-window 
	 (irchat-get-buffer-window 
	  (car (list (nth (- (length irchat-P-buffer) 1) irchat-P-buffer))))))
    (if irchat-private-window
	(if (interactive-p)
	    (let* ((my-IN-win (irchat-get-buffer-window irchat-Command-buffer))
		   (my-OUT-win (irchat-get-buffer-window irchat-Dialogue-buffer))
		   (sw (selected-window))
		   (INwh (window-height my-IN-win))
		   (OUTwh (+ (window-height my-OUT-win)
			     (window-height irchat-private-window))))
	      (delete-window irchat-private-window)
	      (if my-IN-win
		  (progn
		    (select-window my-IN-win)
		    (shrink-window (- (window-height my-IN-win) INwh))))
	      (if my-OUT-win
		  (progn
		    (select-window my-OUT-win)
		    (shrink-window (- (window-height my-OUT-win) OUTwh))))
	      (select-window sw)))
      (if (irchat-get-buffer-window irchat-Dialogue-buffer)
	  (let ((sw (selected-window)))
	    (select-window (irchat-get-buffer-window irchat-Dialogue-buffer))
	    (condition-case err
		(progn
		  (split-window-vertically)
		  (if irchat-swap-private
		      (progn
			(shrink-window 
			 (- irchat-private-window-height (window-height)))
			(other-window 1)
			(set-window-buffer 
			 (irchat-get-buffer-window irchat-Dialogue-buffer)
			 (irchat-get-buffer-create 
			  (nth (- (length irchat-P-buffer) 1)
			       irchat-P-buffer))))
		    (shrink-window 
		     (- (window-height) irchat-private-window-height))
		    (set-window-buffer 
		     (irchat-get-buffer-window irchat-Dialogue-buffer)
		     (irchat-get-buffer-create
		      (nth (- (length irchat-P-buffer) 1)
			   irchat-P-buffer)))))
	      (args-out-of-range 
	       (message "IRCHAT: Command-window-toggle with too small window")))
	    (select-window sw)))))
  (irchat-w-insert irchat-D-buffer "")) ;; recenter if needed

(eval-and-compile (provide 'irchat-commands))

;;;
;;;  eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-copyright.el,v 3.4 2002/06/04 16:03:06 tri Exp $
;;;
(defvar irchat-copyright "
    $Id: irchat-copyright.el,v 3.4 2002/06/04 16:03:06 tri Exp $
    Internet Relay CHAT interface for GNU Emacs

    Copyright  1989 Tor Lillqvist <tml@iki.fi>

    Copyright  1990-2002 Irchat Developer Team    (see below)
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with GNU Emacs; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
    Following people have made major improvements to irchat. Also 
    there are numerous people who have given suggestions and
    improvements. If you think your name should be included here,
    give us a note...

	   Markku Jrvinen	<mta@iki.fi>
	   Kai Keinnen		<kmk@iki.fi>
	   Janne Liimatainen	<jsl@iki.fi>
	   Tero T. Mononen     	<tmo@iki.fi>
	   Kim Nyberg          	<kny@iki.fi>
           Tomi Ollila          <too@iki.fi>
	   Jukka Partanen      	<jtp@iki.fi>
	   Timo J. Rinne	<tri@iki.fi>

    You can ftp the latest version from cs.hut.fi:/pub/irchat

    Current team is:
	   Timo J. Rinne	<tri@iki.fi>
	   Janne Liimatainen	<jsl@iki.fi>
	   Tero T. Mononen     	<tmo@iki.fi>
	   Kim Nyberg          	<kny@iki.fi>
           Tomi Ollila          <too@iki.fi>
	   Jukka Partanen      	<jtp@iki.fi>

    MSN Messenger protocol implementation by Timo J. Rinne <tri@iki.fi>.

    Bug reports and suggestions can be mailed to irchat-developers@cs.hut.fi.
" "Basically GPL")

;;;
;;;  $Log: irchat-copyright.el,v $
;;;  Revision 3.4  2002/06/04 16:03:06  tri
;;;  	Couple of bindings to msn stuff.
;;;  	Acknowledged MSNM code in the copyright notice.
;;;
;;;  Revision 3.3  1997/07/09 13:46:04  tri
;;;  Made some small modifications.  Copyright notice of course
;;;  remains intact.
;;;
;;;  Revision 3.2  1997/02/27 10:19:14  jsl
;;;  changed irchat-nickname to irchat-real-nickname
;;;  users should only change irchat-nickname
;;;
;;;  Revision 3.1  1997/02/24  16:00:02  tri
;;;  Promoted revision to 3.1.
;;;
;;;  Revision 1.2  1997/02/13 09:10:47  tri
;;;  Added Domo to "current" team.
;;;
;;;  Revision 1.1  1996/12/19 14:54:48  tri
;;;  Initial revision
;;;
;;;  Revision 0.1.1  1995/01/15  11:43:37  tmo
;;;  Initial revision
;;; 
;;;  Revision 0.1.2  1993/07/13  11:56:52  tmo
;;;  	made this file to be defvar
;;; 
;;;  Revision 0.1.2  1993/07/13  11:56:52  tmo
;;;  	made this file to be defvar
;;; 
;;;  Revision 0.1.1  1993/07/10  13:23:50  tmo
;;;  Initial revision
;;; 
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-cta.el,v 3.10 1998/06/25 07:35:19 jtp Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;; decode and encode of binary data
;;;
(defun irchat-quote-decode (string-to-decode)
  (interactive)
  (save-excursion
    (set-buffer (get-buffer-create "*IRC DECODE*"))
    (delete-region (point-min) (point-max))
    (insert string-to-decode)
    (goto-char (point-min))
    (replace-string "\\\\" "\\")
    (goto-char (point-min))
    (replace-string "\\a" "")
    (goto-char (point-min))
    (replace-string "\\n" "\n")
    (goto-char (point-min))
    (replace-string "\\r" "\r")
    (setq string-to-decode (buffer-substring (point-min) (point-max)))
    string-to-decode))

(defun irchat-quote-encode (string)
  (interactive)
  (save-excursion
    (set-buffer (get-buffer-create "*IRC ENCODE*"))
    (delete-region (point-min) (point-max))
    (insert string)
    (goto-char (point-min))
    (while (search-forward "\\" nil 1)
      (insert "\\"))
    (goto-char (point-min))
    (while (search-forward "" nil 1)
      (delete-char -1)
      (insert "\\a"))
    (goto-char (point-min))
    (while (search-forward "" nil 1)
      (delete-char -1)
      (insert "\\0"))
    (goto-char (point-min))
    (while (search-forward "\n" nil 1)
      (delete-char -1)
      (insert "\\n"))
    (goto-char (point-min))
    (while (search-forward "\r" nil 1)
      (delete-char -1)
      (insert "\\r"))
    (setq string (buffer-substring (point-min) (point-max)))
    string))


(defun irchat-ctl-a-action-msg (from chnl rest)
  (if (string-ci-equal chnl irchat-current-channel)
      (irchat-w-insert irchat-D-buffer
		       (format "%sAction: %s %s\n" irchat-info-prefix from rest))
    (irchat-w-insert irchat-D-buffer
		     (format "%sAction to %s: %s %s\n"
			     irchat-info-prefix chnl from rest))))


(defun irchat-ctl-a-msg (from chnl rest)
  "It's ctl-a request, act on it."
  (let* (left now right message rest-of-line hook)
    (if (string-match "^\\([^]*\\)\\([^]*\\)\\(.*\\)" rest)
	(progn
	  (setq left (matching-substring rest 1))
	  (setq now (matching-substring rest 2))
	  (setq right (matching-substring rest 3))
	  (setq rest (concat left right))
	  (if (string-match "^\\([^ ]*\\) \\(.*\\)" now)
	      (progn
		(setq message (downcase (matching-substring now 1)))
		(setq rest-of-line (matching-substring now 2)))
	    (if (string-match "^\\([^ ]*\\)" now)
		(progn
		  (setq message (downcase (matching-substring now 1)))
		  (setq rest-of-line nil))
	      (progn
		(setq message "errmsg")
		(setq rest-of-line "Couldn't figure out what was said."))))
	  (if (and (boundp 
		    (setq hook
			  (intern 
			   (concat "irchat-ctl-a-" message "-msg-hook"))))
		   (eval hook)
		   (eq (eval (list hook from rest-of-line)) t))
	      ;; If we have a hook, and it returns T, do nothing more
	      nil
	    ;; else call the handler
	    (if (fboundp (setq fun (intern
				    (concat "irchat-ctl-a-" message "-msg"))))
		(progn
		  (eval (list fun from chnl rest-of-line)))
	      (progn
;		(irchat-send "NOTICE %s :ERRMSG %s :%s"
;			     from
;			     (upcase message)
;			     (format irchat-client-error-msg (upcase message)))
		(message (format "CLIENT %s query from %s (ignored)."
				 (upcase message) from)))
		))))
    rest))


(defun irchat-ctl-a-client-msg (from chnl rest)
  (let* (message rest-of-line hook)
    (if rest
	(progn
	  (if (string-match "^\\([^ ]*\\) \\(.*\\)" rest)
	      (progn
		(setq message (downcase (matching-substring rest 1)))
		(setq rest-of-line (matching-substring rest 2)))
	    (if (string-match "^\\([^ ]*\\)" rest)
		(progn
		  (setq message (downcase (matching-substring rest 1)))
		  (setq rest-of-line nil))
	      (progn
		(setq message "errmsg")
		(setq rest-of-line "Couldn't figure out what was said."))))
	  (if (and (boundp 
		    (setq hook
			  (intern 
			   (concat "irchat-ctl-a-" message "-msg-hook"))))
		   (eval hook)
		   (eq (eval (list hook from rest-of-line)) t))
	      ;; If we have a hook, and it returns T, do nothing more
	      nil
	    ;; else call the handler
	    (if (fboundp (setq fun (intern
				    (concat "irchat-ctl-a-" message "-msg"))))
		(progn
		  (eval (list fun from rest-of-line)))
	      (message (format 
			"IRCHAT: Unknown CLIENT message from %s \"%s\""
			from (upcase message)))))))))	      


(defun irchat-ctl-a-version-msg (from chnl rest)
  (if (not (equal 'none irchat-external-version-string))
      (let ((version-string 
	     (if (not (stringp irchat-external-version-string))
		 (if irchat-use-ancient-version-format
		     (format "%s %s :%s for %s"
			     irchat-version
			     irchat-emacs-version-name
			     irchat-version
			     irchat-emacs-version-name)
		   (format "%s :%s for %s"
			   irchat-version
			   irchat-client-name
			   irchat-emacs-version-name))
	       irchat-external-version-string)))
	(irchat-send-delayed (format "NOTICE %s :VERSION %s" 
				     from
				     version-string))))
  (if (string-ci-equal chnl irchat-real-nickname)
      (message (format "CLIENT VERSION query from %s%s." 
		       from
		       (if (equal 'none irchat-external-version-string)
			   ".  Ignored" "")))
    (message (format "CLIENT VERSION query from %s (%s)%s." 
		     from 
		     chnl
		     (if (equal 'none irchat-external-version-string)
			 ".  Ignored" "")))))


(defun irchat-ctl-a-userinfo-msg (from chnl rest)
  (irchat-send-delayed "NOTICE %s :USERINFO %s"
		       from irchat-client-userinfo)
  (message (format "CLIENT USERINFO query from %s." from)))

(defun irchat-ctl-a-clientinfo-msg (from chnl rest)
  (irchat-send-delayed 
   (format 
    "NOTICE %s :CLIENTINFO :VERSION USERINFO CLIENTINFO X-FACE HELP ERRMSG" from))
  (message (format "CLIENT CLIENTINFO query from %s." from)))


(defun irchat-ctl-a-help-msg (from chnl rest)
  (irchat-send-delayed 
   (format 
    "NOTICE %s :HELP :VERSION gives version of this client" 
    from))
  (irchat-send-delayed
   (format 
    "NOTICE %s :HELP :USERINFO gives user supplied info (if any)" 
    from))
  (irchat-send-delayed 
   (format 
    "NOTICE %s :HELP :CLIENTINFO gives commands this client knows" 
    from))
  (irchat-send-delayed 
   (format 
    "NOTICE %s :HELP :X-FACE gives you user supplied X-Face (if exists)" 
    from))
  (irchat-send-delayed 
   (format 
    "NOTICE %s :HELP :HELP gives this help message" 
    from))
  (irchat-send-delayed 
   (format 
    "NOTICE %s :HELP :ERRMSG tells you your command was not valid" 
    from))
  (message (format "CLIENT HELP query from %s." from)))


(defun irchat-ctl-a-comment-msg (from chnl rest)  
  (message (format "CLIENT COMMENT query from %s." from)))


(defun irchat-ctl-a-xyzzy-msg (from chnl rest)
  (irchat-send-delayed "NOTICE %s :Nothing happens.(xyzzy inactive)" from)
  (message (format "CLIENT XYZZY query from %s." from)))


(defun irchat-ctl-a-ping-msg (from chnl rest)
  (if (not rest)
      (setq rest ""))
  (irchat-send-delayed "NOTICE %s :PING %s" from rest)
  (message (format "CLIENT PING query from %s." from)))

(defun irchat-ctl-a-x-face-msg (from chnl rest)
  (irchat-send-delayed "NOTICE %s :X-FACE %s"
		       from irchat-client-x-face)
  (message (format "CLIENT X-FACE query from %s." from)))

;;;
;;; read CLIENT messages from notice, no postprocessing done
;;;

(defun irchat-ctl-a-notice (prefix rest)
  "Ctl-a notice."
  (let* (message rest-of-line hook) 
    (if (string-match "\\([^ ]*\\) \\(.*\\)" rest)
	  (setq message (downcase (matching-substring rest 1))
		rest-of-line (matching-substring rest 2))
      (if (string-match "\\([^ ]*\\)" rest)
	  (setq message (downcase (matching-substring rest 1))
		rest-of-line nil)
	(setq message "errmsg"
	      rest-of-line "Couldn't figure out what was said.")))

    (if (and (boundp 
	      (setq hook
		    (intern (concat "irchat-ctl-a-" message "-notice-hook"))))
	     (eval hook)
	     (eq (eval (list hook prefix rest-of-line)) t))
	;; If we have a hook, and it returns T, do nothing more
	nil
      ;; else call the handler
      (if (fboundp (setq fun (intern
			      (concat "irchat-ctl-a-" message "-notice"))))
	  (progn
	    (eval (list fun prefix rest-of-line)))
	(message (format 
		  "IRCHAT: Unknown ctl-a notice \":%s %s %s\"" 
		  prefix (upcase message) rest-of-line))))))


(defun irchat-ctl-a-file-notice (prefix rest)
  (if irchat-file-accept
      (let* (message file-name rest-of-line hook)
	(string-match "^\\([^ ]*\\) \\([^ ]*\\) :\\(.*\\)" rest)
	(setq message (downcase (matching-substring rest 1)))
	(setq file-name (matching-substring rest 2))
	(setq rest-of-line (matching-substring rest 3))
	(if (and (boundp 
		  (setq hook
			(intern (concat "irchat-file-" message "-hook"))))
		 (eval hook)
		 (eq (eval (list hook prefix file-name rest-of-line)) t))
	    ;; If we have a hook, and it returns T, do nothing more
	    nil
	  ;; else call the handler
	  (if (fboundp (setq fun (intern
				  (concat "irchat-file-" message))))
	      (progn
		(eval (list fun prefix file-name rest-of-line)))
	    (message (format 
		      "IRCHAT: Unknown FILE message \":%s %s %s %s\""
		      prefix (upcase message) file-name rest-of-line)))))
    (message (format
	      "FILE: %s sending, set irchat-file-accept and ask to resend"
	      prefix))))

(defun irchat-file-start (prefix name data)
  (save-excursion
    (set-buffer (get-buffer-create (format "*IRC R_FILE_%s*" name)))
    (delete-region (point-min) (point-max))
    (insert data)))

(defun irchat-file-cont (prefix name data)
  (save-excursion
    (set-buffer (get-buffer-create (format "*IRC R_FILE_%s*" name)))
    (goto-char (point-max))
    (insert data)))


(defun irchat-file-end (prefix name data)
  (save-excursion
    (set-buffer (get-buffer-create (format "*IRC R_FILE_%s*" name)))
    (goto-char (point-max))
    (insert data)
    (let (str)
      (setq str (buffer-string))
      (delete-region (point-min) (point-max))
      (insert (irchat-quote-decode str)))
    (goto-char (point-min))
    (if (and irchat-file-confirm-save
	     (not (y-or-n-p "Save file?")))
	nil
      (progn
	(if (not (file-exists-p "~/.irchat"))
	    (shell-command (format "mkdir %s"
				   (expand-file-name ".irchat" "$HOME"))))
	(write-region (point-min) (point-max) (expand-file-name 
					       (file-name-nondirectory
						(format "%s(%s)" name prefix))
					       "~/.irchat"))
	(kill-buffer (get-buffer-create (format "*IRC R_FILE_%s*" name)))))))


(defun irchat-ctl-a-client-notice (prefix rest)
  (let* (message rest-of-line hook)
    (if (string-match "^\\([^ ]*\\) \\(.*\\)" rest)
	(setq message (downcase (matching-substring rest 1))
	      rest-of-line (matching-substring rest 2))
      (if (string-match "^\\([^ ]*\\)" rest)
	  (setq message (downcase (matching-substring rest 1))
		rest-of-line nil)
	(setq message "errmsg"
	      rest-of-line "Couldn't figure out what was said.")))
    (if (and (boundp (setq hook
			   (intern (concat 
				    "irchat-client-" message "-notice-hook"))))
	     (eval hook)
	     (eq (eval (list hook prefix rest-of-line)) t))
	;; If we have a hook, and it returns T, do nothing more
	nil
      ;; else call the handler
      (if (fboundp (setq fun (intern
			      (concat "irchat-client-" message "-notice"))))
	  (progn
	    (eval (list fun prefix rest-of-line)))
	(message (format 
		  "IRCHAT: Unknown CLIENT notice \":%s %s %s %s\""
		  prefix (upcase message) rest-of-line))))))


(defvar irchat-client-message "CLIENT@%s: %s\n" 
  "*Message in which info of other clients is displayed.")


(defun irchat-client-version-notice (prefix rest)  
  (if rest
      (if (string-match "^\\([^:]*\\):\\(.*\\)" rest)
	  (irchat-w-insert irchat-D-buffer
			   (format irchat-client-message prefix 
				   (concat (matching-substring rest 1)
					   (matching-substring rest 2))))
	(irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))
    (message (format "Empty CLIENT version notice from \"%s\"." prefix))))


(defun irchat-client-clientinfo-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-client-userinfo-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-client-help-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-client-x-face-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-client-errmsg-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))

(defun irchat-client-comment-notice (from rest)  
  (message (format "CLIENT COMMENT query from %s." from)))

(defun irchat-client-ping-notice (from rest)  
  (let ((timenow (current-time)))
    (irchat-w-insert 
     irchat-Dialogue-buffer 
     (format 
      "PING time from %s: %ds\n" 
      from
      (+ (* 65536 (- (car timenow) (car irchat-ctcp-ping-time)))
	 (- (car (cdr timenow)) (car (cdr irchat-ctcp-ping-time))))))))

;;;
(defun irchat-ctl-a-version-notice (prefix rest)  
  (if rest
      (if (string-match "^\\([^:]*\\):\\(.*\\)" rest)
	  (irchat-w-insert irchat-D-buffer
			   (format irchat-client-message prefix
				   (concat (matching-substring rest 1)
					   (matching-substring rest 2))))
	(if (string-match "^\\([^ ]*\\) \\([^ ]*\\) \\([^ ]*\\) \\(.*\\)" rest)
	    (irchat-w-insert irchat-D-buffer 
			     (format irchat-client-message 
				     prefix
				     (format "%s %s for %s: %s" 
					     (matching-substring rest 1)
					     (matching-substring rest 2)
					     (matching-substring rest 3)
					     (matching-substring rest 4))))
	  (irchat-w-insert irchat-D-buffer 
			   (format irchat-client-message prefix rest))))
    (message (format "Empty CLIENT version notice from \"%s\"." prefix))))


(defun irchat-ctl-a-clientinfo-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-ctl-a-userinfo-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-ctl-a-help-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-ctl-a-errmsg-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-ctl-a-x-face-notice (prefix rest)  
  (irchat-w-insert irchat-D-buffer (format irchat-client-message prefix rest)))


(defun irchat-ctl-a-comment-notice (from rest)  
  (message (format "CLIENT COMMENT query from %s." from)))

(defun irchat-ctl-a-ping-notice (from rest)  
  (let ((timenow (current-time)))
    (irchat-w-insert 
     irchat-Dialogue-buffer 
     (format 
      "PING time from %s: %ds\n" 
      from
      (+ (* 65536 (- (car timenow) (car irchat-ctcp-ping-time)))
	 (- (car (cdr timenow)) (car (cdr irchat-ctcp-ping-time))))))))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-handle.el,v 3.34 2002/06/09 15:35:02 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))
(eval-and-compile  
  (require 'irchat-filter))

(defun irchat-handle-error-msg (parsed-sender parsed-msg prefix rest)
  (message "IRC error: %s" rest))


(defun irchat-ifrest (rest &optional quote)
  (if (= 0 (length rest)) 
      "" 
    (if quote
	(regexp-quote (format " (%s)" rest))
      (format " (%s)" rest))))


(defun irchat-handle-channel-msg (parsed-sender parsed-msg prefix rest)
  (let ((ispart (string= rest "0")))
    (if (string= prefix irchat-real-nickname)
	(progn
	  (setq irchat-current-channel rest)
	  (setq irchat-channel-indicator
		(if ispart
		    "No channel"
		  (format "Channel %s" rest)))
	  (irchat-set-crypt-indicator)))
    (if ispart
	(if (not irchat-ignore-changes)
	    (irchat-w-insert irchat-D-buffer
			     (format "%s%s has left this channel\n" 
				     irchat-change-prefix prefix)))
      (if (not irchat-ignore-changes)
	  (irchat-w-insert irchat-D-buffer 
			   (format "%s%s has joined this channel\n"
				   irchat-change-prefix prefix))))
    (irchat-change-nick-of prefix prefix)))


(defun irchat-handle-nick-msg (parsed-sender parsed-msg prefix rest)
  (irchat-change-nick-of prefix rest)
  (put (intern rest irchat-obarray) 'chnl 
       (get (intern prefix irchat-obarray) 'chnl))
  (put (intern prefix irchat-obarray) 'chnl nil)
  (if (string= prefix irchat-real-nickname)
      (setq irchat-real-nickname rest))
  (irchat-w-insert irchat-D-buffer 
		   (format "%s%s is now known as %s\n" 
			   irchat-change-prefix prefix rest)))


(defun irchat-handle-notice-msg (parsed-sender parsed-msg prefix rest)
  (if (and irchat-shorten-kills
	   (string-match 
	    "Notice[: -]*Received KILL message for \\([^.]*\\)\\. From \\([^ ]*\\) Path: \\([^ ]*\\) ?\\(.*\\)"
	    rest))
      (let ((killed (matching-substring rest 1))
	    (killer (matching-substring rest 2))
	    (reason (matching-substring rest 4))
	    (buf (current-buffer)))
	(set-buffer irchat-KILLS-buffer)
	(goto-char (point-max))
	(irchat-w-insert irchat-K-buffer 
			 (format "%s\n" rest))
	(set-buffer buf)
	(irchat-w-insert irchat-D-buffer 
			 (format "%s%s KILLed %s %s\n" 
				 irchat-notice-prefix
				 killer killed 
				 (if (= (length reason) 0) 
				     "-No reason supplied-" 
				   reason))))
    (let* ((msg-encrypted-p nil)
	   (msg-suspicious-p nil)
	   (msg-garbled-p nil)
	   (msg-fingerprint nil)
	   (msg-timestamp nil)
	   (rest (if (string-match "^\\([^:][^:]*:\\)\\(.*\\)$" rest)
		     (let ((head (matching-substring rest 1))
			   (tail (matching-substring rest 2)))
		       (if (irchat-encrypted-message-p tail)
			   (let* ((clear (irchat-decrypt-message tail))
				  (stat (nth 0 clear))   ;; 'success or 'error
				  (nick (nth 1 clear))   ;;  sender's nick
				  (time (nth 2 clear))   ;;  timestamp
				  (msg (nth 3 clear))    ;;  cleartext msg
				  (fprint (nth 4 clear)) ;;  fingerprint
				  (warn ""))
			     (setq msg-encrypted-p t)
			     (setq msg-fingerprint fprint)
			     (setq msg-timestamp time)
		             ;;; Check timestamp and nick here
			     (if (and (equal 'success stat)
				      (not (irchat-hex-timestamp-valid 
					    time
					    irchat-crypt-timestamp-tolerance)))
				 (progn
				   (setq msg-suspicious-p t)
				   (setq warn 
					 (concat warn 
						 " [Invalid timestamp!]"))))
			     (if (and (equal 'success stat)
				      (not (string-ci-equal nick
							    prefix)))
				 (progn
				   (setq msg-suspicious-p t)
				   (setq warn 
					 (concat warn 
						 (format 
						  " [Invalid sender \"%s\" != \"%s\"]"
						  nick
						  prefix)))))
			     (if (not (equal 'success stat))
				 (progn
				   (setq msg-garbled-p t)
				   (irchat-w-insert irchat-C-buffer
						    (format "-%s- %s [%s]\n"
							    prefix
							    tail
							    msg))))
			     (concat head (format "%s%s" msg warn)))
			 rest))
		   rest)))
      (if (or (not irchat-ignore-extra-notices)
	      (not prefix)
	      (not (string-match "as being away" rest)))
	  (if (let ((irchat-current-message-encrypted-p msg-encrypted-p)
		    (irchat-current-message-suspicious-p msg-suspicious-p)
		    (irchat-current-message-fingerprint msg-fingerprint)
		    (irchat-current-message-timestamp msg-timestamp))
		(irchat-run-message-hook-types 'irchat-notice-cleartext-hook
					       prefix
					       rest))
	      nil
	    (if prefix
		(cond 
					; prefixed clt-a-notice
		 ((string-match "\\(.*\\)" rest) 
		  (irchat-ctl-a-notice prefix rest))
		 ((and irchat-ignore-fakes
		       (string-match ".*Notice.*Fake:.*" rest))
		  t)
					; not a clt-a, but notice
		 ((string-match ".*Notice -- \\(.*\\)" rest) 
		  (irchat-w-insert irchat-D-buffer 
				   (format "%s%s: %s\n" 
					   irchat-notice-prefix
					   prefix 
					   (matching-substring rest 1))))
					; else send user a private message
		 (t 
		  (irchat-handle-privmsglike-msg parsed-sender parsed-msg prefix rest msg-encrypted-p)))
	      (progn
		;; no prefix
		(string-match "^\\([^ ]*\\) :\\(.*\\)" rest)
		(irchat-w-insert irchat-D-buffer 
				 (format "%s%s\n"
					 irchat-notice-prefix 
					 (matching-substring rest 2))))))))))


(defun irchat-handle-ping-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match "^\\([^ ][^ ]*\\)" rest)
      (irchat-send "PONG %s: %s" 
		   irchat-real-nickname
		   (matching-substring rest 1))
    (irchat-send "PONG yourself"))
  (irchat-Command-timestamp-if-interval-expired t)
  (irchat-maybe-poll))


(defun irchat-handle-pong-msg (parsed-sender parsed-msg prefix rest)
  ())

;; Is this still needed?
(defun irchat-handle-msg-msg (parsed-sender parsed-msg prefix rest)
  (irchat-nick-to-uah-append prefix irchat-userathost irchat-userathost-type)
  (if (or (and prefix
	       (irchat-ignore-this-p prefix irchat-userathost)
	       (irchat-msg-from-ignored prefix rest))
	  (and (not prefix)
	       (string= "> " rest)))
      nil
    (if prefix 
	(let ((oma (get (intern prefix irchat-obarray) 'chnl)))
	  (if oma
	      (while (or (string= "#" (substring (car oma) 0 1))
			 (string= "&" (substring (car oma) 0 1)))
		(setq oma (cdr oma)))
	    (setq oma (list irchat-current-channel)))
	  (if (string= (car oma) irchat-current-channel)
	      (irchat-w-insert 
	       irchat-D-buffer 
	       (format "%s %s\n"
		       (format (irchat-format-string 2 nil) prefix)
		       rest))
	    (irchat-w-insert 
	     irchat-D-buffer
	     (format "%s %s\n"
		     (format (irchat-format-string 3 nil) prefix (car oma))
		     rest)))))))

(defun irchat-handle-privmsg-msg (parsed-sender parsed-msg prefix rest)
  (irchat-nick-to-uah-append prefix irchat-userathost irchat-userathost-type)
  (if (and prefix
	   (irchat-ignore-this-p prefix irchat-userathost)
	   (irchat-msg-from-ignored prefix rest))
      nil
    (string-match "^\\([^ ]+\\) :\\(.*\\)" rest)
    (let* ((msg-encrypted-p nil)
	   (msg-suspicious-p nil)
	   (msg-garbled-p nil)
	   (msg-fingerprint nil)
	   (msg-timestamp nil)
	   (chnl (matching-substring rest 1))
	   (temp (matching-substring rest 2))
	   (case-fold-search t)
	   (temp (if (not (irchat-encrypted-message-p temp))
		     (progn
		       (setq msg-encrypted-p nil)
		       temp)
		   (let* ((clear (irchat-decrypt-message temp))
			  (stat (nth 0 clear))   ;; 'success or 'error
			  (nick (nth 1 clear))   ;;  sender's nick (string)
			  (time (nth 2 clear))   ;;  timestamp (hex-string)
			  (msg (nth 3 clear))    ;;  cleartext msg (string)
			  (fprint (nth 4 clear)) ;;  fingerprint (string)
			  (warn ""))
		     (setq msg-encrypted-p t)
		     (setq msg-fingerprint fprint)
		     (setq msg-timestamp time)
		     ;;; Check timestamp and nick here
		     (if (and (equal 'success stat)
			      (not (irchat-hex-timestamp-valid 
				    time
				    irchat-crypt-timestamp-tolerance)))
			 (progn
			   (setq msg-suspicious-p t)
			   (setq warn (concat warn " [Invalid timestamp!]"))))
		     (if (and (equal 'success stat)
			      (not (string-ci-equal nick
						    prefix)))
			 (progn
			   (setq msg-suspicious-p t)
			   (setq warn 
			       (concat warn (format 
					   " [Invalid sender \"%s\" != \"%s\"]"
					   nick
					   prefix)))))
		     (if (not (equal 'success stat))
			 (progn
			   (setq msg-garbled-p t)
			   (irchat-w-insert irchat-C-buffer
					    (format "<%s -> %s> %s [%s]\n"
						    prefix
						    chnl
						    temp
						    msg))))
		     (format "%s%s" msg warn)))))
      (if (and msg-encrypted-p
	       msg-garbled-p
	       irchat-crypt-ignore-defected)
	  (irchat-msg-from-ignored prefix (concat chnl " :" temp))
	(if (and msg-encrypted-p
		 msg-suspicious-p
		 irchat-crypt-ignore-suspicious)
	    (irchat-msg-from-ignored prefix (concat chnl " :" temp))
	  (if (let ((irchat-current-message-encrypted-p msg-encrypted-p)
		    (irchat-current-message-suspicious-p msg-suspicious-p)
		    (irchat-current-message-fingerprint msg-fingerprint)
		    (irchat-current-message-timestamp msg-timestamp))
		(irchat-run-message-hook-types 'irchat-privmsg-cleartext-hook
					       prefix
					       (concat chnl
						       " :"
						       temp)))
	      nil
	    (let ((x-prefix (cond ((and msg-encrypted-p
					msg-suspicious-p)
				   irchat-suspicious-message-prefix)
				  ((and msg-encrypted-p
					msg-garbled-p)
				   irchat-defected-message-prefix)
				  (t ""))))
	      (if (string-match "\\(.*\\)" temp)
		  (setq temp (irchat-ctl-a-msg prefix chnl temp)))
	      (if (not (string= temp ""))
		  (if (and prefix
			   (irchat-ignore-this-p prefix irchat-userathost temp)
			   (irchat-msg-from-ignored prefix rest))
		      (irchat-msg-from-ignored prefix rest)
		    (progn
		      ;; only private messages to us get time-stamp
		      (if (and (string-equal "A" irchat-away-indicator) 
			       (string-ci-equal chnl irchat-real-nickname))
			  (setq temp (format "%s (%s)" temp 
					     (if irchat-format-time-function
						 (apply irchat-format-time-function
							(list (current-time-string)))
					       (current-time-string)))))
		      (cond
		       ((string-ci-equal chnl irchat-real-nickname)
			(irchat-Dialogue-insert-message 
			 irchat-P-buffer 
			 x-prefix
			 (irchat-format-string 1 msg-encrypted-p)
			 prefix
			 temp))
		       ((string-ci-equal chnl (or irchat-current-channel ""))
			(if (irchat-user-on-this-channel prefix chnl)
			    (irchat-Dialogue-insert-message
			     (irchat-pick-buffer chnl)
			     x-prefix
			     (irchat-format-string 2 msg-encrypted-p)
			     prefix
			     temp)
			  (irchat-Dialogue-insert-message 
			   (irchat-pick-buffer chnl)
			   x-prefix
			   (irchat-format-string 4 msg-encrypted-p)
			   prefix
			   temp)))

		       (t;; channel we are joined (not current)
			(if (irchat-user-on-this-channel prefix chnl)
			    (irchat-Dialogue-insert-message 
			     (irchat-pick-buffer chnl)
			     x-prefix
			     (irchat-format-string 3 msg-encrypted-p)
			     prefix
			     temp
			     chnl)
			  (irchat-Dialogue-insert-message 
			   (irchat-pick-buffer chnl)
			   x-prefix
			   (irchat-format-string 5 msg-encrypted-p)
			   prefix
			   temp
			   chnl))))
		      (or (irchat-get-buffer-window (current-buffer))
			  (not (string-ci-equal chnl irchat-real-nickname))
			  (message "IRCHAT: A private message has arrived from %s" prefix))))))))))))


;; NOTICE
(defun irchat-handle-privmsglike-msg (parsed-sender parsed-msg prefix rest &optional msg-encrypted-p)
  (if (and prefix
	   (irchat-ignore-this-p prefix irchat-userathost)
	   (irchat-msg-from-ignored prefix rest))
      nil
    (irchat-nick-to-uah-append prefix irchat-userathost irchat-userathost-type)
    (string-match "^\\([^ ]+\\) :\\(.*\\)" rest)
    (let ((chnl (matching-substring rest 1))
	  (temp (matching-substring rest 2))
	  (case-fold-search t)
	  (x-prefix (cond ((and (boundp 'msg-encrypted-p)
				msg-encrypted-p
				(boundp 'msg-suspicious-p)
				msg-suspicious-p)
			   irchat-suspicious-message-prefix)
			  ((and (boundp 'msg-encrypted-p)
				msg-encrypted-p
				(boundp 'msg-garbled-p)
				msg-garbled-p)
			   irchat-defected-message-prefix)
			  (t ""))))
      (if (string-match "\\(.*\\)" temp)
	  (setq temp (irchat-ctl-a-msg prefix chnl temp)))
      (if (not (string= temp ""))
	  (if (and prefix
		   (irchat-ignore-this-p prefix irchat-userathost temp)
		   (irchat-msg-from-ignored prefix rest))
	      (irchat-msg-from-ignored prefix rest)
	    (cond
	     ((string-ci-equal chnl irchat-real-nickname)
	      (irchat-w-insert irchat-D-buffer 
			       (format "%s%s %s\n" 
				       x-prefix
				       (format 
					(irchat-format-string 
					 0 
					 msg-encrypted-p)
					prefix) 
				       temp)))
	     ((string-ci-equal chnl (or irchat-current-channel ""))
	      (if (irchat-user-on-this-channel prefix chnl)
		  (irchat-w-insert (irchat-pick-buffer chnl) 
				   (format "%s%s %s\n" 
					   x-prefix
					   (format 
					    (irchat-format-string 
					     2
					     msg-encrypted-p)
					    prefix) 
					   temp))
		(irchat-w-insert (irchat-pick-buffer chnl) 
				 (format "%s%s %s\n" 
					 x-prefix
					 (format 
					  (irchat-format-string 
					   4
					   msg-encrypted-p)
					  prefix) 
					 temp))))

	     (t;; channel we are joined (not current)
	      (if (irchat-user-on-this-channel prefix chnl)
		  (irchat-w-insert (irchat-pick-buffer chnl) 
				   (format "%s%s %s\n" 
					   x-prefix
					   (format 
					    (irchat-format-string 
					     3
					     msg-encrypted-p)
					    prefix chnl)
					   temp))
		(irchat-w-insert (irchat-pick-buffer chnl) 
				 (format "%s%s %s\n" 
					 x-prefix
					 (format
					  (irchat-format-string 
					   5
					   msg-encrypted-p)
					  prefix chnl) 
					 temp))))))))))

(defun irchat-handle-wall-msg (parsed-sender parsed-msg prefix rest)
  "Handle the WALL message."
  (irchat-w-insert irchat-D-buffer 
		   (format "%s%s %s\n" 
			   irchat-broadcast-prefix 
			   (if prefix (concat "from " prefix) "") rest)))


(defun irchat-handle-wallops-msg (parsed-sender parsed-msg prefix rest)
  "Handle the WALLOPS message."
  (if irchat-show-wallops
      (irchat-w-insert irchat-D-buffer 
		       (format "%s%s: %s\n" 
			       irchat-wallops-prefix 
			       (if prefix prefix "UNKNOWN") rest)))
  (let ((buf (current-buffer)))
    (set-buffer irchat-WALLOPS-buffer)
    (goto-char (point-max))
    (irchat-w-insert irchat-W-buffer 
		     (format "%s%s %s\n" 
			     irchat-wallops-prefix 
			     (if prefix 
				 (concat "from " prefix) "") 
			     rest))
    (set-buffer buf)))


(defun irchat-handle-quit-msg (parsed-sender parsed-msg prefix rest)
  "Handle the QUIT message."
  (if (and (not irchat-ignore-changes)
	   (or (not irchat-ignore-changes-from-ignored)
	       (not (irchat-ignore-this-p prefix irchat-userathost))))
      (if irchat-compress-changes
	  (let* ((text (format " \\(has\\|have\\) left IRC%s" 
			       (irchat-ifrest rest t)))
		 (match (format "^%s.*%s$" 
				(regexp-quote irchat-change-prefix) text))
		 (default (format "%s%s has left IRC%s\n" 
				  irchat-change-prefix
				  prefix
				  (irchat-ifrest rest))))
	    (irchat-w-replace irchat-D-buffer 
			      match default text
			      (format ", %s have left IRC%s" 
				      prefix (irchat-ifrest rest))))
	(irchat-w-insert irchat-D-buffer
			 (format "%s%s has left IRC%s\n" 
				 irchat-change-prefix
				 prefix
				 (irchat-ifrest rest)))))
  (irchat-change-nick-of prefix nil))


(defun irchat-handle-topic-msg (parsed-sender parsed-msg prefix rest)
  "Handle the TOPIC message."
  (if (string-match "\\([^ :]*\\)[: ]*\\(.*\\)" rest)
      (let ((chnl (matching-substring rest 1))
	    (topic (matching-substring rest 2)))
	(if (not irchat-ignore-changes)
	    (irchat-w-insert (irchat-pick-buffer chnl)
			     (format "%sNew topic on channel %s set by %s: %s\n" 
				     irchat-change-prefix chnl prefix topic))))
    (message "IRCHAT: Strange TOPIC")))


(defun irchat-handle-mode-msg (parsed-sender parsed-msg prefix rest)
  "Handle the MODE message."
  (if (and (not irchat-ignore-changes)
	   (or (not irchat-ignore-changes-from-ignored)
	       (not (irchat-ignore-this-p prefix irchat-userathost))))
      (let ((chnl " ")
	    (str "")
	    (server-change (if (string-match "..*\\...*" prefix) t nil)))
	(if (or (and (string-match "\\([^ ]*\\) :*\\(.*\\)" rest)
		     (setq chnl (matching-substring rest 1)
			   str (matching-substring rest 2)
			   str (if (= (aref str (1- (length str))) 32) 
				   (substring str 0 -1) str)))
		(and (string-match " :\\(.*\\)" rest) 
		     (setq str (matching-substring rest 1))))
	    (if irchat-compress-changes
		(if server-change
		    (let* ((default (format "%sNew mode for %s set by %s: %s\n"
					    irchat-change-prefix 
					    chnl
					    prefix
					    str)))
		      '())
		  (let* ((text (format "\n" rest))
			 (match (format "^%sNew mode for %s set by %s: " 
					(regexp-quote irchat-change-prefix) 
					chnl 
					prefix))
			 (default (format "%sNew mode for %s set by %s: %s\n" 
					  irchat-change-prefix 
					  chnl
					  prefix
					  str)))
		    (irchat-w-replace (irchat-pick-buffer chnl)
				      match 
				      default
				      text
				      (format ", %s\n" str))))
	      (irchat-w-insert (irchat-pick-buffer chnl)
			       (format "%sNew mode for %s set by %s: %s\n" 
				       irchat-change-prefix
				       chnl 
				       prefix
				       str)))
	  (message "IRCHAT: Strange MODE")))))


(defun irchat-handle-kick-msg (parsed-sender parsed-msg prefix rest)
  "Handle the KICK message."
  (if (string-match "^\\([^ ]*\\) \\([^ ]*\\) *:\\(.*$\\)" rest)
      (let ((match1 (matching-substring rest 1))
	    (match2 (matching-substring rest 2))
	    (match3 (matching-substring rest 3)))
	(if (string= match2 irchat-real-nickname)
	  (progn
	    (irchat-w-insert (irchat-pick-buffer match1)
			     (format 
			      "%sYou were kicked off channel %s by %s%s.\n" 
			      irchat-change-prefix 
			      match1 
			      prefix 
			      (if (or (not match3)
				      (string= "" match3)
				      (string= prefix match3))
				  ""
				(concat " (" match3 ")"))))
	    (setq 
	     irchat-current-channels 
	           (string-list-ci-delete match1 irchat-current-channels)
	     irchat-current-channel (car irchat-current-channels)
	     irchat-channel-indicator (if irchat-current-channel 
					  (format "Channel %s" 
						  irchat-current-channel) 
					"No channel"))
	    (irchat-set-crypt-indicator)
	    (irchat-remove-from-thischannel irchat-real-nickname match1))
	  (irchat-w-insert irchat-D-buffer 
			   (format "%s%s has kicked %s out%s%s\n" 
				   irchat-change-prefix prefix match2
				   (if (string= (or irchat-current-channel "") match1) 
				       "" 
				     (format " from channel %s" match1))
				   (if (or (not match3)
					   (string= "" match3)
					   (string= prefix match3))
				       ""
				     (concat " (" match3 ")"))))))
    (message "IRCHAT: Strange KICK.")))


(defun irchat-handle-invite-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match " :\\([^ ]+\\)" rest)
      (let ((chnl (matching-substring rest 1)))
	(irchat-w-insert irchat-D-buffer 
			 (format "%s%s invites you to channel %s\n"
				 irchat-info-prefix prefix chnl))
	(setq irchat-invited-channel chnl))
    (message "IRCHAT: Strange INVITE")))


(defun irchat-handle-kill-msg (parsed-sender parsed-msg prefix rest)
  (if (string-match "[^ ]+ +:\\(.*\\)" rest)
      (let ((path (matching-substring rest 1)))
	(irchat-w-insert irchat-D-buffer 
			 (format
			  "%sIRCHAT: You were killed by %s. Path: %s. RIP\n" 
			  irchat-info-prefix prefix path)))
    (message "IRCHAT: strange KILL"))
  (setq irchat-channel-indicator "No channel")
  (irchat-set-crypt-indicator))


(defun irchat-handle-join-msg (parsed-sender parsed-msg prefix rest) ; kmk 14101990, tri 03071997
  (let* ((parse (irchat-parse-^G-channel-name rest))
	 (flags (cdr parse))
	 (rest (car parse)))
    (if (string= prefix irchat-real-nickname)
        (progn
          (setq irchat-current-channel rest)
          (setq irchat-current-channels
                (cons irchat-current-channel irchat-current-channels))
          (setq irchat-channel-indicator
                (format "Channel %s" rest))
          (irchat-set-crypt-indicator))
      (irchat-add-to-channel prefix rest))
    (if (and (not irchat-ignore-changes)
	     (or (not irchat-ignore-changes-from-ignored)
		 (not (irchat-ignore-this-p prefix irchat-userathost))))
        (if irchat-compress-changes
            (let* ((text (format " \\(has\\|have\\) joined channel %s" 
                                 (regexp-quote rest)))
                   (match (format "^%s.* .*%s$" 
                                  (regexp-quote irchat-change-prefix) text))
                   (default (format "%s%s%s (%s) has joined channel %s\n" 
                                    irchat-change-prefix 
                                    prefix
				    (if flags (format " [%s]" flags) "")
                                    irchat-userathost 
                                    rest)))
              (irchat-w-replace (irchat-pick-buffer rest) match default text
                                (format ", %s%s (%s) have joined channel %s" 
                                        prefix 
					(if flags (format " [%s]" flags) "")
					irchat-userathost 
					rest)))
          (irchat-w-insert (irchat-pick-buffer rest) 
                           (format "%s%s%s (%s) has joined channel %s\n" 
                                   irchat-change-prefix
                                   prefix
				   (if flags (format " [%s]" flags) "")
                                   irchat-userathost 
                                   rest))))
    (irchat-nick-to-uah-append prefix irchat-userathost irchat-userathost-type)
    (irchat-change-nick-of prefix prefix)))


(defun irchat-handle-part-msg (parsed-sender parsed-msg prefix rest) ; kmk, 14101990
  (if (string-match "\\([^ ]*\\)\ .*" rest)
      (setq rest (matching-substring rest 1))) ;; throw away user given info.
  (if (string= prefix irchat-real-nickname)
      (progn
	(setq irchat-current-channels
	      (string-list-ci-delete rest irchat-current-channels)
	      irchat-current-channel (car irchat-current-channels)
	      irchat-channel-indicator (if irchat-current-channel (format "Channel %s" irchat-current-channel) "No channel"))
	(irchat-set-crypt-indicator)))
  (if (and (not irchat-ignore-changes)
	   (or (not irchat-ignore-changes-from-ignored)
	       (not (irchat-ignore-this-p prefix irchat-userathost))))
      (if irchat-compress-changes
	  (let* ((text (format " \\(has\\|have\\) left channel %s"
			       (regexp-quote rest)))
		 (match (format "^%s.*%s$" (regexp-quote irchat-change-prefix) text))
		 (default (format "%s%s has left channel %s\n"
				  irchat-change-prefix prefix rest)))
	    (irchat-w-replace (irchat-pick-buffer rest) 
			      match default text (format ", %s have left channel %s" prefix rest)))
	(irchat-w-insert (irchat-pick-buffer rest) 
			 (format "%s%s has left channel %s\n" irchat-change-prefix prefix rest))))
  (irchat-remove-from-thischannel prefix rest)
  (irchat-change-nick-of prefix prefix))


(defun irchat-pick-buffer (chnl)
  (let ((mylist irchat-buffer-preferences)
	(found nil)
	(result nil))
    (while (and chnl (not found) mylist)
      (if (string-match (car (car mylist)) chnl)
	  (setq result (car (cdr (car mylist)))
		found t))
      (setq mylist (cdr mylist)))
    (if result
	result
      irchat-D-buffer)))

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-misc.el,v 3.52 2002/11/09 19:40:47 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(eval-when-compile (require 'irchat-inlines))
(eval-and-compile  
  (require 'irchat-filter))

(defun irchat-ignore-this-p (nick uah &optional message)
  (if (string-match "\\." nick)
      nil ;;; Server messages can't be ignored by nick.
    (progn
      (let ((mylist irchat-ignore-nickname)
	    (time (current-time)))
	(while mylist
	  (let ((expiretime (if (cdr (car mylist))
				(irchat-time-difference time
							(cdr (car mylist)))
			      1)))
	    (if (< expiretime 0)
		(progn
		  (setq irchat-ignore-nickname (remassoc 
						(car (car mylist))
						irchat-ignore-nickname)
			irchat-save-vars-is-dirty t)
		  (if (= (car (cdr (cdr (cdr (car mylist))))) 0)
		      (irchat-w-insert irchat-D-buffer
				       (format 
					"%sIgnore timeout for %s expired.\n"
					irchat-info-prefix
					(car (car mylist)))))))
	    (setq mylist (cdr mylist)))))
      (let ((mylist irchat-ignore-message-regexp)
	    (time (current-time)))
	(while mylist
	  (let ((expiretime (if (cdr (car mylist))
				(irchat-time-difference time 
							(cdr (car mylist)))
			      1)))
	    (if (< expiretime 0)
		(progn
		  (setq irchat-ignore-message-regexp 
			(remassoc (car (car mylist))
				  irchat-ignore-message-regexp)
			irchat-save-vars-is-dirty t)
		  (if (= (car (cdr (cdr (cdr (car mylist))))) 0)
		      (irchat-w-insert irchat-D-buffer
				       (format
					"%sIgnore timeout for messages matching \"%s\" expired.\n"
					irchat-info-prefix
					(car (car mylist)))))))
	    (setq mylist (cdr mylist)))))
      (if (or (and (fboundp 'irchat-custom-ignore-this-p)
		   (irchat-custom-ignore-this-p nick uah))
	      (and message
		   (fboundp 'irchat-custom-ignore-this-message-p)
		   (irchat-custom-ignore-this-message-p nick uah message)))
	  t
	(let ((killit nil)
	      (case-fold-search t))
	  (mapcar (function 
		   (lambda (kill)
		     (if (or (string-ci-equal (car kill) nick)
			     (and (string-match (upcase (car kill)) 
						(upcase nick))
				  (= (match-beginning 0) 0)
				  (= (match-end 0) (length nick)))
			     (and (string-match "@" (car kill))
				  (or (string-ci-equal (car kill) uah)
				      (and (string-match (upcase (car kill))
							 (upcase uah))
					   (= (match-beginning 0) 0)
					   (= (match-end 0) (length uah))))))
			 (setq killit t))))
		  irchat-ignore-nickname)
	  (if message
	      (mapcar (function 
		       (lambda (kill)
			 (if (string-match (upcase (car kill)) 
					   (upcase message))
			     (setq killit t))))
		      irchat-ignore-message-regexp))
	  killit)))))


(defun irchat-split-string-with-separator (string separator size)
  "Split STRING to list of SIZE sized pieces with SEPARATORs."
  (if (not (stringp string))
      nil
    (if (or (not (stringp separator))
	    (<= size (* 2 (length separator))))
	(list string)
      (let ((s string)
	    (first 1)
	    (r '()))
	(while s
	  (if (<= (+ (length s)
		     (if first 0 (length separator)))
		  size)
	      (progn
		(setq r (cons (concat (if first "" separator) s)
			      r))
		(setq s nil))
	    (let ((l (- size (+ (if first 0 (length separator))
				(length separator)))))
	      (setq r (cons (concat (if first "" separator) 
				    (substring s 0 l)
				    separator)
			    r))
	      (setq s (substring s l (length s)))))
	  (setq first nil))
	(reverse r)))))


(defun irchat-parse-^G-channel-name (chnl)
  "Parse CHNL to pair e.g. #42 -> (#42 . nil) and #42^Gov -> (#42 . +o +v)"
  (if (string-match "^\\([^ ][^ ]*\\)[\007 \t]\\(.*\\)$" chnl)
      (let* ((body (matching-substring chnl 1))
             (args (matching-substring chnl 2))
             (niceargs "")
             (l (length args))
             (i 0))
        (while (< i l)
          (setq niceargs (concat niceargs
                                 (if (> (length niceargs) 0) " +" "+")
                                 (char-to-string (elt args i))))
          (setq i (+ i 1)))
        (if (> (length niceargs) 0)
	    (cons body niceargs)
          (cons body nil)))
    (cons chnl nil)))


(defun irchat-get-^G-channel-flags (chnl)
  "Convert CHNL e.g. #42 -> nil and #42^Go -> +o"
  (cdr (irchat-parse-^G-channel-name chnl)))


(defun irchat-convert-^G-channel-name (chnl)
  "Convert CHNL e.g. #42 -> #42 and #42^Go -> #42 (+o)"
  (let* ((parse (irchat-parse-^G-channel-name chnl))
	 (body (car parse))
	 (flags (cdr parse)))
    (if flags
	(format "%s (%s)" body flags)
      body)))


(defun irchat-insert-to-debug (str prefix)
  "Insert STR to debug buffer prefixed by PREFIX."
  (if (and irchat-debug-buffer (get-buffer irchat-debug-buffer))
      (let ((curbuf (current-buffer)))
	(setq str (irchat-replace-in-string str "\r" "\\r"))
	(setq str (irchat-replace-in-string str "\n\n\n\n\n\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n\n\n\n\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n\n\n\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n\n\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n\n" "\n"))
	(setq str (irchat-replace-in-string str "\n$" ""))
	(set-buffer irchat-debug-buffer)
	(let* ((dbgwin (irchat-get-buffer-window irchat-debug-buffer))
	       (wp (if dbgwin (window-point dbgwin) nil))
	       (pm (point-max)))
	  (save-excursion
	    (goto-char (point-max))
	    (insert (format "%s%s%s\n" 
			    (if (stringp irchat-debugmsg) irchat-debugmsg "")
			    (if (stringp prefix) prefix "")
			    (if (stringp str) str ""))))
	  (if (and wp (>= wp pm))
	      (irchat-scroll-if-visible dbgwin)))
	(set-buffer curbuf)
	t)
    nil))


(defun irchat-Dialogue-insert-message (buffer 
				       absolute-prefix 
				       format-string
				       sender
				       message 
				       &optional channel
				       force)
  "Insert incoming message into dialog buffer(s)."
  (if (and (> (length message) 0)
	   (string-match (concat "^" 
				 "\\("
				 (if irchat-message-split-^C-compat
				     (concat (regexp-quote
					      irchat-message-split-separator)
					     "\\|"
					     (regexp-quote ""))
				   (regexp-quote
				    irchat-message-split-separator))
				 "\\)"
				 "\\(.*\\)")
			 message)
	   (null force))
      (let* ((n (matching-substring message 2))
	     (m (concat "^"
			absolute-prefix
			(regexp-quote (format format-string
					      sender
					      channel))
			" .*\\("
			(if irchat-message-split-^C-compat
			    (concat (regexp-quote
				     irchat-message-split-separator)
				    "\\|"
				    (regexp-quote ""))
			  (regexp-quote
			   irchat-message-split-separator))
		       "\\)$"))
	     (d (concat absolute-prefix
			(format format-string 
				sender 
				channel)
			" ... "
			n
			"\n"))
	     (o (concat "\\("
			(if irchat-message-split-^C-compat
			    (concat (regexp-quote
				     irchat-message-split-separator)
				    "\\|"
				    (regexp-quote ""))
			  (regexp-quote
			   irchat-message-split-separator))
			"\\)"
			"$")))
	(irchat-w-replace buffer m d o n 24))
    (irchat-w-insert buffer (concat absolute-prefix
				    (format format-string sender channel)
				    " "
				    message
				    "\n"))))


(defun irchat-Dialogue-buffer-p (buffer)
  "Is BUFFER the irchat-Dialogue-buffer or it's name?" 
  (cond ((and (stringp buffer)
	      (stringp irchat-Dialogue-buffer)
	      (string= buffer irchat-Dialogue-buffer))
	 t)
	((and (bufferp buffer)
	      (stringp irchat-Dialogue-buffer))
	 (equal buffer (get-buffer irchat-Dialogue-buffer)))
	((and (stringp buffer)
	      (bufferp irchat-Dialogue-buffer))
	 (equal (get-buffer buffer) irchat-Dialogue-buffer))
	(t nil)))


(defun irchat-own-message (message)
  (irchat-w-insert (irchat-pick-buffer irchat-current-channel)
		   (format "%s\n" message))
  (if (not irchat-ownfreeze)
      (let ((chan-buffer (irchat-pick-buffer irchat-current-channel)))
	(irchat-freeze-toggle (car chan-buffer))
	(irchat-w-insert chan-buffer "")
	(irchat-freeze-toggle (car chan-buffer)))))


(defun irchat-own-private-message (message)
  (irchat-w-insert irchat-P-buffer (format "%s\n" message)))

(defun irchat-send-privmsg (&rest args)
  "Send the protocol string to the irc server.  Delayed if irchat-use-delayed-privmsg is non-nil."
  (if irchat-use-delayed-privmsg
      (eval (cons (function irchat-send-delayed) args))
    (eval (cons (function irchat-send) args))))

(defun irchat-send (&rest args)
  "Send the protocol string to the irc server."
  (irchat-reset-idle)
  (let ((item (irchat-encode-coding-string (apply 'format args)))
	ditem)
    (let ((conv-list irchat-send-convert-list))
      (while conv-list
        (setq item (irchat-replace-in-string item (car (car conv-list))
					     (car (cdr (car conv-list)))))
        (setq conv-list (cdr conv-list))))
    (let* ((sndstr (concat item "\r"))
	   (len (length sndstr)))
      (if (> len 512)
	  (progn
	    (message (format "Protocol message too long (%d).  Truncated."
			     (length sndstr)))
	    (if irchat-beep-on-bells
		(beep))))
      (irchat-insert-to-debug sndstr "-> ")
      (process-send-string irchat-server-process sndstr))
    (setq ditem (downcase item))
    (if (string-match "^list" (downcase ditem))
        (if (string-match "\\(^list\\) \\(.+\\)" ditem)
            (setq irchat-channel-filter (matching-substring ditem 2))
          (setq irchat-channel-filter "")))))

;;;
;;; Delayed send.  Same as irchat send.  Sends one message per 3 seconds
;;; saves order between other delayed sends, but direct irchat-send
;;; calls can violate the order.  Use this to send replies for example
;;; to ctcp queries that don't need to be sent immediately and can
;;; cause excess flood.  If string is already in the queue, it's not
;;; added there again.
;;;
;;; //tri 19980523
;;;

(defvar irchat-send-delayed-queue '())
(defvar irchat-send-delayed-timer nil)
(defvar irchat-send-delayed-first-commit nil)

(defun irchat-send-delayed-start-timer ()
  "Start delayed send timer.  This is done automagically in delayed send."
  (if irchat-send-delayed-timer
      nil
    (progn
      (setq irchat-send-delayed-first-commit t)
      (setq irchat-send-delayed-timer
	    (irchat-start-timer 'irchat-commit-delayed-send 
				irchat-send-delayed-interval)))))

(defun irchat-send-delayed-reset ()
  "Wipe delayed sends.  This is done automagically in quit."
  (setq irchat-send-delayed-queue '())
  (irchat-send-delayed-cancel-timer))

(defun irchat-send-delayed-cancel-timer ()
  "Cancel delayed send timer.  This is done automagically when all gets sent."
  (if irchat-send-delayed-timer
      (progn
	(irchat-cancel-timer irchat-send-delayed-timer)
	(setq irchat-send-delayed-timer nil)
	t)
    nil))

(defun irchat-commit-delayed-send ()
  "Send one delayed send line.  This is called by timer."
  (if irchat-send-delayed-first-commit
      (setq irchat-send-delayed-first-commit nil)
    (let ((l (length irchat-send-delayed-queue)))
      (if (> l 0)
	  (let ((r (nth (- l 1) irchat-send-delayed-queue))
		(nq '()))
	    (while (> l 1)
	      (setq nq (cons (nth (- l 2) irchat-send-delayed-queue) nq))
	      (setq l (- l 1)))
	    (setq irchat-send-delayed-queue nq)
	    (irchat-send r)))
      (if (null irchat-send-delayed-queue)
	  (irchat-send-delayed-cancel-timer)))))

(defun irchat-send-delayed (format &rest args)
  "Same as irchat-send but queues send."
  (let ((item (apply 'format format args)))
    (setq item (irchat-replace-in-string item "%" "%%"))
    (if (or (not irchat-send-delayed-interval)
	    (< irchat-send-delayed-interval 1))
	(irchat-send item)
      (progn
	(if (or irchat-send-delayed-allow-duplicates
		(not (string-list-memberp item irchat-send-delayed-queue)))
	    (progn
	      (setq irchat-send-delayed-queue (cons item 
						    irchat-send-delayed-queue))
	      (irchat-send-delayed-start-timer))
	  (message (format "Duplicate delayed message not sent \"%s\""
			   (irchat-truncate-string (format item) 38))))))))
    

;;;
;;; end of delayed send stuff...
;;;

(defun irchat-clean-hostname (hostname)
  "Return the arg HOSTNAME, but if is a dotted-quad, put brackets around it."
  (let ((data (match-data)))
    (unwind-protect
	(if (string-match "[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+" hostname)
	    (concat "[" hostname "]")
	  hostname)
      (store-match-data data))))


(defun irchat-current-nickname ()
  "Our current nickname."
  irchat-real-nickname)


(defun irchat-user-on-my-channels (user)
  "Return the list of channels that user (and I) is on."
  (let ((r '())
	(l irchat-current-channels))
    (while l
      (if (irchat-user-on-this-channel user (car l))
	  (setq r (cons (car l) r)))
      (setq l (cdr l)))
    r))


(defun irchat-replace-in-string (str regexp newtext)
  (if (and (string-match "XEmacs" emacs-version)
	   (fboundp 'replace-in-string))
      (replace-in-string str regexp newtext t)
    (save-excursion
      (let ((buf (get-buffer-create "*replace-in-string*")) res)
        (set-buffer buf)
        (erase-buffer)
        (insert str)
        (goto-char (point-min))
        (while (re-search-forward regexp nil t)
          (replace-match newtext t t))
        (setq res (buffer-substring-no-properties (point-min) (point-max)))
        (kill-buffer buf)
        res))))

(defun irchat-scroll-if-visible (window)
  (if window (set-window-point window (point-max))))


(defun irchat-completing-default-read 
  (prompt table predicate require-match initial-input)
  "completing-read w/ default argument like in 'kill-buffer'"
  (let ((default-read
	  (completing-read
	   (if initial-input
	       (format "%s(default %s) "
		       prompt initial-input)
	     (format "%s" prompt))
	   table predicate require-match nil)))
    (if (and (string= default-read "") initial-input)
	initial-input
      default-read)))


(defun irchat-greet-user (user chnl)
  ;; actually user has already been interned, but who cares. 
  (let ((u (intern user irchat-obarray)))
    (if (get u 'irchat-greeting)
	(progn 
	  (beep t)
	  ;; tell me that she has arrived
	  (message "IRCHAT: %s has entered! (%s)"
		   user
		   (if (string= chnl "0") 
		       "on no channel yet"
		     (concat "on channel " chnl)))
	  ;; and send her a greeting
	  (irchat-send "PRIVMSG %s :%s" user (get u 'irchat-greeting))
	  ;; ... done
	  (put u 'irchat-waited-for nil)
	  (put u 'irchat-greeting nil)))))


(defun irchat-greet-author ()
  (setq irchat-greet-author nil)
  (irchat-send "PRIVMSG %s :%s <%s@%s> is using irchat version %s"
	       irchat-author-nickname 
	       (user-full-name) 
	       (user-login-name) 
	       irchat-system-fqdname
	       irchat-version))


(defun irchat-change-nick-of (old new)
  (if (and (stringp old)
	   (stringp new)
	   (not (string-ci-equal old new)))
      (let ((uah-pair (irchat-nick-to-uah-raw old)))
	(if uah-pair
	    (progn
	      (irchat-nick-to-uah-append new 
					 (nth 0 uah-pair)
					 (nth 1 uah-pair))
	      (irchat-nick-to-uah-append old
					 (nth 0 uah-pair)
					 'invalid)))))
  (let ((pair (assoc old irchat-nick-alist)))
    (if new
	(if pair
	    (rplaca pair new)
	  (setq irchat-nick-alist (cons (cons new nil) irchat-nick-alist)))
      (setq irchat-nick-alist (string-list-ci-delete old 
						     irchat-nick-alist)))))


(defun irchat-convert-seconds (time)
  "Convert seconds to printable string."
  (let* ((seconds (string-to-int time))
	 (minutes (/ seconds 60))
	 (seconds (if minutes (% seconds 60) seconds))
	 (hours (/ minutes 60))
	 (minutes (if hours (% minutes 60) minutes))
	 (days (/ hours 24))
	 (hours (if days (% hours 24) hours))
	 (ds (and (/= 0 days)
		  (format "%d day%s, " days
			  (if (> days 1) "s" ""))))
	 (hs (and (/= 0 hours)
		  (format "%d hour%s, " hours
			  (if (> hours 1) "s" ""))))
	 (ms (and (/= 0 minutes)
		  (format "%d minute%s " minutes
			  (if (> minutes 1) "s" ""))))
	 (ss (format "%d seconds" seconds)))
    (concat ds hs ms (if seconds ss ""))))

(defvar irchat-idle-point nil "Timestamp of last idle reset")

(defun irchat-reset-idle ()
  "Reset idle counter and return last idle."
  (let ((r (irchat-idle)))
    (setq irchat-idle-point (current-time))
    r))

(defun irchat-idle ()
  "How long has irchat been idle"
  (if irchat-idle-point
      (irchat-time-difference irchat-idle-point (current-time))
    9999999))

(defun irchat-ping-if-idle (&optional limit)
  (if (null limit)
      (setq limit 120))
  (if (> (irchat-idle) limit)
      (progn
	(irchat-Command-ping)
	t)
    nil))

(defun irchat-msg-from-ignored (prefix rest)
  (save-excursion
    (let ((buf (current-buffer)))
      (irchat-w-insert irchat-I-buffer 
		       (format "%s%s::%s\n" 
			       (if (not irchat-timestamp-irc-I-buffer-p)
				   ""
				 (format "[%s] " 
					 (current-time-string)))
			       prefix 
			       rest))
      t)))

(defun irchat-window-height (win)   ;;; No w-v-h in GNU-Emacs //tri
  "Get the visible window height portably with GNU-Emacs and XEmacs." 
  (if (fboundp 'window-displayed-height)
      (window-displayed-height win)
    (window-height win)))

(defun irchat-time-difference (t0 t1)
  "Difference in seconds of two `three integer lists' returned by current-time function."
  (+ (* (- (car t1) (car t0)) 65536) (- (car (cdr t1)) (car (cdr t0)))))

(defun irchat-time-add (t0 t1)
  "Add t1 seconds to time t0. t0 is in `three integer lists'-format returned by current-time function."
  (list (+ (car t0) (/ (+ (car (cdr t0)) t1) 65536))
	(% (+ (car (cdr t0)) t1) 65536)
	0))

(defun irchat-generate-hex-timestamp (&optional time)
  "Generate timestamp string as hexadecimal"
  (let ((x (if time time (current-time))))
	(format "%04x%04x" (car x) (car (cdr x)))))

(defun irchat-hex-timestamp-valid (timestamp limit)
  "Is TIMESTAMP valid within LIMIT?"
  (if (not (and (stringp timestamp)
		(string-match "^[0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f]$" timestamp)))
      nil
    (let* ((t1 (irchat-hex-to-int (substring timestamp 0 4)))
	   (t2 (irchat-hex-to-int (substring timestamp 4 8)))
	   (diff (irchat-time-difference (list t1 t2 0)
					 (current-time))))
      (if (> limit 0)
	  (if (and (< diff limit)
		   (> diff (- 0 limit)))
	      t
	    nil)
	t))))

(defun irchat-hex-to-int (x)
  "Convert a HEX-STRING like ffff to the decimal integer"
  (if (string-match "^[0-9a-fA-F][0-9a-fA-F]*$" x)
      (let ((i 0)
	    (l (length x))
	    (r 0))
	(while (< i l)
	  (setq r (+ (* 16 r) (irchat-hex-digit-to-int (elt x i))))
	  (setq i (+ i 1)))
	r)
    -1))

(defun irchat-hex-digit-to-int (x)
  "Convert single HEX-DIGIT (char or string) to integer"
  (cond ((= x ?0) 0)
        ((= x ?1) 1)
        ((= x ?2) 2)
        ((= x ?3) 3)
        ((= x ?4) 4)
        ((= x ?5) 5)
        ((= x ?6) 6)
        ((= x ?7) 7)
        ((= x ?8) 8)
        ((= x ?9) 9)
        ((or (= x ?a) (= x ?A)) 10)
        ((or (= x ?b) (= x ?B)) 11)
        ((or (= x ?c) (= x ?C)) 12)
        ((or (= x ?d) (= x ?D)) 13)
        ((or (= x ?e) (= x ?E)) 14)
        ((or (= x ?f) (= x ?F)) 15)
	((string= x "0") 0)
        ((string= x "1") 1)
        ((string= x "2") 2)
        ((string= x "3") 3)
        ((string= x "4") 4)
        ((string= x "5") 5)
        ((string= x "6") 6)
        ((string= x "7") 7)
        ((string= x "8") 8)
        ((string= x "9") 9)
        ((or (string= x "a") (string= x "A")) 10)
        ((or (string= x "b") (string= x "B")) 11)
        ((or (string= x "c") (string= x "C")) 12)
        ((or (string= x "d") (string= x "D")) 13)
        ((or (string= x "e") (string= x "E")) 14)
        ((or (string= x "f") (string= x "F")) 15)
        (t -1)))

(if (not (fboundp 'remassoc))
    (defun remassoc (key lst)
      (let ((r '())
	    (i (length lst)))
	(while (> i 0)
	  (setq i (- i 1))
	  (let ((current (nth i lst)))
	    (if (not (equal (car current) key))
		(setq r (cons current r)))))
	r)))

(if (not (fboundp 'assoc-if))
    (defun assoc-if (pred lst)
      (let ((r nil))
	(while lst
	  (if (eval (list pred (car (car lst))))
	      (progn
		(setq r (car lst))
		(setq lst '()))
	    (setq lst (cdr lst))))
	r)))

(if (not (fboundp '<<))
    (defun << (x n)
      (while (> n 0)
	(setq x (* x 2))
	(setq n (- n 1)))
      x))

(if (not (fboundp '>>))
    (defun >> (x n)
      (while (> n 0)
	(setq x (/ x 2))
	(setq n (- n 1)))
      x))

(defun irchat-split-protocol-args (str)
  "Split irchat protocol argument STR to list of separate arguments."
  (let ((r '())
	(e nil))
    (if (and (stringp str)
	    (> (length str) 0))
	(while (not e)
	  (cond ((and (> (length str) 0) (= (elt str 0) ?:))
		 (setq r (append r (list (substring str 1 (length str)))))
		 (setq e t))
		((string-match "^\\([^ ]*\\) \\(.*\\)" str)
		 (setq r (append r (list (matching-substring str 1))))
		 (setq str (matching-substring str 2)))
		(t
		 (setq r (append r (list str)))
		 (setq e t)))))
    r))

(defun irchat-concat-protocol-args (ipas &optional with-space skip)
  "Concatenate protocol argument list IPAS optionally WITH-SPACE and SKIP." 
  (let ((r ""))
    (if (numberp skip)
	(while (and (> skip 0) ipas)
	  (setq ipas (cdr ipas))
	  (setq skip (- skip 1)))
      (error "fdoo"))
    (while ipas
      (progn
	(setq r (concat r (car ipas)))
	(setq ipas (cdr ipas))
	(if (and ipas with-space)
	    (setq r (concat r " ")))))
    r))

(defun irchat-encode-coding-string (string &optional coding)
  "String encoding for MULE systems (emacs & xemacs 20 with mule option)."
  (let ((coding (if coding coding irchat-send-coding-system)))
    (if (and coding
	     (fboundp 'encode-coding-string))
	(encode-coding-string string coding)
      string)))

(defun irchat-decode-coding-string (string &optional coding)
  "String decoding for MULE systems (emacs & xemacs 20 with mule option)."
  (let ((coding (if coding coding irchat-receive-coding-system)))
    (if (and coding
	     (fboundp 'decode-coding-string))
	(decode-coding-string string coding)
      string)))

(defun irchat-truncate-string (str limit)
  "Make STR at most LIMIT characters long.  Mark truncation with dots..."
  (if (> (length str) limit)
      (if (< limit 3)
	  (substring str 0 limit)
	(if (< limit 8)
	    (concat (substring str 0 (- limit 2)) "..")
	  (concat (substring str 0 (- limit 3)) "...")))
    str))

(defun irchat-crypt-support-p ()
  "Is crypto support included?"
  (and (boundp 'irchat-crypt-support)
       irchat-crypt-support))

(defun irchat-format-string (num &optional encrypted)
  (cond ((null num) (if (and encrypted (boundp 'irchat-format-string-e))
			irchat-format-string-e
		      irchat-format-string))
	((eq t num) (if (and encrypted (boundp 'irchat-myformat-string))
			irchat-myformat-string-e
		      irchat-myformat-string))
	((= 0 num) (if (and encrypted (boundp 'irchat-format-string0-e))
		       irchat-format-string0-e
		     irchat-format-string0))
	((= 1 num) (if (and encrypted (boundp 'irchat-format-string1-e))
		       irchat-format-string1-e
		     irchat-format-string1))
	((= 2 num) (if (and encrypted (boundp 'irchat-format-string2-e))
		       irchat-format-string2-e
		     irchat-format-string2))
	((= 3 num) (if (and encrypted (boundp 'irchat-format-string3-e))
		       irchat-format-string3-e
		     irchat-format-string3))
	((= 4 num) (if (and encrypted (boundp 'irchat-format-string4-e))
		       irchat-format-string4-e
		     irchat-format-string4))
	((= 5 num) (if (and encrypted (boundp 'irchat-format-string5-e))
		       irchat-format-string5-e
		     irchat-format-string5))
	(t "")))

(defun irchat-set-nth (n list item)
  (if (> n 0)
      (cons (car list) (irchat-set-nth (- n 1) (cdr list) item))
    (cons item (cdr list))))

(defun irchat-list-rnd-int (lst)
  (let ((l lst)
	(r '()))
    (while l
      (let ((c (car l)))
	(setq l (cdr l))
	(if (< (irchat-random-8) 128)
	    (setq r (append r (list c)))
	  (setq r (append (list c) r)))))
    r))

(defun irchat-list-rnd (lst)
  "Return items of LST in random order list.  Doesn't randomize long lists well."
  (irchat-list-rnd-int
   (irchat-list-rnd-int
    (irchat-list-rnd-int
     (irchat-list-rnd-int
      (irchat-list-rnd-int
       (irchat-list-rnd-int
	(irchat-list-rnd-int lst))))))))

(eval-and-compile (provide 'irchat-misc))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-timer.el,v 3.2 1998/05/26 10:35:00 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(defvar irchat-emacs-major-version 
  (string-to-number (substring emacs-version 0 2)))

(if (= irchat-emacs-major-version 18)
    (progn 
      (defun irchat-start-timer (function interval)
	"Timers are not supported under emacs 18."
	nil)
      (defun  irchat-cancel-timer (timer)
	"Timers are not supported under emacs 18."
	nil))
  (if (or (string-match ".*Lucid.*" (emacs-version))
	  (string-match ".*XEmacs.*" (emacs-version)))
      (progn
	(defun irchat-cancel-timer (timer)
	  "Cancel timer."
	  (delete-itimer timer))
	(defun irchat-start-timer (function interval)
	  "Add timer function."
	  (interactive)
	  (if interval
	      (start-itimer "IRCHAT" function interval interval))))
    (progn
;;      (eval-and-compile
;;	(require 'timer))
      (defun irchat-cancel-timer (timer)
	"Cancel timer."
	(cancel-timer timer))
      (defun irchat-start-timer (func interval)
	"Add timer function."
	(interactive)
	(setq timer-dont-exit t)
	(if interval
	    ;; messy implementation as RMS emacs keeps on killing the 
	    ;; timers signalling errors. (actually it may lose timers 
	    ;; without reason as well...)
	    (run-at-time (format "%d sec" interval) 
			 interval
			 '(lambda (f)
			    (run-at-time "1 sec" nil f)) func))))))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-time.el,v 3.1 2002/06/04 15:12:51 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(defun irchat-string-int-div-2 (n)
  "Divide integer represented by STRING by two and return pair of result string and boolean marking nonzero remainder"
  (if (string-match "^[0-9][0-9]*$" n)
      (let ((x '())
	    (y '())
	    (r "")
	    (odd 0))
	(while (string-match "^\\([0-9]*\\)\\([0-9]\\)$" n)
	  (setq x (cons (string-to-int (matching-substring n 2)) x))
	  (setq n (matching-substring n 1)))
	(while x
	  (let ((z (irchat-string-int-div-2-small (car x))))
	    (setq y (cons (car z) y))
	    (if (= 0 (cdr z))
		(setq x (cdr x))
	      (if (null (cdr x))
		  (setq x '()
			odd 1)
		(setq x (cons (+ 10 (car (cdr x))) (cdr (cdr x))))))))
	(while y
	  (setq r (format "%d%s" (car y) r))
	  (setq y (cdr y)))
	(if (string-match "^0*\\([0-9][0-9]*\\)$" r)
	    (setq r (matching-substring r 1)))
	(cons r odd))
    nil))

(defun irchat-string-int-div-2-small (x)
  "Divide INTEGER <20 by two and return pair of result and remainder"
  (nth x (list (cons 0 0) (cons 0 1) (cons 1 0)
	       (cons 1 1) (cons 2 0) (cons 2 1)
	       (cons 3 0) (cons 3 1) (cons 4 0)
	       (cons 4 1) (cons 5 0) (cons 5 1)
	       (cons 6 0) (cons 6 1) (cons 7 0)
	       (cons 7 1) (cons 8 0) (cons 8 1)
	       (cons 9 0) (cons 9 1) (cons 10 0))))

(defun irchat-string-to-16bit-int-list (n)
  "Convert STRING integer to list of 16 bit integers"
  (if (string-match "^[0-9][0-9]*$" n)  
      (let ((r '()))
	(while (not (string= n "0"))
	  (let* ((p n)
		 (m (irchat-string-int-div-2 p))
		 (b0 (if (= 0 (cdr m)) 0 1)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b1 (if (= 0 (cdr m)) 0 2)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b2 (if (= 0 (cdr m)) 0 4)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b3 (if (= 0 (cdr m)) 0 8)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b4 (if (= 0 (cdr m)) 0 16)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b5 (if (= 0 (cdr m)) 0 32)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b6 (if (= 0 (cdr m)) 0 64)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b7 (if (= 0 (cdr m)) 0 128)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b8 (if (= 0 (cdr m)) 0 256)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (b9 (if (= 0 (cdr m)) 0 512)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (ba (if (= 0 (cdr m)) 0 1024)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (bb (if (= 0 (cdr m)) 0 2048)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (bc (if (= 0 (cdr m)) 0 4096)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (bd (if (= 0 (cdr m)) 0 8192)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (be (if (= 0 (cdr m)) 0 16384)) (p (car m))
		 (m (irchat-string-int-div-2 p))
		 (bf (if (= 0 (cdr m)) 0 32768)) (p (car m)))
	    (setq n p)
	    (setq r (cons (+ b0 b1 b2 b3 b4 b5 b6 b7 b8 b9 ba bb bc bd be bf)
			  r))))
	r)
    nil))

(defun irchat-time-val-string-to-current-time-format (x)
  "Convert STRING integer time value to current-time format"
  (let ((n (irchat-string-to-16bit-int-list x)))
    (if (null n)
	(list 0 0 0)
      (append n '(0)))))

;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-main.el,v 3.44 2002/11/09 19:01:57 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))
(eval-and-compile  
  (require 'irchat-filter))

;; Define hooks for each IRC message the server might send us.
;; The newer IRC servers use numeric reply codes instead of words.

(defvar irchat-msg-list
  '(channel error invite linreply msg namreply nick ping pong
    privmsg quit topic wall whoreply kill wallops mode kick part join
    200 203 204 205 206 209 211 212 213 214 215 216 217 218 219
    301 311 312 313 314 315 317 318 321 322 323 331 332 341 
    351 352 353 361 364 365 366 367 368 371 372 381 382 391
    401 402 403 411 412 421 431 432 433 441 451 461 462 463 
    464 465 471 472 473 474 481 482 491)
  "A list of the IRC messages and numeric reply codes irchat can handle.")

(mapcar (function
	 (lambda (sym)
	   ;; not interned at irchat-obarray to get completing reads.
	   (eval (list 'defvar (intern (concat "irchat-"
					       (prin1-to-string sym)
					       "-hook"))
		       nil
		       (concat "*A hook that is executed when the IRC "
			       "message \"" (upcase (prin1-to-string sym))
			       "\" is received.
The hook function is called with two arguments, PREFIX and REST-OF-LINE.
It should return non NIL if no further processing of the message is to be
carried out.")))
	   (eval (list 'defvar (intern (concat "irchat-after-"
					       (prin1-to-string sym)
					       "-hook"))
		       nil
		       (concat "*A hook that is executed after the IRC "
			       "message \"" (upcase (prin1-to-string sym))
			       "\" has been received and handled.
The hook function is called with two arguments, PREFIX and REST-OF-LINE.
It should return non NIL if no further processing of the hook is to be
carried out.")))))
	irchat-msg-list)

(defvar irchat-Command-mode-map nil)
(defvar irchat-Dialogue-mode-map nil)
(defvar irchat-Client-query-map nil)
(defvar irchat-Crypt-map nil)

(put 'irchat-Command-mode 'mode-class 'special)
(put 'irchat-Dialogue-mode 'mode-class 'special)

(defvar irchat-Client-query-keys
  '(("v" 	irchat-Command-client-version)
    ("u" 	irchat-Command-client-userinfo)
    ("h" 	irchat-Command-client-help)
    ("b"        irchat-Command-msn-list-online)
    ("B"        irchat-Command-msn-list-lists)
    ("\C-b"     irchat-Command-show-online)
    ("c" 	irchat-Command-client-clientinfo)
    ("d"        irchat-Command-msn-list-discussions)
    ("g" 	irchat-Command-client-generic)
    ("p" 	irchat-Command-client-ping)
    ("x" 	irchat-Command-client-x-face)
    ("X" 	irchat-Command-client-x-face-from-minibuffer)
    ("\C-x" 	irchat-Command-client-x-face-from-commandbuffer)
    ("U" 	irchat-Command-client-userinfo-from-minibuffer)
    ("\C-u" 	irchat-Command-client-userinfo-from-commandbuffer))
  "Key definition table for Client-query-map")
  
(defvar irchat-Dialogue-keys
  '(("\C-?" 	scroll-down)
    ("\C-h" 	scroll-down)
    (" " 	scroll-up)
    ("$" 	end-of-buffer)
    ("/" 	irchat-Command-generic)
    (">" 	end-of-buffer)
    ("<" 	beginning-of-buffer)
    ("a" 	irchat-Command-away)
    ("b"	irchat-Command-msn-send)
    ("B"	irchat-Command-msn-set-status)
    ("f" 	irchat-Dialogue-freeze)
    ("M" 	irchat-Dialogue-ownfreeze)
    ("i" 	irchat-Command-invite)
    ("I" 	irchat-Command-msn-invite)
    ("j" 	irchat-Command-join)
    ("l" 	irchat-Command-load-vars)
    ("s" 	irchat-Command-save-vars)
    ("m" 	irchat-Dialogue-enter-message)
    ("n" 	irchat-Command-names)
    ("o" 	other-window)
    ("p" 	irchat-Command-part)
    ("P" 	irchat-Command-toggle-private)
    ("r" 	irchat-Command-reconfigure-windows)
    ("x" 	irchat-Dialogue-tag-line)
    ("t" 	irchat-Command-find-timestamp)
    ("T" 	irchat-Command-timestamp)
    ("w" 	irchat-Command-who)
    ("1"        irchat-Command-select-channel-1)
    ("2"        irchat-Command-select-channel-2)
    ("3"        irchat-Command-select-channel-3)
    ("4"        irchat-Command-select-channel-4)
    ("5"        irchat-Command-select-channel-5)
    ("6"        irchat-Command-select-channel-6)
    ("7"        irchat-Command-select-channel-7)
    ("8"        irchat-Command-select-channel-8)
    ("9"        irchat-Command-select-channel-9)
    ("0"        irchat-Command-select-channel-10)
    ("\C-m" 	irchat-Command-message)
;    ("%t"	irchat-Command-toggle-crypt)
;    ("%d"	irchat-Command-set-default-key)
;    ("%a"	irchat-Command-add-new-key)
    )
  "Key definition table for Dialogue mode")


(defvar irchat-Crypt-keys
  '(("t"	irchat-Command-toggle-crypt)
    ("k"	irchat-Command-set-default-key)
    ("a"	irchat-Command-add-new-key)
    ("d"	irchat-Command-delete-key))
  "Key definition table for crypt keys (Dialogue and Command mode).")

(defvar irchat-Command-keys
  '(("\C-cg" 	irchat-Command-dcc-receive)
    ("\C-c\C-a" irchat-Command-send-action)
    ("\C-cs" 	irchat-Command-dcc-send)
    ("\C-cG" 	irchat-Command-dcc-list)
    ("\C-m"  	irchat-Command-enter-message)
    ("\C-j"  	irchat-Command-enter-message)
    ("\M-\C-m" 	irchat-Command-enter-message-opposite-crypt-mode)
    ("\M-\C-j" 	irchat-Command-enter-message-opposite-crypt-mode)
    ("\C-cF" 	irchat-Command-send-file)
    ("\C-c\C-c" irchat-Client-query-prefix)
    ("\C-c\C-d" irchat-Command-debug)
    ("\C-c\C-i" irchat-Command-ison)
    ("\C-c\C-l" irchat-Command-redisplay)
    ("\C-c\C-n" irchat-Command-names)
    ("\C-c\C-r" irchat-Command-caesar-line)
    ("\C-c\C-u" irchat-Command-userhost)
    ("\C-c\C-y" irchat-Command-yank-send)
    ("\C-c\C-?" irchat-Command-scroll-down)
    ("\C-c " 	irchat-Command-scroll-up)
    ("\C-c!" 	irchat-Command-exec)
    ("\C-c2" 	irchat-Command-private-conversation)
    ("\C-ca" 	irchat-Command-away)
    ("\C-cb"	irchat-Command-msn-send)
    ("\C-cB"	irchat-Command-msn-set-status)
    ("\C-cc" 	irchat-Command-inline)
    ("\C-cf" 	irchat-Command-finger)
    ("\C-c\C-f" irchat-Command-freeze)
    ("\C-cM"	irchat-Command-ownfreeze)
    ("\C-ci" 	irchat-Command-invite)
    ("\C-cI" 	irchat-Command-msn-invite)
    ("\C-cj" 	irchat-Command-join)
    ("\C-c\C-p" irchat-Command-part)
    ("\C-ck" 	irchat-Command-ignore)
    ("\C-cK" 	irchat-Command-ignore-by-regexp)
    ("\C-c\M-k"	irchat-Command-global-ignore)
    ("\C-c\C-k" irchat-Command-kick)
    ("\C-cl" 	irchat-Command-list)
    ("\C-cL" 	irchat-Command-load-vars)
    ("\C-cS" 	irchat-Command-save-vars)
    ("\C-cm" 	irchat-Command-message)
    ("\C-c\C-m" irchat-Command-modec)
    ("\C-cn" 	irchat-Command-nickname)
    ("\C-cp" 	irchat-Command-mta-private)
    ("\C-cq" 	irchat-Command-quit)
    ("\C-cQ" 	irchat-Command-msn-quit)
    ("\C-cr" 	irchat-Command-reconfigure-windows)
    ("\C-ct" 	irchat-Command-topic)
    ("\C-cT" 	irchat-Command-timestamp)
    ("\C-c\C-t" irchat-Command-find-timestamp)
    ("\C-cP"	irchat-Command-toggle-private)
    ("\C-cu" 	irchat-Command-lusers)
    ("\C-c\C-v"	irchat-Command-scroll-up)
    ("\C-cw" 	irchat-Command-who)
    ("\C-cW" 	irchat-Command-wait)
    ("\C-c\C-x"	irchat-Command-scroll-down)
    ("\C-c|" 	irchat-Command-show-last-kill)
    ("\C-c/" 	irchat-Command-generic)
    ("\C-i" 	irchat-Command-complete)
    ("\C-[\C-i" lisp-complete-symbol)
    ("\C-c$" 	irchat-Command-eod-buffer)
    ("\C-c>" 	irchat-Command-push)
    ("\C-c<" 	irchat-Command-pop))
  "Key definition table for Command mode")
    
(defun irchat-define-keys (map keys)
  (mapcar (function (lambda (keydef)
		      (define-key map (car keydef) (car (cdr keydef)))))
	  keys))

(if irchat-Client-query-map
    nil
  (define-prefix-command 'irchat-Client-query-map)
  (setq irchat-Client-query-map (make-keymap))
  (irchat-define-keys irchat-Client-query-map irchat-Client-query-keys)
  (fset 'irchat-Client-query-prefix irchat-Client-query-map))

(if irchat-Dialogue-mode-map
    nil
  (setq irchat-Dialogue-mode-map (make-keymap))
  (suppress-keymap irchat-Dialogue-mode-map)
  (irchat-define-keys irchat-Dialogue-mode-map irchat-Dialogue-keys))

(if irchat-Crypt-map
    nil
  (setq irchat-Crypt-map (make-sparse-keymap))
  (irchat-define-keys irchat-Crypt-map irchat-Crypt-keys)
  (define-key irchat-Dialogue-mode-map "%" irchat-Crypt-map))

(if irchat-Command-mode-map
    nil
  (setq irchat-Command-mode-map (make-sparse-keymap))
  (irchat-define-keys irchat-Command-mode-map irchat-Command-keys)
  (define-key irchat-Command-mode-map "\C-c%" irchat-Crypt-map)
  (if irchat-want-traditional
      (define-key irchat-Command-mode-map "/" 'irchat-Command-irc-compatible)))

(defun irchat-read-nickname (str)
  (substring str 0 (min 9 (length str))))

;;;
;;;
;;;

(defun irchat-start-server (&optional confirm)
  "Open network stream to remote irc server.
If optional argument CONFIRM is non-nil, ask the host that the server
is running on."
  (if (irchat-server-opened)
      ;; Stream is already opened.
      nil
    ;; Open IRC server.
    (if (or
	 (and confirm
	      (not (eq confirm 'always)))
	 (null irchat-server))
	(setq irchat-server
	      (read-string "IRC server: " irchat-server)))
    (if (and confirm
	     (not (eq confirm 'always))
	     irchat-ask-for-nickname)
	(setq irchat-nickname
	      (irchat-read-nickname
	       (read-string "Enter your nickname: " irchat-nickname))))
    (if (eq confirm 'always)
	(setq irchat-nickname
	      (irchat-read-nickname
	       (concat irchat-nickname irchat-grow-tail))))
    ;; If no server name is given, local host is assumed.
    (if (string-equal irchat-server "")
	(setq irchat-server (system-name)))
    (message "Connecting to IRC server on %s..." irchat-server)
    (cond ((irchat-open-server irchat-server irchat-service))
	  ((and (stringp irchat-status-message-string)
		(> (length irchat-status-message-string) 0))
	   ;; Show valuable message if available.
	   (error irchat-status-message-string))
	  (t (error "Cannot open IRC server on %s" irchat-server)))))


(defun irchat-open-server (host &optional service)
  "Open chat server on HOST.
If HOST is nil, use value of environment variable \"IRCSERVER\".
If optional argument SERVICE is non-nil, open by the service name."
  (let ((host (or host (getenv "IRCSERVER")))
	(status nil))
    (setq irchat-status-message-string "")
    (cond ((and host (irchat-open-server-internal host service))
	   (irchat-send "USER %s %s %s :%s"
			(or (getenv "IRCUSER")
			    (user-real-login-name))
			(or (getenv "IRCHOST")
			    irchat-system-fqdname)
			irchat-server
			(or (getenv "IRCNAME")
			    (getenv "NAME")
			    (user-full-name)))
	   (if (not irchat-real-nickname)
	       (setq irchat-real-nickname irchat-nickname)
	     (let ((match (string-match "[^_]_+$" irchat-real-nickname)))
	       (if (and irchat-autoremove-virtual-penis
			match)
		   (setq irchat-real-nickname
			 (substring irchat-real-nickname 0 (+ match 1))))))
	   (setq irchat-old-nickname irchat-real-nickname)
	   (setq irchat-nick-accepted 'sent)
	   (irchat-send "NICK %s" irchat-real-nickname)
	   ; (irchat-send "PING %s" host)
	   (setq status t)
	   (if (setq status (irchat-wait-for-response ".*"))
	       (progn
		 (set-process-sentinel irchat-server-process
				       'irchat-sentinel)
		 (set-process-filter irchat-server-process
				     'irchat-filter))
	     ;; We have to close connection here, since the function
	     ;;  `irchat-server-opened' may return incorrect status.
	     (irchat-close-server-internal)))
	  ((null host)
	   (setq irchat-status-message-string "IRC server is not specified.")))
    status))


(defun irchat-server-opened ()
  "Return server process status, T or NIL.
If the stream is opened, return T, otherwise return NIL."
  (and irchat-server-process
       (memq (process-status irchat-server-process) '(open run))))


(defun irchat-open-server-internal (host &optional service)
  "Open connection to chat server on HOST by SERVICE (default is irc)."
  (condition-case err 
      (save-excursion
	;; Initialize communication buffer.
	(setq irchat-server-buffer (get-buffer-create " *IRC*"))
	(set-buffer irchat-server-buffer)
	(kill-all-local-variables)
	(irchat-buffer-disable-undo (current-buffer))
	(erase-buffer)
	(setq irchat-server-process
	      (open-network-stream "IRC" (current-buffer)
				   host (or service "irc")))
	(setq irchat-server-name host)
	(run-hooks 'irchat-server-hook)
	;; Return the server process.
	irchat-server-process)
    (error (message (car (cdr err)))
	   nil)))

(defun irchat-close-server ()
  "Close chat server."
  (unwind-protect
      (progn
	;; Un-set default sentinel function before closing connection.
	(and irchat-server-process
	     (eq 'irchat-sentinel
		 (process-sentinel irchat-server-process))
	     (set-process-sentinel irchat-server-process nil))
	;; We cannot send QUIT command unless the process is running.
	(if (irchat-server-opened)
	    (irchat-send "QUIT")))
    (irchat-close-server-internal)))


(defun irchat-close-server-internal ()
  "Close connection to chat server."
  (if irchat-server-process
      (delete-process irchat-server-process))
  (if irchat-server-buffer
      (kill-buffer irchat-server-buffer))
  (setq irchat-server-buffer nil
	irchat-server-process nil))


(defvar irchat-timers nil
  "Symbol name to store timer, timer-function and timer-interval.")
(defvar irchat-timers-list-initialized-p nil
  "Are irchat internal timers in place?")

(defun irchat (&optional confirm)
  "Connect to the IRC server and start chatting.
If optional argument CONFIRM is non-nil, ask which IRC server to connect.
If already connected, just pop up the windows."
  (interactive "P")
  (if (file-exists-p (expand-file-name irchat-variables-file))
      (progn
	(load (expand-file-name irchat-variables-file))
	(irchat-append-obsolete-vars)))
  (if (irchat-server-opened)
      (irchat-configure-windows)
    (if (irchat-crypt-support-p)
	(progn
	  (message "Initializing crypt keys...")
	  (irchat-init-crypt)
	  (message "Initializing crypt keys...done")))
    (unwind-protect
	(progn
	  (switch-to-buffer (irchat-get-buffer-create irchat-Command-buffer))
	  (irchat-Command-mode)
	  (irchat-start-server confirm))
      (if (not (irchat-server-opened))
	  (irchat-Command-quit)
	;; IRC server is successfully open. 
	(setq mode-line-process (format " {%s}" irchat-server))
	(let ((buffer-read-only nil))
	  (if (not irchat-keep-buffers)
	      (erase-buffer))
	  (sit-for 0))
	(if (irchat-frozen (car irchat-D-buffer))
	    (progn
	      (setq irchat-freeze-indicator "-")
	      (irchat-freeze-toggle (car irchat-D-buffer))))
	(irchat-set-crypt-indicator)
	(irchat-Dialogue-setup-buffer)
	(irchat-Private-setup-buffer)
	(irchat-KILLS-setup-buffer)
	(irchat-IGNORED-setup-buffer)
	(irchat-WALLOPS-setup-buffer)
	(if (irchat-crypt-support-p)
	    (irchat-CRYPT-setup-buffer))
	(irchat-configure-windows)
	(setq irchat-current-channels nil)
	(irchat-send-delayed-reset)
	(if irchat-current-channel
	    (irchat-Command-join irchat-current-channel)
	  (if irchat-startup-channel
	      (irchat-Command-join irchat-startup-channel)))
	(if (not (string-equal irchat-awaymsg ""))
	    (irchat-Command-away irchat-awaymsg))
	(run-hooks 'irchat-Startup-hook)

	(if (not irchat-timers-list-initialized-p)
	    (progn
	      (setq irchat-timers   
		    (append irchat-timers
			    (list
			     ;(list nil (function irchat-Command-timestamp)
			     ;      irchat-timestamp-interval)
			     (list nil (function irchat-Command-pollnames) 
				   irchat-pollnames-interval)
			     (list nil (function irchat-Command-keepalive)
				   irchat-keepalive-interval)
			     ;(list nil (function irchat-check-buffers)
			     ;     irchat-checkbuffer-interval)
			     ))
		    irchat-timers-list-initialized-p t)
	      (if (and (fboundp 'irchat-Command-msn-ping-server)
		       (boundp 'irchat-msn-server-ping-interval)
		       (> irchat-msn-server-ping-interval 0))
		  (setq irchat-timers   
			(append irchat-timers
				(list
				 (list nil (function irchat-Command-msn-ping-server)
				       irchat-msn-server-ping-interval)))))))
	(setq
	 irchat-obarray (make-vector irchat-obarray-size nil)
	 irchat-timers
	      (mapcar 
	       (function 
		(lambda (timer)
		  (if (nth 0 timer)
		      (irchat-cancel-timer (nth 0 timer)))
		  (list (irchat-start-timer (nth 1 timer) (nth 2 timer))
			(nth 1 timer)
			(nth 2 timer))))
	       irchat-timers))

	(irchat-maybe-poll)
	(irchat-Command-timestamp)
	(irchat-Command-describe-briefly)))))


(defun irchat-Command-mode ()
  "Major mode for IRCHAT.  Normal edit function are available.
Typing Return or Linefeed enters the current line in the dialogue.
The following special commands are available:
For a list of the generic commands type \\[irchat-Command-generic] ? RET.
\\{irchat-Command-mode-map}"
  (interactive)
  (kill-all-local-variables)

  (irchat-set-crypt-indicator)
  (setq irchat-nick-alist (list (list irchat-nickname))
	mode-line-modified "--- "
	major-mode 'irchat-Command-mode
	mode-name "IRCHAT Commands"
	irchat-privmsg-partner nil
	irchat-private-indicator nil
	irchat-away-indicator "-"
	irchat-freeze-indicator "-"
	irchat-ownfreeze-indicator "-"

	mode-line-format
	'("--- IRCHAT: Commands "
	  irchat-private-indicator
	  "{" irchat-channel-indicator "} "
	  irchat-away-indicator 
	  irchat-crypt-indicator
	  irchat-freeze-indicator
	  irchat-ownfreeze-indicator 
	  "- " 
	  irchat-real-nickname "@" irchat-server 
	  " %-"))

  (if (irchat-frozen (car irchat-D-buffer))
      (irchat-freeze-toggle (car irchat-D-buffer)))

  (use-local-map irchat-Command-mode-map)
  (if irchat-blink-parens
      nil
    (make-variable-buffer-local 'blink-matching-paren)
    (set-default 'blink-matching-paren t)
    (setq blink-matching-paren nil))
  (setq buffer-read-only nil)
  (run-hooks 'irchat-Command-mode-hook))
  

(defun irchat-Dialogue-mode ()
  "Major mode for displaying the IRC dialogue.
All normal editing commands are turned off.
Instead, these commands are available:
\\{irchat-Dialogue-mode-map}"
  (kill-all-local-variables)
  (setq mode-line-modified "--- "
	major-mode 'irchat-Dialogue-mode
	mode-name "IRCHAT Dialogue"

	mode-line-format
	'("--- IRCHAT: Dialogue " 
	  "{" irchat-channel-indicator "} "
	  irchat-away-indicator 
	  irchat-crypt-indicator
	  irchat-freeze-indicator
	  irchat-ownfreeze-indicator 
	  irchat-msn-indicator
	  "-" (-3 . "%p") "-%-"))

  (use-local-map irchat-Dialogue-mode-map)
  (irchat-buffer-disable-undo (current-buffer))
  (if (not irchat-keep-buffers) 
      (erase-buffer))
  (set-buffer irchat-Dialogue-buffer)
  (setq buffer-read-only t)
  (run-hooks 'irchat-Dialogue-mode-hook))


(defun irchat-configure-windows ()
  "Configure Command mode and Dialogue mode windows.
One is for entering commands and text, the other displays the IRC dialogue."
  (if (or (one-window-p t)
	  (null (irchat-get-buffer-window irchat-Command-buffer))
	  (null (irchat-get-buffer-window irchat-Dialogue-buffer)))
      (progn
	(if irchat-command-window-on-top
	    (progn
	      (switch-to-buffer irchat-Command-buffer)
	      (if irchat-use-full-window
		  (delete-other-windows))
	      (if irchat-one-buffer-mode
		  (switch-to-buffer irchat-Dialogue-buffer)
		(split-window-vertically (max window-min-height 
					      irchat-command-window-height))
		(other-window 1)
		(switch-to-buffer irchat-Dialogue-buffer)
		(other-window 1)))
	  ;; mta@tut.fi wants it like this
	  (switch-to-buffer irchat-Dialogue-buffer)
	  (if irchat-use-full-window
	      (delete-other-windows))
	  (if irchat-one-buffer-mode
	      nil
	    (split-window-vertically
	     (- (window-height) (max window-min-height 
				     irchat-command-window-height)))
	    (other-window 1)
	    (switch-to-buffer irchat-Command-buffer))))))

(fset 'irchat-Dialogue-freeze 'irchat-Command-freeze)
(fset 'irchat-Dialogue-ownfreeze 'irchat-Command-ownfreeze)


(defun irchat-Dialogue-tag-line ()
  "Move current line to kill-ring."
  (interactive)
  (save-excursion
    (let (start)
      (beginning-of-line)
      (setq start (point))
      (end-of-line)
      (kill-ring-save start (point)))))


(defun irchat-set-crypt-indicator ()
  "Set crypt mode indicator"
  (if (irchat-crypt-support-p)
      (setq irchat-crypt-indicator
	    (cond ((and irchat-crypt-mode-active
			(not (null irchat-current-channel))
			(irchat-crypt-address-has-default-key-p 
			 irchat-current-channel))
		   "C")
		  (irchat-crypt-mode-active "c")
		  (t "-")))))


(defun irchat-Dialogue-setup-buffer ()
  "Initialize Dialogue mode buffer."
  (or (get-buffer irchat-Dialogue-buffer)
      (save-excursion
	(set-buffer (get-buffer-create irchat-Dialogue-buffer))
	(irchat-Dialogue-mode))))

(defun irchat-Private-setup-buffer ()
  "Initialize Dialogue mode buffer for private messages."
  (or (get-buffer irchat-Private-buffer)
      (save-excursion
	(set-buffer (get-buffer-create irchat-Private-buffer))
	(irchat-Dialogue-mode))))

(defun irchat-KILLS-setup-buffer ()
  "Initialize KILLS buffer."
  (or (get-buffer irchat-KILLS-buffer)
      (save-excursion
	(set-buffer (irchat-get-buffer-create irchat-KILLS-buffer)))))


(defun irchat-IGNORED-setup-buffer ()
  "Initialize IGNORED buffer."
  (or (get-buffer irchat-IGNORED-buffer)
      (save-excursion
	(set-buffer (irchat-get-buffer-create irchat-IGNORED-buffer)))))


(defun irchat-WALLOPS-setup-buffer ()
  "Initialize WALLOPS buffer."
  (or (get-buffer irchat-WALLOPS-buffer)
      (save-excursion
	(set-buffer (irchat-get-buffer-create irchat-WALLOPS-buffer)))))


(defun irchat-CRYPT-setup-buffer ()
  "Initialize CRYPT buffer."
  (or (get-buffer irchat-CRYPT-buffer)
      (save-excursion
	(set-buffer (irchat-get-buffer-create irchat-CRYPT-buffer))
	(irchat-Dialogue-mode))))


(defun irchat-clear-system ()
  "Clear all IRCHAT variables and buffers."
  (interactive)
  (if (and irchat-Command-buffer (get-buffer irchat-Command-buffer))
      (bury-buffer irchat-Command-buffer))
  (if (and irchat-Dialogue-buffer (get-buffer irchat-Dialogue-buffer))
      (bury-buffer irchat-Dialogue-buffer))
  (if (and irchat-KILLS-buffer (get-buffer irchat-KILLS-buffer))
      (bury-buffer irchat-KILLS-buffer))
  (if (and irchat-IGNORED-buffer (get-buffer irchat-IGNORED-buffer))
      (bury-buffer irchat-IGNORED-buffer))
  (if (and irchat-WALLOPS-buffer (get-buffer irchat-WALLOPS-buffer))
      (bury-buffer irchat-WALLOPS-buffer))
  (if (and irchat-debug-buffer (get-buffer irchat-debug-buffer))
      (bury-buffer irchat-debug-buffer))
  (setq irchat-debug-buffer nil
	irchat-channel-indicator "No channel"))


(defun irchat-wait-for-response (regexp)
  "Wait for server response which matches REGEXP."
  (save-excursion
    (let ((status t)
	  (wait t))
      (set-buffer irchat-server-buffer)
      (irchat-accept-response)
      (while wait
	(goto-char (point-min))
	(cond ((looking-at "ERROR")
	       (setq status nil)
	       (setq wait nil))
	      ((looking-at ".")
	       (setq wait nil))
	      (t (irchat-accept-response))))
      ;; Save status message.
      (end-of-line)
      (setq irchat-status-message-string
	    (buffer-substring (point-min) (point)))
      (if status
	  (progn
	    (setq wait t)
	    (while wait
	      (goto-char (point-max))
	      (forward-line -1)		;(beginning-of-line)
	      (if (looking-at regexp)
		  (setq wait nil)
		(message "IRCHAT: Reading...")
		(irchat-accept-response)
		(message "")))
	    ;; Successfully received server response.
	    t)))))

(defun irchat-accept-response ()
  "Read response of server. Only used at startup time"
  ;; To deal with server process exiting before accept-process-output is called.
  (or (memq (process-status irchat-server-process) '(open run))
      (if (not irchat-reconnect-automagic)
	  (error "IRCHAT: Connection closed.")
	(if irchat-grow-tail
	    (irchat 'always)
	  (irchat))))
  (condition-case errorcode
      (accept-process-output irchat-server-process)
    (error
     (cond ((string-equal "select error: Invalid argument" (nth 1 errorcode))
	    ;; Ignore select error.
	    nil)
	   (t (signal (car errorcode) (cdr errorcode)))))))

(defun irchat-maybe-poll ()
  (irchat-Command-ping))

(defun irchat-w-replace (buffer 
			 match
			 defstring
			 oldstring 
			 newstring 
			 &optional treshold)
  (if (not (integerp treshold)) (setq treshold irchat-compress-treshold))
  (if (or (not buffer) (listp buffer))
      (while buffer
	(progn
	  (irchat-w-replace (car buffer) match defstring oldstring newstring)
	  (setq buffer (cdr buffer))))
    (let* ((obuf (current-buffer))
	   (bufintern (intern (format "%s" (irchat-get-buffer-create buffer)) 
			      irchat-obarray))
	   (frozen (irchat-frozen buffer))
	   (spoint nil))
      (set-buffer (get-buffer buffer))
      (goto-char (point-max))
      (let ((buffer-read-only nil))
	(previous-line treshold)
	(if (re-search-forward match nil t)
	    (progn
	      (while (re-search-forward match nil t))
	      (beginning-of-line)
	      (if (re-search-forward oldstring nil t)
		  (replace-match newstring nil t)
		(irchat-w-insert buffer defstring)) ;This should't happen (kny)
	      (irchat-w-insert buffer ""))
	  (irchat-w-insert buffer defstring))))))


(defun irchat-w-insert (buffer string)
  (if (or (not buffer) (listp buffer))
      (while buffer
	(progn
	  (irchat-w-insert (car buffer) string)
	  (setq buffer (cdr buffer))))
    (let* ((obuf (current-buffer))
	   (bufintern) 
	   (nbuf (get-buffer buffer))
	   (frozen)
	   (spoint nil)
	   (oldwstart nil)
	   (oldwpoint nil))
      ;;
      ;; Deleted or nonexistent buffers are (re)created.
      ;;
      (if (not nbuf)
	  (save-excursion
	    (setq nbuf (irchat-get-buffer-create buffer))
	    (set-buffer nbuf)
	    (irchat-Dialogue-mode)))
      ;;
      ;; Check buffers and possibly insert timestamp.
      ;;
      (irchat-check-buffers-if-interval-expired)
      (if (irchat-Dialogue-buffer-p buffer)
	  (irchat-Command-timestamp-if-interval-expired))

      (if (irchat-get-buffer-window obuf)
	  (progn
	    (setq oldwstart (window-start (irchat-get-buffer-window obuf))
		  oldwpoint (window-point (irchat-get-buffer-window obuf)))))
      (setq bufintern (intern (format "%s" nbuf) irchat-obarray))
      (setq frozen (irchat-frozen nbuf))
      (set-buffer nbuf)
      ;;
      ;; have to use this kludge because save-restriction
      ;; doesn't work right, it keeps distance between point
      ;; and point-max a constant, not the distance between
      ;; point and mark, seesh mta Fri Jan  3 12:43:01 EET 1992
      ;;
      (setq spoint (point))
      (goto-char (point-max))
      (let ((buffer-read-only nil))
	(if (not (irchat-is-message-ignored string nbuf))
	    (progn
	      (if (and (or (eq irchat-beep-on-bells 'always)
			   (and irchat-beep-on-bells
				(null
				 (irchat-get-buffer-window (current-buffer)))))
		       (string-match "\007" string)
		       (irchat-Dialogue-buffer-p (current-buffer)))
		  (beep t))
	      (insert string)
	      (if (and (irchat-Dialogue-buffer-p buffer)
		       irchat-msn-partner
		       (not (string-equal string "")))
		  (irchat-Command-msn-send irchat-msn-partner string))
	      (if (and irchat-use-smiley (fboundp 'smiley-region))
		  (smiley-region spoint (point-max)))))
	(goto-char spoint)
	(let ((win-list (irchat-get-buffer-window-list (get-buffer buffer))))
	  (mapcar
	   '(lambda (win)
	      (if (and win (not frozen)
		       (not (pos-visible-in-window-p (point-max) win)))
		  (progn
		    (goto-char (point-max))
		    (if (and (string-match "Xemacs" emacs-version)
			     (not (string-match "^1[0-9]\\." emacs-version)))
			(progn
			  (set-window-point win (point-max))
			  (vertical-motion (/ (- (or irchat-scroll-step
						     (1+ (/ 
							  (irchat-window-height
							   win)
							  2)))
						 (irchat-window-height win)) 2)
					   win)
			  (set-window-start win (point) t)
			  (goto-char (point-max))
			  )
		      (progn
			(vertical-motion (- (or irchat-scroll-step
						(1+ (/ 
						     (irchat-window-height 
						      win)
						     2)))
					    (irchat-window-height win))
					 win)
			(set-window-start win (point))
			(goto-char (point-max))
			)
		      )))
	      )
	   win-list))
	(set-buffer obuf)
	(if (and frozen (irchat-get-buffer-window obuf) (equal obuf nbuf))
	    (progn
	      (set-window-start (irchat-get-buffer-window obuf) oldwstart)
	      (set-window-point (irchat-get-buffer-window obuf) oldwpoint)
	      (goto-char oldwpoint)
	      ))))))


(defun irchat-get-buffer-create (name)
  "Get or create buffer, keep track on its name so we can kill it."
  (if (not (member name irchat-buffer-list))
      (setq irchat-buffer-list (append (list name) irchat-buffer-list)))
  (get-buffer-create name))


(defun irchat-freeze-toggle (buffer)
  (interactive "b")
  (let* ((intern-buffer (intern (format "%s" (get-buffer buffer)) 
				irchat-obarray))
	 (buffer-frozen (get intern-buffer 'frozen)))
    (if buffer-frozen
	(put intern-buffer 'frozen nil)
      (put intern-buffer 'frozen t))))


(defun irchat-frozen (buffer)
  (interactive "b")
  (let* ((intern-buffer (intern (format "%s" (get-buffer buffer)) 
				irchat-obarray))
	 (buffer-frozen (get intern-buffer 'frozen)))
    buffer-frozen))


(defun irchat-ownfreeze-toggle (buffer)
  (interactive "b")
  (let* ((intern-buffer (intern (format "%s" (get-buffer buffer)) 
				irchat-obarray))
	 (buffer-ownfrozen (get intern-buffer 'ownfrozen)))
    (if buffer-ownfrozen
	(put intern-buffer 'ownfrozen nil)
      (put intern-buffer 'ownfrozen t))))


(defun irchat-ownfrozen (buffer)
  (interactive "b")
  (let* ((intern-buffer (intern (format "%s" (get-buffer buffer)) 
				irchat-obarray))
	 (buffer-ownfrozen (get intern-buffer 'ownfrozen)))
    buffer-ownfrozen))


(defvar irchat-last-checkbuffer-time nil "Last time buffers were checked.")


(defun irchat-check-buffers-if-interval-expired ()
  (if (and (numberp irchat-checkbuffer-interval)
           (> irchat-checkbuffer-interval 0)
           (or (null irchat-last-checkbuffer-time)
               (> (irchat-time-difference irchat-last-checkbuffer-time
                                          (current-time))
                  irchat-checkbuffer-interval)))
      (progn
        (irchat-check-buffers)
        t)
    nil))


(defun irchat-check-buffers ()
  (let ((irchat-checkbuffer-interval 0))
    (if (> irchat-buffer-maxsize 0)
	(let ((buflist irchat-buffer-list)
	      (obuf (current-buffer)))
	  (while buflist
	    (set-buffer (irchat-get-buffer-create (car buflist)))
	    (if (< irchat-buffer-maxsize (buffer-size))
		(let ((buffer-read-only nil))
		  (goto-char (- (buffer-size) irchat-buffer-defsize))
		  (forward-line -1)
		  (delete-region (point-min) (point))))
	    (setq buflist (cdr buflist)))
	  (set-buffer obuf))))
  (setq irchat-last-checkbuffer-time (current-time)))


(defun irchat-is-message-ignored (string buffer)
  (let ((mylist irchat-ignore-list)
	(found nil)
	(case-fold-search t))
    (catch 'ignore
      (if (member buffer irchat-no-ignore-buffers)
	  (throw 'ignore t))
      (while mylist
	(let ((msg nil) 
	      (str nil)
	      (who nil)
	      (mymylist (car mylist)))
	  (if (listp (car mymylist))
	      (setq msg (car (car mymylist))
		    str (car (cdr (car mymylist))))
	    (if (fboundp (car mymylist))
		(let ((msgstr (apply (car mymylist) (list string))))
		  (setq msg (car msgstr)
			str (car (cdr msgstr))))
	      (message "Malformed ignore-list, no msg+str function.")))
	  (if (listp (car (cdr mymylist)))
	      (setq who (car (cdr mymylist)))
	    (if (fboundp (car (cdr mymylist)))
		(setq who (apply (car (cdr mymylist)) (list string)))
	      (if (not (car (cdr mylist)))
		  (message "Malformed ignore-list, no user function."))))
	  (if (or msg str)
	      (if (or
		   (and msg (string-match msg (nth 4 irchat-current-function)))
		   (and str (string-match str string)))
		  (let ((mywho who))
		    (while mywho
		      (if (string-match (car mywho) (nth 2 irchat-current-function))
			  (progn
			    (setq found t)
			    (throw 'ignore t))
			)
		      (setq mywho (cdr mywho))))
		)))
	(setq mylist (cdr mylist))))
    found
    )
  )

(eval-and-compile (provide 'irchat))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-uah-cache.el,v 1.8 1997/10/20 08:08:49 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info
;;;
;;; Nick to user-at-host cache by jsl.
;;; Modifications by tri.
;;;
;;; Created              : Mon Mar 10 17:07:39 1997 tri
;;; Integrated to Irchat : Sun Oct 19          1997 tri
;;;

(eval-when-compile (require 'irchat-inlines))

(defconst irchat-nick-to-uah-vector-length 256
  "*Vector length")

(defvar irchat-nick-to-uah-pos 0
  "Current update position")

(defvar irchat-nick-to-uah-vector nil
  "*Vector to store nick to user-at-host mappings")


(defun irchat-nick-to-uah-init ()
  (setq irchat-nick-to-uah-vector
	(make-vector irchat-nick-to-uah-vector-length (list "" "" 'invalid)))
  (setq irchat-nick-to-uah-pos 0))


(defun irchat-nick-to-uah-append (nick uah &optional type)
  "Append NICK UAH tuple into the vector"
  (if (null irchat-nick-to-uah-vector)
      (irchat-nick-to-uah-init))
  (if (null type)
      (setq type 'unknown))
  (if (and (stringp nick)
	   (stringp uah)
	   (> (length nick) 0)
	   (> (length uah) 0))
      (progn
	(aset irchat-nick-to-uah-vector irchat-nick-to-uah-pos 
	      (list nick uah type))
	(setq irchat-nick-to-uah-pos (mod (+ irchat-nick-to-uah-pos 1)
					  irchat-nick-to-uah-vector-length)))))


(defun irchat-nick-to-uah-raw (nick)
  "Find uah and uah-type associated with NICK from cache if possible."
  (if (null irchat-nick-to-uah-vector)
      (irchat-nick-to-uah-init))
  (let ((pos (if (= irchat-nick-to-uah-pos 0)
		 (- irchat-nick-to-uah-vector-length 1)
	       (- irchat-nick-to-uah-pos 1))))
    (while (and (not (string-ci-equal nick (car (elt irchat-nick-to-uah-vector
						     pos))))
		(not (= pos irchat-nick-to-uah-pos)))
      (setq pos (if (= pos 0) 
		    (- irchat-nick-to-uah-vector-length 1)
		  (- pos 1))))
    (let ((n (nth 0 (elt irchat-nick-to-uah-vector pos)))
	  (u (nth 1 (elt irchat-nick-to-uah-vector pos)))
	  (m (nth 2 (elt irchat-nick-to-uah-vector pos))))
    (if (and (string-ci-equal nick n)
	     (not (equal m 'invalid))
	     (stringp u)
	     (> (length u) 0))
	(list u m)
      nil))))

(defun irchat-nick-to-uah (nick)
  "Find uah associated with NICK from cache if possible."
  (let ((r (irchat-nick-to-uah-raw nick)))
    (if (null r)
	nil
      (nth 0 r))))

(defun irchat-convert-uah-to-ignore-list (uah)
  "Convert UAH-string to list of regexps to be ignored."
  (let ((usr nil)
	(dom nil))
    (cond ((string-match "^\\([^ \t@][^ \t@]*\\)@[^ \t@][^ \t@]*\\.\\([^ \t@.][^ \t@.]*\\.[^ \t@.][^ \t@.]*\\)$"
			 uah)
	   (setq usr (substring uah (match-beginning 1) (match-end 1)))
	   (setq dom (substring uah (match-beginning 2) (match-end 2))))
	  ((string-match "^\\([^ \t@][^ \t@]*\\)@\\([^ \t@.][^ \t@.]*\\.[^ \t@.][^ \t@.]*\\)$"
			 uah)
	   (setq usr (substring uah (match-beginning 1) (match-end 1)))
	   (setq dom (substring uah (match-beginning 2) (match-end 2)))))
    (if (and usr dom)
	(progn
	  (setq usr (regexp-quote usr))
	  (setq dom (regexp-quote dom))
	  (list (format "%s@%s" usr dom) (format "%s@.*\\.%s" usr dom)))
      (list (regexp-quote uah)))))

(eval-and-compile (provide 'irchat-uah-cache))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-global-kill.el,v 1.5 1998/10/06 11:47:06 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info
;;;
;;; Handle global kill ctcp messages by jsl and tri.
;;;
;;; Created              : Mon Mar 10 17:07:39 1997 tri
;;; Integrated to Irchat : Sun Oct 19          1997 tri
;;;

(eval-when-compile (require 'irchat-inlines))

;(setq irchat-global-kill-valid-senders 
;      (list ("mypal@.*\\.hut.fi" . 600) ;; max 600 minutes from my pal
;            (".*@.*\\.hut\\.fi" . 30)   ;; max 30 minutes from .hut.fi
;             \".*@.*\\.fi\")            ;; max default (99) minutes from .fi

(defun irchat-remove-from-ignore-list (name)
  "Remove NAME from ignore list, if there."
  (let ((elem (assoc-ci-string name irchat-ignore-nickname)))
    (if elem
	(setq irchat-ignore-nickname
	      (remassoc (car elem) irchat-ignore-nickname))))
  irchat-ignore-nickname)


(defun irchat-global-kill-valid-sender (uah)
  (let ((l irchat-global-kill-valid-senders)
	(r nil))
    (while l
      (let* ((ca (car l))
	     (ua (if (consp ca) (car ca) ca))
	     (lm (if (consp ca) 
		     (cdr ca) 
		   irchat-global-kill-default-timeout-limit)))
	(if (and (string-match ua uah)
		 (> lm 0))
	    (progn
	      (setq l '())
	      (setq r lm))
	  (setq l (cdr l)))))
    r))


(defun irchat-ctl-a-ignore-msg-hook (from rest)
  (if (and (stringp rest)
	   (string-match "^\\(.+\\) \\([0-9][0-9]*\\) \\(.+\\)" rest))
      (let ((nick (matching-substring rest 1))
	    (timeout (string-to-int (matching-substring rest 2)))
	    (reason (matching-substring rest 3))
	    (valid-sender (irchat-global-kill-valid-sender irchat-userathost)))
	(if (string= nick (regexp-quote nick))
	    (if valid-sender
		(let* ((nick-uah (if irchat-global-kill-use-uah-p
				     (irchat-nick-to-uah nick)
				   nil))
		       (ign-name (if nick-uah
				     nick-uah
				   nick))
		       (ign-list (if nick-uah
				     (irchat-convert-uah-to-ignore-list 
				      nick-uah)
				   '()))
		       (real-timeout (if (> timeout valid-sender)
					 valid-sender
				       timeout))
		       (bad-timeout-p (not (= timeout real-timeout))))
		  (irchat-w-insert irchat-D-buffer
				   (format 
			        "*** %s: Ignoring %s%s for %d minutes%s (%s)\n"
				    from
				    ign-name
				    (if (string= nick ign-name)
					""
				      (concat " (" nick ")"))
				    real-timeout
				    (if bad-timeout-p
					(format " [%d minutes requested]"
						timeout)
				      "")
				    reason))
		  (if (not ign-list)
		      (if (not (assoc-ci-string ign-name 
						irchat-ignore-nickname))
			  (irchat-Command-ignore
			   (downcase ign-name)
			   real-timeout 
			   irchat-global-kill-use-silent-ignore))
		    (let ((l ign-list))
		      (while l
			(if (not (assoc-ci-string (car l)
						  irchat-ignore-nickname))
			    (irchat-Command-ignore
			     (downcase (car l))
			     real-timeout
			     irchat-global-kill-use-silent-ignore))
			(setq l (cdr l))))))
	      (let ((msg (format "Rejected IGNORE from %s <%s> (%s, %d, %s)."
				 from irchat-userathost nick timeout reason)))
		(if (or (equal 'minibuffer
			       irchat-global-kill-report-ignored)
			(equal t
			       irchat-global-kill-report-ignored))
		    (message msg))
		(if (or (equal 'dialogue
			       irchat-global-kill-report-ignored)
			(equal t
			       irchat-global-kill-report-ignored))
		    (irchat-w-insert irchat-D-buffer (concat irchat-info-prefix
							     msg
							     "\n"))))))))
  t)


(defun irchat-Command-global-ignore (nickname timeout reason)
  (interactive (let ((nickname-var nil)
		     (timeout 0)
		     (reason ""))
		 (setq nickname
		       (completing-read "Global ignore nickname: "
					irchat-nick-alist
					'(lambda (s) t) nil nil))
		 (if (not (string= "" nickname))
		     (progn
		       (while (not (and (> timeout 0)
					(< timeout 1000)))
			 (setq timeout
			       (string-to-int
				(read-from-minibuffer "Timeout [1-99]: "))))
		       (setq reason (read-from-minibuffer "Reason: "))))
		 (list nickname timeout reason)))

  (if (not (string= "" nickname))
      (let* ((k1 (format "%s-ignore" irchat-current-channel))
	     (k2 (format "%s-kill" irchat-current-channel))
	     (k3 irchat-current-channel)
	     (key (if (assoc-ci-string k1 irchat-default-idea-key-list)
		      k1
		    (if (assoc-ci-string k2 irchat-default-idea-key-list)
			k2
		      (if (assoc-ci-string k3 irchat-default-idea-key-list)
			k3
			nil)))))
	(if (null key)
	    (irchat-w-insert irchat-D-buffer
			     (format 
			      "*** error: no default key (%s, %s or %s) set\n"
			      k1 k2 k3))
	  (let ((msg (irchat-encrypt-message
		      (format "IGNORE %s %d %s" nickname timeout reason)
		      key t)))
	    ; Ignore
	    (let* ((nick-uah (if irchat-global-kill-use-uah-p
				 (irchat-nick-to-uah nickname)
			       nil))
		   (ign-name (if nick-uah
				 nick-uah
			       nickname))
		   (ign-list (if nick-uah
				 (irchat-convert-uah-to-ignore-list 
				  nick-uah)
			       '())))
	      (irchat-w-insert irchat-D-buffer
			       (format 
			     "*** Globally ignoring %s%s for %d minutes (%s)\n"
				ign-name
				(if (string= nickname ign-name)
				    ""
				  (concat " (" nickname ")"))
				timeout
				reason))
	      (if (not ign-list)
		  (progn
		    (irchat-remove-from-ignore-list ign-name)
		    (irchat-Command-ignore (downcase ign-name)
					   timeout
					 irchat-global-kill-use-silent-ignore))
		(let ((l ign-list))
		  (while l
		    (irchat-remove-from-ignore-list (car l))
		    (irchat-Command-ignore (downcase (car l))
					   timeout 
					 irchat-global-kill-use-silent-ignore)
		    (setq l (cdr l)))))
	      ; Send ctcp
	      (irchat-send-privmsg "PRIVMSG %s :%s"
				   irchat-current-channel
				   msg)))))))


(setq irchat-ctl-a-ignore-msg-hook (function irchat-ctl-a-ignore-msg-hook))

(eval-and-compile (provide 'irchat-global-kill))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-obsolete.el,v 1.4 1997/10/20 05:57:17 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))

;;;
;;;  Ignores
;;;
(defvar irchat-kill-nickname nil
  "*Obsolete.  See irchat-ignore-nickname")

(defvar irchat-kill-message-regexp nil
  "*Obsolete.  See irchat-ignore-message-regexp.")

;(defvar irchat-kill-realname nil
;  "*Obsolete.  See irchat-ignore-realname.")

;(defvar irchat-kill-logon nil
;  "*Obsolete.  See irchat-ignore-logon.")

(defun irchat-append-obsolete-vars ()
  "Append obsolete variable values to the current ones."
  (if irchat-kill-nickname
      (progn
	(message "Obsolete variable irchat-kill-nickname converted.")
	(setq irchat-ignore-nickname (append irchat-ignore-nickname
					     irchat-kill-nickname))
	(setq irchat-kill-nickname nil)
	(setq irchat-save-vars-is-dirty t)))
  (if irchat-kill-message-regexp
      (progn
	(message "Obsolete variable irchat-kill-message-regexp converted.")
	(setq irchat-ignore-message-regexp (append irchat-ignore-message-regexp
						   irchat-kill-message-regexp))
	(setq irchat-kill-message-regexp nil)
	(setq irchat-save-vars-is-dirty t))))

(defun irchat-Command-kill (kill-nickname-var &optional timeout silent)
  "Obsolete front end to irchat-Command-ignore."
  (interactive (let ((kill-nickname-var nil)
		     (timeout nil)
		     (completion-ignore-case t))
		 (setq kill-nickname-var 
		       (completing-read "Ignore nickname or regexp: " 
					(append irchat-nick-alist
						irchat-kill-nickname)
					'(lambda (s) t) nil nil))
		 (if (and (not (string= "" kill-nickname-var))
			  (not (assoc-ci-string kill-nickname-var irchat-kill-nickname)))
		     (setq timeout
			   (string-to-int
			    (read-from-minibuffer "Timeout [RET for none]: "))))
		 (list kill-nickname-var timeout)))

  (let ((msg (concat "Obsolete interface irchat-Command-kill."
		     "  "
		     "Use irchat-Command-ignore.")))
    (message msg)
    (irchat-Command-ignore kill-nickname-var timeout silent)
    (message msg)))
  
  

(defun irchat-Command-kill-by-regexp (kill-regexp-var &optional timeout silent)
  "Obsolete front end to irchat-Command-ignore-by-regexp."
  (interactive (let ((kill-regexp-var nil)
		     (timeout nil)
		     (completion-ignore-case t))
		 (setq kill-regexp-var 
		       (completing-read "Ignore messages matching: " 
					irchat-kill-message-regexp
					'(lambda (s) t) nil nil))
		 (if (and (not (string= "" kill-regexp-var))
			  (not (assoc-ci-string kill-regexp-var
						irchat-kill-message-regexp)))
		     (setq timeout
			   (string-to-int
			    (read-from-minibuffer
			     "Timeout [RET for none]: "))))
		 (list kill-regexp-var timeout)))

  (let ((msg (concat "Obsolete interface irchat-Command-kill-by-regexp."
		     "  "
		     "Use irchat-Command-ignore-by-regexp.")))
    (message msg)
    (irchat-Command-ignore-by-regexp kill-regexp-var timeout silent)
    (message msg)))

(defun irchat-Command-global-kill (nickname timeout reason)
  "Obsolete front end to irchat-Command-global-ignore."
  (interactive (let ((nickname-var nil)
		     (timeout 0)
		     (reason ""))
		 (setq nickname
		       (completing-read "Global ignore nickname: "
					irchat-nick-alist
					'(lambda (s) t) nil nil))
		 (if (not (string= "" nickname))
		     (progn
		       (while (not (and (> timeout 0)
					(< timeout 1000)))
			 (setq timeout
			       (string-to-int
				(read-from-minibuffer "Timeout [1-99]: "))))
		       (setq reason (read-from-minibuffer "Reason: "))))
		 (list nickname timeout reason)))

  (let ((msg (concat "Obsolete interface irchat-Command-global-kill."
		     "  "
		     "Use irchat-Command-global-ignore.")))
    (message msg)
    (irchat-Command-global-ignore nickname timeout reason)
    (message msg)))

(eval-and-compile (provide 'irchat-obsolete))
;;;
;;; eof
;;;
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-vars.el,v 3.9 2002/11/09 21:09:56 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

;;;
;;; Public variables.
;;;
(defvar irchat-msn-server '("messenger.hotmail.com")
  "*Primary MSN Messenger server(s).  List can contain many addresses.")
(defvar irchat-msn-service 1863
  "*MSN Messenger port")
(defvar irchat-msn-uid nil
  "*MSN Messenger user id.  Usually a hotmail address.")
(defvar irchat-msn-uid-password nil
  "*MSN Messenger password of the account.  User will be prompted if nil.")
(defvar irchat-msn-auto-add-reverse-contacts t
  "*Add new reverse contact automatically (if not blocked).")
(defvar irchat-msn-show-lists-in-startup nil
  "*If non-nil, various MSN lists are dumped to screen on startup.")
(defvar irchat-msn-send-typing-notifications nil
  "*If non-nil, send typing notification, while writing a message.")
(defvar irchat-msn-show-typing-notifications nil
  "*If non-nil, show incoming typing notifications in message line.")
(defvar irchat-msn-show-status-changes t
  "*If non-nil, show whenever your contact's status changes.")

;;;
;;; Fake version string to messenger server.  Don't change contents 
;;; of this string.  Anyway, you can set this to nil, in order not to
;;; send fake id to the server.  It seems to work OK without it, but
;;; I don't know what may happen, if you do so.
;;;
(defvar irchat-msn-fake-client-version "0x0409 winnt 5.0 i386 MSMSGS 4.6.0042 MSMSGS"
  "*Send this information as our client version.  Don't change this!")
  
;;;
;;; Format strings
;;;
(defvar irchat-msn-info-prefix "$$$ MSN-Info: "
  "*String to add before any MSN informational message")
(defvar irchat-msn-error-prefix "$$$ MSN-Error: "
  "*String to add before any MSN error message")
(defvar irchat-msn-change-prefix "$$$ Change: "
  "*String to add before any change msg.")
(defvar irchat-msn-format-string-in "<%s>"
  "*Format string for incoming message with one parameter.")
(defvar irchat-msn-format-string-out ">%s<"
  "*Format string for outgoing message with one parameter.")
(defvar irchat-msn-format-string-in2 "<%s:%s>"
  "*Format string for incoming message with two parameters.")
(defvar irchat-msn-format-string-out2 ">%s:%s<"
  "*Format string for outgoing message with two parameters.")
;;;
;;; Format strings for encrypted stuff.
;;;
(defvar irchat-msn-format-string-in-e "<<%s>>"
  "*Format string for encrypted incoming message with one parameter.")
(defvar irchat-msn-format-string-out-e ">>%s<<"
  "*Format string for encrypted outgoing message with one parameter.")
(defvar irchat-msn-format-string-in2-e "<<%s:%s>>"
  "*Format string for encrypted incoming message with two parameters.")
(defvar irchat-msn-format-string-out2-e ">>%s:%s<<"
  "*Format string for encrypted outgoing message with two parameters.")

;;;
;;; Variables for Irchat Messenger gateway (kny@iki.fi)
;;;
;;; This extension provides a way to connect to your favourite
;;; irchat session from any MSN Messenger client. You need
;;; two MSN Messenger accounts for this service:
;;; 1. Your primary MSN Messenger account
;;; 2. A private secondary MSN Messenger account for the gateway service
;;;
;;; There are two ways to use this service:
;;; 1. Primary MSN Messenger account used in irchat, secondary account
;;;    used to contact your irchat session from any MSN Messenger client
;;; 2. Secondary MSN Messenger account used in irchat, primary account
;;;    used in any MSN Messenger client
;;;
;;; To initiate a MSN Messenger - Irchat relay you start a discussion
;;; with the MSN Messenger account used within Irchat. All messages
;;; are forwarded both ways. If you prefix your line with a "/" in the
;;; MSN Messenger client, the rest of the line is sent directly to the
;;; irc server. This can be used to send private messages, join and part
;;; channels etc.
;;;

(defvar irchat-msn-gwcontact nil
  "*Contact that is allowed to use the MSN relay (the MSN login name)")

(eval-and-compile (provide 'irchat-msn-vars))

;;; eof (irchat-msn-vars.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-handle.el,v 3.10 2002/09/02 20:28:19 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(defun irchat-msn-handle-generic (parsed msg)
  (cond ((string-match "^[0-9][0-9]*$" (nth 0 parsed))
	 (progn
	   (irchat-w-insert irchat-MSN-buffer 
			    (format "%s%s.\n"
				    irchat-msn-error-prefix
				    (irchat-msn-protocol-error-string (nth 0 parsed))))
	   t))
	(t t)))

(defun irchat-msn-handle-MSG-message (cmd-len pp-uid pp-name len msg)
  t)

(defun irchat-msn-handle-VER (parsed msg)
  (if (eq irchat-msn-connection-phase 'ver-sent)
      (progn
	(irchat-msn-send "INF %d" (irchat-msn-seqno))
	(setq irchat-msn-connection-phase 'inf-sent))
    (irchat-msn-protocol-error)))

(defun irchat-msn-handle-INF (parsed msg)
  (if (eq irchat-msn-connection-phase 'inf-sent)
      (progn
	(irchat-w-insert irchat-MSN-buffer 
			 (format "%sTrying to sign into MSN Messenger Server %s.\n"
				 irchat-msn-info-prefix
				 irchat-msn-server-int))
	(irchat-msn-send "USR %d MD5 I %s" (irchat-msn-seqno) 
			 irchat-msn-uid)
	(setq irchat-msn-connection-phase 'usr-sent))
    (irchat-msn-protocol-error)))

(defun irchat-msn-handle-OUT (parsed msg)
  (if (> (length parsed) 1)
    (cond ((string-equal "OTH" (nth 1 parsed))
	   (progn
	     (setq irchat-msn-server-closed-because-of-another-login t)
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%sMSN Messenger connection closed because of another login.\n"
				      irchat-msn-info-prefix))))
	  ((string-equal "SSD" (nth 1 parsed))
	   (progn
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%sMSN Messenger connection closed server is going down.\n"
				      irchat-msn-info-prefix))))
	  (t
	   (progn
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%sMSN Messenger connection closed (reason=%s).\n"
				      irchat-msn-info-prefix
				      (nth 1 parsed)))))))
  (irchat-msn-close-server))

(defun irchat-msn-handle-NLN (parsed msg)
  (if (> (length parsed) 3)
      (let ((stat (nth 1 parsed))
	    (pp-uid (nth 2 parsed))
	    (pp-name (irchat-msn-decode-name (nth 3 parsed))))
	(let ((user (irchat-search-contact-list-with-name pp-uid
							  irchat-msn-online-list)))
	  (if (null user)
	      (setq user (irchat-search-contact-list-with-name pp-uid
							       irchat-msn-forward-list)))
	  (if (null user)
	      (setq user (list pp-uid pp-name -1 nil)))
	  (setq user (irchat-set-nth 3 user stat))
	  (setq user (irchat-set-nth 1 user pp-name))
	  (setq irchat-msn-online-list (cons user
					     (irchat-remove-from-contact-list-with-name pp-uid
											irchat-msn-online-list)))
	  (irchat-msn-name-cache-add pp-uid pp-name)
	  (if irchat-msn-show-status-changes
	      (irchat-w-insert irchat-MSN-buffer 
			       (format "%sUser %s <%s> status is %s.\n"
				       irchat-msn-info-prefix
				       pp-name
				       pp-uid
				       (irchat-msn-status-string stat)))))))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-ILN (parsed msg)
  (if (> (length parsed) 4)
      (let ((stat (nth 2 parsed))
	    (pp-uid (nth 3 parsed))
	    (pp-name (irchat-msn-decode-name (nth 4 parsed))))
	(let ((user (irchat-search-contact-list-with-name pp-uid
							  irchat-msn-online-list)))
	  (if (null user)
	      (setq user (irchat-search-contact-list-with-name pp-uid
							       irchat-msn-forward-list)))
	  (if (null user)
	      (setq user (list pp-uid pp-name -1 nil)))
	  (setq user (irchat-set-nth 3 user stat))
	  (setq user (irchat-set-nth 1 user pp-name))
	  (setq irchat-msn-online-list (cons user
					     (irchat-remove-from-contact-list-with-name pp-uid
											irchat-msn-online-list)))
	  (irchat-msn-name-cache-add pp-uid pp-name)
	  (if irchat-msn-show-status-changes
	      (irchat-w-insert irchat-MSN-buffer 
			       (format "%sUser %s <%s> status is %s.\n"
				       irchat-msn-info-prefix
				       pp-name
				       pp-uid
				       (irchat-msn-status-string stat)))))))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-FLN (parsed msg)
  (if (> (length parsed) 1)
      (let ((pp-uid (nth 1 parsed)))
	(setq irchat-msn-online-list (irchat-remove-from-contact-list-with-name pp-uid
										irchat-msn-online-list))
	(if irchat-msn-show-status-changes
	    (irchat-w-insert irchat-MSN-buffer 
			     (format "%sUser %s <%s> goes Offline.\n"
				     irchat-msn-info-prefix
				     (irchat-msn-name-cache-get pp-uid)
				     pp-uid)))))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-USR (parsed msg)
  (cond ((and (eq irchat-msn-connection-phase 'usr-sent)
	      (> (length parsed) 4)
	      (string-equal "MD5" (nth 2 parsed))
	      (string-equal "S" (nth 3 parsed)))
	 (let ((salt (nth 4 parsed))
	       (pass (cond ((not (null irchat-msn-password-cache))
			    (let ((r irchat-msn-password-cache))
			      (setq irchat-msn-password-cache nil)
			      r))
			   ((and irchat-msn-uid-password
				 (or (null irchat-msn-last-password)
				     (not (string-equal irchat-msn-uid-password
							irchat-msn-last-password))))
			    irchat-msn-uid-password)
			   (t (irchat-read-passphrase (format "Password for %s: " irchat-msn-uid))))))
	   (setq irchat-msn-connection-phase 'usr-pass-sent)
	   (setq irchat-msn-last-password pass)
	   (irchat-msn-send "USR %d MD5 S %s" (irchat-msn-seqno) (md5 (concat salt pass)))))
	((and (or (eq irchat-msn-connection-phase 'usr-sent)
		  (eq irchat-msn-connection-phase 'usr-pass-sent))
	      (> (length parsed) 3)
	      (string-equal "OK" (nth 2 parsed)))
	 (setq irchat-msn-password-cache irchat-msn-last-password)
	 (irchat-msn-send "CHG %d NLN" (irchat-msn-seqno))
	 (setq irchat-msn-online-list '()) ;; Server sends online list after CHG
	 (setq irchat-msn-connection-phase 'chg-sent)
	 (if irchat-msn-fake-client-version
	     (irchat-msn-send "CVR %d %s" (irchat-msn-seqno) irchat-msn-fake-client-version)))
       (t (irchat-msn-protocol-error))))

(defun irchat-msn-handle-XFR (parsed msg)
  (cond ((and (eq irchat-msn-connection-phase 'usr-sent)
	      (> (length parsed) 4)
	      (string-equal "NS" (nth 2 parsed))
	      (string-match "^\\([0-9][0-9.]*[0-9]\\):\\([1-9][0-9]*\\)$" (nth 3 parsed)))
	 (let ((host (matching-substring (nth 3 parsed) 1))
	       (service (string-to-int (matching-substring (nth 3 parsed) 2))))
	   (irchat-w-insert irchat-MSN-buffer
			    (format "%sServer %s redirects the connection to %s.\n"
				    irchat-msn-info-prefix
				    irchat-msn-server-int
				    host))
	   (irchat-msn-close-server)
	   (setq irchat-msn-connection-phase 'xfr-received)
	   (irchat-msn-start-server host service)))
	((and (> (length parsed) 5)
	      (string-equal "SB" (nth 2 parsed))
	      (string-equal "CKI" (nth 4 parsed))
	      (string-match "^\\([0-9][0-9.]*[0-9]\\):\\([1-9][0-9]*\\)$" (nth 3 parsed)))
	 (let ((host (matching-substring (nth 3 parsed) 1))
	       (service (string-to-int (matching-substring (nth 3 parsed) 2)))
	       (hash (nth 5 parsed)))
	   (if irchat-msn-messages-pending-sb
	       (let ((pending (car irchat-msn-messages-pending-sb)))
		 (setq irchat-msn-messages-pending-sb (cdr irchat-msn-messages-pending-sb))
		 (irchat-msn-start-calling-sub-server host service hash pending))
	     t)))
	(t (irchat-msn-protocol-error))))

(defun irchat-msn-handle-911 (parsed msg)
  (cond ((eq irchat-msn-connection-phase 'usr-pass-sent)
	 (progn
	   (setq irchat-msn-connection-phase nil)
	   (irchat-msn-close-server)
	   (irchat-msn-start-server)))
	(t 
	 (progn 
	   (message "MSN: 911, Giving up!")
	   (irchat-msn-close-server)))))

(defun irchat-msn-handle-QRY (parsed msg)
  t)

(defun irchat-msn-handle-CVR (parsed msg)
  t)

(defun irchat-msn-handle-CHL (parsed msg)
  (if (> (length parsed) 2)
      (let ((c1 (nth 2 parsed))
	    (c2 irchat-msn-challenge-cookie))
	(irchat-msn-send-raw "QRY %d %s 32\r\n%s"
			     (irchat-msn-seqno)
			     "msmsgs@msnmsgr.com"
			     (md5 (concat c1 c2))))
    (irchat-msn-protocol-error)))

(defun irchat-msn-handle-CHG (parsed msg)
  (cond ((eq irchat-msn-connection-phase 'chg-sent)
	 (progn
	   (setq irchat-msn-connection-phase 'online)
	   (setq irchat-msn-my-online-mode "Online")
	   (irchat-w-insert irchat-MSN-buffer 
			    (format "%sConnected MSN Messenger Server %s on %s.\n"
				    irchat-msn-info-prefix
				    irchat-msn-server-int
				    (if irchat-format-time-function 
					(apply irchat-format-time-function 
					       (list (current-time-string))) 
				      (current-time-string))))
	   (irchat-msn-send "SYN %d 0" (irchat-msn-seqno))))
	((> (length parsed) 2)
	 (let ((stat (nth 2 parsed)))
	   (let ((statstr (irchat-msn-status-string stat)))
	     (setq irchat-msn-my-online-mode statstr)
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%sYour status changed to %s.\n"
				      irchat-msn-info-prefix
				      statstr)))))
	(t t))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-SYN (parsed msg)
  (setq irchat-msn-lists-in-sync nil))

(defun irchat-msn-handle-LST (parsed msg)
  (cond ((> (length parsed) 7)
	 (let ((lstid (nth 2 parsed))
	       (lstver (string-to-int (nth 3 parsed)))
	       (num (string-to-int (nth 4 parsed)))
	       (tot (string-to-int (nth 5 parsed)))
	       (pp-uid (nth 6 parsed))
	       (pp-name (irchat-msn-decode-name (nth 7 parsed)))
	       (gid (if (nth 8 parsed) (string-to-int (nth 8 parsed)) -1)))
	   (cond ((string-equal "FL" lstid)
		  (progn
		    (if (eq num 1)
			(setq irchat-msn-forward-list '()))
		    (setq irchat-msn-forward-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-forward-list)))
		    t))
		 ((string-equal "RL" lstid)
		  (progn
		    (if (eq num 1)
			(setq irchat-msn-reverse-list '()))
		    (setq irchat-msn-reverse-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-reverse-list)))
		    t))
		 ((string-equal "AL" lstid)
		  (progn
		    (if (eq num 1)
			(setq irchat-msn-allow-list '()))
		    (setq irchat-msn-allow-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-allow-list)))
		    t))
		 ((string-equal "BL" lstid)
		  (progn
		    (if (eq num 1)
			(setq irchat-msn-block-list '()))
		    (setq irchat-msn-block-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-block-list)))
		    t))
		 (t t))))
	((and (> (length parsed) 5)
	      (string-equal "0" (nth 4 parsed))
	      (string-equal "0" (nth 5 parsed)))
	 (cond ((string-equal "FL" (nth 2 parsed))
		(setq irchat-msn-forward-list '()))
	       ((string-equal "RL" (nth 2 parsed))
		(setq irchat-msn-reverse-list '()))
	       ((string-equal "AL" (nth 2 parsed))
		(setq irchat-msn-allow-list '()))
	       ((string-equal "BL" (nth 2 parsed))
		(setq irchat-msn-block-list '()))
	       (t t)))
	(t t))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-ADD (parsed msg)
  (cond ((and (> (length parsed) 5)
	      (string-equal "RL" (nth 2 parsed)))
	 (let ((lstver (string-to-int (nth 3 parsed)))
	       (pp-uid (nth 4 parsed))
	       (pp-name (irchat-msn-decode-name (nth 5 parsed)))
	       (gid -1))
	   (irchat-w-insert irchat-MSN-buffer 
			    (format "%sUser %s <%s> adds you to to the contact list.\n"
				    irchat-msn-info-prefix pp-name pp-uid))
	   (setq irchat-msn-reverse-list (cons (list pp-uid pp-name gid nil)
					       irchat-msn-reverse-list))
	   (if (and irchat-msn-lists-in-sync
		    (null (irchat-search-contact-list-with-name pp-uid
								irchat-msn-allow-list))
		    (null (irchat-search-contact-list-with-name pp-uid
								irchat-msn-block-list)))
	       ;;; If it's not blocked, add it to allow list.
	       (irchat-msn-send "ADD %d AL %s %s" (irchat-msn-seqno) pp-uid pp-uid))
	   (if (and irchat-msn-lists-in-sync
		    irchat-msn-auto-add-reverse-contacts
		    (null (irchat-search-contact-list-with-name pp-uid
								irchat-msn-forward-list))
		    (null (irchat-search-contact-list-with-name pp-uid
								irchat-msn-block-list)))
	       ;;; If auto-contact is on and if it's not blocked, add it to forward list.
	       (irchat-msn-send "ADD %d FL %s %s 0" (irchat-msn-seqno) pp-uid pp-uid))
	   t))
	((> (length parsed) 5)
	 (let ((lstid (nth 2 parsed))
	       (lstver (nth 3 parsed))
	       (pp-uid (nth 4 parsed))
	       (pp-name (nth 5 parsed))
	       (gid (if (nth 6 parsed) (string-to-int (nth 6 parsed)) -1)))
	   (cond ((string-equal "FL" lstid)
		  (progn
		    (irchat-w-insert irchat-MSN-buffer
				     (format "%sUser %s <%s> added to the contact list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-forward-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-forward-list)))
		    t))
		 ((string-equal "AL" lstid)
		  (progn
		    (irchat-w-insert irchat-MSN-buffer
				     (format "%sUser %s <%s> added to the allow list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-allow-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-allow-list)))
		    t))
		 ((string-equal "BL" lstid)
		  (progn
		    (irchat-w-insert irchat-MSN-buffer
				     (format "%sUser %s <%s> added to the block list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-block-list 
			  (cons (list pp-uid pp-name gid nil)
				(irchat-remove-from-contact-list-with-name pp-uid
									   irchat-msn-block-list)))
		    t))
		 (t t))))
	(t t)))

(defun irchat-msn-handle-REM (parsed msg)
  (cond ((> (length parsed) 4)
	 (let ((lstid (nth 2 parsed))
	       (lstver (nth 3 parsed))
	       (pp-uid (nth 4 parsed)))
	   (cond ((string-equal "FL" lstid)
		  (let ((pp-data (irchat-search-contact-list-with-name pp-uid
								       irchat-msn-forward-list))
			(pp-name pp-uid))
		    (if pp-data (setq pp-name (nth 1 pp-data)))
		    (irchat-w-insert irchat-MSN-buffer 
				     (format "%sUser %s <%s> removed from the contact list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-forward-list (irchat-remove-from-contact-list-with-name 
						   pp-uid
						   irchat-msn-forward-list))))
		 ((string-equal "RL" lstid)
		  (let ((pp-data (irchat-search-contact-list-with-name pp-uid
								       irchat-msn-reverse-list))
			(pp-name pp-uid))
		    (if pp-data (setq pp-name (nth 1 pp-data)))
		    (irchat-w-insert irchat-MSN-buffer 
				     (format "%sUser %s <%s> removes you from his contact list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-reverse-list (irchat-remove-from-contact-list-with-name 
						   pp-uid
						   irchat-msn-reverse-list))))
		 ((string-equal "AL" lstid)
		  (let ((pp-data (irchat-search-contact-list-with-name pp-uid
								       irchat-msn-allow-list))
			(pp-name pp-uid))
		    (if pp-data (setq pp-name (nth 1 pp-data)))
		    (irchat-w-insert irchat-MSN-buffer 
				     (format "%sUser %s <%s> removed from the allow list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-allow-list (irchat-remove-from-contact-list-with-name 
						   pp-uid
						   irchat-msn-allow-list))))

		 ((string-equal "BL" lstid)
		  (let ((pp-data (irchat-search-contact-list-with-name pp-uid
								       irchat-msn-block-list))
			(pp-name pp-uid))
		    (if pp-data (setq pp-name (nth 1 pp-data)))
		    (irchat-w-insert irchat-MSN-buffer 
				     (format "%sUser %s <%s> removed from the block list.\n"
					     irchat-msn-info-prefix pp-name pp-uid))
		    (setq irchat-msn-block-list (irchat-remove-from-contact-list-with-name 
						   pp-uid
						   irchat-msn-block-list))))
		 (t t))))
	(t t)))

(defun irchat-msn-handle-RNG (parsed msg)
  (cond	((and (> (length parsed) 6)
	      (string-equal "CKI" (nth 3 parsed))
	      (string-match "\\([0-9][0-9.]*[0-9]\\):\\([1-9][0-9]*\\)" (nth 2 parsed)))
	 (let ((host (matching-substring (nth 2 parsed) 1))
	       (service (string-to-int (matching-substring (nth 2 parsed) 2)))
	       (sid (nth 1 parsed))
	       (hash (nth 4 parsed))
	       (pp-uid (nth 5 parsed))
	       (pp-name (irchat-msn-decode-name (nth 6 parsed))))
	   (irchat-msn-start-answering-sub-server host service (concat hash " " sid))))
	(t t)))

(eval-and-compile (provide 'irchat-msn-handle))

;;; eof (irchat-msn-handle.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-proc.el,v 3.5 2002/09/02 20:28:19 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(defun irchat-msn-filter (process output)
  "Filter function for IRC server process."
  (let ((obuf (current-buffer))
	(data (match-data))
	(output (irchat-decode-coding-string output))
	bol)
    ;;
    ;; C-c C-d creates debug buffer for incoming messages...
    ;;
    (irchat-insert-to-debug output "<-MSN- ")
    (set-buffer (process-buffer process))
    (goto-char (point-max))
    (insert output)
    (goto-char (point-min))
    (if (string-match "\n" output)
	(irchat-msn-handle-message))
    (set-buffer obuf)
    (store-match-data data)))

(defun irchat-msn-sentinel (proc status)
  "Sentinel function for Irchat MSN process."
  (setq irchat-msn-connection-phase nil)
  (message "IRCHAT: MSN connection closed.")
  (irchat-msn-kill-all-conversations)
  (irchat-w-insert irchat-MSN-buffer 
		   (format "%sMSN Messenger connection terminated.\n"
			   irchat-msn-info-prefix))
  (irchat-set-msn-indicator))

(defun irchat-msn-handle-message ()
  "Called when we have at least one line of output from the MSN server."
  (let ((obuf (current-buffer)))
    (while (and (buffer-live-p obuf)
		(irchat-msn-consume-message))
      (if (buffer-live-p obuf)
	  (set-buffer obuf)))))

(defun irchat-msn-consume-message ()
  (let ((s (buffer-substring (point-min) (point-max))))
    (cond ((or (string-match "^\\(MSG \\([^ ][^ ]*\\) \\([^ ]*\\) \\([0-9][0-9]*\\)\r\n\\)"
			     s)
	       (string-match "\\(^MSG \\([^ ][^ ]*\\) \\([^ ]*\\) \\([0-9][0-9]*\\)\n\\)"
			     s))
	   (let ((s1 (matching-substring s 1))
		 (s2 (matching-substring s 2))
		 (s3 (matching-substring s 3))
		 (s4 (matching-substring s 4)))
	     (let ((cmd-len (length s1))
		   (pp-uid s2)
		   (pp-name (irchat-msn-decode-name s3))
		   (len (string-to-int s4)))
	       (setq s (substring s cmd-len (length s)))
	       (if (not (< (length s) len))
		   (let ((msg (substring s 0 len)))
		     (delete-region (point-min) (+ (point-min) cmd-len len))
		     (irchat-msn-handle-MSG-message cmd-len pp-uid pp-name len msg)
		     t)
		 nil))))
	  ((and (not (or (string-match "^MSG [^ ][^ ]* [^ ]* [0-9][0-9]*\r\n" s)
			 (string-match "^MSG [^ ][^ ]* [^ ]* [0-9][0-9]*\n" s)))
		(or (string-match "^\\(\\(.*\\)\r\n\\)" s)
		    (string-match "^\\(\\(.*\\)\n\\)" s)))
	   (let ((msg (matching-substring s 2))
		 (len (length (matching-substring s 1))))
	     (delete-region (point-min) (+ (point-min) len))
	     (if (string-match "^\\([^ \t][^ \t]*\\)[ \t]" msg)
		 (progn
		   (setq cmd (upcase (matching-substring msg 1)))
		   (setq handler (intern (concat "irchat-msn-handle-" cmd))))
	       (progn
		 (setq cmd "")
		 (setq handler nil)))
	     (let ((ml (irchat-msn-proto-msg-parse msg)))
	       (if ml
		   (if (and handler (fboundp handler))
		       (apply handler (list ml msg))
		     (irchat-msn-handle-generic ml msg))))
	     t))
	  (t nil))))

(eval-and-compile (provide 'irchat-msn-proc))

;;; eof (irchat-msn-proc.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-cmd.el,v 3.13 2002/11/09 19:01:57 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(defun irchat-Command-msn-quit ()
  (interactive)
  (let ((running (irchat-msn-server-opened)))
    (irchat-msn-close-server)
    (if running
	(message "MSN Messenger connection closed.")
      (message "MSN Messenger connection is not open."))))

(defun irchat-Command-msn-send-proto ()
  (interactive)
  (if (irchat-msn-server-opened)
      (let ((msg (read-string "MSN Messenger Protocol Message: ")))
	(irchat-msn-send msg))
    (message "Not connected to MSN.")))

(defun irchat-Command-msn-add-contact (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (read-string "Add MSN-Contact: " nil)))
  (if (not (string-match "^[^@ ][^@ ]*@[^@ ][^@ ]*$" name))
      (error "Illegal MSN Name."))
  (if (null (irchat-search-contact-list-with-name name 
						  irchat-msn-forward-list))
      (progn
	(let ((data (irchat-search-contact-list-with-name name
							  irchat-msn-block-list)))
	  (if data
	      (irchat-msn-send "REM %d BL %s" (irchat-msn-seqno) (nth 0 data))))
	(if (null (irchat-search-contact-list-with-name name
							irchat-msn-allow-list))
	    (irchat-msn-send "ADD %d AL %s %s" (irchat-msn-seqno) name name))
	(irchat-msn-send "ADD %d FL %s %s 0" (irchat-msn-seqno) name name))
    (message "%s is already on MSN-Contacts list" name)))

(defun irchat-Command-msn-remove-contact (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (irchat-completing-default-read "Remove MSN-Contact: "
						 (irchat-msn-contact-uids-alist 
						  irchat-msn-forward-list) 
						 '(lambda (s) t) nil nil)))
  (let ((data (irchat-search-contact-list-with-name name
						    irchat-msn-forward-list)))
    (if data
	(irchat-msn-send "REM %d FL %s %d" 
			 (irchat-msn-seqno) 
			 (nth 0 data) 
			 (nth 2 data))
      (message "%s is not on MSN-Contacts list" name))))

(defun irchat-Command-msn-add-block (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (read-string "Add name to MSN block list: " nil)))
  (if (not (string-match "^[^@ ][^@ ]*@[^@ ][^@ ]*$" name))
      (error "Illegal MSN Name."))
  (if (null (irchat-search-contact-list-with-name name 
						  irchat-msn-block-list))
      (progn
	(let ((data (irchat-search-contact-list-with-name name
							  irchat-msn-allow-list)))
	  (if data
	      (irchat-msn-send "REM %d AL %s" (irchat-msn-seqno) (nth 0 data))))
	(irchat-msn-send "ADD %d BL %s %s" (irchat-msn-seqno) name name))
    (message "%s is already on the block list" name))
  t)

(defun irchat-Command-msn-remove-block (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (irchat-completing-default-read "Remove name from MSN block list: "
						 (irchat-msn-contact-uids-alist 
						  irchat-msn-block-list) 
						 '(lambda (s) t) nil nil)))
  (let ((data (irchat-search-contact-list-with-name name
						    irchat-msn-block-list)))
    (if data
	(irchat-msn-send "REM %d BL %s" 
			 (irchat-msn-seqno) 
			 (nth 0 data))
      (message "%s is not on MSN-Contacts list" name))))

(defun irchat-Command-msn-add-allow (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (read-string "Add name to MSN allow list: " nil)))
  (if (not (string-match "^[^@ ][^@ ]*@[^@ ][^@ ]*$" name))
      (error "Illegal MSN Name."))
  (if (null (irchat-search-contact-list-with-name name 
						  irchat-msn-allow-list))
      (progn
	(let ((data (irchat-search-contact-list-with-name name
							  irchat-msn-block-list)))
	  (if data
	      (irchat-msn-send "REM %d BL %s" (irchat-msn-seqno) (nth 0 data))))
	(irchat-msn-send "ADD %d AL %s %s" (irchat-msn-seqno) name name))
    (message "%s is already on the allow list" name))
  t)

(defun irchat-Command-msn-remove-allow (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (irchat-completing-default-read "Remove name from MSN allow list: "
						 (irchat-msn-contact-uids-alist 
						  irchat-msn-allow-list) 
						 '(lambda (s) t) nil nil)))
  (let ((data (irchat-search-contact-list-with-name name
						    irchat-msn-allow-list)))
    (if data
	(irchat-msn-send "REM %d AL %s" 
			 (irchat-msn-seqno) 
			 (nth 0 data))
      (message "%s is not on MSN-Contacts list" name))))

(defun irchat-Command-msn-kill-conversation (&optional name)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null name)
      (setq name (irchat-completing-default-read "Kill conversation: "
						 (irchat-msn-conversation-names-alist)
						 '(lambda (s) t) nil nil)))
  (irchat-msn-kill-conversation name))

(defun irchat-Command-msn-send (&optional recipient msg)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null recipient)
      (setq recipient (irchat-completing-default-read "Send message to: "
						      (append
						       (irchat-msn-conversation-names-alist)
						       (irchat-msn-online-users-alist))
						      '(lambda (s) t) nil irchat-msn-recipient-cache)))
  (let ((p (irchat-msn-sub-server-search-with-name recipient))
	(cmsg ""))
    (if (null p)
	(setq p (irchat-msn-sub-server-search-with-user recipient)))
    (if p
	(progn
	  (if (null msg)
	      (progn
		(if irchat-msn-send-typing-notifications
		    (let ((tn (irchat-msn-make-typing-notification)))
		      (irchat-msn-send-sub-raw (nth 0 p)
					       "MSG %d U %d\r\n%s"
					       (irchat-msn-sub-server-seqno (nth 0 p))
					       (length tn)
					       tn)))
		(setq msg (read-string (format "Message to %s: " recipient)))))
	  (setq cmsg msg)
	  (if (fboundp 'irchat-encrypt-message)
	      (let ((irchat-real-nickname irchat-msn-uid))
		(setq msg (irchat-encrypt-message msg recipient))))
	  (let ((m (irchat-msn-make-message msg (not (string-equal msg cmsg)))))
	    (setq irchat-msn-recipient-cache recipient)
	    (if (not irchat-msn-partner)
		(irchat-w-insert irchat-MSN-MSG-buffer
				 (concat 
				  (format (if (string-equal msg cmsg)
					      irchat-msn-format-string-out
					    irchat-msn-format-string-out-e)
					  (if (> (length (nth 6 p)) 1)
					      (nth 1 p)
					    (nth 0 (nth 6 p))))
				  " "
				  cmsg
				  "\n")))
	    (irchat-msn-send-sub-raw (nth 0 p)
				     "MSG %d A %d\r\n%s"
				     (irchat-msn-sub-server-seqno (nth 0 p))
				     (length m)
				     m)))
      (cond ((irchat-search-contact-list-with-name recipient irchat-msn-online-list)
	     (if (null msg)
		 (setq msg (read-string (format "Message to %s: " recipient))))
	     (setq cmsg msg)
	     (if (fboundp 'irchat-encrypt-message)
		 (let ((irchat-real-nickname irchat-msn-uid))
		   (setq msg (irchat-encrypt-message msg recipient))))
	     (setq irchat-msn-recipient-cache recipient)
	     (setq irchat-msn-messages-pending-sb (append irchat-msn-messages-pending-sb
							  (list (list recipient msg cmsg))))
	     (irchat-msn-send "XFR %s SB" (irchat-msn-seqno)))
	    ((irchat-search-contact-list-with-name recipient irchat-msn-forward-list)
	     (message "User %s is not online." recipient))
	    (t
	     (message "Recipient not found."))))))

(defun irchat-msn-list-list (head list)
  (if list
      (let ((l list))
	(if head
	    (irchat-w-insert irchat-MSN-buffer
			     (format "%s%s\n"
				     irchat-msn-info-prefix
				     head)))
	(while l
	  (irchat-w-insert irchat-MSN-buffer
			   (format "    %s <%s>\n"
				   (nth 1 (car l))
				   (nth 0 (car l))))
	  (setq l (cdr l))))))

(defun irchat-Command-msn-list-lists ()
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (irchat-msn-list-list "Contacts online:" irchat-msn-online-list)
  (irchat-msn-list-list "Contacts list:" irchat-msn-forward-list)
  (irchat-msn-list-list "Reverse contacts list:" irchat-msn-reverse-list)
  (irchat-msn-list-list "Allow list:" irchat-msn-allow-list)
  (irchat-msn-list-list "Block list:" irchat-msn-block-list))

(defun irchat-Command-msn-set-status ()
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (let ((i irchat-msn-status-strings)
	(m '())
	(a '()))
    (while i
      (if (not (string-equal "FLN" (car (car i))))
	  (setq m (cons (car i) m)
		a (cons (list (cdr (car i))) a)))
      (setq i (cdr i)))
    (let ((s (irchat-completing-default-read "Set status to: "
					     a
					     '(lambda (s) t)
					     t
					     (if (and (stringp irchat-msn-my-online-mode) 
						      (string-equal "Online" 
								    irchat-msn-my-online-mode)) 
						 "Away" 
					       "Online"))))
      (if (> (length s) 1)
	  (let ((x nil))
	    (while m
	      (if (string-equal s (cdr (car m)))
		  (setq x (car (car m))
			m '()))
	      (setq m (cdr m)))
	    (if x
		(irchat-msn-send "CHG %d %s" (irchat-msn-seqno) x)))))))

(defun irchat-Command-msn-list-discussion ()
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  t)

(defun irchat-Command-msn-invite (&optional discussion user)
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if (null user)
      (setq user (irchat-completing-default-read "Invite user: "
						 (irchat-msn-online-users-alist)
						 '(lambda (s) t) nil nil)))
  (if (null discussion)
      (setq discussion (irchat-completing-default-read "Discussion: "
						       (irchat-msn-conversation-names-alist)
						       '(lambda (s) t) t nil)))
  (let ((p (irchat-msn-sub-server-search-with-name discussion)))
    (cond ((null p)
	   (message "No such discussion."))
	  ((null (irchat-search-contact-list-with-name user irchat-msn-forward-list))
	   (message "User not found."))
	  ((null (irchat-search-contact-list-with-name user irchat-msn-online-list))
	   (message "User not online."))
	  (t 
	   (irchat-msn-send-sub (nth 0 p)
				"CAL %d %s"
				(irchat-msn-sub-server-seqno (nth 0 p))
				user)))))

(defun irchat-Command-msn-list-discussions ()
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (cond ((< (length irchat-msn-sub-servers) 1)
	 (irchat-w-insert irchat-MSN-buffer 
			  (format "%sNo active discussions.\n"
				  irchat-msn-info-prefix)))
	(t
	 (let ((l irchat-msn-sub-servers))
	   (irchat-w-insert irchat-MSN-buffer 
			    (format "%s%d active discussion%s.\n"
				    irchat-msn-info-prefix
				    (length l)
				    (if (> (length l) 1) "s" "")))
	   (while l
	     (let ((u (nth 6 (car l))))
	       (irchat-w-insert irchat-MSN-buffer 
				(format "%sDiscussion %s (%d guest%s).\n"
					irchat-msn-info-prefix
					(nth 1 (car l))
					(length u)
					(if (> (length u) 1) "s" "")))
	       (while u
		 (irchat-w-insert irchat-MSN-buffer 
				  (format "    %s <%s>\n"
					  (car u)
					  (irchat-msn-name-cache-get (car u))))
		 (setq u (cdr u)))
	       (setq l (cdr l))))))))

(defun irchat-Command-msn-ping-server () 
  (interactive) 
  (cond ((not (irchat-msn-server-opened)) 
         nil)
        ((stringp irchat-msn-fake-client-version) 
         (progn 
           (irchat-msn-send "CVR %d %s"  
                            (irchat-msn-seqno) 
                            irchat-msn-fake-client-version) 
           t))
        ((stringp irchat-msn-my-online-mode) 
         (progn 
           (irchat-msn-send "CHG %d %s"  
                            (irchat-msn-seqno) 
                            (irchat-msn-status-code irchat-msn-my-online-mode)))) 
        (t nil)))

(defun irchat-msn-online-list-string ()
  (let ((l (irchat-list-rnd irchat-msn-online-list)) 
        (m nil)
	(l2 nil))
    (setq l2 l)
    (while l 
      (let ((c (car l)) 
            (x nil)) 
        (setq l (cdr l)) 
	(if (string-equal "NLN" (nth 3 c))
	    (progn
	      (if (not (string-equal (nth 0 c) (nth 1 c))) 
		  (setq x (format "%s <%s>" (nth 1 c) (nth 0 c))) 
		(setq x (nth 0 c))) 
	      (if m 
		  (setq m (concat m ", " x)) 
		(setq m x))))))
    (setq l l2)
    (while l 
      (let ((c (car l)) 
            (x nil)) 
        (setq l (cdr l)) 
	(if (not (string-equal "NLN" (nth 3 c)))
	    (progn
	      (if (not (string-equal (nth 0 c) (nth 1 c))) 
		  (setq x (format "%s <%s> (%s)" (nth 1 c) (nth 0 c) (irchat-msn-status-string (nth 3 c)))) 
		(setq x (format "%s (%s)" (nth 0 c) (irchat-msn-status-string (nth 3 c)))))
	      (if m
		  (setq m (concat m ", " x)) 
		(setq m x))))))
    (if m m "None!")))

(defun irchat-Command-msn-list-online ()
  (interactive)
  (if (not (irchat-msn-server-opened))
      (error "MSN Messenger connection is not open."))
  (if irchat-msn-online-list
      (irchat-w-insert irchat-MSN-buffer 
		       (format "%s%d contact%s online: %s\n"
			       irchat-msn-info-prefix
			       (length irchat-msn-online-list)
			       (if (> (length irchat-msn-online-list) 1) "s" "")
			       (irchat-msn-online-list-string)))
    (irchat-w-insert irchat-MSN-buffer 
		     (format "%sNo contacts online.\n"
			     irchat-msn-info-prefix))))

(defun irchat-Command-show-online () 
  (interactive) 
  (if (not (irchat-msn-server-opened)) 
      (error "MSN Messenger connection is not open.")) 
  (message (irchat-msn-online-list-string)))

(eval-and-compile (provide 'irchat-msn-cmd))

;;; eof (irchat-msn-cmd.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-sub.el,v 3.14 2002/11/09 21:09:56 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(defun irchat-msn-start-answering-sub-server (host service session)
  "Open network stream to remote MSN server."
  (if (not (irchat-msn-server-opened))
      nil
    (message "Connecting to MSN Conversation server on %s..." host)
    (if (irchat-msn-open-answering-sub-server host service session)
	(progn
	  (message "")))))
  
(defun irchat-msn-open-answering-sub-server (host service session)
  (let ((p (irchat-msn-open-sub-server-internal host service)))
    (if p
	(progn
	  (setq irchat-msn-sub-servers (cons p irchat-msn-sub-servers))
	  (irchat-msn-send-sub (nth 0 p)
			       "ANS %d %s %s"
			       (irchat-msn-sub-server-seqno (nth 0 p))
			       irchat-msn-uid
			       session)
	  (irchat-w-insert irchat-MSN-buffer 
			   (format "%sJoined conversation %s according to invitation.\n"
				   irchat-msn-info-prefix (nth 1 p)))
	  (set-process-sentinel (nth 0 p)
				'irchat-msn-sub-sentinel)
	  (set-process-filter (nth 0 p)
	  		      'irchat-msn-sub-filter)
	t)
      (progn
	(irchat-close-sub-server-internal)
	nil))))

(defun irchat-msn-start-calling-sub-server (host service session pending)
  "Open network stream to remote MSN server."
  (if (not (irchat-msn-server-opened))
      nil
    (message "Connecting to MSN Conversation server on %s..." host)
    (if (irchat-msn-open-calling-sub-server host service session pending)
	(progn
	  (message "")))))
  
(defun irchat-msn-open-calling-sub-server (host service session pending)
  (let ((p (irchat-msn-open-sub-server-internal host service)))
    (if p
	(progn
	  (setq p (irchat-set-nth 7 p pending))
	  (setq irchat-msn-sub-servers (cons p irchat-msn-sub-servers))
	  (irchat-msn-send-sub (nth 0 p)
			       "USR %d %s %s"
			       (irchat-msn-sub-server-seqno (nth 0 p))
			       irchat-msn-uid
			       session)
	  (irchat-w-insert irchat-MSN-buffer 
			   (format "%sSet up conversation %s.\n"
				   irchat-msn-info-prefix (nth 1 p)))
	  (set-process-sentinel (nth 0 p)
				'irchat-msn-sub-sentinel)
	  (set-process-filter (nth 0 p)
	  		      'irchat-msn-sub-filter)
	t)
      (progn
	(irchat-close-sub-server-internal)
	nil))))

(defun irchat-msn-open-sub-server-internal (host service)
  "Open connection to chat server on HOST by SERVICE."
  (condition-case err 
      (save-excursion
	(let ((conv-name (irchat-msn-generate-conversation-name))
	      (conv-server host)
	      (conv-service service)
	      (conv-process nil)
	      (conv-buffer nil))
	  (setq conv-buffer (get-buffer-create (concat "*"
						       (irchat-replace-in-string conv-name "\\$" "")
						       "*")))
	  (set-buffer conv-buffer)
	  (kill-all-local-variables)
	  (irchat-buffer-disable-undo (current-buffer))
	  (erase-buffer)
	  (setq conv-process
		(open-network-stream "MSN-SUB" 
				     (current-buffer)
				     conv-server
				     conv-service))
	  (list conv-process conv-name conv-server conv-service conv-buffer 1 '() nil)))
    nil))

(defun irchat-msn-sub-server-search-with-process (process)
  (let ((l irchat-msn-sub-servers)
	(r nil))
    (while l
      (if (eq (car (car l)) process)
	  (setq r (car l)
		l nil)
	(setq l (cdr l))))
    r))

(defun irchat-msn-sub-server-search-with-name (name)
  (let ((l irchat-msn-sub-servers)
	(r nil))
    (while l
      (if (string-equal (nth 1 (car l)) name)
	  (setq r (car l)
		l nil)
	(setq l (cdr l))))
    r))

(defun irchat-msn-sub-server-remove-with-process (process)
  (let ((l irchat-msn-sub-servers)
	(p nil)
	(r nil))
    (while l
      (if (not (eq (car (car l)) process))
	  (setq r (cons (car l) r))
	(setq p (car l)))
      (setq l (cdr l)))
    (setq irchat-msn-sub-servers r)
    p))

(defun irchat-msn-sub-server-seqno (process)
  (let ((p (irchat-msn-sub-server-remove-with-process process)))
    (if (null p)
	(error "Process not found."))
    (let ((seq (nth 5 p)))
      (setq p (irchat-set-nth 5 p (+ seq 1)))
      (setq irchat-msn-sub-servers (cons p irchat-msn-sub-servers))
      seq)))

(defun irchat-msn-sub-sentinel (proc status)
  "Sentinel function for Irchat MSN process."
  (setq irchat-msn-partner nil)
  (let ((p (irchat-msn-sub-server-remove-with-process proc)))
    (if p
	(progn
	  (if (nth 7 p)
	      (irchat-w-insert irchat-MSN-buffer 
			       (format "%sCan't send message "%s" to user %s <%s>\n"
				       irchat-msn-error-prefix
				       (if (nth 2 (nth 7 p)) (nth 2 (nth 7 p)) (nth 1 (nth 7 p)))
				       (irchat-msn-name-cache-get (nth 0 (nth 7 p)))
				       (nth 0 (nth 7 p)))))
	  (kill-buffer (nth 4 p))
	  (irchat-w-insert irchat-MSN-buffer 
			   (format "%sConversation %s terminated.\n"
				   irchat-msn-info-prefix (nth 1 p)))))))

(defun irchat-msn-sub-filter (process output)
  "Filter function for IRC server process."
  (let ((obuf (current-buffer))
	(data (match-data))
	(output (irchat-decode-coding-string output))
	bol)
    (irchat-insert-to-debug output "<-MSN-SUB- ")
    (set-buffer (process-buffer process))
    (goto-char (point-max))
    (insert output)
    (goto-char (point-min))
    (if (string-match "\n" output)
	(irchat-msn-handle-sub-message process))
    (set-buffer obuf)
    (store-match-data data)))

(defun irchat-msn-handle-sub-message (process)
  "Called when we have at least one line of output from the MSN server."
  (let ((obuf (current-buffer)))
    (while (and (buffer-live-p obuf)
		(irchat-msn-sub-consume-message process))
      (if (buffer-live-p obuf)
	  (set-buffer obuf)))))

(defun irchat-msn-sub-consume-message (proc)
  (let ((s (buffer-substring (point-min) (point-max))))
    (cond ((or (string-match "^\\(MSG \\([^ ][^ ]*\\) \\([^ ]*\\) \\([0-9][0-9]*\\)\r\n\\)"
			     s)
	       (string-match "\\(^MSG \\([^ ][^ ]*\\) \\([^ ]*\\) \\([0-9][0-9]*\\)\n\\)"
			     s))
	   (let ((s1 (matching-substring s 1))
		 (s2 (matching-substring s 2))
		 (s3 (matching-substring s 3))
		 (s4 (matching-substring s 4)))
	     (let ((cmd-len (length s1))
		   (pp-uid s2)
		   (pp-name (irchat-msn-decode-name s3))
		   (len (string-to-int s4)))
	       (setq s (substring s cmd-len (length s)))
	       (if (not (< (length s) len))
		   (let ((msg (substring s 0 len)))
		     (delete-region (point-min) (+ (point-min) cmd-len len))
		     (irchat-msn-sub-handle-MSG-message proc cmd-len pp-uid pp-name len msg)
		     t)
		 nil))))
	  ((and (not (or (string-match "^MSG [^ ][^ ]* [^ ]* [0-9][0-9]*\r\n" s)
			 (string-match "^MSG [^ ][^ ]* [^ ]* [0-9][0-9]*\n" s)))
		(or (string-match "^\\(\\(.*\\)\r\n\\)" s)
		    (string-match "^\\(\\(.*\\)\n\\)" s)))
	   (let ((msg (matching-substring s 2))
		 (len (length (matching-substring s 1))))
	     (delete-region (point-min) (+ (point-min) len))
	     (if (string-match "^\\([^ \t][^ \t]*\\)[ \t]" msg)
		 (progn
		   (setq cmd (upcase (matching-substring msg 1)))
		   (setq handler (intern (concat "irchat-msn-sub-handle-" cmd))))
	       (progn
		 (setq cmd "")
		 (setq handler nil)))
	     (let ((ml (irchat-msn-proto-msg-parse msg)))
	       (if ml
		   (if (and handler (fboundp handler))
		       (apply handler (list proc ml msg))
		     (irchat-msn-sub-handle-generic proc ml msg))))
	     t))
	  (t nil))))

(defun irchat-msn-sub-handle-generic (process parsed msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (if p
	(cond ((string-match "^[0-9][0-9]*$" (nth 0 parsed))
	       (progn
		 (irchat-w-insert irchat-MSN-buffer 
				  (format "%s[%s] %s.\n"
					  irchat-msn-error-prefix
					  (nth 1 p)
					  (irchat-msn-protocol-error-string (nth 0 parsed))))
		 t))
	      (t t)))
    nil))

(defun irchat-msn-sub-handle-MSG-message (process cmd-len pp-uid pp-name len msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (if p
	(let ((m (irchat-msn-parse-message msg)))
	  (if irchat-msn-show-typing-notifications
	      (message ""))
	  (cond ((and m
		      (> (length (cdr m)) 0)
		      (irchat-msn-message-header-val "Content-type" m)
		      (string-match "TEXT/PLAIN" (upcase (irchat-msn-message-header-val "Content-type" m))))
		 (progn
		   (setq irchat-msn-recipient-cache (if (> (length (nth 6 p)) 1) (nth 1 p) pp-uid))
		   (irchat-msn-name-cache-add pp-uid pp-name)
		   (let ((partner irchat-msn-partner))
		     (setq irchat-msn-partner nil) ;; Avoid reqursion
		     (irchat-w-insert irchat-MSN-MSG-buffer
				      (concat 
				       (if (> (length (nth 6 p)) 1)
					   (format irchat-msn-format-string-in2 pp-uid (nth 1 p))
					 (format irchat-msn-format-string-in pp-uid))
				       " "
				       (irchat-msn-iso8859-1-to-utf8 (cdr m) t)
				       "\n"))
		     (if partner
			 (let ((cmsg (irchat-msn-iso8859-1-to-utf8
				      (irchat-replace-in-string (cdr m) "\n" " ") t)))
			   (if (string-equal (substring cmsg 0 1) "/")
			       (progn
				 (setq cmsg (substring cmsg 1 (length cmsg)))
				 (irchat-send cmsg))
			     (irchat-Command-send-message cmsg 'cleartext ""
							  ""))))
		     (setq irchat-msn-partner partner))))
		((and m
		      (irchat-msn-message-header-val "Content-type" m)
		      (string-match "TEXT/X-MSMSGSCONTROL" (upcase (irchat-msn-message-header-val "Content-type" m)))
		      (irchat-msn-message-header-val "TypingUser" m)
		      (string-ci-equal pp-uid (irchat-msn-message-header-val "TypingUser" m)))
		 (if irchat-msn-show-typing-notifications
		     (message "MSN-User %s is typing a message." pp-uid)))
		((and m
		      (> (length (cdr m)) 0)
		      (irchat-msn-message-header-val "Content-type" m)
		      (string-match "TEXT/IRCHAT-ENCRYPTED" (upcase (irchat-msn-message-header-val "Content-type" m))))
		 (let ((c (irchat-decrypt-message (cdr m))))
		   (irchat-w-insert irchat-MSN-MSG-buffer
				    (format "%sUnsupported encrypted message from %s <%s>\n"
					    irchat-msn-error-prefix
					    pp-name
					    pp-uid))))
		(t t)))
      nil)))

(defun irchat-msn-sub-handle-BYE (process parsed msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (cond ((and p
		(> (length parsed) 1))
	   (let ((pp-uid (nth 1 parsed)))
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%s%s left %s.\n"
				      irchat-msn-info-prefix 
				      pp-uid
				      (nth 1 p)))
	     (irchat-msn-conversation-remove-user (nth 0 p) pp-uid)
	     (setq p (irchat-msn-sub-server-search-with-process (nth 0 p)))
	     (if (and p (> 1 (length (nth 6 p))))
		 (irchat-msn-kill-conversation (nth 1 p)))))
	  (t nil))))

(defun irchat-msn-sub-handle-JOI (process parsed msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (cond ((and p
		(> (length parsed) 2))
	   (let ((pp-uid (nth 1 parsed))
		 (pp-name (irchat-msn-decode-name (nth 2 parsed))))
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%s%s <%s> joined %s.\n"
				      irchat-msn-info-prefix 
				      pp-name
				      pp-uid
				      (nth 1 p)))
	     (irchat-msn-conversation-add-user (nth 0 p) pp-uid)
	     (setq p (irchat-msn-sub-server-search-with-process process))
	     (let ((pending (if p (nth 7 p) nil)))
	       (if pending
		   (let ((m (irchat-msn-make-message (nth 1 pending) 
						     (not (or (null (nth 2 pending))
							      (string-equal (nth 1 pending) 
									    (nth 2 pending)))))))
		     (irchat-msn-sub-server-remove-with-process (nth 0 p))
		     (setq irchat-msn-sub-servers (cons (irchat-set-nth 7 p nil)
							irchat-msn-sub-servers))
		     (irchat-w-insert irchat-MSN-MSG-buffer
				      (concat 
				       (format 
					(if (or (null (nth 2 pending))
						(string-equal (nth 1 pending)
							      (nth 2 pending)))
					    irchat-msn-format-string-out
					  irchat-msn-format-string-out-e)
					(nth 0 pending))
				       " "
				       (if (null (nth 2 pending))
					   (nth 1 pending)
					 (nth 2 pending))
				       "\n"))
		     (irchat-msn-send-sub-raw (nth 0 p)
					      "MSG %d A %d\r\n%s"
					      (irchat-msn-sub-server-seqno (nth 0 p))
					      (length m)
					      m))))))
	  (t nil))))

(defun irchat-msn-sub-handle-IRO (process parsed msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (cond ((and p
		(> (length parsed) 5))
	   (let ((num (string-to-int (nth 2 parsed)))
		 (tot (string-to-int (nth 3 parsed)))
		 (pp-uid (nth 4 parsed))
		 (pp-name (irchat-msn-decode-name (nth 5 parsed))))
	     (irchat-msn-name-cache-add pp-uid pp-name)
	     (irchat-w-insert irchat-MSN-buffer 
			      (format "%s%s <%s> joined %s.\n"
				      irchat-msn-info-prefix 
				      pp-name
				      pp-uid
				      (nth 1 p)))
	     (if (string-equal pp-uid irchat-msn-gwcontact)
		 (setq irchat-msn-partner pp-uid))
	     (irchat-msn-conversation-add-user (nth 0 p) pp-uid)))
	  (t nil))))

(defun irchat-msn-sub-handle-USR (process parsed msg)
  (let ((p (irchat-msn-sub-server-search-with-process process)))
    (if p
	(let ((pending (nth 7 p)))
	  (if pending
	      (irchat-msn-send-sub (nth 0 p)
				   "CAL %d %s"
				   (irchat-msn-sub-server-seqno (nth 0 p))
				   (nth 0 pending))
	    t))
      t)))

(defun irchat-msn-sub-handle-CAL (process parsed msg)
  t)

(defun irchat-msn-kill-all-conversations ()
  (while irchat-msn-sub-servers
    (irchat-msn-kill-conversation (nth 1 (car irchat-msn-sub-servers)) t)))

(defun irchat-msn-kill-conversation (conversation &optional silent)
  (if (and irchat-msn-partner
	   (string-equal irchat-msn-partner irchat-msn-gwcontact))
      (setq irchat-msn-partner nil))
  (let ((p (irchat-msn-sub-server-search-with-name conversation)))
    (if p
	(progn
	  (irchat-msn-sub-server-remove-with-process (nth 0 p))
	  (set-process-sentinel (nth 0 p) nil)
	  (set-process-filter (nth 0 p) nil)
	  (delete-process (nth 0 p))
	  (kill-buffer (nth 4 p))
	  (if (and irchat-msn-recipient-cache
		   (string-equal irchat-msn-recipient-cache 
				 (nth 1 p)))
	      (setq irchat-msn-recipient-cache nil))
	  (if (null silent)
	      (irchat-w-insert irchat-MSN-buffer 
			       (format "%sConversation %s terminated.\n"
				       irchat-msn-info-prefix (nth 1 p))))
	  t)
      (progn
	(message "No such conversation.")
	nil))))

(defun irchat-msn-conversation-names ()
  (let ((l irchat-msn-sub-servers)
	(r '()))
    (while l
      (setq r (cons (nth 1 (car l)) r)
	    l (cdr l)))
    r))

(defun irchat-msn-conversation-names-alist ()
  (let ((l irchat-msn-sub-servers)
	(r '()))
    (while l
      (setq r (cons (list (nth 1 (car l))) r)
	    l (cdr l)))
    r))

(defun irchat-msn-conversation-add-user (proc user)
  (let ((p (irchat-msn-sub-server-remove-with-process proc)))
    (if p
	(let ((users (nth 6 p)))
	  (setq users (cons user (string-list-ci-delete user users)))
	  (setq p (irchat-set-nth 6 p users))
	  (setq irchat-msn-sub-servers (cons p irchat-msn-sub-servers))))))

(defun irchat-msn-conversation-remove-user (proc user)
  (let ((p (irchat-msn-sub-server-remove-with-process proc)))
    (if p
	(let ((users (nth 6 p)))
	  (setq users (string-list-ci-delete user users))
	  (setq p (irchat-set-nth 6 p users))
	  (setq irchat-msn-sub-servers (cons p irchat-msn-sub-servers))))))

(defun irchat-msn-sub-server-search-with-user (name)
  (let ((l irchat-msn-sub-servers)
	(r nil))
    (while l
      (if (and (eq 1 (length (nth 6 (car l))))
	       (string-equal (upcase (nth 0 (nth 6 (car l)))) (upcase name)))
	  (setq r (car l)
		l nil)
	(setq l (cdr l))))
    r))

(eval-and-compile (provide 'irchat-msn-sub))

;;; eof (irchat-msn-sub.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn-msg.el,v 3.9 2002/11/09 19:40:47 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

;;; This is far from being complete.  But it'll do.
;;; Full table would me annoyingly slow.
(defconst irchat-msn-iso8859-1-to-utf8-table 
  '(("\344" . "\303\244")
    ("\345" . "\303\245")
    ("\366" . "\303\266")
    ("\304" . "\303\204")
    ("\305" . "\303\205")
    ("\326" . "\303\226")
    ("\321" . "\303\221")
    ("\361" . "\303\261")
    ("\350" . "\303\250")
    ("\351" . "\303\251")
    ("\374" . "\303\234")
    ("\334" . "\303\234")
    ("\367" . "\303\267")
    ("\244" . "\302\244")
    ("\247" . "\302\247")
    ("\275" . "\302\275")
    ("\243" . "\302\243")
    ("\241" . "\302\241") 
    ("\242" . "\302\242") 
    ("\245" . "\302\245") 
    ("\251" . "\302\251") 
    ("\253" . "\302\253") 
    ("\256" . "\302\256") 
    ("\261" . "\302\261") 
    ("\262" . "\302\262") 
    ("\263" . "\302\263") 
    ("\265" . "\302\265") 
    ("\271" . "\302\271") 
    ("\273" . "\302\273") 
    ("\277" . "\302\277") 
    ("\260" . "\302\260") 
    ("'"    . "\302\264")     ;;; This would actually be acute sign, but not in latin1
    ("EURO" . "\342\202\254") ;;; Kludge alert, there is no euro sign in latin1
    ))

(defun irchat-msn-possible-utf8-combined-char (character)
  (let ((c (cond ((and (fboundp 'char-to-int)
		       (fboundp 'characterp)
		       (characterp character))
		  (char-to-int character))
		 (t (+ 0 character)))))
    (> c 127)))

(defun irchat-msn-iso8859-1-to-utf8 (msg &optional reverse)
  (let ((m msg)
	(r ""))
    (while (> (length m) 0)
      (let ((l irchat-msn-iso8859-1-to-utf8-table)
	    (d nil))
	(if (not (irchat-msn-possible-utf8-combined-char (elt m 0)))
	    (setq l '()))
	(while l
	  (if (and (not (< (length m) 
			   (length (if reverse (cdr (car l)) (car (car l))))))
		   (string-equal (if reverse (cdr (car l)) (car (car l)))
				 (substring m 0 (length (if reverse (cdr (car l)) (car (car l)))))))
	      (setq r (concat r (if reverse (car (car l)) (cdr (car l))))
		    m (substring m (length (if reverse (cdr (car l)) (car (car l)))) (length m))
		    l '()
		    d t)
	    (setq l (cdr l))))
	(if (null d)
	    (setq r (concat r (substring m 0 1))
		  m (substring m 1 (length m))))))
    r))

(defun irchat-msn-make-typing-notification ()
  (let ((msg (concat "MIME-Version: 1.0\n"
		     "Content-Type: text/x-msmsgscontrol\n"
		     "TypingUser: " irchat-msn-uid "\n"
		     "\n"
		     "\n")))
    (setq msg (irchat-replace-in-string msg "\n" "\r\n"))
    msg))

(defun irchat-msn-make-message (msg &optional encrypted)
  (setq msg (irchat-replace-in-string msg "\r\n" "\n"))
  (setq msg (irchat-msn-iso8859-1-to-utf8 msg))
  (let ((cnt (if encrypted "irchat-encrypted" "plain"))
	(len (length msg)))
    (if (string-equal "\n" (substring msg (- len 1) len))
	(setq msg (substring msg 0 (- len 1))))
    (setq msg (concat "MIME-Version: 1.0\n"
		      "Content-Type: text/" cnt "; charset=UTF-8\n"
		      "X-MMS-IM-Format: FN=Microsoft%20Sans%20Serif; EF=; CO=000000; CS=0; PF=22\n"
		      "\n"
		      msg))
    (setq msg (irchat-replace-in-string msg "\n" "\r\n"))
    msg))

(defun irchat-msn-parse-message (msg)
  (let ((m (irchat-replace-in-string msg "\r" "")))
    (let ((z (string-match "\n\n" m)))
      (if z
	  (let ((head (substring m 0 z))
		(body (substring m (+ z 2) (length m)))
		(headers '()))
	    (while (and (> (length body) 0)
			(string-equal (substring body (- (length body) 1) (length body)) "\n"))
	      (setq body (substring body 0 (- (length body) 1))))
	    (while (and (> (length body) 0)
			(string-equal (substring body 0 1) "\n"))
	      (setq body (substring body 1 (length body))))
	    (if (string-match "\n" body)
		(setq body (irchat-replace-in-string (concat "\n" body) "\n" "\n    ")))
	    (while (or (string-match "^\\(\\([^ \t]*\\):[ \t]*\\([^\n]*\\)\n\\)" head)
		       (string-match "^\\(\\([^ \t]*\\):[ \t]*\\([^\n]*\\)\\)" head))
	      (let ((n (matching-substring head 2))
		    (v (matching-substring head 3))
		    (l (length (matching-substring head 1))))
		(setq head (substring head l (length head))
		      headers (cons (cons n v) headers))))
	    (cons headers body))))))

(defun irchat-msn-message-header-val (header-name parsed-message)
  (let ((l (car parsed-message))
	(h (upcase header-name))
	(r nil))
    (while l
      (if (string-equal h (upcase (car (car l))))
	  (setq r (cdr (car l))
		l '())
	(setq l (cdr l))))
    r))

(eval-and-compile (provide 'irchat-msn-msg))

;;; eof (irchat-msn-meg.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-msn.el,v 3.13 2002/11/09 19:01:57 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

;;;
;;; MSN Messenger Client implementation by tri@iki.fi.
;;;

(eval-and-compile (require 'irchat-msn-vars))
(eval-and-compile (require 'irchat-msn-handle))
(eval-and-compile (require 'irchat-msn-proc))
(eval-and-compile (require 'irchat-msn-cmd))
(eval-and-compile (require 'irchat-msn-sub))
(eval-and-compile (require 'irchat-msn-msg))

;;;
;;; Internal variables
;;;
(defvar irchat-msn-server-ping-interval 900)
(defvar irchat-msn-server-closed-because-of-another-login nil)
(defvar irchat-msn-server-int nil)
(defvar irchat-msn-seqno 0)
(defvar irchat-msn-uid-password-int nil)
(defvar irchat-msn-status-message-string nil)
(defvar irchat-msn-server-buffer nil)
(defvar irchat-msn-server-name nil)
(defvar irchat-msn-server-process nil)
(defvar irchat-msn-connection-phase nil)
(defvar irchat-msn-last-password nil)
(defvar irchat-msn-password-cache nil)
(defvar irchat-msn-recipient-cache nil)
(defvar irchat-msn-lists-in-sync nil)
(defvar irchat-msn-conversation-counter '(0))
(defvar irchat-msn-sub-servers '())
(defvar irchat-msn-messages-pending-sb '())
(defvar irchat-msn-name-cache '())
(defvar irchat-msn-indicator "")
(defvar irchat-msn-my-online-mode nil)
(defvar irchat-msn-online-list '() 
  "*Users in our forward list that are currently online")
(defvar irchat-msn-forward-list '() 
  "*FL Forward List - Users on your contact list.  (FL)")
(defvar irchat-msn-reverse-list '()
  "*RL Reverse List - Users who have you on their contact list.  (RL)")
(defvar irchat-msn-allow-list '()
  "*AL Allow List - Users who are allowed to see your status.  (AL)")
(defvar irchat-msn-block-list '()
  "*BL Block List - Users who are not allowed to see your status.  (BL)")
(defvar irchat-MSN-buffer (list irchat-Dialogue-buffer)
  "*A list of buffers where MSN messages and status info are sent.")
(defvar irchat-MSN-MSG-buffer (list irchat-Dialogue-buffer irchat-Private-buffer)
  "*A list of buffers where MSN messages are sent.")
(defvar irchat-msn-partner nil
  "*Current MSN relay partner, do not set this.")

;;;
;;; Constants
;;;
(defconst irchat-msn-challenge-cookie "Q1P7W2E4J9R8U3S5"
  "Don't change this!!!")
(defconst irchat-msn-status-strings '(("NLN" . "Online")
				      ("FLN" . "Offline")
				      ("HDN" . "Appear Offline")
				      ("IDL" . "Idle")
				      ("AWY" . "Away")
				      ("BSY" . "Busy")
				      ("BRB" . "Be Right Back")
				      ("PHN" . "On the Phone")
				      ("LUN" . "Out to Lunch"))
  "Translate table for status message codes.")

(defconst irchat-msn-protocol-error-codes '((200 . "Syntax error")
					    (201 . "Invalid parameter")
					    (205 . "Invalid user")
					    (206 . "Domain name missing")
					    (207 . "Already logged in")
					    (208 . "Invalid username")
					    (209 . "Invalid fusername")
					    (210 . "User list full")
					    (215 . "User already there")
					    (216 . "User already on list")
					    (217 . "User not online")
					    (218 . "Already in mode")
					    (219 . "User is in the opposite list")
					    (280 . "Switchboard failed")
					    (281 . "Transfer to switchboard failed")
					    (300 . "Required field missing")
					    (302 . "Not logged in")
					    (500 . "Internal server error")
					    (501 . "Database server error")
					    (510 . "File operation failed")
					    (520 . "Memory allocation failed")
					    (600 . "Server is busy")
					    (601 . "Server is unavaliable")
					    (602 . "Peer nameserver is down")
					    (603 . "Database connection failed")
					    (604 . "Server is going down")
					    (707 . "Could not create connection")
					    (711 . "Write is blocking")
					    (712 . "Session is overloaded")
					    (713 . "Too many active users")
					    (714 . "Too many sessions")
					    (715 . "Not expected")
					    (717 . "Bad friend file")
					    (911 . "Authentication failed")
					    (913 . "Not allowed when offline")
					    (920 . "Not accepting new users"))
  "Translate table for protocol error messages sent by the server.")

(defun irchat-msn ()
  (interactive)
  (if (not (fboundp 'md5))
      (progn
	(require 'greed-md5)
	(defun md5 (object)
	  (greed-md5 object))))
  (setq irchat-msn-seqno 0)
  (irchat-msn-start-server))

(defun irchat-msn-start-server (&optional host service)
  "Open network stream to remote MSN server."
  (if (irchat-msn-server-opened)
      (irchat-Command-msn-quit))
  (setq irchat-msn-online-list '())
  (setq irchat-msn-forward-list '())
  (setq irchat-msn-reverse-list '())
  (setq irchat-msn-allow-list '())
  (setq irchat-msn-block-list '())
  (setq irchat-msn-lists-in-sync nil)
  (setq irchat-msn-conversation-counter '(0))
  (setq irchat-msn-sub-servers '())
  (setq irchat-msn-messages-pending-sb '())
  (setq irchat-msn-recipient-cache nil)
  (setq irchat-msn-my-online-mode "Offline")
  (setq irchat-msn-server-closed-because-of-another-login nil)
  (setq irchat-msn-server-int nil)
  (irchat-set-msn-indicator)
  (irchat-msn-name-cache-flush)
  (if host
      (setq irchat-msn-server-int host))
  (if (null irchat-msn-server-int)
      (setq irchat-msn-server-int (irchat-msn-server)))
  (if (null irchat-msn-server-int)
      (setq irchat-msn-server-int
	    (read-string "MSN Messenger server: " irchat-msn-server-int)))
  (if (null irchat-msn-server)
      (setq irchat-msn-server irchat-msn-server-int))
  (if (null irchat-msn-uid)
      (setq irchat-msn-uid (read-string "MSN Messenger UID: " 
					irchat-msn-uid)))
  (if (eq irchat-msn-connection-phase 'xfr-received)
      (message "Transfering MSN Messenger connection to server on %s..." irchat-msn-server-int)
    (message "Connecting to MSN Messenger server on %s..." irchat-msn-server-int))
  (setq irchat-msn-connection-phase nil)
  (cond ((irchat-msn-open-server irchat-msn-server-int (if service service irchat-msn-service))
	 (progn
	   (message "")
	   (setq irchat-msn-connection-phase 'ver-sent)))
	((and (stringp irchat-msn-status-message-string)
	      (> (length irchat-msn-status-message-string) 0))
	 ;; Show valuable message if available.
	 (error irchat-msn-status-message-string))
	(t (error "Cannot open MSN Messenger server on %s" 
		  irchat-msn-server))))

(defun irchat-msn-server-opened ()
  "Return server process status, T or NIL.  T if running."
  (let ((running (and irchat-msn-server-process
		      (memq (process-status irchat-msn-server-process) 
			    '(open run)))))
    running))

(defun irchat-msn-open-server (host &optional service)
  (setq irchat-msn-status-message-string "")
  (if (and host (irchat-msn-open-server-internal host service))
      (progn
	(irchat-msn-send "VER %d MSNP7 MSNP6 MSNP5 MSNP4 CVRO"
			 (irchat-msn-seqno))
	(set-process-sentinel irchat-msn-server-process
			      'irchat-msn-sentinel)
	(set-process-filter irchat-msn-server-process
			    'irchat-msn-filter)
	t)
    (irchat-close-server-internal)
    nil))

(defun irchat-msn-open-server-internal (host &optional service)
  "Open connection to chat server on HOST by SERVICE (default is 1863)."
  (condition-case err 
      (save-excursion
	;; Initialize communication buffer.
	(setq irchat-msn-server-buffer (get-buffer-create " *MSN*"))
	(set-buffer irchat-msn-server-buffer)
	(kill-all-local-variables)
	(irchat-buffer-disable-undo (current-buffer))
	(erase-buffer)
	(setq irchat-msn-server-process
	      (open-network-stream "MSN" (current-buffer)
				   host (or service 1863)))
	(setq irchat-msn-server-name host)
	(run-hooks 'irchat-msn-server-hook)
	;; Return the server process.
	irchat-msn-server-process)
    (error (message (car (cdr err)))
	   nil)))

(defun irchat-msn-close-server ()
  "Close MSN chat server."
  (unwind-protect
      (progn
        ;; Un-set default sentinel function before closing connection.
	(irchat-msn-kill-all-conversations)
        (and irchat-msn-server-process
             (eq 'irchat-msn-sentinel
                 (process-sentinel irchat-msn-server-process))
             (set-process-sentinel irchat-msn-server-process nil))))
  (irchat-msn-close-server-internal))

(defun irchat-msn-close-server-internal ()
  "Close connection to chat server."
  (irchat-insert-to-debug "Closing server" "*MSN* ")
  (setq irchat-msn-connection-phase nil)
  (if irchat-msn-server-process
      (delete-process irchat-msn-server-process))
  (setq irchat-msn-server-process nil)
  (if irchat-server-buffer
      (kill-buffer irchat-msn-server-buffer))
  (setq irchat-msn-server-buffer nil)
  (irchat-set-msn-indicator))

(defun irchat-msn-send (&rest args)
  "Send the protocol string to the MSN server."
  (let ((item (irchat-encode-coding-string (concat (apply 'format args) 
						   "\r"))))
    (irchat-insert-to-debug item "-MSN-> ")
    (process-send-string irchat-msn-server-process (concat item "\n"))))

(defun irchat-msn-send-sub (proc &rest args)
  "Send the protocol string to the MSN server."
  (let ((item (irchat-encode-coding-string (concat (apply 'format args) 
						   "\r"))))
    (irchat-insert-to-debug item "-MSN-SUB-> ")
    (process-send-string proc (concat item "\n"))))

(defun irchat-msn-send-raw (&rest args)
  "Send the protocol string to the MSN server."
  (let ((item (irchat-encode-coding-string (apply 'format args))))
    (irchat-insert-to-debug item "-MSN-> ")
    (process-send-string irchat-msn-server-process item)))

(defun irchat-msn-send-sub-raw (proc &rest args)
  "Send the protocol string to the MSN server."
  (let ((item (irchat-encode-coding-string (apply 'format args))))
    (irchat-insert-to-debug item "-MSN-SUB-> ")
    (process-send-string proc item)))

(defun irchat-msn-seqno ()
  (let ((seqno irchat-msn-seqno))
    (setq irchat-msn-seqno (+ 1 seqno))
    seqno))

(defun irchat-msn-status-string (status)
  "Translate user status code to English."
  (let ((l irchat-msn-status-strings)
	(s (upcase status))
	(r nil))
    (while l
      (if (string-equal (car (car l)) s)
	  (setq r (cdr (car l))
		l nil)
	(setq l (cdr l))))
    (if r
	r
      (format "Unknown Status (%s)" s))))

(defun irchat-msn-status-code (status)
  "Translate user status code to English."
  (let ((l irchat-msn-status-strings)
	(s (upcase status))
	(r nil))
    (while l
      (if (string-equal (upcase (cdr (car l))) s)
	  (setq r (car (car l))
		l nil)
	(setq l (cdr l))))
    r))

(defun irchat-msn-decode-name (name) 
  (let ((r name)) 
    (setq r (irchat-replace-in-string r "%20" " ")) 
    (setq r (irchat-msn-iso8859-1-to-utf8 r t))
    r))
 
(defun irchat-msn-encode-name (name) 
  (let ((r name))   
    (setq r (irchat-replace-in-string r  " " "%20")) 
    r)) 

(defun irchat-search-contact-list-with-name (name lst)
  (let ((l lst)
	(r nil)
	(n (upcase name)))
    (while l
      (if (string-equal (upcase (car (car l))) n)
	  (setq r (car l)
		l nil)
	(setq l (cdr l))))
    r))

(defun irchat-remove-from-contact-list-with-name (name lst)
  (let ((l lst)
	(r nil)
	(n (upcase name)))
    (while l
      (if (not (string-equal (upcase (car (car l))) n))
	  (setq r (cons (car l) r)))
      (setq l (cdr l)))
    r))

(defun irchat-msn-contact-uids (lst)
  (let ((l lst)
	(r '()))
    (while l
      (setq r (cons (car (car l)) r)
	    l (cdr l)))
    r))

(defun irchat-msn-contact-uids-alist (lst)
  (let ((l lst)
	(r '()))
    (while l
      (setq r (cons (list (car (car l))) r)
	    l (cdr l)))
    r))

(defun irchat-msn-conversation-counter-increment ()
  (let ((x irchat-msn-conversation-counter)
	(y '()))
    (while (and x
		(eq (car x) 9))
      (setq y (cons 0 y)
	    x (cdr x)))
    (if x
	(progn
	  (setq y (cons (+ (car x) 1) y)
		x (cdr x))
	  (while x
	    (setq y (cons (car x) y)
		  x (cdr x))))
      (setq y (cons 1 y)))
    (while y
      (setq x (cons (car y) x)
	    y (cdr y)))
    (setq irchat-msn-conversation-counter x))
  irchat-msn-conversation-counter)

(defun irchat-msn-generate-conversation-name ()
  (irchat-msn-conversation-counter-increment)
  (let ((n "")
	(x irchat-msn-conversation-counter))
    (while x
      (setq n (concat (format "%d" (car x)) n)
	    x (cdr x)))
    (format "$MSN[%s]" n)))

(defun irchat-msn-protocol-error ()
  (message "MSN Connection closed because of a protocol error.")
  (irchat-msn-close-server))


(defun irchat-msn-online-users ()
  (let ((l irchat-msn-online-list)
	(r '()))
    (while l
      (if (string-equal "NLN" (nth 3 (car l)))
	  (setq r (cons (nth 0 (car l)) r)))
      (setq l (cdr l)))
    r))

(defun irchat-msn-online-users-alist ()
  (let ((l irchat-msn-online-list)
	(r '()))
    (while l
      (if (string-equal "NLN" (nth 3 (car l)))
	  (setq r (cons (list (nth 0 (car l))) r)))
      (setq l (cdr l)))
    r))

(defun irchat-msn-server ()
  (if (listp irchat-msn-server)
      (nth (random (length irchat-msn-server)) irchat-msn-server)
    irchat-msn-server))

(defun irchat-set-msn-indicator ()
  (cond ((and irchat-msn-uid
	      (not (irchat-msn-server-opened)))
	 (setq irchat-msn-indicator " {MSN: Offline} "))
	((not (irchat-msn-server-opened))
	 (setq irchat-msn-indicator ""))
	(t
	 (setq irchat-msn-indicator
	       (concat "{MSN: "
		       irchat-msn-uid
		       " "
		       irchat-msn-my-online-mode
		       (if (or (> (length irchat-msn-online-list) 0)
			       (> (length irchat-msn-forward-list) 0))
			   (format " (%d/%d)" 
				   (length irchat-msn-online-list)
				   (length irchat-msn-forward-list))
			 "")
		       "}"))))
  irchat-msn-indicator)

(defun irchat-msn-name-cache-delete (name)
  (let ((l irchat-msn-name-cache)
	(n (upcase name))
	(r '()))
    (while l
      (if (not (string-equal (car (car l)) n))
	  (setq r (cons (car l) r)))
      (setq l (cdr l)))
    (setq irchat-msn-name-cache r))
  t)

(defun irchat-msn-name-cache-add (name visible)
  (let ((n (upcase name)))
    (irchat-msn-name-cache-delete n)
    (setq irchat-msn-name-cache (cons (cons n visible)
				      irchat-msn-name-cache)))
  t)

(defun irchat-msn-name-cache-get (name)
  (let ((l irchat-msn-name-cache)
	(n (upcase name))
	(r name))
    (while l
      (if (string-equal (car (car l)) n)
	  (setq r (cdr (car l))
		l '())
	(setq l (cdr l))))
    r))

(defun irchat-msn-name-cache-flush ()
  (setq irchat-msn-name-cache '()))

(defun irchat-msn-proto-msg-parse (msg)
  "Parse string of form \"<cmd> <param1> <param2> ...\" to list (\"<cmd>\" \"<param1>\" \"<param2> ...\")"
  (let ((m msg)
	(r '()))
    (while (or (string-match "^\\([^ ].*\\) \\([^ ]*\\)$" m)
	       (string-match "^\\(\\)\\([^ ][^ ]*\\)$" m))
      (setq r (cons (matching-substring m 2) r)
	    m (matching-substring m 1)))
    r))

(defun irchat-msn-protocol-error-string (errorcode)
  (let ((l irchat-msn-protocol-error-codes)
	(r nil))
    (while l
      (if (or (eq (car (car l)) errorcode)
	      (and (stringp errorcode)
		   (string= (format "%03d" (car (car l))) errorcode)))
	  (setq r (cdr (car l))
		l nil)
	(setq l (cdr l))))
    (if (null r)
	(setq r (format "Unknown error #%s" errorcode)))
    r))

(eval-and-compile (provide 'irchat-msn))

;;; eof (irchat-msn.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-crypt-vars.el,v 1.1 1998/11/04 10:54:33 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(defconst irchat-crypt-support t
  "*Does current version support crypto?")

;;;
;;; user modifiable variables needed for crypto stuff
;;;

;;;
;;;  Format strings for various messages
;;;
(defvar irchat-myformat-string-e ">>"
  "*Format for own messages. (encrypted)")

(defvar irchat-format-string-e "-> **%s**"
  "*Format string for private messages being sent. (encrypted)")

(defvar irchat-format-string0-e "--%s--"
  "*Format string for arriving NOTICE messages. (encrypted)")

(defvar irchat-format-string1-e "**%s**"
  "*Format string for arriving private messages. (encrypted)")

(defvar irchat-format-string2-e "<<%s>>"
  "*Format string for arriving messages to current channel. (encrypted)")

(defvar irchat-format-string3-e "<<%s:%s>>"
  "*Format string for arriving messages to current channel from outside the channel. (encrypted)")

(defvar irchat-format-string4-e "((%s))"
  "*Format string for arriving messages to other channel from outside the channel. (encrypted)")

(defvar irchat-format-string5-e ">>%s:%s<<"
  "*Format string for arriving messages to other channel from outside the channel. (encrypted)")

(defvar irchat-defected-message-prefix ""
  "Prefix to attach before the defected crypt message.")

(defvar irchat-suspicious-message-prefix ""
  "Prefix to attach before the suspicious crypt message.")

;;;
;;;  Buffers
;;;
(defvar irchat-CRYPT-buffer "*IRC CRYPT*")

;;;
;;;  Crypt
;;;
(defvar irchat-crypt-known-keys '() 
  "String list containing encryption keys.  e.g. '(\"foo\" \"bar\")")
(defvar irchat-crypt-default-keys '() 
  "List containing pairs (or lists) of addresses and associated default keys (and possibly key expand version).  e.g '((\"#42\" . \"asd\") (\"#xyz\" . \"zappa\") (\"friend\" \"moebius\" 1))")
(defvar irchat-crypt-timestamp-tolerance 300
  "Allow incoming messages to have N seconds old timestamp.")
(defvar irchat-crypt-ignore-suspicious nil
  "If t, ignore messages with invalid nick or timestamp.")
(defvar irchat-crypt-mode-active t
  "If t, irchat encrypts all messages it has a default key for.")
(defvar irchat-crypt-ignore-defected nil
  "If t, ignore encrypted messages that cannot be opened.")
(defvar irchat-crypt-secure-passphrase-read nil
  "If t, read passphrases with secured keyboard (if possible).")

;;;
;;; Buffer display variables
;;;
(defvar irchat-C-buffer (list irchat-CRYPT-buffer)
  "*A list of buffers where messages that were not decrypted are sent.")

(eval-and-compile (provide 'irchat-crypt-vars))

;;;
;;; eof
;;;
;;;   -*- lisp -*-
;;; 
;;;  ----------------------------------------------------------------------
;;;  b64 - almost base64
;;;  ----------------------------------------------------------------------
;;;  Created      : Thu Feb 20 16:21:28 1997 tri
;;;  Last modified: Mon Feb 24 15:43:23 1997 tri
;;;  ----------------------------------------------------------------------
;;;  Copyright  1997
;;;  Timo J. Rinne <tri@iki.fi>
;;; 
;;;  Address: Cirion oy, PO-BOX 250, 00121 Helsinki, Finland
;;;  ----------------------------------------------------------------------
;;;  Any express or implied warranties are disclaimed.  In no event
;;;  shall the author be liable for any damages caused (directly or
;;;  otherwise) by the use of this software.
;;;  
;;;  irchat-copyright.el applies only if used with irchat IRC client.
;;;  Contact the author for additional copyright info.
;;;
;;;  $Id: b64.el,v 3.1 1997/02/24 16:00:02 tri Exp $
;;;

(eval-and-compile  
  (provide 'b64))

(defconst b64-alphabet
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
  "Alphabet used in b64 encoding")
(defvar b64-decode-vector 
  nil
  "Vector used in b64 decoding.  Generated in run-time.")

(defun b64-build-decode-vector ()
  "Generate b64-decode-vector from b64-alphabet"
  (let ((v (make-vector 256 -1))
	(l (length b64-alphabet))
	(i 0))
    (while (< i l)
      (aset v (elt b64-alphabet i) i)
      (setq i (+ 1 i)))
    (aset v ?= -2)
    v))

(defun b64-encode-string (str)
  "B64 encode STR"
  (let* ((l (length str))
	 (n (/ l 3))
	 (i 0)
	 (r ""))
    (while (< i n)
      (let* ((c0 (elt str (* i 3)))
	     (c1 (elt str (+ 1 (* i 3))))
	     (c2 (elt str (+ 2 (* i 3))))
	     (e0 (>> c0 2))
	     (e1 (logior (logand 63 (<< c0 4)) (>> c1 4)))
	     (e2 (logior (logand 63 (<< c1 2)) (>> c2 6)))
	     (e3 (logand 63 c2)))
	(setq r (concat r 
			(char-to-string (elt b64-alphabet e0))
			(char-to-string (elt b64-alphabet e1))
			(char-to-string (elt b64-alphabet e2))
			(char-to-string (elt b64-alphabet e3))))
	(setq i (+ 1 i))))
    (cond ((= (% l 3) 2)
	   (let* ((c0 (elt str (* i 3)))
		  (c1 (elt str (+ 1 (* i 3))))
		  (e0 (>> c0 2))
		  (e1 (logior (logand 63 (<< c0 4)) (>> c1 4)))
		  (e2 (logand 63 (<< c1 2))))
	     (setq r (concat r 
			     (char-to-string (elt b64-alphabet e0))
			     (char-to-string (elt b64-alphabet e1))
			     (char-to-string (elt b64-alphabet e2))
			     "="))))
	  ((= (% l 3) 1)
	   (let* ((c0 (elt str (* i 3)))
		  (e0 (>> c0 2))
		  (e1 (logand 63 (<< c0 4))))
	     (setq r (concat r 
			     (char-to-string (elt b64-alphabet e0))
			     (char-to-string (elt b64-alphabet e1))
			     "==")))))
    r))

(defun b64-decode-string (str)
  "B64 decode STR"
  (if (not b64-decode-vector)
      (setq b64-decode-vector (b64-build-decode-vector)))
  (let* ((l (length str))
	 (n (/ l 4))
	 (i 0)
	 (r ""))
    (if (> (% l 4) 0)
	(message "b64-decode: Trailing garbage ignored"))
    (while (< i n)
      (let ((e0 (elt b64-decode-vector (elt str (* i 4))))
	    (e1 (elt b64-decode-vector (elt str (+ 1 (* i 4)))))
	    (e2 (elt b64-decode-vector (elt str (+ 2 (* i 4)))))
	    (e3 (elt b64-decode-vector (elt str (+ 3 (* i 4))))))
	(cond ((and (>= e0 0) (>= e1 0) (>= e2 0) (>= e3 0))
	       (let ((c0 (logior (logand 255 (<< e0 2)) (>> e1 4)))
		     (c1 (logior (logand 255 (<< e1 4)) (>> e2 2)))
		     (c2 (logior (logand 255 (<< e2 6)) e3)))
		 (setq r (concat r 
				 (char-to-string c0)
				 (char-to-string c1)
				 (char-to-string c2)))))
	      ((and (>= e0 0) (>= e1 0) (>= e2 0) (= e3 -2))
	       (let ((c0 (logior (logand 255 (<< e0 2)) (>> e1 4)))
		     (c1 (logior (logand 255 (<< e1 4)) (>> e2 2))))
		 (setq r (concat r 
				 (char-to-string c0)
				 (char-to-string c1)))))
	      ((and (>= e0 0) (>= e1 0) (= e2 -2) (= e3 -2))
	       (let ((c0 (logior (logand 255 (<< e0 2)) (>> e1 4))))
		 (setq r (concat r (char-to-string c0)))))
	      (t (progn 
		   (setq r nil)
		   (setq i n))))
	(setq i (+ 1 i))))
    r))

;;; eof (b64.el)
;;;   -*- lisp -*-
;;; 
;;;  ----------------------------------------------------------------------
;;;  CRC32 compatible crc in elisp.  Cool, ha?
;;;  ----------------------------------------------------------------------
;;;  Created      : Fri Dec 16 13:47:25 1995 tri
;;;  Last modified: Tue Oct  7 13:27:46 1997 tri
;;;  ----------------------------------------------------------------------
;;;  Copyright  1995-1997
;;;  Timo J. Rinne <tri@iki.fi>
;;; 
;;;  Address: Cirion oy, PO-BOX 250, 00121 Helsinki, Finland
;;;  ----------------------------------------------------------------------
;;;  Any express or implied warranties are disclaimed.  In no event
;;;  shall the author be liable for any damages caused (directly or
;;;  otherwise) by the use of this software.
;;;  
;;;  irchat-copyright.el applies only if used with irchat IRC client.
;;;  Contact the author for additional copyright info.
;;;
;;;  $Id: crc32.el,v 3.3 1997/10/07 10:28:17 tri Exp $
;;;

(eval-and-compile  
  (provide 'crc32))

;;; Interactive functions
(defun crc32-region (&optional raw)
  "Calculate crc32 of the current region"
  (interactive)
  (if mark-active
      (let* ((s (buffer-substring (mark) (point)))
	     (crc (crc32-string s raw)))
	(message "Length = %d, CRC32 = %s" (length s) crc)
	crc)
    (error "Mark is not set")))

(defun crc32-buffer (&optional raw)
  "Calculate crc32 of the current buffer"
  (interactive)
  (let* ((s (buffer-substring (point-min) (point-max)))
	 (crc (crc32-string s raw)))
    (message "Length = %d, CRC32 = %s" (length s) crc)
    crc))

;;; Lookup table
(defconst crc32-lookup-table
  [ (0 . 0) (30471 . 12438) (60942 . 24876) (39177 . 20922) (1901 . 50201)
    (28778 . 62607) (59747 . 42293) (40548 . 38307) (3803 . 34866)
    (31196 . 47268) (57557 . 59678) (38866 . 55688) (2486 . 19499)
    (32433 . 31933) (59320 . 11527) (37055 . 7569) (7607 . 4196) (27312 . 8434)
    (62393 . 29000) (33982 . 16862) (6874 . 54397) (28125 . 58603)
    (62676 . 46417) (33747 . 34247) (4972 . 38998) (25707 . 43200)
    (64866 . 63866) (35429 . 51692) (5121 . 23631) (25350 . 27865)
    (64015 . 15715) (36104 . 3573) (15214 . 8392) (19561 . 4190)
    (54624 . 16868) (41575 . 29042) (15363 . 58577) (19204 . 54343)
    (53773 . 34301) (42250 . 46443) (13749 . 43258) (17074 . 39020)
    (56251 . 51670) (44220 . 63808) (13016 . 27875) (17887 . 23669)
    (56534 . 3535) (43985 . 15705) (9945 . 12460) (20958 . 58) (51415 . 20864)
    (49104 . 24854) (8628 . 62645) (22195 . 50211) (53178 . 38297)
    (47293 . 42255) (10242 . 47262) (24325 . 34824) (50700 . 55730)
    (45323 . 59684) (12143 . 31879) (22632 . 19473) (49505 . 7595)
    (46694 . 11581) (30428 . 16784) (475 . 28934) (39122 . 8380) (61397 . 4138)
    (29105 . 34185) (1718 . 46367) (40895 . 58533) (59576 . 54323)
    (30727 . 51618) (3840 . 63796) (38409 . 43150) (57614 . 38936)
    (32618 . 3515) (2157 . 15661) (37220 . 27799) (58979 . 23553)
    (27499 . 20980) (7276 . 24930) (34149 . 12504) (62050 . 78) (27654 . 38381)
    (6913 . 42363) (33288 . 62657) (62735 . 50263) (26032 . 55750)
    (4791 . 59728) (35774 . 47338) (64697 . 34940) (25309 . 7647)
    (5594 . 11593) (36051 . 31987) (64468 . 19557) (19890 . 24920)
    (15029 . 20942) (41916 . 116) (54459 . 12514) (19167 . 42305)
    (15832 . 38359) (42193 . 50285) (54230 . 62715) (17257 . 59754)
    (13422 . 55804) (44391 . 34886) (55904 . 47312) (17412 . 11635)
    (13059 . 7653) (43530 . 19551) (56589 . 31945) (20485 . 28988)
    (9986 . 16810) (48651 . 4112) (51468 . 8326) (22376 . 46373) (8303 . 34227)
    (47462 . 54281) (52833 . 58527) (24286 . 63758) (10713 . 51608)
    (45264 . 38946) (51159 . 43188) (22963 . 15639) (11956 . 3457)
    (47037 . 23611) (49338 . 27821) (60856 . 33568) (39615 . 46006)
    (950 . 57868) (29873 . 53914) (60117 . 18233) (40402 . 30639) (1243 . 9749)
    (29660 . 5763) (58211 . 2834) (37988 . 15236) (3437 . 27198)
    (31338 . 23208) (58382 . 53003) (37641 . 65437) (2560 . 44583)
    (32007 . 40625) (61455 . 37700) (34568 . 41938) (7681 . 62056)
    (26886 . 49918) (63330 . 22365) (32869 . 26571) (6508 . 13937)
    (28267 . 1767) (65236 . 7030) (35283 . 11232) (4314 . 31322)
    (26589 . 19148) (63929 . 57199) (36542 . 61433) (6071 . 48707)
    (24752 . 36565) (54998 . 41960) (41425 . 37758) (14552 . 49860)
    (20447 . 62034) (53691 . 26609) (42684 . 22375) (16309 . 1757)
    (18610 . 13899) (55309 . 11226) (44810 . 6988) (13827 . 19190)
    (16644 . 31328) (57184 . 61379) (43111 . 57173) (12654 . 36591)
    (18025 . 48761) (52065 . 45964) (48230 . 33562) (9583 . 53920)
    (21096 . 57910) (52236 . 30613) (47883 . 18179) (8706 . 5817)
    (21765 . 9775) (50618 . 15294) (45757 . 2856) (11188 . 23186)
    (23731 . 27140) (49879 . 65447) (46544 . 53041) (11481 . 40587)
    (23518 . 44573) (39780 . 49840) (60515 . 61990) (30058 . 41884)
    (621 . 37642) (39945 . 1705) (60174 . 13887) (29191 . 26501) (1280 . 22291)
    (38335 . 19074) (58040 . 31252) (31665 . 11182) (3254 . 6968)
    (37586 . 36507) (58837 . 48653) (31964 . 61367) (3035 . 57121)
    (34515 . 53972) (61908 . 57922) (26845 . 46072) (8154 . 33646)
    (33214 . 5837) (63161 . 9819) (28592 . 30689) (6327 . 18295)
    (34824 . 23270) (65295 . 27248) (26118 . 15306) (4353 . 2908)
    (36709 . 40703) (63586 . 44649) (24939 . 65491) (5740 . 53061)
    (40970 . 57976) (55053 . 53998) (19972 . 33620) (14595 . 46018)
    (42855 . 9825) (53344 . 5879) (18793 . 18253) (15982 . 30683)
    (44753 . 27210) (55766 . 23260) (16607 . 2918) (14296 . 15344)
    (43452 . 44627) (57019 . 40645) (18354 . 53119) (12469 . 65513)
    (48573 . 61980) (51898 . 49802) (21427 . 37680) (9396 . 41894)
    (47824 . 13829) (52695 . 1683) (21726 . 22313) (9177 . 26559)
    (45926 . 31278) (50273 . 19128) (23912 . 6914) (10863 . 11156)
    (46091 . 48695) (49932 . 36513) (23045 . 57115) (11522 . 61325) ]
  "Lookup table for crc32 routines")

;;; Arithmetics
(defun crc32-^ (x1 x2)
  "Bitwise xor for crc32 routine"
  (let ((a (if (consp x1) x1 (cons 0 x1)))
	(b (if (consp x2) x2 (cons 0 x2))))
    (cons (logxor (car a) (car b)) (logxor (cdr a) (cdr b)))))

(defun crc32-& (x1 x2)
  "Bitwise xor for crc32 routine"
  (let ((a (if (consp x1) x1 (cons 0 x1)))
	(b (if (consp x2) x2 (cons 0 x2))))
    (cons (logand (car a) (car b)) (logand (cdr a) (cdr b)))))

(defun crc32-mask-8bit (x)
  "Mask crc32 PAIR to 8bit integer"
  (logand 255 (cdr x)))

(defun crc32-shr-8 (x)
  "Shift crc32 PAIR left 8 bits"
  (let* ((h (car x)) 
	 (l (cdr x)))
    (cons (/ h 256)
	  (logior (* (logand h 255) 256) (/ l 256)))))

;;; CRC32 function
(defun crc32-string (str &optional raw)
  "Calculate 32bit crc of STRING"
  (let ((r (cons 0 0))
	(l (length str))
	(j 0))
    (while (< j l)
      (let* ((i (crc32-mask-8bit (crc32-^ r (elt str j)))))
	(setq r (crc32-^ (elt crc32-lookup-table i)
			 (crc32-shr-8 r)))
	(setq j (+ 1 j))))
    (cond ((or (null raw) (equal raw 'hex-string))
	   (format "%04.4x%04.4x" (car r) (cdr r)))
	  ((equal raw 'raw-vector)
	   (let ((v [0 0 0 0]))
	     (aset v 0 (logand (/ (car r) 256) 255))
	     (aset v 1 (logand (car r) 255))
	     (aset v 2 (logand (/ (cdr r) 256) 255))
	     (aset v 3 (logand (cdr r) 255))
	     v))
	  ((or (equal raw 'raw-string) (equal raw t))
	   (let ((v "xxxx"))
	     (aset v 0 (logand (/ (car r) 256) 255))
	     (aset v 1 (logand (car r) 255))
	     (aset v 2 (logand (/ (cdr r) 256) 255))
	     (aset v 3 (logand (cdr r) 255))
	     v))
	  (t (error "Unknown return type in crc32-string")))))

;;; eof (crc32.el)






;;;   -*- lisp -*-
;;; 
;;;  ----------------------------------------------------------------------
;;;  RC4 encryption in elisp.  Cool, ha?
;;;  ----------------------------------------------------------------------
;;;  Created      : Tue Jun 23 15:02:41 1998 tri
;;;  Last modified: Mon Aug 10 17:58:50 1998 tri
;;;  ----------------------------------------------------------------------
;;;  Copyright  1998
;;;  Timo J. Rinne <tri@iki.fi>
;;; 
;;;  Address: Cirion oy, PO-BOX 250, 00121 Helsinki, Finland
;;;  ----------------------------------------------------------------------
;;;  Any express or implied warranties are disclaimed.  In no event
;;;  shall the author be liable for any damages caused (directly or
;;;  otherwise) by the use of this software.
;;;  
;;;  irchat-copyright.el applies only if used with irchat IRC client.
;;;  Contact the author for additional copyright info.
;;;
;;;  $Id: rc4.el,v 1.2 1998/08/10 15:04:17 tri Exp $
;;;

(eval-and-compile  
  (provide 'rc4))

(defun rc4-elt-swap (vector idx1 idx2)
  "Swap elements in VECTOR indexed IDX1 and IDX2."
  (let ((x (elt vector idx1)))
    (aset vector idx1 (elt vector idx2))
    (aset vector idx2 x)))

(defun rc4-make-state (key &optional limit)
  "Build rc4 state from string or integer vector KEY."
  (let* ((limit (if (and (numberp limit) (> limit 0)) limit 256))
	 (S (make-vector limit 0))
	 (K (make-vector limit 0))
	 (l (length key))
	 (i 0)
	 (j 0))
    (if (= l 0)
	(progn
	  (setq l 1)
	  (setq key (make-vector 1 0))))
    (setq i 0)
    (while (< i limit)
      (aset K i (% (elt key (% i l)) limit))
      (aset S i i)
      (setq i (+ i 1)))
    (setq i 0)
    (setq j 0)
    (while (< i limit)
      (setq j (% (+ j (elt S i) (elt K i)) limit))
      (rc4-elt-swap S i j)
      (setq i (+ i 1)))
    (vector 0 0 S)))

(defun rc4-random (state)
  "Generate rc4 stream byte from STATE and update state accordingly."
  (let ((limit (length (elt state 2))))
    (aset state 0 
	  (% (+ (elt state 0) 1) limit))
    (aset state 1 
	  (% (+ (elt state 1) (elt (elt state 2) (elt state 0))) limit))
    (rc4-elt-swap (elt state 2) (elt state 0) (elt state 1))
    (elt (elt state 2)
	 (% (+ (elt (elt state 2) (elt state 0))
	       (elt (elt state 2) (elt state 1))) limit))))

(defun rc4-encrypt (str state)
  "Encrypt STRING with rc4 STATE and update state accordingly."
  (let* ((l (length str))
	 (r (if (stringp str) (make-string l 0) (make-vector l 0)))
	 (i 0))
    (while (< i l)
      (aset r i (logxor (elt str i) (rc4-random state)))
      (setq i (+ i 1)))
    r))

(defun rc4-random-vector-complex (str len &optional skip level limit)
  (if (null skip)
      (setq skip 0))
  (if (null level)
      (setq level 0))
  (let ((r (make-vector len 0))
	(s (rc4-make-state str limit))
	(i 0)
	(j 0))
    (setq i skip)
    (while (> i 0)
      (rc4-random s)
      (setq i (- i 1)))
    (setq j level)
    (while (> j 0)
      (setq i (rc4-random s))
      (while (> i 0)
	(rc4-random s)
	(setq i (- i 1)))
      (setq j (- j 1)))
    (rc4-encrypt r s)))

(defun rc4-random-vector (str len &optional limit)
  (rc4-random-vector-complex str len 0 0 limit))

;;; eof (rc4.el)
;;;   -*- lisp -*-
;;; 
;;;  ----------------------------------------------------------------------
;;;  IDEA encryption in elisp.  Cool, ha?
;;;  ----------------------------------------------------------------------
;;;  Created      : Thu Jun 29 08:11:25 1995 tri
;;;  Last modified: Tue Sep 14 20:09:11 1999 tri
;;;  ----------------------------------------------------------------------
;;;  Copyright  1995-1999
;;;  Timo J. Rinne <tri@iki.fi>
;;; 
;;;  Address: Cirion oy, PO-BOX 250, 00121 Helsinki, Finland
;;;  ----------------------------------------------------------------------
;;;  Any express or implied warranties are disclaimed.  In no event
;;;  shall the author be liable for any damages caused (directly or
;;;  otherwise) by the use of this software.
;;;  
;;;  irchat-copyright.el applies only if used with irchat IRC client.
;;;  Contact the author for additional copyright info.
;;;
;;;  $Id: idea.el,v 3.23 1999/09/14 17:09:17 tri Exp $
;;;

(eval-and-compile  
  (require 'b64)
  (require 'crc32))

(eval-and-compile  
  (provide 'idea))

(defvar idea-default-key-expand-version 3
  "Which version of key expand is used as default (1 or 2 or 3).")

;;;; 16bit basic arithmetic operations for IDEA.
(defun idea-mask-16bit (x)
  "Mask X with 0xffff"
  (logand 65535 x))

(defun idea-+ (a b)
  "16 bit add A and B.  Ignore overflow."
  (idea-mask-16bit (+ a b)))

(defun idea-- (a b)
  "16 bit add A and B.  Ignore overflow."
  (idea-mask-16bit (- a b)))

(defun idea-/ (a b)
  "16 bit add A and B.  Ignore overflow."
  (idea-mask-16bit (/ a b)))

(defun idea-% (a b)
  "16 bit add A and B.  Ignore overflow."
  (idea-mask-16bit (%  a b)))

(defun idea-* (a b)
  "16 bit multiply A and B.  Ignore overflow."
  (let* ((al (logand 255 a))
	 (ah (logand 255 (/ a 256)))
	 (bl (logand 255 b))
	 (bh (logand 255 (/ b 256)))
	 (r1 (* al bl))
	 (r2 (+ (* al bh) (* ah bl)))
	 (l (idea-mask-16bit (+ r1 (* r2 256)))))
    l))

(defun idea-xor-blocks (a b)
  "Perform logical xor operation between encryption blocks A and B"
  (list (idea-^ (nth 0 a) (nth 0 b))
	(idea-^ (nth 1 a) (nth 1 b))
	(idea-^ (nth 2 a) (nth 2 b))
	(idea-^ (nth 3 a) (nth 3 b))))

(defun idea-& (a b)
  "16 bit AND A and B.  Ignore overflow."
  (logand (idea-mask-16bit a) (idea-mask-16bit b)))

(defun idea-| (a b)
  "16 bit OR A and B.  Ignore overflow."
  (logior (idea-mask-16bit a) (idea-mask-16bit b)))

(defun idea-^ (a b)
  "16 bit XOR A and B.  Ignore overflow."
  (logxor (idea-mask-16bit a) (idea-mask-16bit b)))

(defun idea-mul (a b)
  "A * B in idea style."
  (let* ((al (logand 255 a))
	 (ah (logand 255 (/ a 256)))
	 (bl (logand 255 b))
	 (bh (logand 255 (/ b 256)))
	 (r1 (* al bl))
	 (r2 (+ (* al bh) (* ah bl)))
	 (r3 (* ah bh))
	 (l (+ r1 (* (logand 255 r2) 256)))
	 (h (+ (/ r2 256) (/ l 65536) r3))
	 (l (idea-mask-16bit l)))
    (if (not (and (= 0 l) (= 0 h)))
	(idea-mask-16bit (idea-+ (idea-- l h) (if (< l h) 1 0)))
      (if (= 0 a)
	  (idea-mask-16bit (idea-- 1 b))
	(idea-mask-16bit (idea-- 1 a))))))

(defun idea-random ()
  "Generate 16 bit random value"
  (if (fboundp 'irchat-random-16)
      (irchat-random-16)
    (idea-mask-16bit (abs (random)))))

(defun idea-random-char ()
  "Generate 8 bit random value"
  (if (fboundp 'irchat-random-8)
      (irchat-random-8)
    (idea-& 255 (idea-random))))

(defun idea-<< (x n)
  "Shift integer X left N bits"
  (if (> n 16)
      (setq n 16))
  (setq x (idea-mask-16bit x))
  (if (or (< n 0)
	  (< x 0))
      0
    (while (> n 0)
      (setq x (idea-mask-16bit (* x 2)))
      (setq n (- n 1))))
  x)

(defun idea->> (x n)
  "Shift integer X right N bits"
  (if (> n 16)
      (setq n 16))
  (setq x (idea-mask-16bit x))
  (if (or (< n 0)
	  (< x 0))
      0
    (while (> n 0)
      (setq x (/ x 2))
      (setq n (- n 1))))
  x)

(defun idea-additive-inverse (x)
  (idea-mask-16bit (- 0 x)))

(defun idea-multiplicative-inverse (n)
  "16 bit multiplicative inverse of N"
  (let ((x (idea-mask-16bit n)))
    (if (< x 2)
	x
      (let ((hlp2 (idea-/ 65537 x))
	    (y  (idea-% 65537 x))
	    (do-brk '()))
	(if (= 1 y)
	    (idea-mask-16bit (idea-- 1 hlp2))
	  (let ((hlp1 1)
		(q)
		(ret))
	    (while (not do-brk)
	      (setq q (idea-/ x y))
	      (setq x (idea-% x y))
	      (setq hlp1 (idea-+ hlp1 (idea-* q hlp2)))
	      (if (= 1 x)
		  (progn
		    (setq ret (idea-mask-16bit hlp1))
		    (setq do-brk t))
		(progn
		  (setq q (idea-/ y x))
		  (setq y (idea-% y x))
		  (setq hlp2 (idea-+ hlp2 (idea-* q hlp1)))
		  (if (= 1 y)
		      (progn
			(setq ret (idea-mask-16bit (idea-- 1 hlp2)))
			(setq do-brk t))))))
	    ret))))))

(defun idea-shift-key-25bits-left (key)
  "Shift expanded (128bit) idea key left 25 bits"
  (list (idea-| (idea-<< (nth 1 key) 9) (idea->> (nth 2 key) 7))
	(idea-| (idea-<< (nth 2 key) 9) (idea->> (nth 3 key) 7))
	(idea-| (idea-<< (nth 3 key) 9) (idea->> (nth 4 key) 7))
	(idea-| (idea-<< (nth 4 key) 9) (idea->> (nth 5 key) 7))
	(idea-| (idea-<< (nth 5 key) 9) (idea->> (nth 6 key) 7))
	(idea-| (idea-<< (nth 6 key) 9) (idea->> (nth 7 key) 7))
	(idea-| (idea-<< (nth 7 key) 9) (idea->> (nth 0 key) 7))
	(idea-| (idea-<< (nth 0 key) 9) (idea->> (nth 1 key) 7))))

;;; [(1 2 3 4) (5 6 7 8) (9 10 11 12)]
;;; transforms to
;;; [(1 5 9 2) (6 10 3 7) (11 4 8 12)]
;;; and
;;; [(1 2 3 4) (5 6 7 8) (9 10 11 12) (13 14 15 16)]
;;; transforms to
;;; [(1 5 9 13) (2 6 10 14) (3 7 11 15) (4 8 12 16)]
;;; transforms to
;;; [(1 2 3 4) (5 6 7 8) (9 10 11 12) (13 14 15 16)]
(defun idea-interlace-blocklist (bl)
  "Make interlaced version of BLOCKLIST."
  (let* ((l (length bl))
	 (r (make-vector l 0))
	 (i 0))
    (while (< i l)
      (aset r i
	    (let ((j (* 4 i)))
	      (list (nth (/ j l) (elt bl (% j l)))
		    (nth (/ (+ j 1) l) (elt bl (% (+ j 1) l)))
		    (nth (/ (+ j 2) l) (elt bl (% (+ j 2) l)))
		    (nth (/ (+ j 3) l) (elt bl (% (+ j 3) l))))))
      (setq i (+ i 1)))
    r))

(defun idea-expand-string-to-key (string &optional version)
  "Expand string to full 128bit key (list of 8 16bit ints)"
  (if (null version) (setq version idea-default-key-expand-version))
  (cond ((= version 1) (idea-expand-string-to-key-version-1 string))
	((= version 2) (idea-expand-string-to-key-version-2 string))
	((= version 3) (idea-expand-string-to-key-version-3 string))
	(t (error "Unknown key expansion version"))))

(defun idea-expand-string-to-key-version-1 (string)
  "Expand string to full 128bit key (list of 8 16bit ints) (version 1)"
  (if (= (length string) 0)
      '(0 0 0 0 0 0 0 0)
    (let* ((s (if (< (length string) 64)
		  (let* ((ss string)
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss)))
			 (ss (concat ss (crc32-string ss))))
		    ss)
		string))
	   (l (length s))
	   (s (concat s (format "%d" l)))
	   (l (length s))
	   (x1 0)
	   (x2 (/ l 4))
	   (x3 (* 2 (/ l 4)))
	   (x4 (* 3 (/ l 4)))
	   (v1 (crc32-string (substring s x1 l)))
	   (v2 (crc32-string (substring s x2 l)))
	   (v3 (crc32-string (substring s x3 l)))
	   (v4 (crc32-string (substring s x4 l)))
	   (k1 (idea-4hex-to-int (substring v1 0 4)))
	   (k2 (idea-4hex-to-int (substring v1 4 8)))
	   (k3 (idea-4hex-to-int (substring v2 0 4)))
	   (k4 (idea-4hex-to-int (substring v2 4 8)))
	   (k5 (idea-4hex-to-int (substring v3 0 4)))
	   (k6 (idea-4hex-to-int (substring v3 4 8)))
	   (k7 (idea-4hex-to-int (substring v4 0 4)))
	   (k8 (idea-4hex-to-int (substring v4 4 8))))
      (list k1 k2 k3 k4 k5 k6 k7 k8))))

(defun idea-expand-string-to-key-version-2 (string)
  "Expand string to full 128bit key (list of 8 16bit ints) (version 2)"
  (if (= (length string) 0)
      '(0 0 0 0 0 0 0 0)
    (let* ((s (if (> (length string) 3) 
		  string 
		(concat string (crc32-string string 'raw-string))))
	   (l (length s))
	   (s1 (idea-expand-substring (concat (format "%c%c" 0 (logand l 255)) 
					      (substring s 0 (/ l 4)))))
	   (s2 (idea-expand-substring (concat (format "%c%c" 85 (logand l 255))
					      (substring s (/ l 4) (/ l 2)))))
	   (s3 (idea-expand-substring (concat (format "%c%c" 
						      170 
						      (logand l 255))
					      (substring s  
							 (/ l 2) 
							 (+ (/ l 2)
							    (/ l 4))))))
	   (s4 (idea-expand-substring (concat (format "%c%c" 
						      255 
						      (logand l 255))
					      (substring s
							 (+ (/ l 2) (/ l 4)) 
							 l))))
	   (v1 (crc32-string s1))
	   (v2 (crc32-string s2))
	   (v3 (crc32-string s3))
	   (v4 (crc32-string s4))
	   (k1 (idea-4hex-to-int (substring v1 0 4)))
	   (k2 (idea-4hex-to-int (substring v1 4 8)))
	   (k3 (idea-4hex-to-int (substring v2 0 4)))
	   (k4 (idea-4hex-to-int (substring v2 4 8)))
	   (k5 (idea-4hex-to-int (substring v3 0 4)))
	   (k6 (idea-4hex-to-int (substring v3 4 8)))
	   (k7 (idea-4hex-to-int (substring v4 0 4)))
	   (k8 (idea-4hex-to-int (substring v4 4 8))))
      (list k1 k2 k3 k4 k5 k6 k7 k8))))

;;;
;;; Mystical constants in following function are just digits from pi
;;; taken with following scheme.
;;;   - Digits are split into sequences of five 31415 58979 32384 ...
;;;   - Any sequence forming a number > 65535 is dropped
;;;
(defun idea-expand-string-to-key-version-3 (str)
  "Generate idea key from STRING."
  (if (= (length str) 0)
      '(0 0 0 0 0 0 0 0)
    (let* ((kk '(31415 58979 32384 62643 38327 16939 5820 45923))
	   (ek (idea-build-encryption-key kk 666))
	   (bl (idea-cleartext-string-to-block-list str nil))
	   (bl (idea-interlace-blocklist bl))
	   (r1 '(7816 40628 62089 3482))
	   (r2 '(53421 17067 13282 30664))
	   (r3 '(44609 55058 22317 25359))
	   (r4 '(40812 17450 28410 27019))
	   (i 0)
	   (l (length bl)))
      (while (< i l)
	(setq r1 (idea-xor-blocks 
		  (idea-crypt-transform-block (idea-xor-blocks 
					       (elt bl i)
					       r2)
					      ek) 
		  r1))
	(if (< (+ i 1) l)
	    (setq r2 (idea-xor-blocks
		      (idea-crypt-transform-block (idea-xor-blocks 
						   (elt bl (+ i 1))
						   r1)
						  ek)
		      r2)))
	(setq i (+ i 1)))
      (setq ek (idea-build-encryption-key (list (nth 0 r1) (nth 0 r2)
						(nth 1 r1) (nth 1 r2)
						(nth 2 r1) (nth 2 r2)
						(nth 3 r1) (nth 3 r2))
					  666))
      (setq bl (idea-interlace-blocklist bl))
      (setq i 0)
      (while (< i l)
	(setq r3 (idea-xor-blocks 
		  (idea-crypt-transform-block (idea-xor-blocks 
					       (elt bl i)
					       r4)
					      ek) 
		  r3))
	(if (< (+ i 1) l)
	    (setq r4 (idea-xor-blocks
		      (idea-crypt-transform-block (idea-xor-blocks 
						   (elt bl (+ i 1))
						   r3)
						  ek)
		      r4)))
	(setq i (+ i 1)))
      (list (nth 0 r3) (nth 0 r4) (nth 1 r3)
	    (nth 1 r4) (nth 2 r3) (nth 2 r4)
	    (nth 3 r3) (nth 3 r4)))))

(defun idea-expand-substring (string)
  (if (= (length string) 0)
      string
    (progn
      (setq string (concat (crc32-string string 'raw-string) string))
      (let ((i (+ 3 (logand 3 (elt string 0)))))
	(while (> i 0)
	  (setq i (- i 1))
	  (setq string (concat (crc32-string string 'raw-string) string))))
      string)))

(defun idea-random-key ()
  "Generate random IDEA key (list of 8 16bit values)"
  (list (idea-random) (idea-random) (idea-random) (idea-random)
	(idea-random) (idea-random) (idea-random) (idea-random)))

(defun idea-build-encryption-key (passphrase &optional version)
  "Build idea encryption context from string or keylist (list of 8 16bit ints)"
  (let* ((s1 (if (stringp passphrase) 
		 (idea-expand-string-to-key passphrase version)
	       (if (listp passphrase)
		   passphrase
		 (error "IDEA key can be built from string or keylist."))))
	 (annotation (idea-build-key-annotation s1 "e" version))
	 (s2 (idea-shift-key-25bits-left s1))
	 (s3 (idea-shift-key-25bits-left s2))
	 (s4 (idea-shift-key-25bits-left s3))
	 (s5 (idea-shift-key-25bits-left s4))
	 (s6 (idea-shift-key-25bits-left s5))
	 (s7 (idea-shift-key-25bits-left s6))
	 ;;; (s8 (idea-shift-key-25bits-left s7))
	 (r1 (list (nth 0 s1) (nth 1 s1) (nth 2 s1) 
		   (nth 3 s1) (nth 4 s1) (nth 5 s1)))
	 (r2 (list (nth 6 s1) (nth 7 s1) (nth 0 s2) 
		   (nth 1 s2) (nth 2 s2) (nth 3 s2)))
	 (r3 (list (nth 4 s2) (nth 5 s2) (nth 6 s2) 
		   (nth 7 s2) (nth 0 s3) (nth 1 s3)))
	 (r4 (list (nth 2 s3) (nth 3 s3) (nth 4 s3) 
		   (nth 5 s3) (nth 6 s3) (nth 7 s3)))
	 (r5 (list (nth 0 s4) (nth 1 s4) (nth 2 s4) 
		   (nth 3 s4) (nth 4 s4) (nth 5 s4)))
	 (r6 (list (nth 6 s4) (nth 7 s4) (nth 0 s5) 
		   (nth 1 s5) (nth 2 s5) (nth 3 s5)))
	 (r7 (list (nth 4 s5) (nth 5 s5) (nth 6 s5) 
		   (nth 7 s5) (nth 0 s6) (nth 1 s6)))
	 (r8 (list (nth 2 s6) (nth 3 s6) (nth 4 s6) 
		   (nth 5 s6) (nth 6 s6) (nth 7 s6)))
	 (r9 (list (nth 0 s7) (nth 1 s7) (nth 2 s7) 
		   (nth 3 s7))))
    (list r1 r2 r3 r4 r5 r6 r7 r8 r9 annotation)))

(defun idea-build-decryption-key (passphrase &optional version)
  "Build idea decryption context from string or keylist (list of 8 16bit ints)"
  (let* ((k (idea-build-encryption-key passphrase version))
	 (annotation (idea-build-key-annotation (list (nth 0 (nth 0 k))
						      (nth 1 (nth 0 k))
						      (nth 2 (nth 0 k))
						      (nth 3 (nth 0 k))
						      (nth 4 (nth 0 k))
						      (nth 5 (nth 0 k))
						      (nth 0 (nth 1 k))
						      (nth 1 (nth 1 k))) 
						"d"
						version)))
    (list
     (list (idea-multiplicative-inverse (nth 0 (nth 8 k)))
	   (idea-additive-inverse (nth 1 (nth 8 k)))
	   (idea-additive-inverse (nth 2 (nth 8 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 8 k)))
	   (nth 4 (nth 7 k))
	   (nth 5 (nth 7 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 7 k)))
	   (idea-additive-inverse (nth 2 (nth 7 k)))
	   (idea-additive-inverse (nth 1 (nth 7 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 7 k)))
	   (nth 4 (nth 6 k))
	   (nth 5 (nth 6 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 6 k)))
	   (idea-additive-inverse (nth 2 (nth 6 k)))
	   (idea-additive-inverse (nth 1 (nth 6 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 6 k)))
	   (nth 4 (nth 5 k))
	   (nth 5 (nth 5 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 5 k)))
	   (idea-additive-inverse (nth 2 (nth 5 k)))
	   (idea-additive-inverse (nth 1 (nth 5 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 5 k)))
	   (nth 4 (nth 4 k))
	   (nth 5 (nth 4 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 4 k)))
	   (idea-additive-inverse (nth 2 (nth 4 k)))
	   (idea-additive-inverse (nth 1 (nth 4 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 4 k)))
	   (nth 4 (nth 3 k))
	   (nth 5 (nth 3 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 3 k)))
	   (idea-additive-inverse (nth 2 (nth 3 k)))
	   (idea-additive-inverse (nth 1 (nth 3 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 3 k)))
	   (nth 4 (nth 2 k))
	   (nth 5 (nth 2 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 2 k)))
	   (idea-additive-inverse (nth 2 (nth 2 k)))
	   (idea-additive-inverse (nth 1 (nth 2 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 2 k)))
	   (nth 4 (nth 1 k))
	   (nth 5 (nth 1 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 1 k)))
	   (idea-additive-inverse (nth 2 (nth 1 k)))
	   (idea-additive-inverse (nth 1 (nth 1 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 1 k)))
	   (nth 4 (nth 0 k))
	   (nth 5 (nth 0 k)))
     (list (idea-multiplicative-inverse (nth 0 (nth 0 k)))
	   (idea-additive-inverse (nth 1 (nth 0 k)))
	   (idea-additive-inverse (nth 2 (nth 0 k)))
	   (idea-multiplicative-inverse (nth 3 (nth 0 k))))
     annotation)))

(defun idea-encryption-round (data subkey)
  "Basic round of IDEA encryption.  Method 1."
  (let* ((v1  (idea-mul (nth 0 data) (nth 0 subkey)))
	 (v2  (idea-+ (nth 1 data) (nth 1 subkey)))
	 (v3  (idea-+ (nth 2 data) (nth 2 subkey)))
	 (v4  (idea-mul (nth 3 data) (nth 3 subkey)))
	 (v5  (idea-^ v1 v3))
	 (v6  (idea-^ v2 v4))
	 (v7  (idea-mul v5 (nth 4 subkey)))
	 (v8  (idea-+ v6 v7))
	 (v9  (idea-mul v8 (nth 5 subkey)))
	 (v10 (idea-+ v7 v9))
	 (v11 (idea-^ v1 v9))
	 (v12 (idea-^ v3 v9))
	 (v13 (idea-^ v2 v10))
	 (v14 (idea-^ v4 v10)))
    (list v11 v12 v13 v14)))

(defun idea-final-transform (data subkey)
  "Final transformation round of IDEA encryption."
  (let* ((v1 (idea-mul (nth 0 data) (nth 0 subkey)))
	 (v2 (idea-+ (nth 2 data) (nth 1 subkey)))
	 (v3 (idea-+ (nth 1 data) (nth 2 subkey)))
	 (v4 (idea-mul (nth 3 data) (nth 3 subkey))))
    (list v1 v2 v3 v4)))

(defun idea-hex-to-key (h)
  "Expand 32 character HEXSTRING to idea key."
  (let ((h1 (idea-4hex-to-int (substring h 0 4)))
	(h2 (idea-4hex-to-int (substring h 4 8)))
	(h3 (idea-4hex-to-int (substring h 8 12)))
	(h4 (idea-4hex-to-int (substring h 12 16)))
	(h5 (idea-4hex-to-int (substring h 16 20)))
	(h6 (idea-4hex-to-int (substring h 20 24)))
	(h7 (idea-4hex-to-int (substring h 24 28)))
	(h8 (idea-4hex-to-int (substring h 28 32))))
    (if (and (> h1 -1) (> h2 -1) (> h3 -1) (> h4 -1)
	     (> h5 -1) (> h6 -1) (> h7 -1) (> h8 -1))
	(list h1 h2 h3 h4 h5 h6 h7 h8)
      (list 0 0 0 0 0 0 0 0))))

(defun idea-4hex-to-int (hex)
  "Convert four character long HEXSTRING to int"
  (let* ((x0 (idea-hex-char-to-int (elt hex 3)))
	 (x1 (idea-hex-char-to-int (elt hex 2)))
	 (x2 (idea-hex-char-to-int (elt hex 1)))
	 (x3 (idea-hex-char-to-int (elt hex 0))))
    (if (and (> x0 -1) (> x1 -1) (> x2 -1) (> x3 -1))
	(+ x0
	   (* x1 16)
	   (* x2 256)
	   (* x3 4096))
      -1)))

(defun idea-2hex-to-int (hex)
  "Convert two character long HEXSTRING to int"
  (let* ((x0 (idea-hex-char-to-int (elt hex 1)))
	 (x1 (idea-hex-char-to-int (elt hex 0))))
    (if (and (> x0 -1) (> x1 -1))
	(+ x0
	   (* x1 16))
      -1)))

(defun idea-crypt-transform-block (data key)
  "Make IDEA decryption transformation to DATA with key context KEY"
  ;;; Make 8 basic encryption rounds.
  ;;; Swap inner subblocks between basic rounds.
  ;;; Perform final transformation.
  (idea-final-transform
   (idea-encryption-round  
     (idea-encryption-round  
       (idea-encryption-round  
	 (idea-encryption-round  
	   (idea-encryption-round  
	     (idea-encryption-round  
	       (idea-encryption-round  
		 (idea-encryption-round data (nth 0 key))
		(nth 1 key))
	      (nth 2 key))
	    (nth 3 key))
	  (nth 4 key))
	(nth 5 key))
      (nth 6 key))
    (nth 7 key))
   (nth 8 key)))

(defun idea-encrypt-block (data key &optional version)
  "Encrypt single data block (4 * 16bits) with key"
  (let ((my-key (if (listp key) key 
		  (if (stringp key) (idea-build-encryption-key key version)
		    (error 
		     "IDEA key has to be key-context or string.")))))
    (idea-crypt-transform-block data my-key)))

(defun idea-decrypt-block (data key &optional version)
  "Decrypt single data block (4 * 16bits) with key"
  (let ((my-key (if (listp key) key 
		  (if (stringp key) (idea-build-decryption-key key version)
		    (error 
		     "IDEA key has to be key-context or string.")))))
    (idea-crypt-transform-block data my-key)))

(defun idea-fixed-string-to-block (str)
  "Convert 8 byte STRING to idea block"
  (list (idea-| (idea-<< (elt str 0) 8) (elt str 1))
	(idea-| (idea-<< (elt str 2) 8) (elt str 3))
	(idea-| (idea-<< (elt str 4) 8)	(elt str 5))
	(idea-| (idea-<< (elt str 6) 8)	(elt str 7))))

(defun idea-block-to-fixed-string (block)
  "Convert idea BLOCK to 8 byte string"
  (let ((c1 (idea->> (nth 0 block) 8))
	(c2 (idea-& (nth 0 block) 255))
	(c3 (idea->> (nth 1 block) 8))
	(c4 (idea-& (nth 1 block) 255))
	(c5 (idea->> (nth 2 block) 8))
	(c6 (idea-& (nth 2 block) 255))
	(c7 (idea->> (nth 3 block) 8))
	(c8 (idea-& (nth 3 block) 255)))
    (concat (char-to-string c1) (char-to-string c2) (char-to-string c3)
	    (char-to-string c4) (char-to-string c5) (char-to-string c6)
	    (char-to-string c7) (char-to-string c8))))

(defun idea-add-padding (str rndpad)
  "Add block padding to the STRING."
  (let* ((len (length str))
	 (pad (- 8 (% len 8)))
	 (pad (if (> pad 0) pad 8))
	 (p1 (idea-| (idea-& (if rndpad (idea-random-char) 0) 31) 
		     (idea-<< (- pad 1) 5)))
	 (p2 (if rndpad (idea-random-char) 0))
	 (p3 (if rndpad (idea-random-char) 0))
	 (p4 (if rndpad (idea-random-char) 0))
	 (p5 (if rndpad (idea-random-char) 0))
	 (p6 (if rndpad (idea-random-char) 0))
	 (p7 (if rndpad (idea-random-char) 0))
	 (p8 (if rndpad (idea-random-char) 0))
	 (ps (concat (char-to-string p1) (char-to-string p2)
		     (char-to-string p3) (char-to-string p4)
		     (char-to-string p5) (char-to-string p6)
		     (char-to-string p7) (char-to-string p8)))
	 (ps (substring ps 0 pad)))
    (concat ps str)))

(defun idea-remove-padding (str)
  "Remove block padding from the STRING"
  (let* ((len (length str)))
    (if (< len 8)
	nil
      (substring str (+ 1 (idea->> (elt str 0) 5)) len))))

(defun idea-cleartext-string-to-block-list (str &optional rndpad)
  "Convert cleartext STRING the list of encryption blocks (CLEARTEXT)"
  (let* ((str (idea-add-padding (idea-add-crc str) rndpad))
	 (l (length str))
	 (r (make-vector (/ l 8) 0))
	 (i 0))
    (while (< i l)
      (aset r (/ i 8) (idea-fixed-string-to-block (substring str i (+ i 8))))
      (setq i (+ 8 i)))
    r))

(defun idea-ciphertext-string-to-block-list (str)
  "Convert ciphertext STRING the list of encryption blocks (CIPHERTEXT)"
  (let* ((l (length str))
	 (r (make-vector (/ l 8) 0))
	 (i 0))
    (while (< i l)
      (aset r (/ i 8) (idea-fixed-string-to-block (substring str i (+ i 8))))
      (setq i (+ 8 i)))
    r))

(defun idea-ecb-encrypt-string (str key &optional version)
  "Encrypt STRING with KEY (either key context or string)"
  (let* ((my-key-type (idea-legal-key key))
	 (my-key (cond ((equal 'key-complete-decryption my-key-type)
			(error "KEY is not an encryption key."))
		       ((equal 'key-complete-encryption my-key-type)
			key)
		       ((or (equal 'key-string my-key-type)
			   (equal 'key-intlist my-key-type))
			(idea-build-encryption-key key version))
		       (t (error "Invalid key."))))
	 (data (idea-cleartext-string-to-block-list str '()))
	 (l (length data))
	 (r (make-string (* l 8) 0))
	 (i 0))
    (while (< i l)
      (let ((b (idea-block-to-fixed-string (idea-encrypt-block (elt data i) 
							       my-key)))
	    (so (* i 8)))
	(aset r so (elt b 0))
	(aset r (+ so 1) (elt b 1))
	(aset r (+ so 2) (elt b 2))
	(aset r (+ so 3) (elt b 3))
	(aset r (+ so 4) (elt b 4))
	(aset r (+ so 5) (elt b 5))
	(aset r (+ so 6) (elt b 6))
	(aset r (+ so 7) (elt b 7)))
      (setq i (+ 1 i)))
    (b64-encode-string r)))

(defun idea-ecb-decrypt-string (cipstr key &optional version)
  "Decrypt STRING with KEY (either key context or string)"
  (let ((str (b64-decode-string cipstr)))
    (if (and str (= 0 (% (length str) 8)))
	(let* ((my-key-type (idea-legal-key key))
	       (my-key (cond ((equal 'key-complete-encryption my-key-type)
			      (error "KEY is not an decryption key."))
			     ((equal 'key-complete-decryption my-key-type)
			      key)
			     ((or (equal 'key-string my-key-type)
				  (equal 'key-intlist my-key-type))
			      (idea-build-decryption-key key version))
			     (t (error "Invalid key."))))
	       (data (idea-ciphertext-string-to-block-list str))
	       (l (length data))
	       (r (make-string (* l 8) 0))
	       (i 0))
	  (while (< i l)
	    (let ((b (idea-block-to-fixed-string (idea-encrypt-block 
						  (elt data i) 
						  my-key)))
		  (so (* i 8)))
	      (aset r so (elt b 0))
	      (aset r (+ so 1) (elt b 1))
	      (aset r (+ so 2) (elt b 2))
	      (aset r (+ so 3) (elt b 3))
	      (aset r (+ so 4) (elt b 4))
	      (aset r (+ so 5) (elt b 5))
	      (aset r (+ so 6) (elt b 6))
	      (aset r (+ so 7) (elt b 7)))
	    (setq i (+ 1 i)))
	  (let ((ret (idea-check-crc (idea-remove-padding r))))
	    (if ret
		ret
	      '())))
      '())))

(defun idea-cbc-encrypt-string (str key &optional version)
  "Encrypt STRING with KEY (either key context or string) cbc mode."
  (let* ((my-key-type (idea-legal-key key))
	 (my-key (cond ((equal 'key-complete-decryption my-key-type)
			(error "KEY is not an encryption key."))
		       ((equal 'key-complete-encryption my-key-type)
			key)
		       ((or (equal 'key-string my-key-type)
			   (equal 'key-intlist my-key-type))
			(idea-build-encryption-key key version))
		       (t (error "Invalid key."))))
	 (data (idea-cleartext-string-to-block-list str t))
	 (context '(0 0 0 0))
	 (l (length data))
	 (r (make-string (* l 8) 0))
	 (i 0))
    (while (< i l)
      (let* ((c (idea-encrypt-block (idea-xor-blocks (elt data i) context) 
				    my-key))
	     (b (idea-block-to-fixed-string c))
	     (so (* i 8)))
	(aset r so (elt b 0))
	(aset r (+ so 1) (elt b 1))
	(aset r (+ so 2) (elt b 2))
	(aset r (+ so 3) (elt b 3))
	(aset r (+ so 4) (elt b 4))
	(aset r (+ so 5) (elt b 5))
	(aset r (+ so 6) (elt b 6))
	(aset r (+ so 7) (elt b 7))
	(setq context c))
      (setq i (+ 1 i)))
    (b64-encode-string r)))

(defun idea-cbc-decrypt-string (cipstr key &optional version)
  "Decrypt STRING with KEY (either key context or string) cbc mode."
  (let ((str (b64-decode-string cipstr)))
    (if (and str (= 0 (% (length str) 8)))
	(let* ((my-key-type (idea-legal-key key))
	       (my-key (cond ((equal 'key-complete-encryption my-key-type)
			      (error "KEY is not an decryption key."))
			     ((equal 'key-complete-decryption my-key-type)
			      key)
			     ((or (equal 'key-string my-key-type)
				  (equal 'key-intlist my-key-type))
			      (idea-build-decryption-key key version))
			     (t (error "Invalid key."))))
	       (data (idea-ciphertext-string-to-block-list str))
	       (context '(0 0 0 0))
	       (l (length data))
	       (r (make-string (* l 8) 0))
	       (i 0))
	  (while (< i l)
	    (let* ((c (elt data i))
		   (b (idea-block-to-fixed-string 
		       (idea-xor-blocks (idea-encrypt-block c my-key) 
					context)))
		   (so (* i 8)))
	      (aset r so (elt b 0))
	      (aset r (+ so 1) (elt b 1))
	      (aset r (+ so 2) (elt b 2))
	      (aset r (+ so 3) (elt b 3))
	      (aset r (+ so 4) (elt b 4))
	      (aset r (+ so 5) (elt b 5))
	      (aset r (+ so 6) (elt b 6))
	      (aset r (+ so 7) (elt b 7))
	      (setq context c))
	    (setq i (+ 1 i)))
	  (let ((ret (idea-check-crc (idea-remove-padding r))))
	    (if ret
		ret
	      '())))
      '())))

(defun idea-add-crc (str)
  "Add 32bit crc to STRING"
  (concat (crc32-string str) str))

(defun idea-check-crc (str)
  "Check crc of the STRING.  Return string without crc if ok, otherwise '()"
  (if (< (length str) 8)
      '()
    (let ((realstr (substring str 8 (length str))))
      (if (string= (substring str 0 8) (crc32-string realstr))
	  realstr
	'()))))

(defun idea-hex-char-to-int (x)
  (if (and (>= x ?0) (<= x ?9))
      (- x ?0)
    (if (and (>= x ?a) (<= x ?f))
	(+ (- x ?a) 10)
      (if (and (>= x ?A) (<= x ?F))
	  (+ (- x ?A) 10)
	-1))))

(defun idea-build-key-annotation (key type &optional version)
  "Build annotation table of KEY that is of TYPE \"e\" or \"\d\"."
  (if (null version) (setq version idea-default-key-expand-version))
  (cond ((= version 1) (idea-build-key-annotation-version-1 key type))
	((= version 2) (idea-build-key-annotation-version-2 key type))
	((= version 3) (idea-build-key-annotation-version-3 key type))
	((= version 666) (concat type ":xxxxxxxxxx"))
	(t (error "Unknown key expansion version"))))

(defun idea-build-key-annotation-version-1 (key type)
  "Build annotation table of KEY that is of TYPE \"e\" or \"\d\"."
  (let ((r (make-string 16 0)))
    (aset r 15 (idea-& (nth 0 key) 255))
    (aset r 14 (idea-& (idea->> (nth 0 key) 8) 255))
    (aset r 13 (idea-& (nth 1 key) 255))
    (aset r 12 (idea-& (idea->> (nth 1 key) 8) 255))
    (aset r 11 (idea-& (nth 2 key) 255))
    (aset r 10 (idea-& (idea->> (nth 2 key) 8) 255))
    (aset r 9  (idea-& (nth 3 key) 255))
    (aset r 8  (idea-& (idea->> (nth 3 key) 8) 255))
    (aset r 7  (idea-& (nth 4 key) 255))
    (aset r 6  (idea-& (idea->> (nth 4 key) 8) 255))
    (aset r 5  (idea-& (nth 5 key) 255))
    (aset r 4  (idea-& (idea->> (nth 5 key) 8) 255))
    (aset r 3  (idea-& (nth 6 key) 255))
    (aset r 2  (idea-& (idea->> (nth 6 key) 8) 255))
    (aset r 1  (idea-& (nth 7 key) 255))
    (aset r 0  (idea-& (idea->> (nth 7 key) 8) 255))
    (concat type ":" (crc32-string r))))

(defun idea-build-key-annotation-version-2 (key type)
  "Build annotation table of KEY that is of TYPE \"e\" or \"\d\"."
  (if (and (= 0 (nth 0 key))
	   (= 0 (nth 1 key))
	   (= 0 (nth 2 key))
	   (= 0 (nth 3 key))
	   (= 0 (nth 4 key))
	   (= 0 (nth 5 key))
	   (= 0 (nth 6 key))
	   (= 0 (nth 7 key)))
      (concat type ":000000000000")
    (let ((r (make-string 9 0))
	  (s (make-string 9 0)))
      (aset r 8 0)
      (aset r 7 (idea-& (nth 0 key) 255))
      (aset r 6 (idea-& (idea->> (nth 0 key) 8) 255))
      (aset r 5 (idea-& (nth 1 key) 255))
      (aset r 4 (idea-& (idea->> (nth 1 key) 8) 255))
      (aset r 3 (idea-& (nth 2 key) 255))
      (aset r 2 (idea-& (idea->> (nth 2 key) 8) 255))
      (aset r 1 (idea-& (nth 3 key) 255))
      (aset r 0 (idea-& (idea->> (nth 3 key) 8) 255))
      (aset s 8 255)
      (aset s 7 (idea-& (nth 4 key) 255))
      (aset s 6 (idea-& (idea->> (nth 4 key) 8) 255))
      (aset s 5 (idea-& (nth 5 key) 255))
      (aset s 4 (idea-& (idea->> (nth 5 key) 8) 255))
      (aset s 3 (idea-& (nth 6 key) 255))
      (aset s 2 (idea-& (idea->> (nth 6 key) 8) 255))
      (aset s 1 (idea-& (nth 7 key) 255))
      (aset s 0 (idea-& (idea->> (nth 7 key) 8) 255))
      (setq s (concat (crc32-string (concat r s) 'raw-string) s))
      (setq r (concat (crc32-string (concat s r) 'raw-string) r))
      (concat type 
	      ":"
	      (substring (crc32-string r) 0 6)
	      (substring (crc32-string s) 0 6)))))

(defun idea-build-key-annotation-version-3 (key type)
  "Build annotation table of KEY that is of TYPE \"e\" or \"\d\"."
  (if (and (= 0 (nth 0 key))
           (= 0 (nth 1 key))
           (= 0 (nth 2 key))
           (= 0 (nth 3 key))
           (= 0 (nth 4 key))
           (= 0 (nth 5 key))
           (= 0 (nth 6 key))
           (= 0 (nth 7 key)))
      (concat type ":0000000000000000")
    (let ((r (make-string 16 0)))
      (aset r 15 (idea-& (nth 0 key) 255))
      (aset r 14 (idea-& (idea->> (nth 0 key) 8) 255))
      (aset r 13 (idea-& (nth 1 key) 255))
      (aset r 12 (idea-& (idea->> (nth 1 key) 8) 255))
      (aset r 11 (idea-& (nth 2 key) 255))
      (aset r 10 (idea-& (idea->> (nth 2 key) 8) 255))
      (aset r 9  (idea-& (nth 3 key) 255))
      (aset r 8  (idea-& (idea->> (nth 3 key) 8) 255))
      (aset r 7  (idea-& (nth 4 key) 255))
      (aset r 6  (idea-& (idea->> (nth 4 key) 8) 255))
      (aset r 5  (idea-& (nth 5 key) 255))
      (aset r 4  (idea-& (idea->> (nth 5 key) 8) 255))
      (aset r 3  (idea-& (nth 6 key) 255))
      (aset r 2  (idea-& (idea->> (nth 6 key) 8) 255))
      (aset r 1  (idea-& (nth 7 key) 255))
      (aset r 0  (idea-& (idea->> (nth 7 key) 8) 255))
      (let ((ra (idea-expand-string-to-key-version-3 r)))
	(concat type ":" (format "%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c"
				 (+ ?a (% (nth 0 ra) 26))
				 (+ ?a (% (nth 1 ra) 26))
				 (+ ?a (% (nth 2 ra) 26))
				 (+ ?a (% (nth 3 ra) 26))
				 (+ ?a (% (nth 4 ra) 26))
				 (+ ?a (% (nth 5 ra) 26))
				 (+ ?a (% (nth 6 ra) 26))
				 (+ ?a (% (nth 7 ra) 26))
				 (+ ?a (% (/ (nth 0 ra) 256) 26))
				 (+ ?a (% (/ (nth 1 ra) 256) 26))
				 (+ ?a (% (/ (nth 2 ra) 256) 26))
				 (+ ?a (% (/ (nth 3 ra) 256) 26))
				 (+ ?a (% (/ (nth 4 ra) 256) 26))
				 (+ ?a (% (/ (nth 5 ra) 256) 26))
				 (+ ?a (% (/ (nth 6 ra) 256) 26))
				 (+ ?a (% (/ (nth 7 ra) 256) 26))))))))

(defun idea-legal-subkey-p (subkey)
  "Is SUBKEY a legal subkey structure?"
  (if (and (listp subkey)
	   (= 6 (length subkey))
	   (integerp (nth 0 subkey))
	   (integerp (nth 1 subkey))
	   (integerp (nth 2 subkey))
	   (integerp (nth 3 subkey))
	   (integerp (nth 4 subkey))
	   (integerp (nth 5 subkey))
	   (<= 0 (nth 0 subkey))
	   (>= 65535 (nth 0 subkey))
	   (<= 0 (nth 1 subkey))
	   (>= 65535 (nth 1 subkey))
	   (<= 0 (nth 2 subkey))
	   (>= 65535 (nth 2 subkey))
	   (<= 0 (nth 3 subkey))
	   (>= 65535 (nth 3 subkey))
	   (<= 0 (nth 4 subkey))
	   (>= 65535 (nth 4 subkey))
	   (<= 0 (nth 5 subkey))
	   (>= 65535 (nth 5 subkey)))
      t
    nil))

(defun idea-legal-finalkey-p (finalkey)
  "Is FINALKEY a legal finalkey structure?"
  (if (and (listp finalkey)
	   (= 4 (length finalkey))
	   (integerp (nth 0 finalkey))
	   (integerp (nth 1 finalkey))
	   (integerp (nth 2 finalkey))
	   (integerp (nth 3 finalkey))
	   (<= 0 (nth 0 finalkey))
	   (>= 65535 (nth 0 finalkey))
	   (<= 0 (nth 1 finalkey))
	   (>= 65535 (nth 1 finalkey))
	   (<= 0 (nth 2 finalkey))
	   (>= 65535 (nth 2 finalkey))
	   (<= 0 (nth 3 finalkey))
	   (>= 65535 (nth 3 finalkey)))
      t
    nil))

(defun idea-legal-key (key)
  "Is KEY a idea key or can one be generated from it?"
  (cond ((stringp key)
	 'key-string)
	((and (listp key)
	      (= 8 (length key))
	      (integerp (nth 0 key))
	      (integerp (nth 1 key))
	      (integerp (nth 2 key))
	      (integerp (nth 3 key))
	      (integerp (nth 4 key))
	      (integerp (nth 5 key))
	      (integerp (nth 6 key))
	      (integerp (nth 7 key))
	      (<= 0 (nth 0 key))
	      (>= 65535 (nth 0 key))
	      (<= 0 (nth 1 key))
	      (>= 65535 (nth 1 key))
	      (<= 0 (nth 2 key))
	      (>= 65535 (nth 2 key))
	      (<= 0 (nth 3 key))
	      (>= 65535 (nth 3 key))
	      (<= 0 (nth 4 key))
	      (>= 65535 (nth 4 key))
	      (<= 0 (nth 5 key))
	      (>= 65535 (nth 5 key))
	      (<= 0 (nth 6 key))
	      (>= 65535 (nth 6 key))
	      (<= 0 (nth 7 key))
	      (>= 65535 (nth 7 key)))
	 'key-intlist)
	((and (listp key)
	      (= 10 (length key))
	      (idea-legal-subkey-p (nth 0 key))
	      (idea-legal-subkey-p (nth 1 key))
	      (idea-legal-subkey-p (nth 2 key))
	      (idea-legal-subkey-p (nth 3 key))
	      (idea-legal-subkey-p (nth 4 key))
	      (idea-legal-subkey-p (nth 5 key))
	      (idea-legal-subkey-p (nth 6 key))
	      (idea-legal-subkey-p (nth 7 key))
	      (idea-legal-finalkey-p (nth 8 key)))
	 (if (equal 'encryption (idea-key-type key))
	     'key-complete-encryption
	   'key-complete-decryption))
	(t nil)))

(defun idea-key-version (key)
  (let* ((my-key-type (idea-legal-key key))
	 (r (cond ((or (equal 'key-complete-decryption my-key-type)
		       (equal 'key-complete-encryption my-key-type))
		   (let ((fingerprint (idea-key-fingerprint key)))
		     (cond ((= (length fingerprint) 8) 1)
			   ((= (length fingerprint) 12) 2)
			   ((= (length fingerprint) 16) 3)
			   (t nil))))
		  ((or (equal 'key-string my-key-type)
		       (equal 'key-intlist my-key-type))
		   'any)
		  (t nil))))
    r))

(defun idea-key-fingerprint (key &optional version)
  "Get an IDEA-KEY fingerprint."
  (let* ((my-key-type (idea-legal-key key))
	 (my-key (cond ((equal 'key-complete-decryption my-key-type)
			key)
		       ((equal 'key-complete-encryption my-key-type)
			key)
		       ((or (equal 'key-string my-key-type)
			   (equal 'key-intlist my-key-type))
			(idea-build-encryption-key key version))
		       (t nil))))
    (if my-key
	(let ((annotation (nth 9 my-key)))
	  (if (not (and (stringp annotation)
			(> (length annotation) 9)))
	      nil
	    (let ((ty (elt annotation 0))
		  (de (elt annotation 1)))
	      (if (= de ?:)
		  (if (or (= ty ?e)
			  (= ty ?d))
		      (substring annotation 2 (length annotation))
		    nil)
		nil))))
      nil)))

(defun idea-key-type (key)
  "Is idea KEY encryption or decryption key?"
  (if (listp key)
      (let ((annotation (nth 9 key)))
	(if (not (and (stringp annotation)
		      (> (length annotation) 9)))
	    nil
	  (let ((ty (elt annotation 0))
		(de (elt annotation 1)))
	    (if (= de ?:)
		(if (= ty ?e)
		    'encryption
		  (if (= ty ?d)
		      'decryption
		    nil))
	      nil))))
    nil))

;;; eof (idea.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-crypt.el,v 3.16 1998/06/25 06:48:27 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info

(eval-when-compile (require 'irchat-inlines))
(eval-and-compile  
  (require 'irchat-vars)
  (require 'irchat-misc)
  (require 'idea))

(defvar irchat-crypt-version-default 3
  "Which encryption version to use as default (1-3).  1 is proved insecure.")

(defvar irchat-default-idea-key-list '() 
  "List to save ADDRESS KEY pairs")

(defvar irchat-known-idea-key-list '()
  "List of known IDEA decryption keys.")

(defconst irchat-idea-encrypt-msg-format-1 "|*E*|IDEA|1.0|%s|%s|")
(defconst irchat-idea-encrypt-msg-format-2 "|*E*|IDEA|2.0|%s|%s|")
(defconst irchat-idea-encrypt-msg-format-3 "|*E*|IDEA|3.0|%s|%s|")

(defun irchat-idea-encrypt-msg-format (&optional version)
  (if (not (numberp version)) (setq version irchat-crypt-version-default))
  (cond ((= version 1) irchat-idea-encrypt-msg-format-1)
	((= version 2) irchat-idea-encrypt-msg-format-2)
	((= version 3) irchat-idea-encrypt-msg-format-3)
	(t nil)))

(defun irchat-encrypted-message-p (message)
  (if (string-match "^|\\*E\\*|[^|]*|[0-9][0-9]*\\.[0-9][0-9]*|[^|]*|[^|]*|$"
		    message)
      t
    nil))

(defun irchat-crypt-address-has-default-key-p (address)
  "Is there a default key for ADDRESS."
  (if (assoc-ci-regexp-rev address irchat-default-idea-key-list)
      t
    nil))

(defun irchat-init-crypt ()
  "Initialize crypt variables"
  (while irchat-crypt-known-keys
    (irchat-Command-add-new-key (car irchat-crypt-known-keys))
    (setq irchat-crypt-known-keys (cdr irchat-crypt-known-keys)))
  (while irchat-crypt-default-keys
    (let ((c (car irchat-crypt-default-keys)))
      (if (and (consp c)
	       (stringp (car c))
	       (stringp (cdr c)))
	  (irchat-Command-set-default-key (car c) (cdr c))
	(if (and (listp c)
		 (stringp (nth 0 c))
		 (stringp (nth 1 c))
		 (or (null (nth 2 c)) (numberp (nth 2 c))))
	    (irchat-Command-set-default-key (nth 0 c) (nth 1 c) (nth 2 c)))))
    (setq irchat-crypt-default-keys (cdr irchat-crypt-default-keys)))
  t)

(defun irchat-read-passphrase (&optional prompt)
  "PROMPT for passphrase.  Use secure keyboard if possible."
  (if (null prompt)
      (setq prompt ""))
  (if (and (fboundp 'read-passwd)
	   irchat-crypt-secure-passphrase-read)
      (read-passwd prompt)
    (read-from-minibuffer prompt)))

(defun irchat-Command-add-new-key (key-var &optional interactive-p)
  "Add new KEY to known decryption keys list"
  (interactive (list (irchat-read-passphrase "Add passphrase: ") t))
  (let* ((my-key-type (idea-legal-key key-var))
	 (my-key (cond ((equal 'key-complete-encryption my-key-type)
			(error "KEY is not an decryption key."))
		       ((equal 'key-complete-decryption my-key-type)
			key-var)
		       ((or (equal 'key-string my-key-type)
			    (equal 'key-intlist my-key-type))
			(idea-build-decryption-key key-var 1))
		       (t (error "Invalid key."))))
	 (my-key-2 (if (or (equal 'key-string my-key-type)
			   (equal 'key-intlist my-key-type))
		       (idea-build-decryption-key key-var 2)
		     nil))
	 (my-key-3 (if (or (equal 'key-string my-key-type)
			   (equal 'key-intlist my-key-type))
		       (idea-build-decryption-key key-var 3)
		     nil))
	 (fingerprint (idea-key-fingerprint my-key))
	 (fingerprint-2 (if my-key-2 (idea-key-fingerprint my-key-2) nil))
	 (fingerprint-3 (if my-key-3 (idea-key-fingerprint my-key-3) nil)))
    (setq irchat-known-idea-key-list 
	  (cons (cons fingerprint my-key)
		(remassoc fingerprint
			  irchat-known-idea-key-list)))
    (if (and my-key-2 fingerprint-2)
	(setq irchat-known-idea-key-list 
	      (cons (cons fingerprint-2 my-key-2)
		    (remassoc fingerprint-2
			      irchat-known-idea-key-list))))
    (if (and my-key-3 fingerprint-3)
	(setq irchat-known-idea-key-list 
	      (cons (cons fingerprint-3 my-key-3)
		    (remassoc fingerprint-3
			      irchat-known-idea-key-list))))
    (if interactive-p
	(message (format "Added new decryption key (%s%s%s)."
			 fingerprint
			 (if fingerprint-2 
			     (concat " & " fingerprint-2)
			   "")
			 (if fingerprint-3
			     (concat " & " fingerprint-3)
			   ""))))))

(defun irchat-Command-delete-key (key-var &optional interactive-p)
  "Delete a KEY from known decryption keys list"
  (interactive (list (irchat-read-passphrase "Delete passphrase: ") t))
  (let ((fingerprint-1 (idea-key-fingerprint key-var 1))
	(fingerprint-2 (idea-key-fingerprint key-var 2))
	(fingerprint-3 (idea-key-fingerprint key-var 3)))
    (setq irchat-known-idea-key-list (remassoc fingerprint-1
					       irchat-known-idea-key-list))
    (setq irchat-known-idea-key-list (remassoc fingerprint-2
					       irchat-known-idea-key-list))
    (setq irchat-known-idea-key-list (remassoc fingerprint-3
					       irchat-known-idea-key-list))
    (if interactive-p
	(message (format "Removed decryption keys (%s and %s and %s)." 
			 fingerprint-1 
			 fingerprint-2
			 fingerprint-3)))))

(defun irchat-get-idea-decryption-key (fingerprint)
  "Find decryption key associated with FINGERPRINT"
  (let ((k (assoc fingerprint irchat-known-idea-key-list)))
    (if k
	(cdr k)
      nil)))

(defun irchat-get-idea-encryption-key (address &optional version)
  (let ((r (assoc-ci-regexp-rev address 
				irchat-default-idea-key-list)))
    (if r
	(let* ((ve (cond ((numberp (nth 7 r)) (nth 7 r)) ; In default key
			 ((numberp version) version)
			 (t irchat-crypt-version-default))))
	  (cond ((= ve 1) (nth 2 r))
		((= ve 2) (nth 4 r))
		((= ve 3) (nth 6 r))
		(t nil)))
      nil)))

(defun irchat-Command-set-default-key (addr-var 
				       pass-var
				       &optional version-var
				       interactive-p)
  "Set a default key for ADDRESS (channel/nick) to be KEY"
  (interactive (let (addr-var pass-var)
		 (setq addr-var (irchat-completing-default-read
				 "Default key for channel/user: "
				 (append irchat-nick-alist
					 irchat-channel-alist)
				 '(lambda (s) t) nil irchat-privmsg-partner))
		 (setq pass-var (irchat-read-passphrase "Passphrase: "))
		 (if (string= pass-var "")
		     (setq pass-var nil))	 
		 (list addr-var pass-var nil t)))
  (if (null pass-var)
      (let ((addr-var (upcase addr-var)))
	(setq irchat-default-idea-key-list
	      (remassoc addr-var irchat-default-idea-key-list))
	(if interactive-p
	    (message (format "Removed a default key from \"%s\"." addr-var))))
    (let* ((addr-var (upcase addr-var))
	   (e-key-1 (if (or (null version-var)
			    (= 1 version-var))
			(idea-build-encryption-key pass-var 1)
		      nil))
	   (fingerprint-1 (if e-key-1 
			      (idea-key-fingerprint e-key-1)
			    nil))
	   (e-key-2 (if (or (null version-var)
			    (= 2 version-var))
			(idea-build-encryption-key pass-var 2)
		      nil))
	   (fingerprint-2 (if e-key-2
			      (idea-key-fingerprint e-key-2)
			    nil))
	   (e-key-3 (if (or (null version-var)
			    (= 3 version-var))
			(idea-build-encryption-key pass-var 3)
		      nil))
	   (fingerprint-3 (if e-key-3
			      (idea-key-fingerprint e-key-3)
			    nil)))
      (irchat-Command-add-new-key pass-var)
      (setq irchat-default-idea-key-list 
	    (cons (list addr-var
			fingerprint-1 e-key-1
			fingerprint-2 e-key-2
			fingerprint-3 e-key-3
			version-var
			pass-var)
		  (remassoc addr-var 
			    irchat-default-idea-key-list)))
      (if interactive-p
	  (message (format "Added a default key for \"%s\"." addr-var)))
      (irchat-set-crypt-indicator))))

(defun irchat-make-encrypted-message (message key)
  "Build an encrypted message from MESSAGE with KEY"
  (let ((version (idea-key-version key)))
    (format (irchat-idea-encrypt-msg-format (if (numberp version) version))
	    (idea-key-fingerprint key)
	    (idea-cbc-encrypt-string message key))))

(defun irchat-crypt-valid-version-p (method major minor)
  "Is METHOD, MAJOR, MINOR a valid encryption method?"
  (and (string= method "IDEA")
       (or (= major 1)
	   (= major 2)
	   (= major 3))
       (>= minor 0)))

(defun irchat-encrypt-message (message address &optional no-clear-text)
  "Encrypt MESSAGE to ADDRESS.  NO-CLEAR-TEXT prohibits cleartext output"
  (let ((key (irchat-get-idea-encryption-key address)))
    (cond ((and no-clear-text
		(null key))
	   (error (format "No default key associated with \"%s\"." address)))
	  ((null key)
	   message)
	  (t (irchat-make-encrypted-message 
	      (concat irchat-real-nickname 
		      "" 
		      (irchat-generate-hex-timestamp)
		      ""
		      message)
	      key)))))

(defun irchat-decrypt-message (message)
  "Decrypt MESSAGE"
  (if (string-match "^|\\*E\\*|\\([^|]*\\)|\\([0-9][0-9]*\\)\\.\\([0-9][0-9]*\\)|\\([^|]*\\)|\\([^|]*\\)|$"
		    message)
      (let ((method (substring message (match-beginning 1) (match-end 1)))
	    (version-major (string-to-number (substring message
							(match-beginning 2)
							(match-end 2))))
	    (version-minor (string-to-number (substring message
							(match-beginning 3)
							(match-end 3))))
	    (fingerprint (substring message (match-beginning 4) (match-end 4)))
	    (msg (substring message (match-beginning 5) (match-end 5))))
	; (list method version-major version-minor fingerprint msg)
	(if (irchat-crypt-valid-version-p method version-major version-minor)
	    (let ((key (irchat-get-idea-decryption-key fingerprint)))
	      (if key
		  (let ((r (idea-cbc-decrypt-string msg key)))
		    (if r
			(if (string-match 
			     "^\\([^][^]*\\)\\([^][^]*\\)\\(.*\\)$"
			     r)
			    (let ((nick (substring r 
						   (match-beginning 1)
						   (match-end 1)))
				  (time (substring r 
						   (match-beginning 2)
						   (match-end 2)))
				  (msg (substring r 
						   (match-beginning 3)
						   (match-end 3))))

			      (list 'success nick time msg fingerprint))
			  (list
			   'error nil nil
			   "** Unable to decrypt: Invalid cleartext format!"
			   fingerprint))
		      (list
		       'error nil nil
		       "** Unable to decrypt: Decryption failed!"
		       fingerprint)))
		(list
		 'error nil nil
		 "** Unable to decrypt: No key!"
		 fingerprint)))
	  (list
	   'error nil nil
	   "** Unable to decrypt: Unknown version!"
	   fingerprint)))
    (list
     'error nil nil
     "** Unable to decrypt: Invalid message!"
     nil)))


(eval-and-compile (provide 'irchat-crypt))

;;; eof (irchat-crypt.el)
;;;  -*- emacs-lisp -*-
;;;
;;;  $Id: irchat-random.el,v 1.2 1998/06/24 23:08:31 tri Exp $
;;;
;;; see file irchat-copyright.el for change log and copyright info
;;;
(eval-and-compile  
  (require 'rc4))

(defvar irchat-random-state nil
  "State of the random generator in irchat.")

(defun irchat-random-8 ()
  (if (null irchat-random-state)
      (setq irchat-random-state (rc4-make-state 
				 (concat (if (fboundp 'current-time)
					     (prin1-to-string 
					      (current-time))
					   "*")
					 (format 
					  "%c%c%c%c%c%c%c%c%c%c"
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255))
					 (if (fboundp 'emacs-pid)
					     (prin1-to-string 
					      (emacs-pid))
					   "*")
					 (format 
					  "%c%c%c%c%c%c%c%c%c%c"
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255))
					 (if (fboundp 'user-login-name)
					     (prin1-to-string 
					      (user-login-name))
					   "*")
					 (format 
					  "%c%c%c%c%c%c%c%c%c%c"
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255))
					 (if (fboundp 'emacs-version)
					     (prin1-to-string 
					      (emacs-version))
					   "*")
					 (format 
					  "%c%c%c%c%c%c%c%c%c%c"
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255)
					  (logand (random) 255))))))
  (rc4-random irchat-random-state))

(defun irchat-random-16 ()
  (let ((x1 (irchat-random-8))
	(x2 (irchat-random-8)))
    (+ x1 (* x2 256))))

(defun irchat-random-24 ()
  (let ((x1 (irchat-random-8))
	(x2 (irchat-random-8))
	(x3 (irchat-random-8)))
    (+ x1 (* x2 256) (* x3 65536))))

(eval-and-compile (provide 'irchat-random))
;;;
;;; eof (irchat-random.el)
;;;
